function NumberToBitsTable (v)
  if v < 0 then
    -- negative
    return NumberToBitsTable (BitwiseNot (math.abs (v)) + 1)
  end
  
  -- to bits table
  local tbl = {}
  local cnt = 1
  
  while v > 0 do
    local last = math.mod (v, 2)
    
    if last == 1 then
      tbl [cnt] = 1
    else
      tbl [cnt] = 0
    end
    
    v   = (v-last) / 2
    cnt = cnt + 1
  end

  return tbl
end

function BitsTableToNumber (tbl)
  local n = #tbl

  local result = 0
  local power  = 1
  
  for i = 1, n do
    result = result + tbl [i] * power
    power = power * 2
  end

  return result
end

function BitwiseNot (v)
  local tbl  = NumberToBitsTable (v)
  local size = math.max(#tbl, 32)
  
  for i = 1, size do
    if tbl [i] == 1 then 
      tbl [i] = 0
    else
      tbl [i] = 1
    end
  end
  
  return BitsTableToNumber (tbl)
end

local function ExpandTable (tbl_m, tbl_n)
  local big   = {}
  local small = {}
  
  if #tbl_m > #tbl_n then
    big   = tbl_m
    small = tbl_n
  else
    big   = tbl_n
    small = tbl_m
  end
  
  -- expand small
  for i = #small + 1, #big do
    small [i] = 0
  end
end

function BitwiseOr (v1, v2)
  local tbl_v1 = NumberToBitsTable (v1)
  local tbl_v2 = NumberToBitsTable (v2)
  
  ExpandTable (tbl_v1, tbl_v2)

  local tbl  = {}
  local bits = math.max (#tbl_v1, #tbl_v2)
  
  for i = 1, bits do
    if tbl_v1 [i] == 0 and tbl_v2 [i] == 0 then
      tbl [i] = 0
    else
      tbl [i] = 1
    end
  end

  return BitsTableToNumber (tbl)
end

function BitwiseAnd (v1, v2)
  local tbl_v1 = NumberToBitsTable (v1)
  local tbl_v2 = NumberToBitsTable (v2)
  
  ExpandTable (tbl_v1, tbl_v2) 

  local tbl  = {}
  local bits = math.max (#tbl_v1, #tbl_v2)
  
  for i = 1, bits do
    if tbl_v1 [i] == 0 or tbl_v2 [i] == 0 then
      tbl [i] = 0
    else
      tbl [i] = 1
    end
  end

  return BitsTableToNumber (tbl)
end

function BitwiseShiftRight (v, amount)
  if (v < 0) then
    -- negative
    v = BitwiseNot (math.abs (v)) + 1
  end
  
  for i = 1, amount do
    v = v / 2
  end
  
  return math.floor(v)
end
