local BONUS_DOWNLOADED_VERSION_PATH = "bonus_downloaded_version.txt"
local DEFAULT_QUEUE_NAME            = "default"
local DOWNLOADED_SLOT_INFO_FORMAT   = "<?xml version='1.0' encoding='utf-8'?> <info version='%s' platform='%s' resolution='%s' format='%s'/>"
local MIN_PROGRESS_CALL_INTERVAL    = 1/60
local SLOT_DOWNLOADED_VERSION_PATH  = "slot_downloaded_version.txt"
local SLOTS_DOWNLOAD_PATH           = DOWNLOADS_PATH .. "slots/"
local TMP_PATH                      = "/system/temp/Plarium/Slots"

local BACKGROUND_COPY_FILE_CANCELED_STATUS = -1

GAMES_RESOURCES_SERVER_DEV  = "http://static.indigo-dev.com/slots_mobile/dev/"
GAMES_RESOURCES_SERVER_QA   = "http://plarium.vo.llnwd.net/v1/slots_mobile/live/"
GAMES_RESOURCES_SERVER_LIVE = "http://plarium.vo.llnwd.net/v1/slots_mobile/live/"

--[[GAMES_RESOURCES_SERVER_DEV  = "http://untgames.com/files/slots/"
GAMES_RESOURCES_SERVER_QA   = "http://untgames.com/files/slots/"
GAMES_RESOURCES_SERVER_LIVE = "http://untgames.com/files/slots/"]]

local function GetGamesResourcesServerPath () 
  local hs_port = GetHsPort ()

  if hs_port == HS_LIVE_PORT then
    return GAMES_RESOURCES_SERVER_LIVE
  elseif hs_port == HS_QA_PORT then
    return GAMES_RESOURCES_SERVER_QA
  else
    return GAMES_RESOURCES_SERVER_DEV
  end
end

local function OnActiveDownloadCanceled (queue)
  if queue.active_download_callback then
    local callback = queue.active_download_callback
  
    queue.active_download_callback = nil

    game.main_timer.DelayedAction (0.25, function ()
      callback (BACKGROUND_COPY_FILE_CANCELED_STATUS)
    end)
  end
end

local function IsFileDownloaded (impl, save_path)
  local downloaded_flag = save_path .. "_downloaded"

  return Common.File.IsFileExist (save_path) and Common.File.IsFileExist (downloaded_flag)
end

local function DownloadFileImpl (impl, url, save_path, callback, force_download, queue)
  local downloaded_flag = save_path .. "_downloaded"

  if not force_download and IsFileDownloaded (impl, save_path) then
    callback (Common.BackgroundCopyStateStatus.Finished)
    return
  end

  if Common.File.IsFileExist (downloaded_flag) then
    Common.File.Remove (downloaded_flag)
  end

  queue.active_download_callback = callback

  local function CopyHandler (state)
    local status = state.Status

    queue.download_started = true

    if status == Common.BackgroundCopyStateStatus.Finished or status == Common.BackgroundCopyStateStatus.Failed then
      queue.active_download_callback = nil
    
      if DEBUG then
        print ("Finished downloading file " .. url .. " with status " .. status .. " state " .. state.StatusText)
      end
      
      System.Application.ScreenSaverState = true
    end
    
    if status == Common.BackgroundCopyStateStatus.InProgress or status == Common.BackgroundCopyStateStatus.Started then
      callback (Common.BackgroundCopyStateStatus.InProgress, state.FileSize, state.BytesCopied)
    elseif status == Common.BackgroundCopyStateStatus.Finished then
      Common.File.PostString (downloaded_flag, "")
      
      callback (Common.BackgroundCopyStateStatus.Finished)
    elseif status == Common.BackgroundCopyStateStatus.Failed then
      callback (Common.BackgroundCopyStateStatus.Failed, state.StatusText)
    end     
  end

  System.Application.ScreenSaverState = false

  if DEBUG then
    print ("Start downloading file " .. url .. " to " .. save_path)
  end
  
  queue.download_started = false
  
  return Common.File.BackgroundCopyFile (url, save_path, Common.File.CreateBackgroundCopyFileCallback (CopyHandler))
end

local function DownloadNextFile (impl, queue)
  if #queue.items == 0 then return end

  local current_download = queue.items [1]

  local callbacks     = current_download.callbacks
  local download_list = current_download.list
  
  if not download_list then
    download_list = { { url = current_download.url, save_path = current_download.save_path } }
  end
  
  local function DownloadFileFromList (index)
    if index > #download_list then
      for i = 1, #callbacks do
        local status, return_value = pcall (function ()
          callbacks [i] (Common.BackgroundCopyStateStatus.Finished)
        end)
        
        if not status then
          print ("Download callback error '" .. return_value .. "'")
        end
      end

      table.remove (queue.items, 1)
      DownloadNextFile (impl, queue)
      current_download = nil
      return
    end
  
    current_download.action = DownloadFileImpl (impl, download_list [index].url, download_list [index].save_path, function (status, arg2, arg3)
      if status == Common.BackgroundCopyStateStatus.InProgress or status == Common.BackgroundCopyStateStatus.Started then
        local current_time = game.main_timer.GetTime ()
    
        if current_time - impl.last_progress_call_time > MIN_PROGRESS_CALL_INTERVAL then
          for i = 1, #callbacks do
            callbacks [i] (Common.BackgroundCopyStateStatus.InProgress, arg2 * #download_list, arg2 * (index - 1) + arg3)
          end
        
          impl.last_progress_call_time = current_time
        end
      elseif status == Common.BackgroundCopyStateStatus.Finished then
        DownloadFileFromList (index + 1)
      elseif status == Common.BackgroundCopyStateStatus.Failed then
        for i = 1, #callbacks do
          callbacks [i] (Common.BackgroundCopyStateStatus.Failed, arg2)
        end
      end     

      if status == BACKGROUND_COPY_FILE_CANCELED_STATUS or status == Common.BackgroundCopyStateStatus.Failed then
        table.remove (queue.items, 1)
        DownloadNextFile (impl, queue)
        current_download = nil
      end
    end, current_download.force_download, queue)
  end
  
  DownloadFileFromList (1)
end

local function GetDownloadId (impl)
  local return_value = impl.download_id
  
  impl.download_id = impl.download_id + 1
  
  return return_value
end

local function MakeGenericCallback (progress_callback, finish_callback, error_callback)
  local last_progress = -1

  return function (status, arg2, arg3)
    if status == Common.BackgroundCopyStateStatus.InProgress then
      local progress = math.floor (arg3 / arg2 * 100)
    
      if progress > last_progress then
        last_progress = progress
        progress_callback (progress / 100)
      end
    elseif status == Common.BackgroundCopyStateStatus.Finished then
      finish_callback ()
    elseif status == Common.BackgroundCopyStateStatus.Failed then
      if error_callback then
        error_callback ()
      end
    end
  end
end

local function GetDownloadQueue (impl, name)
  if not name then
    name = DEFAULT_QUEUE_NAME
  end
  
  if not impl.download_queues [name] then
    impl.download_queues [name] = { items = {} }
  end
  
  return impl.download_queues [name]
end

local function DownloadFiles (impl, list, progress_callback, finish_callback, error_callback, first_priority, queue_name)
  local download_list = {}
  
  for i = 1, #list do
    local download = {}
    
    download.url       = list [i].url
    download.save_path = list [i].save_path
    
    if not download.url then
      error ("download url not specified")
    end
    if not download.save_path then
      error ("download save path not specified")
    end
    
    table.insert (download_list, download)
  end
  
  local download_id = GetDownloadId (impl)
  local queue       = GetDownloadQueue (impl, queue_name)
  local queue_item  = { id = download_id, list = download_list, callbacks = { MakeGenericCallback (progress_callback, finish_callback, error_callback) } }
 
  if first_priority then
    if #queue.items > 0 then
      table.insert (queue.items, 2, queue_item)
    else
      table.insert (queue.items, queue_item)
    end
  else
    table.insert (queue.items, queue_item)
  end

  if #queue.items == 1 then --no downloads in progress
    DownloadNextFile (impl, queue)
  end  
  
  return download_id
end

local function DownloadFile (impl, url, save_path, callback, force_download, queue_name, try_local_first, first_priority, params)
  if not params then
    params = {}
  end

  if params.check_already_downloaded_immediately and not force_download then
    if IsFileDownloaded (impl, save_path) then
      callback (Common.BackgroundCopyStateStatus.Finished)
      return
    end
  end

  local queue = GetDownloadQueue (impl, queue_name)

  local insert_position   = #queue.items + 1
  local is_local_download = false

  if try_local_first then
    local local_file_name = GetLastPathComponent (url)
    
    if Common.File.IsFileExist (local_file_name) then
      url               = local_file_name
      is_local_download = true

      insert_position = math.min (2, insert_position)
      
      for i = 2, #queue.items do
        if queue.items [i].is_local_download then
          insert_position = i + 1
        end
      end
    end
  end
  
  if first_priority then
    insert_position = math.min (2, #queue.items + 1)
  end

  table.insert (queue.items, insert_position, { url = url, save_path = save_path, callbacks = { callback }, force_download = force_download, is_local_download = is_local_download } )

  if #queue.items > 1 then --download in progress
    return
  end  

  DownloadNextFile (impl, queue)
end

local function CancelDownload (impl, id)
  print ("CANCEL DOWNLOAD " .. tostring (id))

  for _, queue in pairs (impl.download_queues) do
    for i = 1, #queue.items do
      local item = queue.items [i]
    
      if item.id and item.id == id then
        if i > 1 then
          table.remove (queue.items, i)
        else
          item.action:Cancel ()
 
          OnActiveDownloadCanceled (queue)
        end
      
        return
      end
    end
  end  
end

local function GetSlotByPath (path)
  local slots = ApplicationDataGet (SLOTS_DESCS)
  
  for i = 1, #slots do
    if slots [i].path == path then
      return slots [i]
    end
  end 
end

local function GetSlotPath (impl, path, is_bonus)
  local package_download_path = SLOTS_DOWNLOAD_PATH .. path .. "/"
  
  if is_bonus then
    return package_download_path .. "bonus/bonus.zip"
  else
    return package_download_path .. "slot/slot.zip"
  end
end

local function GetDownloadedSlotInfo (impl, path, is_bonus)
  local download_path = SLOTS_DOWNLOAD_PATH .. path .. "/"
  
  local downloaded_version_path
  
  if is_bonus then
    downloaded_version_path = download_path .. BONUS_DOWNLOADED_VERSION_PATH
  else
    downloaded_version_path = download_path .. SLOT_DOWNLOADED_VERSION_PATH
  end
  
  if Common.File.IsFileExist (downloaded_version_path) then
    local status, return_value = pcall (function ()
      local slot_info = Common.StringNode.LoadXmlFromFile (downloaded_version_path)

      local return_value = {}
    
      return_value.version    = slot_info:Get ("version") + 0
      return_value.platform   = slot_info:Get ("platform")
      return_value.resolution = slot_info:Get ("resolution")
      return_value.format     = slot_info:Get ("format")
    
      return return_value    
    end)
      
    if status then
      return return_value
    else
      print (return_value)
      return nil
    end
  end
end

local function IsSlotResourceDownloaded (impl, path, is_bonus)
  local slot          = GetSlotByPath (path)
  local pak_path      = GetSlotPath (impl, path, is_bonus)
  local pak_flag      = pak_path .. "_downloaded"
    
  if not Common.File.IsFileExist (pak_path) or not Common.File.IsFileExist (pak_flag) then return false end --Check that data files actually exists
  
  local slot_info = GetDownloadedSlotInfo (impl, path, is_bonus)
  
  if not slot_info then return false end
  
  return slot_info.version > slot.update_from_version and slot_info.platform == GetPlatform () and slot_info.format == ApplicationDataGet (TEXTURE_FORMATS) [1]
end

local function IsSlotDownloaded (impl, path)
  return IsSlotResourceDownloaded (impl, path, false)
end

local function IsBonusDownloaded (impl, path)
  return IsSlotResourceDownloaded (impl, path, true)
end

local function AddSlotDownloadCallbacks (impl, path, progress_callback, finish_callback, error_callback, in_url, in_download_path)
  local slot          = GetSlotByPath (path)
  local url           = in_url or GetGamesResourcesServerPath () .. slot.slot_resource_path
  local download_path = in_download_path or SLOTS_DOWNLOAD_PATH .. path .. "/slot/slot.zip"

  for _, queue in pairs (impl.download_queues) do
    for i = 1, #queue.items do
      local item = queue.items [i]
    
      if item.url == url and item.save_path == download_path then
        local last_progress = 0
  
        table.insert (item.callbacks, function (status, arg2, arg3)
          if status == Common.BackgroundCopyStateStatus.InProgress then
            local progress = math.floor (arg3 / arg2 * 100)
    
            if progress > last_progress then
              last_progress = progress

              if progress_callback then
                progress_callback (progress / 100)
              end
            end
          elseif status == Common.BackgroundCopyStateStatus.Finished then
            if finish_callback then
              finish_callback ()
            end
          elseif status == Common.BackgroundCopyStateStatus.Failed then
            if error_callback then
              error_callback ()
            end
          end
        end)
      
        return queue, item, i
      end
    end
  end  
  
  return nil
end

local function MoveSlotDownloadToFront (impl, path, progress_callback, finish_callback, error_callback, ignore_no_slot_error)
  local queue, item, item_index = AddSlotDownloadCallbacks (impl, path, progress_callback, finish_callback, error_callback)

  if queue then    
    if item_index > 2 then
      table.remove (queue.items, item_index)
      table.insert (queue.items, 2, item)
    end
      
    return
  end
  
  print ("ERROR: slot for moving to front not found")
  if error_callback and not ignore_no_slot_error then
    error_callback ("Slot not found")
  end
end

local function MoveFileDownloadToFront (impl, url, save_path)
  for _, queue in pairs (impl.download_queues) do
    for i = 1, #queue.items do
      local item = queue.items [i]
    
      if item.url == url and item.save_path == save_path then
        if i > 2 then
          table.remove (queue.items, i)
          table.insert (queue.items, 2, item)
        end
        
        return
      end
    end
  end  
  
  print ("ERROR: file for moving to front not found")
end

local function HasLocalSlotResource (impl, path, is_bonus)
  local slot = GetSlotByPath (path)
  local local_path
  
  if is_bonus then
    local_path = slot.bonus_resource_path
  else
    local_path =  slot.slot_resource_path
  end
  
  return Common.File.IsFileExist (GetLastPathComponent (local_path))
end

local function DownloadSlot (impl, path, progress_callback, finish_callback, error_callback)
  local queue, item, item_index = AddSlotDownloadCallbacks (impl, path, progress_callback, finish_callback, error_callback)

  if queue then return end

  local slot                     = GetSlotByPath (path)
  local download_path            = SLOTS_DOWNLOAD_PATH .. path .. "/"
  local downloaded_version_path  = download_path .. SLOT_DOWNLOADED_VERSION_PATH
  local package_download_path    = download_path .. "slot"
  
  if not Common.File.IsFileExist (download_path) then
    Common.File.Mkdir (download_path)
  end

  local slot_info = GetDownloadedSlotInfo (impl, path, false)

  if slot_info then
    if IsSlotDownloaded (impl, path) then
      finish_callback ()
      return
    else
      Common.File.Remove (downloaded_version_path)
      Common.File.Remove (package_download_path)
    end
  else
    Common.File.Remove (downloaded_version_path)
    Common.File.Remove (package_download_path)
  end

  if not Common.File.IsFileExist (package_download_path) then
    Common.File.Mkdir (package_download_path)
  end
  
  package_download_path = package_download_path .. "/slot.zip"
  
  local last_progress = 0
  
  DownloadFile (impl, GetGamesResourcesServerPath () .. slot.slot_resource_path, package_download_path, function (status, arg2, arg3)
    if status == Common.BackgroundCopyStateStatus.InProgress then
      local progress = math.floor (arg3 / arg2 * 100)
    
      if progress > last_progress then
        last_progress = progress
        progress_callback (progress / 100)
      end
    elseif status == Common.BackgroundCopyStateStatus.Finished then
      Common.File.PostString (downloaded_version_path, string.format (DOWNLOADED_SLOT_INFO_FORMAT, tostring (slot.version), GetPlatform (), slot.resolution, ApplicationDataGet (TEXTURE_FORMATS) [1]))
      finish_callback ()
    elseif status == Common.BackgroundCopyStateStatus.Failed then
      print ("Download " .. slot.bonus_resource_path .. " failed, error " .. arg2)
      if error_callback then
        error_callback ()
      end
    end
  end, nil, nil, true)
end

local function DownloadBonus (impl, path, progress_callback, finish_callback, error_callback)
  local slot                     = GetSlotByPath (path)
  local download_path            = SLOTS_DOWNLOAD_PATH .. path .. "/"
  local downloaded_version_path  = download_path .. BONUS_DOWNLOADED_VERSION_PATH
  local package_download_path    = download_path .. "bonus"
  local url                      = GetGamesResourcesServerPath () .. slot.bonus_resource_path
  
  local queue, item, item_index = AddSlotDownloadCallbacks (impl, path, progress_callback, finish_callback, error_callback, url, package_download_path .. "/bonus.zip")

  if queue then return end

  if not Common.File.IsFileExist (download_path) then
    Common.File.Mkdir (download_path)
  end

  local slot_info = GetDownloadedSlotInfo (impl, path, true)

  local status, error = pcall (function ()
    if slot_info then
      if IsBonusDownloaded (impl, path) then
        finish_callback ()
        return
      else
        Common.File.Remove (downloaded_version_path)
        Common.File.Remove (package_download_path)
      end 
    else
      Common.File.Remove (downloaded_version_path)
      Common.File.Remove (package_download_path)
    end

    if not Common.File.IsFileExist (package_download_path) then
      Common.File.Mkdir (package_download_path)
    end
  end)

  if not status then
    print ("Download bonus error '" .. error .. "'")
    
    if error_callback then
      error_callback ()
    end
    
    return
  end
  
  package_download_path = package_download_path .. "/bonus.zip"
  
  local last_progress = 0
  
  DownloadFile (impl, url, package_download_path, function (status, arg2, arg3)
    if status == Common.BackgroundCopyStateStatus.InProgress then
      local progress = math.floor (arg3 / arg2 * 100)
    
      if progress > last_progress then
        last_progress = progress
        progress_callback (progress / 100)
      end
    elseif status == Common.BackgroundCopyStateStatus.Finished then
      Common.File.PostString (downloaded_version_path, string.format (DOWNLOADED_SLOT_INFO_FORMAT, tostring (slot.version), GetPlatform (), slot.resolution, ApplicationDataGet (TEXTURE_FORMATS) [1]))
      finish_callback ()
    elseif status == Common.BackgroundCopyStateStatus.Failed then
      print ("Download " .. slot.bonus_resource_path .. " failed, error" .. arg2)
      if error_callback then
        error_callback ()
      end
    end
  end, nil, nil, true)
end

local function DownloadSlotResource (impl, path, progress_callback, finish_callback, is_bonus, error_callback)
  if is_bonus then
    DownloadBonus (impl, path, progress_callback, finish_callback, error_callback)
  else
    DownloadSlot (impl, path, progress_callback, finish_callback, error_callback)
  end
end

local function DeleteSlotResource (impl, path, is_bonus)
  local slot                    = GetSlotByPath (path)
  local download_path           = SLOTS_DOWNLOAD_PATH .. path .. "/"
  local downloaded_version_path
  local package_download_path
  
  if is_bonus then
    downloaded_version_path = download_path .. BONUS_DOWNLOADED_VERSION_PATH
    package_download_path   = download_path .. "bonus"
  else
    downloaded_version_path = download_path .. SLOT_DOWNLOADED_VERSION_PATH
    package_download_path   = download_path .. "slot"
  end

  Common.File.Remove (downloaded_version_path)
  Common.File.Remove (package_download_path)
end

local function CancelAllDownloads (impl)
  print ("CANCEL ALL DOWNLOADS")

  for _, queue in pairs (impl.download_queues) do
    if #queue.items > 0 then
      local current_download = queue.items [1]
  
      queue.items = {}
  
      OnActiveDownloadCanceled (queue)
  
      current_download.action:Cancel ()
    end
  end
end

local function CancelDownloadFile (impl, save_path)
  print ("CANCEL DOWNLOAD PATH " .. tostring (save_path))

  for _, queue in pairs (impl.download_queues) do
    for i = 1, #queue.items do
      local item = queue.items [i]
    
      if item.save_path == save_path then
        if i > 1 then
          table.remove (queue.items, i)
        else
          item.action:Cancel ()
  
          OnActiveDownloadCanceled (queue)
        end
      
        break
      end
    end
  end  
end

local function GetTmpFilePath (impl)
  impl.tmp_file_index = impl.tmp_file_index + 1

  return TMP_PATH .. "/" .. impl.tmp_file_index .. ".tmp"
end

local function LoadString (impl, path, finish_callback, error_callback)
  local save_path = GetTmpFilePath (impl)
 
  local function CopyHandler (state)
    local status = state.Status

    if status == Common.BackgroundCopyStateStatus.Finished or status == Common.BackgroundCopyStateStatus.Failed then
      if DEBUG then
        print ("Finished downloading file " .. path .. " with status " .. status .. " state " .. state.StatusText)
      end
    end    
    
    if status == Common.BackgroundCopyStateStatus.Finished then
      finish_callback (Common.File.LoadString (save_path))
      
      Common.File.Remove (save_path)
    elseif status == Common.BackgroundCopyStateStatus.Failed then
      error_callback ()
    end     
  end
  
  if DEBUG then
    print ("Loading string from '" .. path .. "' to '" .. save_path .. "'")
  end  

  Common.File.BackgroundCopyFile (path, save_path, Common.File.CreateBackgroundCopyFileCallback (CopyHandler))
end

local function CreateDownloadsManager ()
  local impl = {}

  impl.download_id     = 1
  impl.download_queues = {}
  
  impl.last_progress_call_time = 0

  if not Common.File.IsFileExist (DOWNLOADS_PATH) then
    Common.File.Mkdir (DOWNLOADS_PATH)
  end

  if ApplicationDataGet (OS_TYPE) == "iOS" then
    Common.File.SetFileStringAttribute (DOWNLOADS_PATH, "com.apple.MobileBackup", string.char (1))
  end 

  if Common.File.IsFileExist (TMP_PATH) then
    pcall (function ()
      Common.File.Remove (TMP_PATH)
    end)
  end

  if not Common.File.IsFileExist (TMP_PATH) then
    Common.File.Mkdir (TMP_PATH)
  end

  impl.tmp_file_index = 0

  local public_methods = {}

  public_methods.CancelDownload              = function (id) CancelDownload (impl, id) end
  public_methods.CancelDownloadFile          = function (save_path) CancelDownloadFile (impl, save_path) end
  public_methods.CancelAllDownloads          = function () CancelAllDownloads (impl) end
  public_methods.DeleteSlotResource          = function (path, is_bonus) DeleteSlotResource (impl, path, is_bonus) end
  public_methods.DownloadFile                = function (url, save_path, callback, force_download, queue_name, try_local_first, first_priority, params) return DownloadFile (impl, url, save_path, callback, force_download, queue_name, try_local_first, first_priority, params) end
  public_methods.DownloadFiles               = function (list, progress_callback, finish_callback, error_callback, first_priority, queue_name) return DownloadFiles (impl, list, progress_callback, finish_callback, error_callback, first_priority, queue_name) end
  public_methods.DownloadBonus               = function (path, progress_callback, finish_callback, error_callback) DownloadBonus (impl, path, progress_callback, finish_callback, error_callback) end
  public_methods.DownloadSlot                = function (path, progress_callback, finish_callback, error_callback) DownloadSlot (impl, path, progress_callback, finish_callback, error_callback) end
  public_methods.DownloadSlotResource        = function (path, progress_callback, finish_callback, is_bonus, error_callback) DownloadSlotResource (impl, path, progress_callback, finish_callback, is_bonus, error_callback) end
  public_methods.GetDownloadedSlotInfo       = function (path, is_bonus) return GetDownloadedSlotInfo (impl, path, is_bonus) end 
  public_methods.GetGamesResourcesServerPath = function () return GetGamesResourcesServerPath () end
  public_methods.GetSlotPath                 = function (path, is_bonus) return GetSlotPath (impl, path, is_bonus) end 
  public_methods.HasLocalSlotResource        = function (path, is_bonus) return HasLocalSlotResource (impl, path, is_bonus) end
  public_methods.IsSlotResourceDownloaded    = function (path, is_bonus) return IsSlotResourceDownloaded (impl, path, is_bonus) end
  public_methods.IsBonusDownloaded           = function (path) return IsBonusDownloaded (impl, path) end
  public_methods.IsSlotDownloaded            = function (path) return IsSlotDownloaded (impl, path) end
  public_methods.LoadString                  = function (path, finish_callback, error_callback) LoadString (impl, path, finish_callback, error_callback) end
  public_methods.MoveSlotDownloadToFront     = function (path, progress_callback, finish_callback, error_callback, ignore_no_slot_error) MoveSlotDownloadToFront (impl, path, progress_callback, finish_callback, error_callback, ignore_no_slot_error) end
  public_methods.MoveFileDownloadToFront     = function (url, save_path) MoveFileDownloadToFront (impl, url, save_path) end
    
  return public_methods
end

local DownloadsManagerInstance

function GetDownloadsManager ()
  if not DownloadsManagerInstance then
    DownloadsManagerInstance = CreateDownloadsManager ()
  end
  
  return DownloadsManagerInstance
end
