-- TODO - копирование переменных из таблицы экрана, а не ссылка

PROFILER_ENABLED = false
GAME_FPS="FPS: "
FPS_LINE=nil

local BATTLE_LOADING_SCREEN      = "loader_battles"
local DEFAULT_SOUND_FADE_IN_TIME = 0.5
local DEFAULT_SOUND_GROUP_NAME   = "Default"
local DEFAULT_STROKE_ALPHA       = "0.75"
local DONT_PAUSE_SOUNDS          = { 'click', 'good_ui_window', 'click_game', 'battle_win_window', 
                                     'battle_lose_window', 'battle_draw_window', "lobby_background",
                                     'battle_room_click', "spin_click", "booster_buy" }
local EPS                        = 0.0001
local GAME_DESC                  = "game.xml"
local SCREEN_WIDTH               = 10000
local SCREEN_HEIGHT              = 10000
local SCRIPTS_DIR                = "data/scripts"
local SCRIPT_SUFFIX              = ".lua"
local TOUCH_SIZE                 = SCREEN_WIDTH * 0.1

local game_paused = false
local screen_to_set = nil  --set this screen on update end

-- Выполнение функции для всех элементов таблицы
local function ForEachInTable (t, func, ...)
  if not (t) then return end
  local index, value = next (t)
  while index do
    func (value, ...)

    index, value = next (t, index)
  end
end

-- получение значения переменной по ее имени(с возможностью вычисления функций)
function GetVarValue(vars, name)
  local var=vars[name]
  if type(var)=="function" then
    var=var(vars)
  end
  return var
end

function TryGetVar(vars, name, default)
  local v=GetVarValue(vars, name)
  if v==nil then
    v=default
  end
  return v
end

-- Получение таймера
local function GetTimer (screen, name)
  for timer_name, timer in pairs (screen.timers) do
    if timer_name == name then
      return timer
    end
  end

  error (string.format ("Timer '%s' not found", name))
end

-- Получение имён в анимации
local function GetAnimationNames (str, step)
  local delimiters_indices = {}
  local delimiters_count = 0

  for i = 1, string.len (str), 1 do
    if string.byte (str, i, i) == string.byte (";", 1, 1) then
      delimiters_indices [delimiters_count] = i
      delimiters_count = delimiters_count + 1
    end

    if delimiters_count == 2 then
      break
    end
  end  

  local mask = string.sub (str, 1, delimiters_indices[0] - 1)
  local first, last

  if delimiters_indices[0] then
    first = string.sub (str, delimiters_indices[0] + 1, delimiters_indices[1] - 1) + 0
  end
  
  if delimiters_indices[1] then
    last = string.sub (str, delimiters_indices[1] + 1) + 0
  end  

  if not (first) or not (last) then return {} end
  
  local materials = {}
  
  if not step then step = 1 end
  
  if last < first then
    step = -step
  end
  
  for i = first, last, step do
    local material_name = string.format (mask, i)            

    table.insert (materials, material_name)
  end
  
  return materials
end

-- Загрузка ключа трека линейной анимации
local function LoadAnimationTrackKey (xml_node, vars)
  local key   = {}
  local event = xml_node:Get("Event", "")
  local time  = xml_node:Get ("Time")
  
  if string.gmatch (time, "[0-9+-.]+")() == time then
    key.time  = time + 0
  else
    if not GetVarValue(vars, time) then error (string.format ("No var '%s'", time)) end    
    
    key.time = GetVarValue(vars, time)
    time     = key.time
  end
  
  if event ~= "" then
    key.value = loadstring ("return function (time, event, sprite) " .. event .. " end")()
    key.type  = "event"
  else
    local value_string = xml_node:Get ("Value")
  
    if string.find (value_string, ";") == nil then
      -- Конвертация скалярного значения
      
      key.value = value_string + 0.0
      key.type  = "scalar"
    else
      -- Конвертация вектора
      
      key.value = StringToVec (value_string)
      key.type  = "vec"
    end  
  end  

  if key.time < 0 then
    error ("key.time < 0")    

    return nil
  end

  return key
end

-- Обновление времени анимации на треке
local function UpdateAnimationTrack (track)
  local last_key = nil

  if track.spline then
    track.length = track.spline.MaxTime
    
    if track.spline.KeysCount > 0 then
      last_key = {}
      
      last_key.time = track.spline.MaxTime
    end
  else
    local index, key = next (track.keys)

    while index do
      if not (last_key) then
        last_key = key
      else
        if key.time > last_key.time then
          last_key = key
        end
      end

      index, key = next (track.keys, index)
    end
  
    track.length = last_key.time
  end

  if track.loop then
    if last_key then
      track.loop_time = last_key.time
    else
      track.loop_time = 0.0
      track.loop      = false
    end
  end
end

-- Линейная интерполяция
local function ScalarLerp (t, start_value, end_value)
  return start_value + t * (end_value - start_value)
end

local function VectorLerp (t, start_value, end_value)
  return {x = ScalarLerp (t, start_value.x, end_value.x), y = ScalarLerp (t, start_value.y, end_value.y)}
end

-- Сложение векторов
local function ScalarAdd (a, b)
  return a + b
end

local function VectorAdd (a, b)
  if type (a) == "userdata" then
    if a.__library_name == "Math.Vector2" then
      return a + vec2 (b.x + 0, b.y + 0)
    elseif a.__library_name == "Math.Vector3" then
      return a + vec3 (b.x + 0, b.y + 0, b.z + 0)
    elseif a.__library_name == "Math.Vector4" then
      return a + vec4 (b.x + 0, b.y + 0, b.z + 0, b.w + 0)
    end
  end

  return {x=a.x+b.x, y=a.y+b.y}
end

-- Загрузка трека линейной анимации
local function LoadAnimationTrack (xml_node, default_offset, vars, static_spans)
  local track = {}

  track.name   = xml_node:Get ("Name")
  track.offset = xml_node:Get ("Offset", tostring (default_offset)) + 0
  track.loop   = xml_node:Get ("Loop", "false")
  track.loop   = track.loop == "true"
  track.link   = xml_node:Get ("Link", "")

  local children_count = xml_node.ChildrenCount
  local is_vectors     = false
  local is_scalar      = false
  
  local spline

  if track.name == "scale" or track.name == "position" or track.name == "size" or track.name == "offset" then
    spline = Math.LinearSplineVec2.Create ()
  elseif track.name == "alpha" or track.name == "angle" then
    spline = Math.LinearSplineFloat.Create ()
  end
      
  if spline then  
    if track.loop then
      spline:Wrap (Math.SplineWrap.Repeat, Math.SplineWrap.Repeat)
    end
        
    spline:Reserve (children_count)
  
    track.spline = spline
  else
    track.keys = {}
  end
  
  local previous_value
  local previous_time

  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)
    
    if child_node.Name == "Key" then
      local key          = LoadAnimationTrackKey (child_node, vars)
      local value_string = child_node:Get ("Value", "")
      
      if previous_value == value_string then
        local span = {}
        
        span.from_time = previous_time
        span.to_time   = key.time
        
        local pos
        local found
        
        for _, check_span in pairs (static_spans) do
          if check_span.from_time == span.from_time and check_span.to_time == span.to_time then
            found = true
            break
          end
        end
        
        if not found then
          table.insert (static_spans, span)
        end
      end

      previous_value = value_string
      previous_time  = key.time

      if spline then
        local key_value = key.value
        
        if key.type == "scalar" and track.name == "scale" then
          key_value = vec2 (key_value)
          is_vectors = true
        end
        
        spline:AddKey (key.time, key_value)
      else
        table.insert (track.keys, key)
      end

      if key.type == "scalar" then is_scalar  = true end
      if key.type == "vec"    then is_vectors = true end
    end
  end

  if ((track.keys and (#track.keys < 2)) or (spline and spline.KeysCount < 2)) and track.name ~= "events" then
    track.static = true
  end

  if is_scalar then
    track.lerp = ScalarLerp
    track.add  = ScalarAdd
  end

  if is_vectors then
    track.lerp = VectorLerp
    track.add  = VectorAdd
  end

  UpdateAnimationTrack (track)
  
  return track
end

-- Получение обработчика событий спрайта
function GetSpriteEventHandler (code, params)
  if code == "" then return nil end
  
--  print (code)
    
  if params then
    return loadstring ("return function (sprite, event, " .. params .. ") " .. code .. " end")()
  else
    return loadstring ("return function (sprite, event) " .. code .. " end")()
  end
end

-- Получние длительности трека
local function GetTrackDuration (track)
  if track.spline then return track.spline.MaxTime end

  local max_time = -1
  
  local index, key = next (track.keys)

  while index do
    if key.time > max_time then
      max_time = key.time
    end

    index, key = next (track.keys, index)
  end
  
  return max_time
end

-- Получение длительности петли
local function GetSpriteLoopDuration (sprite)
  local duration = 0
  
  local function UpdateDuration (track)
    local track_duration = GetTrackDuration (track)        
    
    if track_duration > duration then
      duration = track_duration
    end
  end
  
  ForEachInTable (sprite.tracks, UpdateDuration)
  
  local frames_duration = sprite.frames_count / sprite.fps    
    
  if sprite.repeat_mode == "mirror" then
    frames_duration = frames_duration * 2
  end
  
  if frames_duration > duration then
    duration = frames_duration
  end
  
  return duration
end

-- Загрузка спрайта
function AttachInputZoneToNode (node, notification, handler, scale)
  if handler == "" then return end
  
  local zone = game.current_screen.input_zones [node.Id]
  
  if not zone then 
    zone = Scene.InputZone.Create () 
  
    if scale then
      zone.Scale = scale
    end
  
    zone:BindToParent (node, Scene.NodeBindMode.AddRef)
    
    game.current_screen.input_zones [node.Id] = zone
  end
  
  local item = {}
  
  item.callback   = Scene.InputZone.CreateNotificationHandler (handler)    
  item.connection = zone:RegisterNotificationHandler (notification, item.callback)
  
  table.insert (game.current_screen.input_handlers, item)
  
  return item
end

function AttachInputZone (sprite, notification, handler, scale)
  if handler == "" then return end
  
  local zone = sprite.input_zone
  
  if not zone then
    zone = Scene.InputZone.Create ()
  
    if scale then
      zone.Scale = scale
    end
  
    zone.Name = string.format ("%s.InputZone", sprite.name)
  
    zone:BindToParent (sprite.node, Scene.NodeBindMode.WeakRef)
    
    sprite.input_zone = zone    
  end
  
  local input_handler
  
  if type (handler) == "function" then
    input_handler = handler
  else
    local compiled_handler = GetSpriteEventHandler (handler, "viewport, touch, x, y")    
  
    input_handler = function (viewport, notification, context)
--    if notification == Scene.InputZoneNotification.OnClick or notification == Scene.InputZoneNotification.OnPress then  
--      print (string.format ("touch sprite '%s', notification=%d, x=%f y=%f", sprite.name, notification, context.TouchWorldPosition.x, context.TouchWorldPosition.y))
--    end
  
      compiled_handler (sprite, game.events, viewport, context.TouchId, context.TouchWorldPosition.x, context.TouchWorldPosition.y)
    end
  end
  
  local item = {}
  
  item.callback   = Scene.InputZone.CreateNotificationHandler (input_handler)    
  item.connection = zone:RegisterNotificationHandler (notification, item.callback)
  
  table.insert (sprite.input_handlers, item)
end

local function UpdateScissor (scissor, left, top, width, height)
  local pos_x = left + width / 2
  local pos_y = top - height / 2

  --TODO bug workaround
  if width < EPS then width = EPS end
  if height < EPS then height = EPS end

  scissor.WorldPosition = vec3 (pos_x, pos_y, 0)
  scissor.WorldScale    = vec3 (width, height, 1)
end

function CreateScissor (left, top, width, height)
  local scissor = Scene.Scissor.Create ()

  UpdateScissor (scissor, left, top, width, height)

  return scissor
end

function BindOrUpdateScissor (model, left, top, width, height)
  if not model.scissor then
    model.scissor = CreateScissor (left, top, width, height)
  
    model.node.Scissor = model.scissor
  
    if model.is_active then
      model.scissor:BindToScene (model.node.Scene)
    end
  else
    UpdateScissor (model.scissor, left, top, width, height)
  end
end

function LoadSprite (xml_node, default_z, sprite, animation_screen)
  local update_sprite = sprite ~= nil
  
  if not sprite then 
    local layout = xml_node:Get ("Layout", "")
    
    if layout ~= "" then
      local layout_sprite = animation_screen.sprites [layout]
      
      if not layout_sprite then
        error (string.format ("No layout sprite '%s'", layout))      
      end
      
      local layout_node = layout_sprite.xml_node
      
      if not layout_node then
        error (string.format ("No layout '%s'", layout))
      end
      
      default_z     = layout_sprite.z - 0.001      

      sprite        = LoadSprite (layout_node, default_z, nil, animation_screen)
      sprite.name   = xml_node:Get ("Name", sprite.name)
      update_sprite = true      
    else
      sprite = {}    

      local name      = xml_node:Get ("Name", "")
      local material  = xml_node:Get ("Material", "")
      local position  = xml_node:Get ("Position", "")
      local size      = xml_node:Get ("Size", "")
      local texoffset = xml_node:Get ("TexOffset", "")
      local texsize   = xml_node:Get ("TexSize", "")
      local alpha     = xml_node:Get ("Alpha", "")
      
      if name ~= "" and material ~= "" and position ~= "" and size ~= "" and texoffset ~= "" and texsize ~= "" and alpha ~= "" then
        local layout_node = Common.StringNode.Create ()

        local function AddNode (name, value)
          local child = Common.StringNode.Create ()

          child.Name = name

          child:AddAttribute (value)

          layout_node:AddChild (child)
        end

        AddNode ("Name", name)
        AddNode ("Material", material)
        AddNode ("Opacity", tostring (255 * (alpha + 0)))
        AddNode ("TexOffset", texoffset)
        AddNode ("TexSize", texsize)

        local position_vec  = StringToVec (position)
        local size_vec      = StringToVec (size)
        local texoffset_vec = StringToVec (texoffset)
        local texsize_vec   = StringToVec (texsize)

        AddNode ("Left", tostring (position_vec.x - size_vec.x / 2))
        AddNode ("Top", tostring (-position_vec.y - size_vec.y / 2))
        AddNode ("Width", tostring (size_vec.x))
        AddNode ("Height", tostring (size_vec.y))

        sprite.layout = layout_node
      end

      sprite.xml_node = xml_node
    end
  end

  local material_animation = xml_node:Get ("MaterialAnimation", "")
  local resource_group     = xml_node:Get ("ResourceGroup", "")

  local on_end_touch   = xml_node:Get ("OnEndTouch", "")
  local on_enter_touch = xml_node:Get ("OnEnterTouch", "")  
  local on_leave_touch = xml_node:Get ("OnLeaveTouch", "")
  local on_click       = xml_node:Get ("OnClick", "")
  local on_pressed     = xml_node:Get ("OnPressed", "")
  local on_drag        = xml_node:Get ("OnDrag", "")
  local on_mouse_enter = xml_node:Get ("OnMouseEnter", "")
  local on_mouse_move  = xml_node:Get ("OnMouseMove", "")
  local on_mouse_leave = xml_node:Get ("OnMouseLeave", "")    
  local on_activate    = xml_node:Get ("OnActivate", "")
  local on_deactivate  = xml_node:Get ("OnDeactivate", "")
  local on_show        = xml_node:Get ("OnShow", "")
  local on_hide        = xml_node:Get ("OnHide", "")
  local on_mouse_screen_move = xml_node:Get ("OnMouseScreenMove", "")
  local on_mouse_screen_up   = xml_node:Get ("OnMouseScreenUp", "")
  local on_mouse_screen_down = xml_node:Get ("OnMouseScreenDown", "")
  local input_zone_scale     = StringToVec (xml_node:Get ("InputZoneScale", "1;1"))

  input_zone_scale = vec3 (input_zone_scale.x, input_zone_scale.y, 1)

  local after_sprite = xml_node:Get ("AfterSprite", "") 

  if not update_sprite then
    sprite.name = xml_node:Get ("Name", "")
  end

  sprite.z                = xml_node:Get ("Z", tostring (default_z)) + 0.0
  sprite.after_sprite     = after_sprite
  sprite.scale            = {x=1, y=1}
  sprite.start_is_active  = xml_node:Get ("Active", "true") == "true"
  sprite.start_is_visible = xml_node:Get ("Visible", "true") == "true" and sprite.start_is_active
  sprite.cursor           = xml_node:Get ("Cursor", "")
  sprite.on_activate      = GetSpriteEventHandler (on_activate)
  sprite.on_deactivate    = GetSpriteEventHandler (on_deactivate)
  sprite.on_show          = GetSpriteEventHandler (on_show)
  sprite.on_hide          = GetSpriteEventHandler (on_hide)

  sprite.is_manual_position    = xml_node:Get ("ManualPosition", "false") == "true"
  sprite.is_manual_orientation = xml_node:Get ("ManualOrientation", "false") == "true"  
  sprite.is_manual_scale       = xml_node:Get ("ManualScale", "false") == "true"    
  
  sprite.invisible_alpha = xml_node:Get ("InvisibleAlpha", "0") + 0.0
  sprite.alpha_factor    = 1
  sprite.opacity         = 1
  sprite.alpha           = 1  
  
  sprite.scale_factor = {x=1, y=1}

  sprite.angle            = 0
  sprite.intensity        = 0
  sprite.way              = xml_node:Get ("Way", "")   
  
  if not update_sprite then
    sprite.node       = Scene.Sprite.Create ()
    sprite.node.Color = vec4 (1, 1, 1, 1)    
  end
  
  sprite.node.Name               = sprite.name
  sprite.node.ScaleInherit       = xml_node:Get ("ScaleInherit", "true") == "true"
  sprite.node.OrientationInherit = xml_node:Get ("OrientationInherit", "true") == "true"
  sprite.bind_space              = xml_node:Get ("BindSpace", "world")
  sprite.is_world_position       = xml_node:Get ("PositionSpace", "world") == "world"
  sprite.is_world_orientation    = xml_node:Get ("OrientationSpace", "local") == "world"
  sprite.is_world_scale          = xml_node:Get ("ScaleSpace", "local") == "world"
  
  local orientation_pivot_enabled = xml_node:Get ("OrientationPivotEnabled", "true") == "true"
  local scale_pivot_enabled       = xml_node:Get ("ScalePivotEnabled", "false") == "true"
  
  sprite.node.OrientationPivotEnabled = orientation_pivot_enabled
  sprite.node.ScalePivotEnabled       = scale_pivot_enabled

  local pivot = StringToVec (xml_node:Get ("PivotPosition", "0; 0"))
  
  sprite.node.PivotPosition = vec3 (pivot.x, pivot.y, 0)
  
  sprite.input_handlers = {}
  
  local function AttachInputZoneToSprite (notification, handler)
    AttachInputZone (sprite, notification, handler, input_zone_scale)
  end

  AttachInputZoneToSprite (Scene.InputZoneNotification.OnPress,           on_pressed)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnClick,           on_click)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnPressMove,       on_drag)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnTouchEnter,      on_enter_touch)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnTouchLeave,      on_leave_touch)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnTouchUpInside,   on_end_touch)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnScreenTouchMove, on_mouse_screen_move)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnScreenTouchDown, on_mouse_screen_down)
  AttachInputZoneToSprite (Scene.InputZoneNotification.OnScreenTouchUp,   on_mouse_screen_up)  
  
  sprite.frame_step   = xml_node:Get ("FrameStep", "1") + 0
  sprite.first_frame  = xml_node:Get ("FirstFrame", "0") + 0
  sprite.last_frame   = xml_node:Get ("LastFrame", "0") + 0
  sprite.fps          = (xml_node:Get ("Fps", "30") + 0) / sprite.frame_step
  sprite.repeat_mode  = xml_node:Get ("RepeatMode", "clamp_to_end")
  sprite.static       = true
  sprite.parent       = xml_node:Get ("Parent", "")
  
  if sprite.last_frame >= sprite.first_frame then
    sprite.frame_order_sign = 1
  else
    sprite.frame_order_sign = -1
  end
  
  local material_name
  
  if sprite.layout and string.find (xml_node:Get ("Material", ""), "/", 1, true) then
    material_name = sprite.layout:Get ("Material", "")
  else
    material_name = xml_node:Get ("Material", "")
  end

  if material_animation ~= "" then
    sprite.materials    = GetAnimationNames (material_animation, sprite.frame_step)
    
    local frames_count = 0
    local index, value = next (sprite.materials)

    while index do
      frames_count = frames_count + 1
--      print (">>> " .. tostring (index) .. " " .. tostring (value))      
      index, value = next (sprite.materials, index)
    end    
    
    sprite.frames_count = frames_count
  elseif material_name ~= "" then
    sprite.node.Material = material_name
    
    if material_name == "" then
      error (string.format ("No material in sprite '%s'", sprite.name))
    end
    
    if material_name == "transparent" then
      sprite.node.Visible = false
    end
    
    sprite.frames_count  = math.floor ((math.abs (sprite.last_frame - sprite.first_frame) + 1) / sprite.frame_step)
  elseif not update_sprite then
    error (string.format ("No material in sprite '%s'", sprite.name))
  end  
  
  if resource_group ~= "" then
    sprite.resource_binding = Engine.ResourceManager.CreateFileGroupBinding ("data/res/" .. resource_group .. ".xreslist")
  end

  local default_offset = xml_node:Get ("Offset", "0") + 0.0

  if not update_sprite then sprite.tracks = {} end

  local children_count = xml_node.ChildrenCount    

  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)

    if child_node.Name == "Track" then
      local static_spans = {}
    
      local track = LoadAnimationTrack (child_node, default_offset, animation_screen.vars, static_spans)
      
      track.static_spans = static_spans

      sprite.tracks [track.name] = track
      
      if (not track.static or track.link ~= "") then
        sprite.static = false
      end
    end    
  end

  if not update_sprite and sprite.layout then
    local function AddStaticVectorTrack (track_name, value)
      local track = {}
      
      track.name   = track_name
      track.offset = 0
      track.loop   = false
      track.link   = ""
      track.keys   = {}   
      
      local key = {}
      
      key.time  = 0
      key.value = {x=value.x, y=value.y}
      key.type  = "vec"
      
      track.keys [0]             = key
      sprite.tracks [track_name] = track
    end

    local left      = sprite.layout:Get ("Left") + 0
    local top       = sprite.layout:Get ("Top") + 0
    local width     = sprite.layout:Get ("Width") + 0
    local height    = sprite.layout:Get ("Height") + 0
    local opacity   = sprite.layout:Get ("Opacity") + 0
    local sprite_x  = left + width / 2
    local sprite_y  = -top - height / 2
    local texoffset = StringToVec (sprite.layout:Get ("TexOffset"))
    local texsize   = StringToVec (sprite.layout:Get ("TexSize"))

    AddStaticVectorTrack ("position", vec2 (sprite_x, sprite_y))
    AddStaticVectorTrack ("size", vec2 (width, height))
    AddStaticVectorTrack ("texoffset", texoffset)
    AddStaticVectorTrack ("texsize", texsize)

    sprite.opacity = opacity

    if opacity < 255 then
      local alpha_track = {}
      
      alpha_track.name   = "alpha"
      alpha_track.offset = 0
      alpha_track.loop   = false
      alpha_track.link   = ""
      alpha_track.keys   = {}
      
      local alpha_track_key = {}
    
      alpha_track_key.time  = 0
      alpha_track_key.value = opacity / 255.0
      alpha_track_key.type  = "scalar"

      alpha_track.keys [0] = alpha_track_key
    
      sprite.tracks ["alpha"] = alpha_track
    end
  else
    if sprite.layout then
      local texoffset = xml_node:Get ("TexOffset", "")
    
      if texoffset ~= "" then
        sprite.layout:Set ("TexOffset", texoffset)
  
        sprite.tracks ["texoffset"].keys [0].value = StringToVec (texoffset)
      end
    
      local texsize = xml_node:Get ("TexSize", "")
    
      if texsize ~= "" then
        sprite.layout:Set ("TexSize", texsize)
  
        sprite.tracks ["texsize"].keys [0].value = StringToVec (texsize)
      end
      
      local material = xml_node:Get ("Material", "")
      
      if material ~= "" then
        sprite.layout:Set ("Material", material)
        sprite.node.Material = material
      end
    
      local position = xml_node:Get ("Position", "")
      local size     = xml_node:Get ("Size", "")
    
      if position ~= "" and size ~= "" then
        local position_vec  = StringToVec (position)
        local size_vec      = StringToVec (size)
        
        local left      = position_vec.x - size_vec.x / 2
        local top       = -position_vec.y - size_vec.y / 2
        local width     = size_vec.x
        local height    = size_vec.y
        local sprite_x  = left + width / 2
        local sprite_y  = -top - height / 2
        

        sprite.layout:Set ("Left", tostring (left))
        sprite.layout:Set ("Top", tostring (top))
        sprite.layout:Set ("Width", tostring (width))
        sprite.layout:Set ("Height", tostring (height))
        
        sprite.tracks ["position"].keys [0].value = vec2 (sprite_x, sprite_y)
        sprite.tracks ["size"].keys [0].value = vec2 (width, height)
      end
    end
  end
  
  if sprite.frames_count > 1 then
    sprite.static = false
  end
  
  -- Добавление встроенного обработчика событий конца анимации
    
  local track    = sprite.tracks ["events"]  
  local duration = GetSpriteLoopDuration (sprite)

  if not track then
    track                    = {}
    sprite.tracks ["events"] = track

    track.name   = "events"
    track.offset = 0
    track.keys   = {}
  end
    
  if sprite.repeat_mode == "loop" or sprite.repeat_mode == "mirror" then
    track.loop = true
  end

  local key = {}

  key.time  = duration
  key.value = loadstring ("return function (time, sprite) " .. string.format ("OnAnimationEnd('%s')", sprite.name) .. " end")()

  table.insert (track.keys, key)

  UpdateAnimationTrack (track)
  
  -- Связывание с таймером
  
  local timer_name = xml_node:Get ("Timer", "default")
  
  sprite.timer = GetTimer (animation_screen, timer_name)

  return sprite
end

-- Загрузка групп спрайтов
local function LoadSpriteGroup (node, animation_screen)
  local group = {}
  
  group.name           = node:Get ("Name", "")
  group.fps            = node:Get ("Fps", "25") + 0
  group.repeat_mode    = node:Get ("RepeatMode", "clamp_to_end")
  group.blend_frames   = node:Get ("BlendFrames", "false") == "true"
  group.animation_mode = node:Get ("AnimationMode", "pairs")
  group.sprites        = {}
  group.start_visible  = node:Get ("Visible", "false")=="true"
  group.need_visible   = group.start_visible
  
  group.visible=group.start_visible  
  
  local timer_name = node:Get ("Timer", "default")
  
  group.timer = GetTimer (animation_screen, timer_name)

  local sprite_names = GetAnimationNames (node:Get ("Sprites"), 1)

  local index, value = next (sprite_names)
  local sprite_index = 0

  while index do
    local sprite = animation_screen.sprites [value]        

    if sprite then      
      LoadSprite (node, sprite.z, sprite, animation_screen)      
      
      sprite.start_is_visible = false
      sprite.start_is_active  = false
    
--      sprite.tracks ["alpha"] = nil
      
      local children_count = node.ChildrenCount
      local default_offset = node:Get ("Offset", "0") + 0.0      

      for i=0, children_count-1 do
        local child_node = node:Child (i)

        if child_node.Name == "Track" then
          local static_spans = {}
        
          local track = LoadAnimationTrack (child_node, default_offset, animation_screen.vars, static_spans)
          
          sprite.static_spans = static_spans

          sprite.tracks [track.name] = track
          
          if (not track.static or track.link ~= "") then
            sprite.static = false
          end
        end
      end            

      group.sprites [sprite_index] = sprite
      sprite_index                 = sprite_index + 1
      
      table.insert (animation_screen.sprite_list, sprite)      
    end

    index, value = next (sprite_names, index)
  end    

  group.frames_count = sprite_index
  
  if group.frames_count > 0 then
    local duration = group.frames_count / group.fps
  
    local sprite = group.sprites [0]
    local track  = sprite.tracks ["events"]    
    
    local key = {}

    key.time  = duration
    key.value = function (time)
      OnAnimationEnd (group.name)
    end
    
    if group.repeat_mode == "loop" or group.repeat_mode == "mirror" then
      track.loop = true
    end        

    table.insert (track.keys, key)

    UpdateAnimationTrack (track)
    
    sprite.static = false
  end
  
  return group
end

-- Получение обработчика событий спрайта
local function GetSpriteEventHandler (code, params)
  if code == "" then return nil end
  
  if params then
    return loadstring ("return function (sprite, event, ".. params .. ") " .. code .. " end")()
  else
    return loadstring ("return function (sprite, event) " .. code .. " end")()
  end

  sprite.tracks = {}

  sprite.tracks ["position"] = position_track
  sprite.tracks ["size"]     = size_track   

  return sprite
end

-- Компиляция функции обработчика событий состояния
local function GetStateHandler (code)
  if code == "" then return nil end

  return loadstring ("return function (time, state, event) " .. code .. " end")()
end

-- Загрузка перехода конечного автомата
local function LoadFsmJunction (xml_node)
  local junction  = {}
  local condition = xml_node:Get ("Condition", "true")
  local action    = xml_node:Get ("Action", "")  
  
  junction.condition_fn = loadstring ("return function (time, event, state, junction) return " .. condition .. " end")()
  junction.state        = xml_node:Get ("State")
  junction.action       = GetStateHandler (action)

  return junction
end

-- Загрузка состояния конечного автомата
local function LoadFsmState (xml_node, sprites, fsm)
  local state = {}
  
  state.fsm       = fsm
  state.name      = xml_node:Get ("Name")
  state.junctions = {}
  
  local on_enter  = xml_node:Get ("OnEnter", "")
  local on_exit   = xml_node:Get ("OnExit", "")
  local on_update = xml_node:Get ("OnUpdate", "")
  
  state.on_enter  = GetStateHandler (on_enter)
  state.on_exit   = GetStateHandler (on_exit)
  state.on_update = GetStateHandler (on_update)

  local animation_sprite_name = xml_node:Get ("AnimationSprite", "")

  if animation_sprite_name ~= "" then
    state.animation_sprite = animation_sprite_name
  end
  
  local animation_name = xml_node:Get ("Animation", "")  
  
  if animation_name ~= "" then
    state.animation = animation_name
  end

  local children_count = xml_node.ChildrenCount

  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)

    if child_node.Name == "Junction" then
      local junction = LoadFsmJunction (child_node)

      table.insert (state.junctions, junction)
    end
  end
  
  -- Добавление перехода на конец проигрывания анимации

  local function NextStateJunction (time, event, state, junction)
    return state.events ["on_end_animation"]
  end
  
  local next_state = xml_node:Get ("NextState", "")
  
  if next_state ~= "" then
    local next_state_junction = {}
    
    next_state_junction.condition_fn = NextStateJunction
    next_state_junction.state        = next_state
    next_state_junction.action       = nil

    table.insert (state.junctions, next_state_junction)
  end

  return state
end

-- Загрузка конечного автомата
local function LoadFsm (xml_node, sprites, animation_screen)
  local fsm = {}
  
  fsm.name        = xml_node:Get ("Name", "")
  fsm.states      = {}
  fsm.start_state = xml_node:Get ("StartState")

  local timer_name = xml_node:Get ("Timer", "default")

  fsm.timer = GetTimer (animation_screen, timer_name)

  local children_count = xml_node.ChildrenCount

  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)

    if child_node.Name == "State" then
      local state = LoadFsmState (child_node, sprites, fsm)

      fsm.states [state.name] = state
    end
  end

  return fsm
end

local function PrecalcWayPoint (way, point1, point2)
  local dx     = point2.position.x - point1.position.x
  local dy     = point2.position.y - point1.position.y
  local length = math.sqrt (dx * dx + dy * dy)
 
  point2.time = point1.time + length / point1.speed
end

-- Загрузка пути
local function LoadWay (xml_node, ways)
  local way = {}
  
  local way_speed = xml_node:Get ("Speed")

  way.name             = xml_node:Get ("Name")
  way.speed            = way_speed
  way.loop             = xml_node:Get ("Loop", "false")=="true"
  way.look_time_offset = xml_node:Get ("LookTimeOffset", "0.01")
  way.const_angle      = xml_node:Get ("ConstAngle", "")
  way.points = {}

  if way.const_angle == "" then way.const_angle = nil end

  local children_count = xml_node.ChildrenCount
  local current_point  = 0
  local way_time       = 0

  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)

    if child_node.Name == "Point" then
      local way_point = {}
      
      way_point.position = StringToVec (child_node:Get ("Position"))
      way_point.scale    = StringToVec (child_node:Get ("Scale", "1;1"))
      way_point.speed    = child_node:Get ("Speed", way_speed) + 0
      
      if way_point.speed == "0" then error ("Way '" .. way.name .. "' has 0 speed") end
      
      way.points [current_point] = way_point

      current_point = current_point + 1
    end
  end
  
  if current_point < 2 then return nil end
  
  way.points [0].time = 0

  for i=0, current_point-2, 1 do
    PrecalcWayPoint (way, way.points [i], way.points [i + 1])
  end

  if (way.loop) then
    PrecalcWayPoint (way, way.points [current_point - 1], way.points [0])
    way.time = way.points [0].time
    way.points [0].time = 0
  else
    way.time = way.points [current_point - 1].time
  end

  return way
end

-- Загрузка переменных
local function LoadVar (xml_node, vars)
  local name              = xml_node:Get ("Name")
  local value_string      = xml_node:Get ("Value", "")
  local expression_string = xml_node:Get ("Expression", "")
  
  if value_string == "" and expression_string == "" then
    error (string.format ("Nor value neither expression found in var '%s'", name))
  end
  
  if expression_string ~= "" then
    vars [name] = loadstring ("return function (var) return " .. expression_string .. " end")()
  end

  if value_string ~= "" then  
    --print (string.format ("value is '%s'", value_string))
  
    if string.find (value_string, ";") == nil then
      -- Конвертация скалярного значения
      
      vars [name] = value_string + 0.0
    else
      -- Конвертация вектора
      
      vars [name] = StringToVec (value_string)
    end  
  end
end

-- Загрузка групп ресурсов
local function LoadResourceGroup (name)
  return { name = name, group = Engine.ResourceManager.Group.Create ("data/res/" .. name .. ".xreslist") }
end

-- Загрузка скрипта
local function LoadScript (node)
  local script = {}
  local file_name = node:Get ("Source")
  
  script.name    = node:Get ("Name")
  
  local script_path         = SCRIPTS_DIR .. '/' .. file_name .. SCRIPT_SUFFIX
  local mounted_script_path = "/lobby/" .. script_path
  
  if Common.File.IsFileExist (mounted_script_path) then
    script_path = mounted_script_path
  end
  
  script.creator = dofile (script_path)
  script.data    = node

  return script
end

local function CreateVecTrackFromString (track_name, str)
  local track = {}
    
  track.name   = track_name
  track.offset = 0
  track.loop   = false
  track.link   = ""
  track.keys   = {}   
    
  local track_key = {}
    
  track_key.time  = 0
  track_key.value = StringToVec (str)
  track_key.type  = "vec"
    
  track.keys [0] = track_key
  
  track.static = true
  
  return track
end

local function CreateScalarTrackFromString (track_name, str)
  local track = {}
    
  track.name   = track_name
  track.offset = 0
  track.loop   = false
  track.link   = ""
  track.keys   = {}   
    
  local track_key = {}

  track_key.time  = 0
  track_key.value = str + 0
  track_key.type  = "scalar"
    
  track.keys [0] = track_key
  
  return track
end

-- Загрузка текстовой линии
local function LoadTextLine (node, default_z, vars, animation_screen)
  local text_line = {}

  text_line.name             = node:Get ("Name", "")
  text_line.node             = Scene.TextLine.Create (game.font_library)
  text_line.start_is_active  = node:Get ("Active", "true") == "true"
  text_line.start_is_visible = node:Get ("Visible", "true") == "true" and text_line.start_is_active
  text_line.z                = tonumber (node:Get ("Z", tostring (default_z)))
  text_line.tracks           = {}
  text_line.text             = node:Get ("Text", "")
  text_line.node.Font        = node:Get ("Font", "")
  text_line.parent           = node:Get ("Parent", "")
  
  text_line.node.FontCreationParams = game.font_creation_params
  text_line.node.Material = "text_material"
  
  text_line.chars_color_factors = {}
  
  local chars_color_factor = node:Get ("CharsColorFactor", "")
  
  if chars_color_factor ~= "" then
    local desc = SplitString (chars_color_factor, ";", true)
    
    if #desc ~= 6 then
      error (string.format ("Invalid 'CharsColorFormat' attribute format at text line '%s', must be 6 numbers separated by ';'", text_line.name))
    end
    
    table.insert (text_line.chars_color_factors, { from = desc [1] + 0, count = desc [2] + 0, factor = vec4 (desc [3] + 0, desc [4] + 0, desc [5] + 0, desc [6] + 0) })
  end
  
  local timer_name = node:Get ("Timer", "default")
  
  text_line.timer = GetTimer (animation_screen, timer_name)
  
  local horizontal_alignment = node:Get ("HorizontalAlignment", "Center")
  local vertical_alignment   = node:Get ("VerticalAlignment", "BaseLine")
  
  if     horizontal_alignment == "Center" then text_line.node.HorizontalAlignment = Scene.TextLineAlignment.Center
  elseif horizontal_alignment == "Left"   then text_line.node.HorizontalAlignment = Scene.TextLineAlignment.Left
  elseif horizontal_alignment == "Right"  then text_line.node.HorizontalAlignment = Scene.TextLineAlignment.Right end
  
  if     vertical_alignment == "Center"   then text_line.node.VerticalAlignment = Scene.TextLineAlignment.Center
  elseif vertical_alignment == "Top"      then text_line.node.VerticalAlignment = Scene.TextLineAlignment.Top
  elseif vertical_alignment == "Bottom"   then text_line.node.VerticalAlignment = Scene.TextLineAlignment.Bottom
  elseif vertical_alignment == "BaseLine" then text_line.node.VerticalAlignment = Scene.TextLineAlignment.BaseLine end
  
  text_line.after_sprite = node:Get ("AfterSprite", "") 

  text_line.static = true
  
  local layout = node:Get ("Layout", "")
  
  if layout ~= "" then
    local layout_sprite = animation_screen.sprites [layout]
    
    if not layout_sprite then
      error (string.format ("No layout sprite '%s'", layout))      
    end
    
    local layout_node = layout_sprite.layout
    
    if not layout_node then
      error (string.format ("No layout '%s'", layout))
    end
    
    local left     = layout_node:Get ("Left") + 0
    local top      = layout_node:Get ("Top") + 0
    local width    = layout_node:Get ("Width") + 0
    local height   = layout_node:Get ("Height") + 0
    local opacity  = layout_node:Get ("Opacity") + 0
    local sprite_x = left + width / 2
    local sprite_y = -(top + height / 2)        
    default_z      = layout_sprite.z - 0.001      

    local position_track = {}
    
    position_track.name   = "position"
    position_track.offset = 0
    position_track.loop   = false
    position_track.link   = ""
    position_track.keys   = {}   
    
    local position_track_key = {}
    
    position_track_key.time  = 0
    position_track_key.value = {x=sprite_x, y=sprite_y}
    position_track_key.type  = "vec"
    
    position_track.keys [0] = position_track_key
        
    text_line.tracks ["position"] = position_track
  end

  local size = node:Get ("Size", "");

  if size ~= "" then
    text_line.tracks ["size"] = CreateVecTrackFromString ("size", size)
  end
  
  local position = node:Get ("Position", "");

  if position ~= "" then
    text_line.tracks ["position"] = CreateVecTrackFromString ("position", position)
  end
  
  local color = node:Get ("Color", "");

  if color ~= "" then
    text_line.tracks ["color"] = CreateVecTrackFromString ("color", color)
  end
  
  local alpha = node:Get ("Alpha", "");

  if alpha ~= "" then
    text_line.tracks ["alpha"] = CreateScalarTrackFromString ("alpha", alpha)
  end
  
  local children_count = node.ChildrenCount

  for i=0, children_count-1 do
    local child_node = node:Child (i)

    if child_node.Name == "Track" then
      local static_spans = {}
    
      local track = LoadAnimationTrack (child_node, 0, vars, static_spans)
      
      track.static_spans = static_spans

      text_line.tracks [track.name] = track
      
      if (not track.static or track.link ~= "") then
        text_line.static = false
      end      
    elseif child_node.Name == "CharsColorFactor" and child_node.ChildrenCount ~= 0 then
      table.insert (text_line.chars_color_factors, { from = child_node:Get ("From") + 0, count = child_node:Get ("Count") + 0, factor = StringToVec (child_node:Get ("Factor")) })
    end    
  end

  local stroke_color = node:Get ("StrokeColor", "");
  local shadow_color = node:Get ("ShadowColor", "");

  if stroke_color ~= "" and shadow_color ~= "" then
    error ("Text line can't have both stroking and shadow")
  end

  if stroke_color ~= "" then
    local stroked_text_line = Scene.TextLine.Create (game.font_library)
    
    local font = text_line.node.Font
    
    local suffix_start = 0
    
    while true do
      local next_start = string.find (font, ".", suffix_start + 1, true)
     
      if next_start then
        suffix_start = next_start
      else
        break
      end
    end
    
    if suffix_start == 0 then
      font = font .. "-stroked"
    else
      font = string.sub (font, 1, suffix_start - 1) .. "-stroked" .. string.sub (font, suffix_start) 
    end
    
    local horizontal_stroke_offset = 0
    local vertical_stroke_offset = 0
    
    if text_line.node.VerticalAlignment == Scene.TextLineAlignment.BaseLine then
      vertical_stroke_offset = 0.04
    elseif text_line.node.VerticalAlignment == Scene.TextLineAlignment.Bottom then
      vertical_stroke_offset = -0.04
    elseif text_line.node.VerticalAlignment == Scene.TextLineAlignment.Top then
      vertical_stroke_offset = 0.04
    end
    
    if text_line.node.HorizontalAlignment == Scene.TextLineAlignment.Left then
      horizontal_stroke_offset = -0.05
    elseif text_line.node.HorizontalAlignment == Scene.TextLineAlignment.Right then
      horizontal_stroke_offset = 0.04
    end
    
    stroked_text_line.FontCreationParams  = game.font_creation_params  --TODO stroked params
    stroked_text_line.Font                = font
    stroked_text_line.HorizontalAlignment = text_line.node.HorizontalAlignment
    stroked_text_line.VerticalAlignment   = text_line.node.VerticalAlignment
    stroked_text_line.Position            = vec3 (horizontal_stroke_offset, vertical_stroke_offset, EPS)
    stroked_text_line.Material            = "text_material"
    
    stroke_color = StringToVec (stroke_color)
    
    local stroke_alpha = node:Get ("StrokeAlpha", DEFAULT_STROKE_ALPHA) + 0
    
    stroked_text_line.Color = vec4 (stroke_color.x, stroke_color.y, stroke_color.z, stroke_alpha)

    local function UpdateStrokedText ()
      stroked_text_line.Text = text_line.node.Text
    end

    text_line.stroked_text_line_update_connection = text_line.node:RegisterEventHandler (Scene.NodeEvent.AfterUpdate, Scene.Node.CreateEventHandler (UpdateStrokedText))

    stroked_text_line:BindToParent (text_line.node, Scene.NodeBindMode.WeakRef)
    
    text_line.stroked_text_line = stroked_text_line
  end

  if shadow_color ~= "" then
    local shadow_offset = StringToVec (node:Get ("ShadowOffset"))
  
    local shadow_text_line = Scene.TextLine.Create (game.font_library)
    
    shadow_text_line.FontCreationParams = game.font_creation_params
    shadow_text_line.Material = "text_material"
    
    shadow_text_line.Font                = text_line.node.Font
    shadow_text_line.HorizontalAlignment = text_line.node.HorizontalAlignment
    shadow_text_line.VerticalAlignment   = text_line.node.VerticalAlignment
    shadow_text_line.Position            = vec3 (shadow_offset.x, shadow_offset.y, EPS)
    
    shadow_color = StringToVec (shadow_color)
    
    text_line.shadow_alpha = node:Get ("ShadowAlpha", "1") + 0
    
    shadow_text_line.Color = vec4 (shadow_color.x, shadow_color.y, shadow_color.z, text_line.shadow_alpha)

    local function UpdateShadowText ()
      shadow_text_line.Text = text_line.node.Text
    end

    text_line.shadow_text_line_update_connection = text_line.node:RegisterEventHandler (Scene.NodeEvent.AfterUpdate, Scene.Node.CreateEventHandler (UpdateShadowText))

    shadow_text_line:BindToParent (text_line.node, Scene.NodeBindMode.WeakRef)
    
    text_line.shadow_text_line = shadow_text_line
  end

  return text_line
end

-- Загрузка камеры
local function LoadCamera (node, parent, vars, animation_screen)
  local camera = {}
  
  camera.name            = node:Get ("Name", "")
  camera.type            = node:Get ("Type")  
  camera.start_is_active = node:Get ("Active", "false") == "true"
  camera.params          = {}
  
  local timer_name = node:Get ("Timer", "default")
  
  camera.timer = GetTimer (animation_screen, timer_name)

  if camera.type == "ortho" then
    camera.node = Scene.OrthoCamera.Create ()
  elseif camera.type == "perspective" then
    camera.node = Scene.PerspectiveCamera.Create ()
  else
    error (string.format ("unknown camera type '%s'", camera.type))
  end
  
  camera.node.Name = camera.name
  
  camera.tracks = {}
  camera.viewports = {}
  camera.center_viewports = {}
  
  local children_count = node.ChildrenCount
  
  for i=0, children_count-1 do
    local child_node = node:Child (i)

    if child_node.Name == "Track" then
      local static_spans = {}
    
      local track = LoadAnimationTrack (child_node, 0, vars, static_spans)
      
      track.static_spans = static_spans

      camera.tracks [track.name] = track      
    elseif child_node.Name == "Viewport" then
      local viewport = Scene.Viewport.Create ()
      
      viewport.Name       = child_node:Get ("Name", "")
      viewport.Technique  = child_node:Get ("RenderPath")
      viewport.ZOrder     = child_node:Get ("ZOrder") + 0
      viewport.MinDepth   = child_node:Get ("MinDepth", "0") + 0
      viewport.MaxDepth   = child_node:Get ("MaxDepth", "1") + 0      
      
      local area_string = child_node:Get ("Area")
      
      local area = SplitString (area_string, " ", true)
      
      if area_string == "centered" and camera.type == "ortho" then
        table.insert (camera.center_viewports, viewport)
      else
        local area = SplitString (area_string, " ", true)
      
        if #area ~= 4 then
          error (string.format ("Invalid camera '%s' viewport '%s' area format", camera.name, viewport.Name))
        end
      
        viewport:SetArea (area [1] + 0, area [2] + 0, area [3] + 0, area [4] + 0)
      end
      
      viewport:SetArea (area [1] + 0, area [2] + 0, area [3] + 0, area [4] + 0)

      viewport.Camera = camera.node
      
      table.insert (camera.viewports, viewport)
    end    
  end
  
  if #camera.viewports == 0 then
    error (string.format ("camera '%s' has no viewports", camera.name))
  end
  
  camera.node:BindToParent (parent)

  return camera
end

local function CreateScreenEventHandler(handler_string)
  if handler_string=="" then 
    return nil
  end
  
  return loadstring("return function(screen) "..handler_string.." end")()
end

-- Загрузка таймера
local function LoadTimer (xml_node, animation_screen)
  local parent_timer = GetTimer (animation_screen, xml_node:Get ("Parent", "screen"))  
  local started      = xml_node:Get ("Started", "true") == "true"
  local timer        = CreateTimer (true, started, parent_timer)  
  
  return timer
end

local function LoadAnimationScreenPart (animation_screen, xml_node, scene)
  local children_count = xml_node.ChildrenCount

  local LoadChildByName={}

  LoadChildByName["Include"]=function (child_node)
    local xml_file_name  = child_node:Get ("Source")
    local include_scene  = child_node:Get ("Scene", "")
    local first_z        = child_node:Get ("FirstZ", "")    
    local last_z         = child_node:Get ("LastZ", "")
    local xml            = Common.StringNode.LoadXml (xml_file_name)
    local children_count = xml.ChildrenCount
    local current_z      = animation_screen.current_z
    local delta_z        = animation_screen.delta_z

    if xml.Name ~= "AnimationScreenPart" then
      error (string.format ("Bad include '%s'. No AnimationScreenPart found", xml_file_name))
    end

    if include_scene == "" then
      include_scene = nil
    else
      if not animation_screen.scenes [include_scene] then
        animation_screen.scenes [include_scene] = Scene.Scene.Create ()

        local scene_light = Scene.SpotLight.Create ()
  
        scene_light:BindToScene (animation_screen.scenes [include_scene], Scene.NodeBindMode.AddRef)    
      end
      
      include_scene = animation_screen.scenes [include_scene]
    end

    if first_z == "" or last_z == "" then
      first_z = nil
      last_z  = nil
    else
      first_z   = first_z + 0
      last_z    = last_z + 0
      current_z = first_z
      
      if children_count > 0 then
        delta_z = (first_z - last_z) / children_count
      end

      animation_screen.current_z = first_z
      animation_screen.delta_z   = delta_z
    end

    LoadAnimationScreenPart (animation_screen, xml, include_scene)

    if first_z then
      animation_screen.current_z = current_z
      animation_screen.delta_z   = delta_z
    end
  end
  
  LoadChildByName["Animation"]=function (child_node)
    local file_name      = child_node:Get ("Source")
    local resource_group = Engine.ResourceManager.Group.Create ()
    
    resource_group:Add (file_name)
        
    table.insert (animation_screen.resource_groups, { name = file_name, group = resource_group })            
  end  
  
  LoadChildByName["Way"]=function (child_node)
    local way = LoadWay (child_node, animation_screen.ways)
    
    if way then 
      animation_screen.ways [way.name] = way
    end
  end  

  LoadChildByName["Sprite"]=function (child_node)
    local name = child_node:Get ("Name", "")       
    local sprite
    
    if name ~= "" then
      sprite = animation_screen.sprites [name]
    end
    
    local update_sprite = sprite ~= nil
    
    if update_sprite then
      LoadSprite (child_node, sprite.z, sprite, animation_screen)
      
      local found = nil
      
      for index, s in pairs (animation_screen.sprite_list) do
        if s == sprite then found = index end      
      end

      if found then
        table.remove (animation_screen.sprite_list, found)
      end
      
      table.insert (animation_screen.sprite_list, sprite)      
    else     
      sprite = LoadSprite (child_node, animation_screen.current_z, sprite, animation_screen)
      
      if sprite.name ~= "" then
        animation_screen.sprites [sprite.name] = sprite        
      else
        table.insert (animation_screen.sprites, sprite)
      end
      
      table.insert (animation_screen.sprite_list, sprite)

      animation_screen.current_z = animation_screen.current_z - animation_screen.delta_z 
    end
    
    if scene then
      sprite.scene = scene
    end
  end
  
  LoadChildByName["SpriteGroup"]=function (child_node)
    local sprite_group = LoadSpriteGroup (child_node, animation_screen)
    if sprite_group.name ~= "" then
      animation_screen.sprite_groups [sprite_group.name] = sprite_group
    else
      table.insert (animation_screen.sprite_groups, sprite_group)
    end
  end
  
  LoadChildByName["TextLine"]=function (child_node)
    local text_line = LoadTextLine (child_node, animation_screen.current_z, animation_screen.vars, animation_screen)

    if text_line.name ~= "" then
      animation_screen.text_lines [text_line.name] = text_line
    else
      table.insert (animation_screen.text_lines, text_line)
    end
    
    animation_screen.current_z = animation_screen.current_z - animation_screen.delta_z        
    
    if scene then
      text_line.scene = scene
    end
  end

  LoadChildByName["FSM"]=function (child_node)
    local fsm = LoadFsm (child_node, animation_screen.sprites, animation_screen)
    if fsm.name ~= "" then
      animation_screen.fsms [fsm.name] = fsm
    else
      table.insert (animation_screen.fsms, fsm)
    end
  end
  
  LoadChildByName["Var"]=function (child_node)
    LoadVar (child_node, animation_screen.vars)
  end
  
  LoadChildByName["GameScript"] = function (child_node)
    local script = LoadScript (child_node)
    
    if script.name ~= "" then
      animation_screen.scripts [script.name] = script
    else
      table.insert (animation_screen.scripts, script)
    end

    table.insert (animation_screen.start_scripts, script)        
  end
  
  LoadChildByName["Camera"] = function (child_node)
    local camera = LoadCamera (child_node, animation_screen.node, animation_screen.vars, animation_screen)
    
    if camera.name ~= "" then
      animation_screen.cameras [camera.name] = camera
    else
      table.insert (animation_screen.cameras, camera)
    end    
    
    if scene then
      camera.node:BindToScene (scene)
    end
  end
  
  LoadChildByName["Timer"] = function (child_node)
    local timer = LoadTimer (child_node, animation_screen)

    local timer_name = child_node:Get ("Name")

    animation_screen.timers [timer_name] = timer
  end

  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)
    local func=LoadChildByName[child_node.Name]
    if func then 
      func(child_node)
    end
  end
end

-- Загрузка одного экрана анимации
local function LoadAnimationScreenResourceGroups (xml_node)
  local resource_groups_names = SplitString (xml_node:Get ("ResourceGroups", ""), " ", true)
  local return_value          = {}
  
  for i = 1, #resource_groups_names do
    table.insert (return_value, LoadResourceGroup (resource_groups_names [i]))
  end

  return return_value
end

local function LoadAnimationScreen (xml_node, default_resource_group, main_timer)
  local animation_screen = {}
  
  animation_screen.name                = xml_node:Get ("Name")  
  animation_screen.shared_resources    = xml_node:Get ("SharedResources", "true") == "true"
  animation_screen.wait_resources      = xml_node:Get ("WaitResources", "false") == "true"
  animation_screen.init_cursor_visible = xml_node:Get ("CursorVisible", "true") == "true"
  animation_screen.fsms              = {}
  animation_screen.vars              = {}
  animation_screen.ways              = {}
  animation_screen.scenes            = {}
  animation_screen.cameras           = {}
  animation_screen.sprites           = {}
  animation_screen.sprite_list       = {}
  animation_screen.text_lines        = {}
  animation_screen.sprite_groups     = {}
  animation_screen.particles         = {}    
  animation_screen.start_scripts     = {}      
  animation_screen.scripts           = {}
  animation_screen.resource_groups   = {}
  animation_screen.sound_emitters    = {}
  animation_screen.waters            = {}
  animation_screen.timers            = {}
  animation_screen.node              = Scene.Node.Create ()

  local localization = xml_node:Get ("Localization", "")

  if localization ~= "" then
    animation_screen.localization = localization
  end

--[[  animation_screen.global_input_zone = Scene.InputZone.Create ()

  animation_screen.global_input_zone.Scale = vec3 (-100000, -100000, -100000)
  animation_screen.global_input_zone:BindToParent (animation_screen.node, Scene.NodeBindMode.AddRef)
    
  local function GlobalInputHandler (viewport, notification, context)

  end
    
  animation_screen.global_input_zone_connection = animation_screen.global_input_zone:RegisterNotificationHandler (Scene.InputZoneNotification.OnScreenTouchMove, Scene.InputZone.CreateNotificationHandler (GlobalInputHandler)) ]]

  animation_screen.timers ["screen"]  = CreateTimer (false, true, main_timer)
  animation_screen.timers ["default"] = CreateTimer (false, true, main_timer)
  animation_screen.timers ["ui"]      = CreateTimer (false, true, main_timer)  

  animation_screen.sound_volume_correction = 0
  animation_screen.sound_fade_in_time      = xml_node:Get ("SoundFadeInTime", "")
  
  if animation_screen.sound_fade_in_time == "" then
    animation_screen.sound_fade_in_time = DEFAULT_SOUND_FADE_IN_TIME
  else
    animation_screen.sound_fade_in_time = animation_screen.sound_fade_in_time + 0
  end
 
  local on_set_current     = xml_node:Get ("OnSetCurrent", "")
  local on_lost_current    = xml_node:Get ("OnLostCurrent", "")

  if on_set_current ~= "" then
    animation_screen.on_set_current=CreateScreenEventHandler(on_set_current)
  end

  if on_lost_current ~= "" then
    animation_screen.on_lost_current=CreateScreenEventHandler(on_lost_current)
  end

  print ("Load animation screen '" .. animation_screen.name .. "'...")

  -- Загрузка связанных с экраном ресурсов
  
  local resource_groups_names
  
  if default_resource_group then
    resource_groups_names = xml_node:Get ("ResourceGroups", default_resource_group)
  else
    resource_groups_names = xml_node:Get ("ResourceGroups")
  end

  resource_groups_names = SplitString (resource_groups_names, " ", true)

  for i = 1, #resource_groups_names do
    table.insert (animation_screen.resource_groups, LoadResourceGroup (resource_groups_names [i]))
  end
  
  -- Загрузка потомков

  animation_screen.current_z = 90
  animation_screen.delta_z   = animation_screen.current_z / 2000

  --animation_screen.test_sprites_count = 0
  
  print ("Load screen" .. animation_screen.name)
  
  LoadAnimationScreenPart (animation_screen, xml_node)  
  
  local scene_manager = Engine.SceneManagers.Get ("SceneManager")            
  
  if scene_manager then    
    table.insert (animation_screen.resource_groups, { name = "scene_manager", group = scene_manager.Resources })
  end

  return animation_screen
end

local function InitScreenObjects (screen)
  -- Сброс счётчиков срабатываний
  
  local function ResetSprites (sprite)
    if sprite.initialized then return end
    
    sprite.initialized = true
  
    sprite.position      = {x=0, y=0}
    sprite.size          = {x=1, y=1}
  
    if (sprite.static) then 
      sprite.static_initialized = false
    end  
    
    if sprite.start_is_active then
      ActivateSprite (sprite.name, true)
    else
      DeactivateSprite (sprite.name)
    end
    
    if sprite.start_is_visible then
      ShowSprite (sprite)
    else
      HideSprite (sprite)
    end
  end

  ForEachInTable (screen.sprite_list, ResetSprites)

  local function ResetSpriteGroup (group)
    if group.initialized then return end
    
    group.initialized = true
  
    if group.start_visible then
      ShowSpriteGroup (group)
    else
      HideSpriteGroup (group)
    end
  
    group.start_time = group.timer.GetTime ()
  end

  ForEachInTable (screen.sprite_groups, ResetSpriteGroup)
  
  local function ResetTextLine (text_line)
    if text_line.initialized then return end
    
    text_line.initialized = true
  
    if text_line.start_is_visible then
      ShowTextLine (text_line.name)
    end
    
    if text_line.start_is_active then
      ActivateTextLine (text_line.name)
    end
  end

  ForEachInTable (screen.text_lines, ResetTextLine)

  -- Запуск камер
  
  local function ResetCamera (camera)
    if camera.initialized then return end
    
    camera.initialized = true
  
    SetCameraState (camera.name, camera.start_is_active)
  end
  
  ForEachInTable (screen.cameras, ResetCamera)
  
  -- Старт конечных автоматов

  local function StartFsm (fsm)
    if fsm.initialized then return end
    
    fsm.initialized = true
  
    fsm.current_state = nil
    
    local function InitFsmState (state)
      state.events = {}
    end
    
    ForEachInTable (fsm.states, InitFsmState)
  end
  
  ForEachInTable (screen.fsms, StartFsm)

  -- Старт скрипта уровня
  
  local function InitScript (script)
    if script.initialized then return end
    
    script.initialized = true
  
    local instance = script.creator (script.data, screen)
    
    if not instance then
      error (string.format ("Can't create script '%s'", script.name))
    end
    
    if script.name ~= "" then
      game.scripts [script.name] = instance
    else
      table.insert (game.scripts, instance)
    end
    
    if instance.Update then
      table.insert (game.update_scripts, instance)
    end
  end

  ForEachInTable (screen.start_scripts, InitScript)

  local function CallSpritesOnActivate (sprite)
    if sprite.on_activate_initialized then return end
    
    sprite.on_activate_initialized = true
  
    if sprite.start_is_active and sprite.on_activate then
      sprite.on_activate (sprite, game.events)
    end
  end

  ForEachInTable (screen.sprite_list, CallSpritesOnActivate)
end

function PrepareAnimationScreenPart (xml_file_name, scene)
  local xml = Common.StringNode.LoadXml (xml_file_name)

  collectgarbage ("setpause", 20)
  
  if scene and not game.current_screen.scenes [scene] then
     game.current_screen.scenes [scene] = Scene.Scene.Create ()

     local scene_light = Scene.SpotLight.Create ()
  
     scene_light:BindToScene (game.current_screen.scenes [scene], Scene.NodeBindMode.AddRef)    
  end
  
  scene = game.current_screen.scenes [scene]
  
  LoadAnimationScreenPart (game.current_screen, xml, scene)

  collectgarbage ("collect")
  InitScreenObjects (game.current_screen)  
  
  collectgarbage ("setpause", 100)
  collectgarbage ("collect")
  
  PerformUnloadUnusedResources ()
end

-- Загрузка всех экранов с анимациями
local function LoadAnimationScreens (screen_decls, path, main_timer)
  local root_node              = Common.StringNode.LoadXml (path)
  local children_count         = root_node.ChildrenCount

  for i=0, children_count-1 do
    local child_node = root_node:Child (i)

    if child_node.Name == "AnimationScreen" then
      screen_decls [child_node:Get ("Name")] = child_node
      
      local sub_children_count = child_node.ChildrenCount
      
      for j=0, sub_children_count-1 do
        local sub_child_node = child_node:Child (j)
        
        if sub_child_node.Name == "Include" then
          local xml_file_name = sub_child_node:Get ("Source")
          local xml           = Common.StringNode.LoadXml (xml_file_name)
            
          if xml.Name ~= "AnimationScreenPart" then
            error (string.format ("Bad include '%s'. No AnimationScreenPart found", xml_file_name))
          end
        end
      end
    end
  end
end

-- Пересчёт состояния трека
local function EvalTrackState (track, vars, time)
  if not (track) then return nil end  
  
  local self_value=track.self_value
  
  if not(track.static and self_value) then
    -- Применение смещение по времени трека и повторений
    time = time - track.offset  
    
    if time < 0 then
      time = 0
    end
   
    if track.spline then
      self_value = track.spline:Eval (time)
    else
      if track.loop then
        time = math.fmod (time, track.loop_time)
      end  
    
      -- Поиск ключа  
    
      local key_range = FindKeyRange (track, time)  
     
      if not (key_range) then return nil end
    
      -- Если интервал пуст (время вне интервала)
    
      if key_range [1] == key_range [2] then
        if track.link ~= "" then
          return track.add (key_range [1].value, GetVarValue(vars, track.link))
        else
          return key_range [1].value    
        end
      end

      -- Если время находится в интервале - линейная интерполяция значений на интервале
    
      self_value = track.lerp ((time - key_range [1].time) / (key_range [2].time - key_range [1].time), key_range [1].value, key_range [2].value)
    end
  end
  
  if track.link ~= "" then
    local var_value = GetVarValue(vars, track.link)
  
    if track.name == "scale" and type (var_value) == "number" then
      self_value = track.add (self_value, vec2 (var_value))
    else
      self_value = track.add (self_value, var_value)
    end
    
    if track.spline and track.spline.KeysCount == 1 then  --если в треке один ключ, то возвращаем значение не сохраняя в self_value
      return self_value
    end
  end
  
  track.self_value=self_value
  
  return self_value
end

-- Выполнение событий
local function ProcessTrackEvents (track, time, sprite)
  if not (track) then return nil end  

  -- Применение смещение по времени трека и повторений
  time = time - track.offset  
  
  if time < 0 then return end
  
  -- Если число срабатываний ключа меньше указанного - выполнение события

  local index, key = next (track.keys)

  while index do
    if not key.call_count then
      key.call_count = 0
    end    
  
    if math.floor ((time - key.time) / track.length) >= key.call_count then
      if track.loop or key.call_count < 1 then
        key.call_count = key.call_count + 1      
      
        local event = key.value      

        if event then
          event (time, game.events, sprite)
        end
      end
    end  

    index, key = next (track.keys, index)
  end      
end

local function EvalSpriteTrackState (sprite, track_name, time, vars, default_value)
  local value = EvalTrackState (sprite.tracks [track_name], vars, time)  

  if value then 
    if type(value) ~= "table" then return value end

    local return_value = {}

    local index, key = next (value)

    while index do
    return_value [index] = key
  
      index, key = next (value, index)
  end

    return return_value 
  end  

  return default_value
end

function EvalAngleToPoint (const_angle, point1, point2)
  if const_angle then
    return const_angle
  else         
    local dx                      = point2.x - point1.x
    local dy                      = point2.y - point1.y
    local length                  = math.sqrt (dx * dx + dy * dy)
    local sin_angle_to_this_point = dx / length
    local cos_angle_to_this_point = dy / length
    local acos_angle              = math.acos (cos_angle_to_this_point) * 180 / math.pi
  
    if sin_angle_to_this_point > 0 then
      return -acos_angle
    else 
      return acos_angle
    end        
  end
end

function EvalWayPointForTime (time, way)  --возвращает {x, y}
  local return_value = {x=0, y=0, scale_x=1, scale_y=1}

  if not way then return return_value end
  
  if (time > way.time) and not way.loop then
    local last_point = way.points [#way.points]
    
    return_value.x       = last_point.position.x
    return_value.y       = last_point.position.y
    return_value.scale_x = last_point.scale.x
    return_value.scale_y = last_point.scale.y    
    
    return return_value
  end

  if (time < 0) and not way.loop then
    local first_point = way.points [0]
    
    return_value.x       = first_point.position.x
    return_value.y       = first_point.position.y
    return_value.scale_x = first_point.scale.x
    return_value.scale_y = first_point.scale.y
    
    return return_value
  end
  
  if (time > 0) then
    time = time % way.time
  elseif time ~= 0 then
    time = way.time - time % way.time
  end  

  local base_point = nil
  local next_point = nil

  for i = 1, #way.points, 1 do
    if way.points[i].time > time then
      base_point = way.points [i - 1]
      next_point = way.points [i]
      break
    end
  end
  
  local section_time = 1
  
  if not base_point then                   --возможно только в случае loop трека
    base_point = way.points [#way.points]
    next_point = way.points [0]
    section_time = way.time - base_point.time
  else
    section_time = next_point.time - base_point.time
  end
  
  time = time - base_point.time
  
  local delta = 0
  
  if section_time ~= 0 then
    delta = time / section_time
  end

  return_value.x       = base_point.position.x + (next_point.position.x - base_point.position.x) * delta  
  return_value.y       = base_point.position.y + (next_point.position.y - base_point.position.y) * delta
  return_value.scale_x = base_point.scale.x + (next_point.scale.x - base_point.scale.x) * delta
  return_value.scale_y = base_point.scale.y + (next_point.scale.y - base_point.scale.y) * delta
  
  return return_value
end

function EvalWayState (time, way)
  local return_value = EvalWayPointForTime (time, way)

  return_value.angle = 0

  if not way then return return_value end

  if time >= way.time and not way.loop then
    local prev_point = EvalWayPointForTime (way.time - way.look_time_offset, way)
    local look_point = EvalWayPointForTime (time, way)

    return_value.angle = EvalAngleToPoint (way.const_angle, prev_point, look_point)
  else
    local look_point = EvalWayPointForTime (time + way.look_time_offset, way)

    return_value.angle = EvalAngleToPoint (way.const_angle, return_value, look_point)
  end
  
  return return_value
end

function SetSpriteWay (sprite_name, way_name)
  local sprite = GetSprite (sprite_name)
  
  if not sprite then
    print ("Error: can't find sprite '" .. sprite_name .. "'") 
    return 
  end 
  
  sprite.way            = way_name
  sprite.start_way_time = nil
end

-- Вызов событий спрайта
local function EvalSpriteEvents (sprite, vars, time)
  time = time - sprite.start_time
  
  if (sprite.tracks["time_factor"]) then 
    time = time * EvalSpriteTrackState (sprite, "time_factor", time, vars, 0)
  end

  local events = sprite.tracks ["events"]

  if events then
    ProcessTrackEvents (events, time, sprite)
  end  
end

-- Пересчёт состояния спрайта
function EvalSpriteState (sprite, vars, time, do_events)
  if not (sprite.is_active) then return end         
  
  local function FastRecompute ()
    local world_pos  = sprite.node.WorldPosition
    
    sprite.position.x = world_pos.x
    sprite.position.y = world_pos.y  
  end
  
  if sprite.manual_updated or (sprite.static and sprite.static_initialized and (not sprite.way or sprite.way == "") and not sprite.need_change_visibility) then
    FastRecompute ()
    return
  end        

  time = time - sprite.start_time
  
  if (sprite.tracks["time_factor"]) then 
    time = time * EvalSpriteTrackState (sprite, "time_factor", time, vars, 0)
  end
  
  if sprite.tracks["time_offset"] then
    time = time + EvalSpriteTrackState (sprite, "time_offset", time, vars, 0)
  end
  
  local can_skip_eval
  
  for _, track in pairs (sprite.tracks) do
    local static_spans = track.static_spans
    
    if static_spans then
      local can_skip = false
    
      for _, span in pairs (static_spans) do
        if time >= span.from_time and time <= span.to_time then
          can_skip = true
        end
      end
      
      if can_skip then
        can_skip_eval = true
      else
        can_skip_eval = false
        
        break        
      end      
    end
  end  
  
  if can_skip_eval then
    FastRecompute ()
    return
  end

  local frame

  if sprite.frames_count > 1 then
    frame = time * sprite.fps

    if frame < 0 then frame = 0 end
    
    if frame >= sprite.frames_count then
      if sprite.repeat_mode == "loop" then
        frame = frame % sprite.frames_count
      elseif sprite.repeat_mode == "mirror" then
        frame = frame % (sprite.frames_count * 2)
        
        if frame >= sprite.frames_count then
          frame = sprite.frames_count - 1 - math.floor (frame) % sprite.frames_count
        end
      elseif sprite.repeat_mode == "clamp_to_start" then
        frame = 0
      else
        frame = sprite.frames_count - 1
      end
    end
    
    frame = math.floor (frame)
  else
    frame = 0
  end

  if sprite.materials then  
    local material = sprite.materials [frame]
    
    if not material then
      material = ""
      print (string.format ("Bad frame %u (frames=%u)", frame, sprite.frames_count))
    end
  
    sprite.node.Material = material
  else
--    sprite.node.Frame = sprite.first_frame + sprite.frame_order_sign * sprite.frame_step * frame
  end

  local way_state = {x=0, y=0, angle=0, scale_x=1, scale_y=1}

  if sprite.way ~= "" then
    if not sprite.start_way_time then sprite.start_way_time = time end
  
    way_state = EvalWayState (time - sprite.start_way_time, game.current_screen.ways[sprite.way])
  end
  
  local offset = { x = 0, y = 0 }
  
  if (sprite.tracks["offset"]) then  
    offset = EvalSpriteTrackState (sprite, "offset", time, vars, {x=0, y=0})
  end
  
  local with_parent = sprite.node.Parent and sprite.parent ~= ""

  if not (with_parent and sprite.is_world_position) and not sprite.is_manual_position then  
    if (sprite.tracks["position"] or sprite.way ~= "" or sprite.tracks ["offset"]) then
      local position = {x=0, y=0}
      
      if (sprite.tracks["position"]) then
        position = EvalSpriteTrackState (sprite, "position", time, vars, {x = 0, y = 0})
      end

      sprite.position.x = position.x + way_state.x + offset.x
      sprite.position.y = position.y + way_state.y + offset.y

      if sprite.is_world_position then
        sprite.node:SetWorldPosition (sprite.position.x, sprite.position.y, sprite.node.WorldPosition.z)
      else
        sprite.node:SetPosition (sprite.position.x, sprite.position.y, sprite.node.Position.z)
      end
    end    
  end
  
  local world_pos = sprite.node.WorldPosition
        
  if not sprite.start_position then
    sprite.start_position = sprite.node.Position
  end
  
  sprite.position.x = world_pos.x
  sprite.position.y = world_pos.y
    
  if not (with_parent and sprite.is_world_scale) and not sprite.is_manual_scale then
    if (sprite.tracks["size"] or sprite.way ~= "") then
      if (sprite.tracks["scale"]) then
        sprite.scale = EvalSpriteTrackState (sprite, "scale", time, vars, nil)

        if type (sprite.scale) == "number" then
          local scalar = sprite.scale
        
          sprite.scale = {x=scalar, y=scalar}
        end

        if math.abs (sprite.scale.x) < EPS then sprite.scale.x = EPS end
        if math.abs (sprite.scale.y) < EPS then sprite.scale.y = EPS end
      end

      sprite.size = EvalSpriteTrackState (sprite, "size", time, vars, {x = 1, y = 1})
      
      sprite.size.x = sprite.size.x * sprite.scale_factor.x * way_state.scale_x
      sprite.size.y = sprite.size.y * sprite.scale_factor.y * way_state.scale_y
      
      if math.abs (sprite.size.x) < EPS then sprite.size.x = EPS end
      if math.abs (sprite.size.y) < EPS then sprite.size.y = EPS end
      
      if sprite.is_world_scale then
        sprite.node:SetWorldScale (sprite.size.x * sprite.scale.x, sprite.size.y * sprite.scale.y, 1.0)    
      else
        sprite.node:SetScale (sprite.size.x * sprite.scale.x, sprite.size.y * sprite.scale.y, 1.0)
      end
    end
  end

  if not sprite.manual_alpha then
    if (sprite.tracks["alpha"]) then
      sprite.alpha = EvalSpriteTrackState (sprite, "alpha", time, vars, sprite.node.Alpha) * sprite.alpha_factor * sprite.opacity
      sprite.node.Alpha = sprite.alpha
    else
      sprite.alpha      = sprite.alpha_factor * sprite.opacity
      sprite.node.Alpha = sprite.alpha
    end
  end
  
  if sprite.node.Alpha < sprite.invisible_alpha then -- отсечение по минимальному значению альфа-канала
    sprite.node.Alpha = 0
  end
  
  if (sprite.tracks["alpha_clamp"]) then
    sprite.node.AlphaReference = EvalSpriteTrackState (sprite, "alpha_clamp", time, vars, sprite.node.AlphaReference)
  end

  if not (with_parent and sprite.is_world_orientation) and not sprite.is_manual_orientation then
    if (sprite.tracks["angle"] or sprite.way ~= "") then
      sprite.angle = 0
    
      if (sprite.tracks["angle"]) then
        sprite.angle = EvalSpriteTrackState (sprite, "angle", time, vars, sprite.angle)
      end
        
      sprite.angle = sprite.angle + way_state.angle
      
      if sprite.is_world_orientation then
        sprite.node:SetWorldOrientation (sprite.angle, 0, 0, 1)
      else
        sprite.node:SetOrientation (sprite.angle, 0, 0, 1)
      end
    end
  end
  
  if (sprite.tracks["scroll"]) then
    local properties = sprite.node.Properties

    if not properties then
      properties = Scene.NodeProperties.Create ()
      
      sprite.node.Properties = properties
    end

    local scroll = EvalSpriteTrackState (sprite, "scroll", time, vars, {x = 0, y = 0})

    properties:SetVector ("render.texture_scroll", vec4 (scroll.x, scroll.y, 0, 0))
  end

  if sprite.tracks["texoffset"] then
    local value = EvalSpriteTrackState (sprite, "texoffset", time, vars, sprite.node.TexOffset)

    sprite.node.TexOffset = vec2 (value.x, value.y)
  end

  if sprite.tracks["texsize"] then
    local value = EvalSpriteTrackState (sprite, "texsize", time, vars, sprite.node.TexSize)

    sprite.node.TexSize = vec2 (value.x, value.y)
  end
  
  if sprite.tracks ["intensity"] then
    sprite.intensity = EvalSpriteTrackState (sprite, "intensity", time, vars, sprite.intensity)    

    sprite.node:SetColor (sprite.intensity, sprite.intensity, sprite.intensity)
  elseif sprite.tracks ["color"] then
    sprite.color = EvalSpriteTrackState (sprite, "color", time, vars, sprite.color)

    sprite.node:SetColor (sprite.color.x, sprite.color.y, sprite.color.z)  
  end
  
  local pos        = vec3 (0)
  local world_pos  = sprite.node.WorldTM * pos

  sprite.position.x = world_pos.x
  sprite.position.y = world_pos.y  
  
  -- Выполнение событий

  local events = sprite.tracks ["events"]
  
  if events and do_events then
    ProcessTrackEvents (events, time, sprite)
  end

  if (sprite.static) then 
    sprite.static_initialized = true 
  end    
end

-- Пересчёт состояния камеры
local function EvalCameraState (camera, time)
  if not (camera.is_active) then return end 

  time = time - camera.start_time
  
  if (camera.tracks["time_factor"]) then  
    time = time * EvalSpriteTrackState (camera, "time_factor", time, vars, 1.0)
  end  
--[[
  if (camera.tracks["position"]) then
    camera.position = {x=0, y=0}
  
    if (camera.tracks["position"]) then
      camera.position = EvalSpriteTrackState (camera, "position", time, vars, {x = 0, y = 0})
    end
  
    camera.position.x = camera.position.x
    camera.position.y = camera.position.y
        
    camera.node:SetPosition (camera.position.x, camera.position.y, 0)
  end]]  
    
  if (camera.tracks["size"]) then
    camera.size     = EvalSpriteTrackState (camera, "size", time, vars, {x = 1, y = 1})
    camera.node:SetScale (camera.size.x, camera.size.y, 1.0)
  end
  
  local params = {"Left", "Right", "Top", "Bottom", "ZNear", "ZFar", "FovX", "FovY"}  

  ForEachInTable (params, function (param) 
    local value = EvalTrackState (camera.tracks [string.lower (param)], vars, time)

    if camera.params [param] ~= value then
      camera.node [param] = value
      camera.params [param] = value
    end
  end)

  -- Выполнение событий

  local events = camera.tracks ["events"]

  if events then
    ProcessTrackEvents (events, time, camera)
  end
end

-- Выполнение действия FSM
local function DoFsmAction (action, state, fsm)
  if not action then return nil end  
  
  action (fsm.timer.GetTime (), state, game.events)
end

-- Вход в состояние
local function ProcessFsmOnEnter (state)
  if state.animation_sprite then 
    local group = game.current_screen.sprite_groups [state.animation_sprite]      
    
    if group then
      ShowSpriteGroup (group.name)
    else
      ActivateSprite (state.animation_sprite)
    end
  elseif state.animation then 
    PlayAnimation (state.animation)
  end  

  DoFsmAction (state.on_enter, state, state.fsm)
  
  state.start_time = state.fsm.timer.GetTime ()
end

-- Выход из состояния
local function ProcessFsmOnExit (state)
  state.events = {}

  if state.animation_sprite then
    local group = game.current_screen.sprite_groups [state.animation_sprite]
    
    if group then
      HideSpriteGroup (group.name)
    else
      DeactivateSprite (state.animation_sprite) 
    end
  end  

  if not state.on_exit_process then
    state.on_exit_process = true  

    DoFsmAction (state.on_exit, state, state.fsm)
  
    state.on_exit_process = false
  end
end

-- Пересчёт FSM
local function ProcessFsms (fsms)
  local function ProcessFsm (fsm)
    local state = fsm.current_state

    if not state then    
      state             = fsm.states [fsm.start_state]
      fsm.current_state = state

      ProcessFsmOnEnter (state)

      return
    end
    
    local function ProcessFsmJunction (junction, index, time)
      if not junction.condition_fn (time, game.events, state, junction) then return false end
      
      local next_state = fsm.states [junction.state]      
      
      if junction.state ~= state.name then
        local current_state_name = "<NIL>"
        local next_state_name    = "<NIL>"

        if fsm.current_state then current_state_name = fsm.current_state.name end
        if next_state then next_state_name = next_state.name end

--        print (string.format ("FSM %s: %s -> %s (junction %s#%u)", fsm.name, current_state_name, next_state_name, current_state_name, index))

        ProcessFsmOnExit (state)
      end

      DoFsmAction (junction.action, nil, fsm)

      if junction.state ~= state.name then
        fsm.current_state = next_state
        state             = fsm.current_state

        if not state then
          error ("State '" .. junction.state .. "' not found in FSM '" .. fsm.name .. "'")
          return true
        end

        ProcessFsmOnEnter (state)
      end

      return true
    end
    
    local time = fsm.timer.GetTime ()

    local index, junction = next (state.junctions)

    while index do
      if ProcessFsmJunction (junction, index, time) then break end

      index, junction = next (state.junctions, index)    
    end
    
    if fsm.current_state then
      DoFsmAction (fsm.current_state.on_update, fsm.current_state, fsm)
    end
  end

  ForEachInTable (fsms, ProcessFsm)
end

-- Установка события FSM
local function SetEvent (name, value, auto_reset)
  local old_value = game.events [name]

  game.events [name] = value

  if old_value ~= value then
--    print (string.format ("SetEvent ('%s', %s, %s)", name, tostring (value), tostring (auto_reset)))  
  end
  
  if auto_reset then
    if not (game.auto_reset_events) then
      game.auto_reset_events = {}
    end
    
    local event = {}
    
    event.time = game.main_timer.GetTime ()
    event.name = name

    table.insert (game.auto_reset_events, event)
  end
end

function SetManualEvent (name, value)
  SetEvent (name, value, false)
end

function SetAutoEvent (name)
  SetEvent (name, true, true)
end

-- Реакция на событие завершения анимации
function OnAnimationEnd (name)
--  print (string.format ("OnAnimationEnd('%s')", name))
  if not game.current_screen then return end

  local function ProcessFsmAnimationCallback (fsm)
    if not fsm.current_state then return end    
    
    if not fsm.current_state.animation_sprite and not fsm.current_state.animation then return end
    if fsm.current_state.animation_sprite and fsm.current_state.animation_sprite ~= name then return end
    if fsm.current_state.animation and fsm.current_state.animation ~= name then return end    

    fsm.current_state.events ["on_end_animation"] = true
  end

  ForEachInTable (game.current_screen.fsms, ProcessFsmAnimationCallback)
end

-- Пересчёт состояния группы спрайтов
local function EvalSpriteGroupState (group, time)
  if group.need_change_visibility then 
    group.visible = group.need_visible
  
    if not group.visible then    
      ForEachInTable (group.sprites, DeactivateSprite)
    else
      ForEachInTable (group.sprites, ActivateSprite)
      ForEachInTable (group.sprites, HideSprite)
    end
    
    group.need_change_visibility = false
  end  

  if not(group.visible) then
    return
  end  

  time = time - group.start_time

  local frame
  local next_frame
  local alpha
  
  local function ClampFrame (frame)
    if frame < 0 then return 0 end
    
    if frame >= group.frames_count then
      if group.repeat_mode == "loop" then
        return frame % group.frames_count
      elseif group.repeat_mode == "mirror" then
        frame = frame % (group.frames_count * 2)
        
        if frame >= group.frames_count then
          frame = group.frames_count - 1 - math.floor (frame) % group.frames_count
        end
        
        return frame
      elseif group.repeat_mode == "mirror_once" then
        if frame >= group.frames_count * 2 then
          return 0
        end
      
        if frame >= group.frames_count then
          frame = group.frames_count - 1 - math.floor (frame) % group.frames_count
        end
        
        return frame
      elseif group.repeat_mode == "clamp_to_start" then
        return 0
      else
        return group.frames_count - 1
      end
    end
    
    return frame
  end

  if group.frames_count > 1 then
    local frame_pos = time * group.fps
    local temp
  
    frame       = math.floor (ClampFrame (frame_pos))
    next_frame  = math.floor (ClampFrame (frame_pos + 1))
    temp, alpha = math.modf (frame_pos)
    
    if frame == group.frames_count - 1 and group.repeat_mode == "clamp_to_end" and frame_pos >= group.frames_count then
      alpha = 1
    end
    
    if frame == 0 and group.repeat_mode == "clamp_to_start" and frame_pos >= group.frames_count then
      alpha = 0
    end
  else
    frame      = 0
    next_frame = 0
  end  

  if not group.sprites then return end
  
  local sprite = group.sprites [frame]

  if not sprite then
    print (string.format ("Bad frame %u (frames=%u)", frame, group.frames_count))
  end

  ForEachInTable (group.sprites, HideSprite)     
        
  if group.animation_mode == "all" then   
    for index=0,frame do
      ShowSprite (group.sprites [index])
    end
  else
    ShowSprite (sprite.name)  
  end
  
  if group.animation_mode == "pairs" and group.blend_frames and next_frame ~= frame then
    local next_sprite = group.sprites [next_frame]
           
    if next_sprite then
      ShowSprite (next_sprite.name)
        
      sprite.alpha_factor      = 1 - alpha
      next_sprite.alpha_factor = alpha      
      sprite.static_initialized = false      
      next_sprite.static_initialized = false
      
--      print (string.format ("blend %s %.2f with %s %.2f", sprite.name, sprite.node.Alpha, next_sprite.name, next_sprite.node.Alpha))
    end
  end
  
  if group.animation_mode == "all" and group.blend_frames then
    sprite.alpha_factor = alpha
    sprite.static_initialized = false
  end  
end

local function EvalTextLineTrackState (text_line, track_name, time, vars, default_value)
  local value = EvalTrackState (text_line.tracks [track_name], vars, time)  

  if value then return value end  

  return default_value
end

-- Обновление линий текста
function EvalTextLineState (text_line, time, vars)
  if not text_line.is_active then return end    
  
  if text_line.static and text_line.static_initialized then return end    

  time = time - text_line.start_time
  
  local offset = { x = 0, y = 0 }
  
  if (text_line.tracks["offset"]) then  
    offset = EvalTextLineTrackState (text_line, "offset", time, vars, {x=0, y=0})
  end
  
  if (text_line.tracks["position"]) then
    local position = {x=0, y=0}

    if (text_line.tracks["position"]) then
      position = EvalTextLineTrackState (text_line, "position", time, vars, {x = 0, y = 0})
    end

    text_line.position = { x = position.x + offset.x, y = position.y + offset.y } 
        
    text_line.node:SetWorldPosition (text_line.position.x, text_line.position.y, text_line.z)
  end  

  local scale = { x = 1, y = 1 }

  if text_line.tracks["scale"] then
    scale = EvalTextLineTrackState (text_line, "scale", time, vars, nil)

    if type (scale) == "number" then
      scale = { x=scale, y=scale}
    end

    if math.abs (scale.x) < EPS then scale.x = EPS end
    if math.abs (scale.y) < EPS then scale.y = EPS end
  end

  if text_line.tracks["size"] then
    text_line.size     = EvalTextLineTrackState (text_line, "size", time, vars, {x = 1, y = 1})
    text_line.node:SetWorldScale (text_line.size.x * scale.x, text_line.size.y * scale.y, 1.0)
  end

  if (text_line.tracks["alpha"]) then
    local alpha = EvalTextLineTrackState (text_line, "alpha", time, vars, text_line.node.Color.w)
   
    text_line.node:SetColor (1, 1, 1, alpha)
    
    if text_line.shadow_text_line then
      local shadow_color = text_line.shadow_text_line.Color
    
      shadow_color.w = alpha * text_line.shadow_alpha
    
      text_line.shadow_text_line.Color = shadow_color
    end
  end

  if (text_line.tracks["angle"]) then
    text_line.angle = 0
  
    if (text_line.tracks["angle"]) then
      text_line.angle = EvalTextLineTrackState (text_line, "angle", time, vars, text_line.angle)
    end
      
    text_line.angle = text_line.angle 
      
    text_line.node:SetOrientation (text_line.angle, 0, 0, 1)
  end  
  
  if text_line.tracks ["color"] then
    text_line.color = EvalTextLineTrackState (text_line, "color", time, vars, text_line.color)

    text_line.node:SetColor (text_line.color.x, text_line.color.y, text_line.color.z)  
  end
  
  
  if text_line.static then
    text_line.static_initialized = true
  end
end

-- Пересчёт прозрачности для спрайта с учётом родительской прозрачности
local function UpdateSpriteFullAlpha (sprite, cache)
  local value = cache [sprite]
  
  if value then return value end
  
  if not sprite.manual_alpha then
    if not sprite.parent_sprite then
      value = sprite.alpha
    else
      value = sprite.alpha * UpdateSpriteFullAlpha (sprite.parent_sprite, cache)
    
      sprite.node.Alpha = value
    end
  else
    value = sprite.node.Alpha
  end
  
  cache [sprite] = value
  
  return value
end

-- Пересчёт состояния анимационного экрана на указанный момент времени
local function EvalAnimationScreenState (screen)
  -- Обновление таймеров  
  
  for _, timer in pairs (screen.timers) do
    if timer ~= game.main_timer then
      timer.Update ()
    end
  end
  
  -- Обработка конечных автоматов и скриптов

  local dt    = screen.timers.default.GetTime () - screen.last_update_time
  local ui_dt = screen.timers.ui.GetTime () - screen.ui_last_update_time
  
  screen.last_update_time    = screen.timers.default.GetTime ()  
  screen.ui_last_update_time = screen.timers.ui.GetTime ()  

  -- Обновеление скриптов

  if screen ~= game.current_screen then return end
  
  for _, script in pairs (game.update_scripts) do
    script.Update (dt, ui_dt)
  end

  -- Изменение параметров спрайтов  
  
  if screen ~= game.current_screen then return end

  for _, sprite_group in pairs (screen.sprite_groups) do
    EvalSpriteGroupState (sprite_group, sprite_group.timer.GetTime ())
  end
  
  local new_update_sprite_list = {}
  
  for _, sprite in pairs (screen.update_sprite_list) do  
    EvalSpriteState (sprite, screen.vars, sprite.timer.GetTime (), true)
    
    local need_update = true
    
    if not sprite.is_active then need_update = false end
    if not sprite.manual_updated and (sprite.static and sprite.static_initialized and (not sprite.way or sprite.way == "") and not sprite.need_change_visibility) and sprite.parent == "" then need_update = false end
    
    sprite.is_in_update_sprite_list = need_update
    
    if need_update then
      table.insert (new_update_sprite_list, sprite)
    end
  end  

  ProcessFsms (screen.fsms)
  
  local alpha_cache = {}
  
  for _, sprite in pairs (screen.update_sprite_list) do
    UpdateSpriteFullAlpha (sprite, alpha_cache)
  end
  
  screen.update_sprite_list = new_update_sprite_list

  -- Изменение параметров текстовых линий

  for _, text_line in pairs (screen.text_lines) do
    EvalTextLineState (text_line, text_line.timer.GetTime (), screen.vars)
  end

  -- Изменение параметров камер
  
  for _, camera in pairs (screen.cameras) do
    EvalCameraState (camera, camera.timer.GetTime ())
  end

  -- Сброс автосбрасываемых событий
  
  if game.auto_reset_events then
    local index, event = next (game.auto_reset_events)

    while index do
      if event.time < game.main_timer.GetTime () then
        game.events [event.name] = nil

        table.remove (game.auto_reset_events, index)
      end  

      index, event = next (game.auto_reset_events, index)
    end
  end
  
  -- Обновление контроллеров сцены      

  screen.node:Update (screen.timers.default.GetTime ())
  
  for _, scene in pairs (screen.scenes) do
    scene.Root:Update (screen.timers.default.GetTime ())
  end
end

function IsPointInsideSprite (sprite, x, y)
  if not sprite.position then return false end
  
  local min_x1 = sprite.position.x - sprite.size.x / 2
  local max_x1 = sprite.position.x + sprite.size.x / 2
  local min_y1 = sprite.position.y - sprite.size.y / 2
  local max_y1 = sprite.position.y + sprite.size.y / 2    
  
  if min_x1 < x and max_x1 < x then return false end
  if x < min_x1 and x < min_x1 then return false end
  if min_y1 < y and max_y1 < y then return false end
  if y < min_y1 and y < min_y1 then return false end

  return true
end

local function IsRectIntersectSprite (sprite, min_x2, min_y2, max_x2, max_y2)
  if not sprite.position then return false end
  
  local min_x1 = sprite.position.x - sprite.size.x / 2
  local max_x1 = sprite.position.x + sprite.size.x / 2
  local min_y1 = sprite.position.y - sprite.size.y / 2
  local max_y1 = sprite.position.y + sprite.size.y / 2    
  
  if min_x1 < min_x2 and max_x1 < min_x2 then return false end
  if min_x2 < min_x1 and max_x2 < min_x1 then return false end
  if min_y1 < min_y2 and max_y1 < min_y2 then return false end
  if min_y2 < min_y1 and max_y2 < min_y1 then return false end
  
--  if math.min (max_x1, max_x2) - math.max (min_x1, min_x2) <= 0 then return false end
--  if math.min (max_y1, max_y2) - math.max (min_y1, min_y2) <= 0 then return false end

  return true
end

function SetInputState (state)
  if not game.current_screen then return end

  for _, camera in pairs (game.current_screen.cameras) do
    for i = 1, #camera.viewports do
      camera.viewports [i].InputState = state
    end
  end
end

function ProcessInputEvent (event, type)
  if game.input_locked then return end

  local queued = false

  local function QueueEvent ()
    table.insert (game.input_queue, { event = event, type = type })
    queued = true
  end

  if type then
    if game.touch_begin_active then
      if game.touch_end_active then  --на текущем кадре произошло окончание касания, все последующие события необходимо обрабатывать только на следующем кадре
        QueueEvent ()
      end

      if not game.touch_begin_updated then   --на текущем кадре произошло касание
        if not game.touch_end_active and type ~= 'moved' then  --обрабатываем события движения сразу, чтобы жесты работали без задержки на кадр
          QueueEvent ()
        end
      end
      
      if type == 'ended' then
        game.touch_end_active = true
      end
    else
      if type == 'began' then
        game.touch_begin_active  = true
        game.touch_begin_updated = false
      end
    end
  end

  if not queued then
    game.input.manager:ProcessEvent (event)
  end
end

-- Инициализация графики
local function InitRender ()  
  game.render_state = {}

  game.render_state.screen = Scene.Screen.Create ()
  game.render_state.screen.BackgroundColor = vec4 (0, 0, 0, 1)
  game.render_state.screen:SetArea (0, 0, 10000, 10000)

  Engine.Screens.Register ("Screen1", game.render_state.screen)
end

-- Инициализация ввода
local function InitInput ()
  Engine.InputEventHandlers.Register ("game", Input.CreateEventHandler ())
  
  game.input                = {}
  game.input.manager        = Scene.InputManager.Create ()
  game.input.manager.Screen = game.render_state.screen
  
  game.input.manager:SetTouchSize (TOUCH_SIZE, Scene.InputTransformSpace.Screen)
end

-- Обновление спрайта в списке спрайтов
local function AddSpriteToUpdateList (screen, sprite)
  if sprite.is_in_update_sprite_list then return end
  
  table.insert (screen.update_sprite_list, sprite)
  
  sprite.is_in_update_sprite_list = true
end

-- Активация спрайта
function SetSpriteActive(sprite, state, no_on_activate)
  if not (game.current_screen) then return end  

  if type (sprite) == "string" then
    sprite = game.current_screen.sprites [sprite]
  end

  if not (sprite) then return end      
  
  if not sprite.is_active and not state then return end
  
  if sprite.is_active and state then
    SetSpriteActive (sprite, false)
  end
  
  sprite.is_active  = state  
  sprite.start_time = sprite.timer.GetTime ()
  sprite.position   = {x=0, y=0}
  sprite.size       = {x=1, y=1}  
  sprite.static_initialized = false  
    
  if state then 
    -- Сброс счётчиков событий
    
    local function ResetEvents (key)
      key.call_count = nil
    end

    local events = sprite.tracks ["events"]

    if events then
      ForEachInTable (events.keys, ResetEvents)
    end
    
    -- Получение начального состяоняи
    
    EvalSpriteState (sprite, game.current_screen.vars, sprite.start_time, false)  

    ShowSprite (sprite)
        
    -- Загрузка ресурсов
    
    if sprite.resource_binding then
      sprite.resource_binding:Load ()
    end        
    
    if sprite.on_activate and not no_on_activate then
      sprite.on_activate (sprite, game.events)
    end    
    
    -- Выполнение событий      

    EvalSpriteEvents (sprite, game.current_screen.vars, sprite.start_time)    
    
    -- Обновление в списке спрайтов
    
    AddSpriteToUpdateList (game.current_screen, sprite)
  else
    if sprite.on_deactivate then
      sprite.on_deactivate (sprite, game.events)
    end  
  
    HideSprite (sprite)

    -- Выгрузка ресурсов

    if sprite.resource_binding then
      sprite.resource_binding:Unload ()
    end
  end    
end

function SetSpritesActive(state, ...)
  ForEachInTable({...}, function (sprite)
    SetSpriteActive (sprite, state)
  end)
end

function ActivateSprite (sprite, no_on_activate)
  SetSpriteActive (sprite, true, no_on_activate)
end

function DeactivateSprite (sprite)
  SetSpriteActive (sprite, false)
end

function IsSpriteActive (sprite)
  local screen_sprite = GetSprite (sprite, true)
  
  if not screen_sprite then return false end
  
  return screen_sprite.is_active
end

--Биндинг одного спрайта к другому
function SetSpriteParent (child_sprite_name, parent_sprite_name)
  local child_sprite = GetSprite (child_sprite_name)
  
  if not child_sprite then return end

  HideSprite (child_sprite)
  child_sprite.parent = parent_sprite_name
  ShowSprite (child_sprite)    
  
  -- Обновление в списке спрайтов
    
  AddSpriteToUpdateList (game.current_screen, sprite)  
end

-- Показ узла
local function ShowNode (node)
  if not (game.current_screen) then return end

  if not (node) then return end
  
  if node.after_sprite ~= "" then
    local above_sprite = game.current_screen.sprites [node.after_sprite]
    
    if not above_sprite then
      error (string.format ("After sprite '%s' not found", node.after_sprite))
    end

    node.z = above_sprite.z - 0.001
  end  
  
  if node.is_world_position then  
    node.node:SetWorldPosition (node.node.WorldPosition.x, node.node.WorldPosition.y, node.z)
  else
    node.node:SetPosition (node.node.Position.x, node.node.Position.y, node.z)
  end
  
  local bind_space
  
  if node.bind_space == "parent" then
    bind_space = Scene.NodeTransformSpace.Local        
  else
    bind_space = Scene.NodeTransformSpace.World
  end  
  
  if node.parent == "" then  
    if node.scene then
      node.node:BindToScene (node.scene, Scene.NodeBindMode.Default, bind_space)
    else
      node.node:BindToParent (game.current_screen.node, Scene.NodeBindMode.Default, bind_space)
    end
  else  
    local parent_sprite = game.current_screen.sprites [node.parent]
    local parent_node

    if not parent_sprite then    
      parent_node = game.current_screen.node:FindChild (node.parent, Scene.NodeSearchMode.OnAllSublevels)

      if not parent_node then
        error (string.format ("Parent sprite '%s' not found for sprite '%s'", node.parent, node.name))
      end
    else
      parent_node = parent_sprite.node
      node.parent_sprite = parent_sprite      
    end    
    
    node.node:BindToParent (parent_node, Scene.NodeBindMode.Default, bind_space)    
  end
  
  if node.scissor then
    node.scissor:BindToScene (node.node.Scene)
  end
end

-- Скрытие узла
function HideNode (node)
  if not (node) then return end
  
  local bind_space
  
  if node.bind_space == "parent" then
    bind_space = Scene.NodeTransformSpace.Local
  else
    bind_space = Scene.NodeTransformSpace.World
  end    

  node.node:Unbind (bind_space)
  
  node.parent_sprite = nil
end

-- Скрытие спрайта
function HideSprite (sprite)
  if not (game.current_screen) then return end

  if type (sprite) == "string" then
    local sprite_name = sprite
  
    sprite = game.current_screen.sprites [sprite]
    
    if not sprite then
      local node = game.current_screen.node:FindChild (sprite_name, Scene.NodeSearchMode.OnAllSublevels)

      if node then
        node.Visible = false
      end

      return
    end
  end  

  if not (sprite) then return end
  
  if sprite and not sprite.is_shown then return end
  
  HideNode (sprite)
  
  if sprite.on_hide then
    sprite.on_hide (sprite, game.events)
  end
  
  sprite.is_shown = false
end

-- Показ спрайта
function ShowSprite (sprite)
  if not (game.current_screen) then return end

  if type (sprite) == "string" then
    local sprite_name = sprite
  
    sprite = GetSprite (sprite, true)
    
    if not sprite then
      local node = game.current_screen.node:FindChild (sprite_name, Scene.NodeSearchMode.OnAllSublevels)
      
      if node then
        node.Visible = true
      else
        error("ShowSprite ("..tostring(sprite_name)..") - sprite not found")
      end
      
      return
    end
  end  
  
  if sprite and sprite.is_shown then return end  
  
  ShowNode (sprite)
  
  if sprite.on_show then
    sprite.on_show (sprite, game.events)
  end  
  
  sprite.is_shown = true  
  
  -- Обновление унаследованного значения прозрачности
  
  local alpha_cache = {}
  
  UpdateSpriteFullAlpha (sprite, alpha_cache)          
end

function ShowSpriteGroup(sprite_group)
  if not (game.current_screen) then return end  
  
  if type (sprite_group) == "string" then
    sprite_group = game.current_screen.sprite_groups [sprite_group]
  end    
  
  if not (sprite_group) then return end

  sprite_group.need_visible           = true
  sprite_group.start_time             = sprite_group.timer.GetTime ()
  sprite_group.need_change_visibility = true
end

function HideSpriteGroup(sprite_group)
  if not (game.current_screen) then return end

  if type (sprite_group) == "string" then
    sprite_group = game.current_screen.sprite_groups [sprite_group]
  end
  
  if not (sprite_group) then return end

  sprite_group.need_visible           = false  
  sprite_group.need_change_visibility = true
end

-- Поиск спрайта
function GetSprite (sprite, ignore_errors)
  if type (sprite) == "string" then
    local result = game.current_screen.sprites [sprite]
    
    if not result and not ignore_errors then 
      error("GetSprite ("..tostring(sprite)..") - sprite not found") 
    end  

    return result
  else
    return sprite
  end
end

function GetSpritesByPosition (x, y, width, height, sprites)
  local result = {}
  
  if not width  then width = 0 end
  if not height then height = 0 end  
  
  local min_x = x - width / 2
  local max_x = x + width / 2
  local min_y = y - height / 2
  local max_y = y + height / 2
  
  if not sprites then sprites = game.current_screen.sprites end

  ForEachInTable (sprites,
    function (sprite)
      if not sprite.node.Parent or not sprite.is_active then
        return
      end
          
      if IsRectIntersectSprite (sprite, min_x, min_y, max_x, max_y) then
        table.insert (result, sprite)
      end
    end
  )

  return result
end

function GetSpriteNode (sprite)
  sprite=GetSprite (sprite)

  if not sprite then
    error("GetSpriteNode ("..tostring(name)..") - sprite not found")
  end

  return sprite.node
end

function SetSpriteVariablesByItsNode(sprite)
  -- устанавливает дублирующие переменные параметров спрайта равными его текущему значению узла
  sprite=GetSprite(sprite)
  local node=sprite.node
  
  if not sprite.position then sprite.position = {} end
  if not sprite.size then sprite.size = {} end
  
  sprite.position.x=node.WorldPosition.x
  sprite.position.y=node.WorldPosition.y
  sprite.size.x=node.WorldScale.x
  sprite.size.y=node.WorldScale.y
end

function SetSpritePosition(sprite, x, y, is_world)
  sprite = GetSprite (sprite)
  local node=GetSpriteNode (sprite)
  if node then
    if is_world then
      node.WorldPosition=vec3(x, y, node.WorldPosition.z)    
    else
      node.Position=vec3(x, y, node.Position.z)
    end

    SetSpriteVariablesByItsNode (sprite)
  end

  sprite.tracks ["position"] = nil
  sprite.tracks ["offset"]   = nil
  sprite.way                 = ""
end

function SetSpriteAngle (sprite, angle, is_world)
  sprite=GetSprite(sprite)
  local node=GetSpriteNode(sprite)
  if node then
    if is_world then
      sprite.node:SetWorldOrientation (angle, 0, 0, 1)
    else
      sprite.node:SetOrientation (angle, 0, 0, 1)
    end

    sprite.angle = angle
  end
  
  sprite.tracks ["angle"] = nil
end

function SetSpriteScaleFactor(sprite, sx, sy)
  sprite=GetSprite(sprite)
  sprite.scale_factor = {x=sx,y=sy}
  sprite.static_initialized=false
end

-- доп. функции работы со множеством спрайтов
function HideSprites(...)
  ForEachInTable({...}, function(name)
  HideSprite(name)
  end)
end

function ShowSprites(...)
  ForEachInTable({...}, ShowSprite)
end

function DeactivateSprites(...)
  ForEachInTable({...}, DeactivateSprite)
end

function ActivateSprites(...)
  ForEachInTable({...}, ActivateSprite)
end

function ActivateTextLines(...)
  ForEachInTable({...}, ActivateTextLine)
end

function DeactivateTextLines(...)
  ForEachInTable({...}, DeactivateTextLine)
end

-- Поиск текстовой линии
function GetTextLine(line)
  if type (line) == "string" then
    return game.current_screen.text_lines [line]
  else
    return line
  end
end

-- Показ текстовой линии
function ShowTextLine (name)
  if not (game.current_screen) then return end

  local text_line = GetTextLine (name)
  
  if not (text_line) then return end
  
  if text_line.after_sprite ~= "" then
    local above_sprite = game.current_screen.sprites [text_line.after_sprite]
    
    if not above_sprite then
      error (string.format ("After sprite '%s' not found", text_line.after_sprite))
    end

    text_line.z = above_sprite.z - 0.001
  end  

  if text_line.parent == "" then
    if text_line.scene then
      text_line.node:BindToScene (text_line.scene)
    else
      text_line.node:BindToParent (game.current_screen.node)
    end  
  else
    local parent_sprite = game.current_screen.sprites [text_line.parent]

    if not parent_sprite then
      error (string.format ("Parent sprite '%s' not found for text-line '%s'", tostring (text_line.parent), tostring (text_line.name)))
    end
  --print (text_line.parent)
    text_line.node:BindToParent (parent_sprite.node) --, Scene.NodeBindMode.Default, Scene.NodeTransformSpace.World)
  end
  
  text_line.static_initialized = false
  
  EvalTextLineState (text_line, text_line.timer.GetTime (), game.current_screen.vars)
end

function HideTextLine (name)
  if not (game.current_screen) then return end

  local text_line = GetTextLine (name)

  if not (text_line) then return end

  text_line.node:Unbind (Scene.NodeTransformSpace.World)
end

function ActivateTextLine (name)
  if not (game.current_screen) then return end

  local text_line = GetTextLine (name)

  if not (text_line) then return end
  
  text_line.start_time         = text_line.timer.GetTime ()
  text_line.is_active          = true
  text_line.static_initialized = false
  
  if text_line.text ~= "" then
    text_line.node.Text = GetLocalizedString (text_line.text)
  end
  
  for i = 1, #text_line.chars_color_factors do
    local factor = text_line.chars_color_factors [i]
  
    text_line.node:SetCharsColorFactors (factor.from, factor.count, factor.factor)
  end
  
  ShowTextLine (text_line)
end

function DeactivateTextLine (name)
  if not (game.current_screen) then return end

  local text_line = GetTextLine (name)

  if not (text_line) then return end

  text_line.start_time = text_line.timer.GetTime ()
  text_line.is_active  = false

  HideTextLine (text_line)
end

function SetTextLineActive (name, state)
  if state then
    ActivateTextLine (name)
  else
    DeactivateTextLine (name)
  end
end

function SetTextLinesActive(state, ...)
  ForEachInTable({...}, function (line)
    SetTextLineActive (line, state)
  end)
end

local function UpdateCameraViewports (camera)
  local window          = Engine.Windows.Find (WINDOW_NAME)
  local window_width    = window.ClientWidth
  local window_height   = window.ClientHeight
  local camera_width    = math.abs (camera.node.Right - camera.node.Left)
  local camera_height   = math.abs (camera.node.Top - camera.node.Bottom)
  local x_offset        = (window_width - camera_width) / 2
  local y_offset        = (window_height - camera_height) / 2
  local x_screen_factor = SCREEN_WIDTH / window_width
  local y_screen_factor = SCREEN_HEIGHT / window_height

  ForEachInTable (camera.center_viewports, function (viewport)
    viewport:SetArea (x_offset * x_screen_factor, y_offset * y_screen_factor, camera_width * x_screen_factor, camera_height * y_screen_factor)
  end)
end

function SetCameraState (name, state)
  if not game.current_screen then return end

  local camera = name

  if type (name) == "string" then
    camera = game.current_screen.cameras [name]
  end

  if not camera then return end

  camera.is_active  = state
  camera.start_time = camera.timer.GetTime ()

  ForEachInTable (camera.viewports, function (viewport)
    if state then
      game.render_state.screen:Attach (viewport)
    else
      game.render_state.screen:Detach (viewport)
    end
  end)  
  
  EvalCameraState (camera, camera.start_time)

  if state then
    UpdateCameraViewports (camera)
  end
end

-- Загрузка ресурсов
local function LoadResources (resource_groups, callback)
  local total_groups = #resource_groups

  for i = 1, #resource_groups do  
    local group   = resource_groups [i]
    local binding = Engine.ResourceManager.CreateBinding (group.group)

    local function ProgressHandler (progress)
      local progress_text_line 

      if game.current_screen then
        progress_text_line = game.current_screen.text_lines ["progress_text"]
        if game.scripts.level and game.scripts.level.SetProgress then
          game.scripts.level.SetProgress ((i - 1 + progress.Value) / #resource_groups)
        end
      end
      
      if progress_text_line then
        progress_text_line.node.Text = string.format ("%d%%", (i - 1 + progress.Value) / total_groups * 100.0)
      end    
    
      if progress.State == Engine.ResourceManager.ProgressState.Processing then
--        print (string.format ("%s: '%s' %.1f%% TOTAL = %f", progress.Stage, progress.Resource, progress.Value * 100.0, (i - 1 + progress.Value) / #resource_groups * 100.0))
      elseif progress.State == Engine.ResourceManager.ProgressState.Processed then
        if i >= total_groups then 
          print ("Load complete")
          
          if callback then 
            DelayedAction ("ui", 0.2, callback) 
          end
        end
      else      
        print (string.format ("Failed to load resource '%s'", progress.Resource))
      end
    end

    local resource_loaded = false

    for j = 1, #game.loaded_resources do
      if game.loaded_resources [j].name == group.name then
        resource_loaded = true
        total_groups = total_groups - 1
        break
      end
    end

    if not resource_loaded then
      binding:AsyncLoad (Engine.ResourceManager.Binding.CreateCallbackHandler (ProgressHandler)) 

      table.insert (game.loaded_resources, { name = group.name, binding = binding } )
    end
  end
  
  if total_groups == 0 then
    if callback then 
      DelayedAction ("ui", 0.2, callback) 
    end
  end
end

-- Выгрузка ресурсов
local function UnloadResources ()
  global.bindings = {}

  collectgarbage ("collect")

  Engine.ResourceManager.FlushUnusedResources ()
end

function UnloadUnnededResources (needed_groups)
  for i = #game.loaded_resources, 1, -1 do
    local resource = game.loaded_resources [i]
    local unload   = true
    
    for j = 1, #needed_groups do
      if resource.name == needed_groups [j].name then
        unload = false
        break
      end
    end
    for j = 1, #game.shared_resources do
      if resource.name == game.shared_resources [j].name then
        unload = false
        break
      end
    end
    
    if unload then
      table.remove (game.loaded_resources, i)
    end
  end

  collectgarbage ("collect")

  Engine.ResourceManager.FlushUnusedResources ()
  Engine.ResourceManager.WaitForAsyncComplete ()
end

-- Обработка изменения размеров окна
function OnWindowResize ()
  if game.current_screen then
    for _, camera in pairs (game.current_screen.cameras) do
      if camera.is_active then
        UpdateCameraViewports (camera)
      end
    end
  end

  local window        = Engine.Windows.Find (WINDOW_NAME)
  local window_width  = window.ClientWidth
  local window_height = window.ClientHeight

  if window_width > 0 and window_height > 0 then --обход возможных проблем
    for _, script in pairs (game.scripts) do
      if script.OnSizeChanged then
        script.OnSizeChanged (window_width, window_height)
      end
    end
  end
end

-- Установка текущего экрана
function IsSlotMounted ()
  return game.mounted_slot
end

function UnmountSlotResources ()
  for i = 1, #game.slot_previous_mount_points do
    Common.File.Unmount (game.slot_previous_mount_points [i])
  end

  for i = 1, #game.slot_previous_pathes do
    Common.File.RemoveSearchPath (game.slot_previous_pathes [i])
  end
end

function UnmountCurrentSlot ()
  for i = 1, #game.slot_mount_points do
    Common.File.Unmount (game.slot_mount_points [i])
  end

  for i = 1, #game.slot_pathes do
    Common.File.RemoveSearchPath (game.slot_pathes [i])
  end
end

function MountCurrentSlot ()
  game.slot_previous_mount_points = game.slot_mount_points
  game.slot_previous_pathes       = game.slot_pathes

  game.mount_point_index = game.mount_point_index or 0
  game.mount_point_index = game.mount_point_index + 1

  local mount_point = "/pack" .. game.mount_point_index
  
  game.slot_mount_points = { mount_point }

  local status, err = pcall (function ()
    Common.File.Mount (mount_point, GetDownloadsManager ().GetSlotPath (game.mounted_slot.slot.path, game.mounted_slot.is_bonus))
  end)
  
  if not status then
    print (err)
    GetDownloadsManager ().DeleteSlotResource (game.mounted_slot.slot.path, game.mounted_slot.is_bonus)
    SetCurrentScreen ("lobby")
    return
  end

  local slot_info = GetDownloadsManager ().GetDownloadedSlotInfo (game.mounted_slot.slot.path, game.mounted_slot.is_bonus)

  game.slot_pathes = { mount_point,
                       string.format (mount_point .. "/resolutions/%s-%s-%s", slot_info.platform, slot_info.format, slot_info.resolution),
                       string.format (mount_point .. "/resolutions/%s-generic-%s", slot_info.platform, slot_info.resolution) }

  for i = 1, #game.slot_pathes do
    Common.File.AddSearchPath (game.slot_pathes [i])
  end
  
  local path = LoadPackLocalization (mount_point)

  local add_pathes = { string.format (path .. "/resolutions/%s-%s-%s", slot_info.platform, slot_info.format, slot_info.resolution), 
                       string.format (path .. "/resolutions/%s-generic-%s", slot_info.platform, slot_info.resolution) }

  for i = 1, #add_pathes do
    Common.File.AddSearchPath (add_pathes [i])
    table.insert (game.slot_pathes, add_pathes [i])
  end
end

function SetSlotScreen (name, skip_loading_screen)
  print ("SET SLOT SCREEN!!!!!!!")

  UnmountSlotResources ()

  if NeedsDownloadSlots () then
    local path     = name
    local is_bonus = false
  
    if string.find (name, BONUS_SCREEN_NAME_SUFFIX) then
      path     = string.sub (name, 1, #name - #BONUS_SCREEN_NAME_SUFFIX)
      is_bonus = true
    end
  
    local slots = ApplicationDataGet (SLOTS_DESCS)
    local slot
  
    for i = 1, #slots do
      if slots [i].path == path then
        slot = slots [i]
        break
      end
    end

    if not GetDownloadsManager ().IsSlotResourceDownloaded (slot.path, is_bonus) then
      if GetDownloadsManager ().HasLocalSlotResource (slot.path, is_bonus) then
        ApplicationDataSet (SLOT_TO_DOWNLOAD, name)
        SetCurrentScreen ("loader", true)
        return
      end
    
      game.scripts.ui_windows.ShowLoadingWindow (0, nil, nil, nil, slot.id)

      local finish_processed = false

      local function ProgressCallback (progress)
        game.scripts.ui_windows.ShowLoadingWindow (progress * 100, nil, nil, nil, slot.id)
      end
      
      local function FinishCallback ()
        if finish_processed then return end
        
        finish_processed = true
      
        ResumeGame ()
        SetSlotScreen (name, skip_loading_screen)
      end
      
      local function ErrorCallback ()
        game.scripts.ui_windows.HideWindow ()
        game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
      end

      GetDownloadsManager ().DownloadSlotResource (slot.path, ProgressCallback, FinishCallback, is_bonus, ErrorCallback)
      GetDownloadsManager ().MoveSlotDownloadToFront (slot.path, ProgressCallback, FinishCallback, ErrorCallback, true)
      
      return
    end

    game.mounted_slot = { slot = slot, is_bonus = is_bonus }

    MountCurrentSlot () 
  end
  
  local after_mount_script = "data/scripts/screens/" .. name .. "_after_mount.lua"
  
  if Common.File.IsFileExist (after_mount_script) then
    local status, error = pcall (function ()
      local mount_function = dofile (after_mount_script)
          
      mount_function ()
    end)
        
    if not status then
      print ("Slot after mount error " .. error)
    end
  end
  
  ApplicationDataSet (IS_SLOT_SCREEN_LOADING, true)
  
  SetCurrentScreen (name, skip_loading_screen)
end

function SetCurrentScreen (name, skip_loading_screen, finish_load, really_set)
  if screen_to_set then return end
  
  if not really_set then
    screen_to_set = {}
    
    screen_to_set.name                = name
    screen_to_set.skip_loading_screen = skip_loading_screen
    screen_to_set.finish_load         = finish_load
    return
  end

  print ("SetCurrentScreen(" .. name .. ")")
  
  game.default_startup_screen = nil
  
  -- Отсоединение старого экрана
    
  if game.current_screen then    
    if game.current_screen.on_lost_current then
      game.current_screen.on_lost_current(game.current_screen)
    end    
    
    ForEachInTable (GetReverseTable (game.scripts), function (script)
      if script.OnLostScreen then
        script.OnLostScreen (game.current_screen)
      end
    end)
    
    ForEachInTable (game.current_screen.cameras, function (camera) SetCameraState (camera, false) end)
    
    -- Отсоединение из сцены
  
    game.current_screen.node:UnbindAllChildren ()
    game.current_screen.node:Unbind (Scene.NodeTransformSpace.World)
    
    -- Остановка конечных автоматов
    
    local function StopFsm (fsm)
      if fsm.current_state then
        ProcessFsmOnExit (fsm.current_state)
      end
    end
    
    ForEachInTable (game.current_screen.fsms, StopFsm)
    
    -- Закрытие уровня
    
    for _, timer in pairs (game.current_screen.timers) do
      timer.Reset ()    
    end
    
    local active_animations = game.current_screen.playing_animations
    
    for _, animation in pairs (active_animations) do
      animation.Stop ()
    end

    for _, item in pairs (game.current_screen.input_handlers) do
      item.connection:Disconnect ()
      item.connection = nil
      item.callback = nil
    end
    
    for _, line in pairs (game.current_screen.text_lines) do
      if line.stroked_text_line_update_connection then
        line.stroked_text_line_update_connection:Disconnect ()
        line.stroked_text_line_update_connection = nil
      end
      if line.shadow_text_line_update_connection then
        line.shadow_text_line_update_connection:Disconnect ()
        line.shadow_text_line_update_connection = nil
      end
      
      line.scene = nil
    end
    
    for _, sprite in pairs (game.current_screen.sprite_list) do
      if sprite.input_handlers then
        for _, item in pairs (sprite.input_handlers) do
          item.connection:Disconnect ()
          item.connection = nil
          item.callback = nil
        end
      end
      
      sprite.input_zone = nil
      sprite.scene = nil
      sprite.node = nil
    end
    
    if game.current_screen.global_input_zone_connection then
      game.current_screen.global_input_zone_connection:Disconnect ()
      game.current_screen.global_input_zone_connection = nil
    end
    
    game.current_screen.sprite_list = {}
    game.current_screen.sprites = {}
    game.current_screen.scenes = {}
    game.current_screen.fsms = {}
    game.current_screen.input_handlers = {}
    game.current_screen.input_zones = {}
    game.current_screen.loaded_parts = {}
    
    for _, camera in pairs (game.current_screen.cameras) do
      SetCameraState (camera.name, false)
    end

    if game.current_screen.localization then
      UnloadLocalizationResource (game.current_screen.localization)
    end
    
    ClearUrlSpritesCache ()

    game.current_screen = nil        
    game.scripts = {}
  end
  
  -- Установка нового экрана
  
  if game.loading_screen_name == "" then
    skip_loading_screen = true
  end

  local screen
  
  local screen_xml_node = game.screen_decls [name]
  
  if not screen_xml_node then
    local screen_path    = "data/animations/" .. name .. ".xml"
    local root_node      = Common.StringNode.LoadXml (screen_path)
    local children_count = root_node.ChildrenCount
    
    for i=0, children_count-1 do
      local child_node = root_node:Child (i)

      if child_node.Name == "AnimationScreen" and child_node:Get ("Name") == name then
        screen_xml_node = child_node
        break
      end
    end
  end  
  
  if not screen_xml_node then
    error ("Can't load animation screen " .. name)
  end
  
  if skip_loading_screen then
    screen = LoadAnimationScreen (screen_xml_node, "", game.main_timer)

    if not (screen) then
      error ("Screen '" .. name .. "' not found!")
      return
    end

    finish_load = true
    
    if #screen.resource_groups == 0 then
      error ("Screen " .. screen.name .. " has no resources")
    end

    if not screen.shared_resources then
      game.shared_resources = {}
      UnloadUnnededResources (screen.resource_groups)
    else
      game.shared_resources = screen.resource_groups
    end
    
    LoadResources (screen.resource_groups, nil)
  end
  
  if game.loading_screen_name == "" then
    skip_loading_screen = true
  end

  local load_callback
  
  if not skip_loading_screen then
    load_callback = function ()
      if game.scripts.level and game.scripts.level.OnResourcesLoaded then
        game.scripts.level.OnResourcesLoaded (name)
      else
        local status, err = pcall (SetCurrentScreen, name, true, true, true)
      
        if not status then
          print (err)
        end
      end
    end
    
    local loading_screen = game.loading_screen_name
    
    if ApplicationDataGet (IS_BATTLE_ACTIVE) and not ApplicationDataGet (IS_BATTLE_LOADING_STARTED) then
      ApplicationDataSet (IS_BATTLE_LOADING_STARTED, true)
    
      loading_screen = BATTLE_LOADING_SCREEN
    end
    
    SetCurrentScreen (loading_screen, true, true, true)
    
    Engine.ResourceManager.WaitForAsyncComplete ()
  end 

  if not finish_load then
    local resource_groups = LoadAnimationScreenResourceGroups (screen_xml_node)

    if #resource_groups == 0 then
      error ("Screen " .. name .. " has no resources")
    end

    ExecuteOnNextUpdate (function () 
      UnloadUnnededResources (resource_groups)
      LoadResources (resource_groups, load_callback)
    end)
  else
    Engine.ResourceManager.WaitForAsyncComplete ()
  end

  if not skip_loading_screen then
    return
  end
  
  if not screen then
    screen = LoadAnimationScreen (game.screen_decls [name], "", game.main_timer)

    if not (screen) then
      error ("Screen '" .. name .. "' not found!")
      return
    end
  end

  game.events = {}  

  local prev_actions = game.next_update_actions

  game.next_update_actions = {}
  
  for i = 1, #prev_actions do
    local action = prev_actions [i]
  
    if action.dont_cancel_on_set_screen then
      table.insert (game.next_update_actions, action)
    end
  end
  
  game.input_locked = true --Ввод запрещен до отрисовки первого кадра нового экрана
  ExecuteOnNextUpdate (function ()
    ExecuteOnNextUpdate (function ()
      game.input_locked = false
    end)
  end)
  
  game.current_screen = screen

  if game.current_screen.localization then
    LoadLocalizationResource (game.current_screen.localization)
  end

  game.current_screen.sound_groups = {}
  game.current_screen.playing_animations = {}
  game.current_screen.playing_animations_index = 0
  game.current_screen.update_sprite_list = {}
  game.current_screen.input_handlers = {}
  game.current_screen.input_zones = {}
  game.current_screen.loaded_parts = {}

  screen.node:BindToScene (game.scene)
  
  -- Обновление таймера
  
  game.main_timer.Update ()

  -- Сброс таймеров

  for _, timer in pairs (screen.timers) do
    timer.Reset ()
    timer.Stop ()
  end

  game.update_scripts = {}
  
  InitScreenObjects (screen)

  if game.current_screen.on_set_current then
    game.current_screen.on_set_current(game.current_screen)
  end

  if finish_load then
    collectgarbage ("collect")

    Engine.ResourceManager.FlushUnusedResources ()  

    for _, timer in pairs (screen.timers) do
      timer.Start ()
    end    
  end
  
  screen.last_update_time    = screen.timers.default.GetTime ()
  screen.ui_last_update_time = screen.timers.ui.GetTime ()
  
  OnWindowResize ()
  
  EvalAnimationScreenState (game.current_screen)
  
--[[  if IsGamePaused () then
    PauseGame ()
  end]]--?????????
  
  game.current_screen.sound_volume_correction = 0
  UpdateSoundsVolume ()

  SlowAction ("default", game.current_screen.sound_fade_in_time, function (t)
    game.current_screen.sound_volume_correction = t
    UpdateSoundsVolume ()
  end)
  
  local initial_music_volume = game.music_volume_correction or 0
  local music_volume_change  = 1 - initial_music_volume
  
  game.music_fade_action = SlowAction ("default", game.current_screen.sound_fade_in_time * music_volume_change, function (t)
    game.music_volume_correction = initial_music_volume + music_volume_change * t
    UpdateMusicVolume ()
  end)

  collectgarbage ("collect")
  PerformUnloadUnusedResources ()
end

--Пауза
function PauseGame (dont_pause_music)
  PauseSounds (dont_pause_music)
  
  if game.current_screen then
    game.current_screen.timers.screen.Stop ()
    game.current_screen.timers.default.Stop ()
    
    if game.scripts.in_game_ui and not game.current_screen.game_manually_paused then
      game.scripts.in_game_ui.ProcessPauseClick ()
    end
  end      
  
  local window = Engine.Windows.Find (WINDOW_NAME)
  
  if window then
    local is_full_screen = GetGameProperty ("Fullscreen") == "1" 
    
    if is_full_screen then
      window:Minimize ()
    end
  end
  
  game_paused = true
end
  
function ResumeGame ()
  ResumeSounds () 
  
  local window = Engine.Windows.Find (WINDOW_NAME)
  
  if window then
    local is_full_screen = GetGameProperty ("Fullscreen") == "1" 
    
    if is_full_screen then
      window:Maximize ()
    end
  end  

  if game.current_screen then
    game.current_screen.timers.screen.Start ()
    game.current_screen.timers.default.Start ()
  end  
  
  game_paused = false
end  

function OnWindowChangedHandle (handle)
  print ("OnWindowChangedHandle " .. tostring (handle))

  Engine.SubsystemManager:Remove ("*Render* WindowRenderer")
  
  if not handle or handle + 0 == 0 then return end
  
  Engine.SubsystemManager:Start ("config.xml", "*Render* WindowRenderer")
end

function OnWindowShown ()
  print ("OnWindowShown")

  local os_type = ApplicationDataGet (OS_TYPE)
  
  if os_type == "iOS" then return end

  Engine.SubsystemManager:Remove ("*Render* WindowRenderer")
  Engine.SubsystemManager:Start ("config.xml", "*Render* WindowRenderer")
end

function OnWindowHidden ()
  print ("OnWindowHidden")

  local os_type = ApplicationDataGet (OS_TYPE)
  
  if os_type == "iOS" then return end

  Engine.SubsystemManager:Remove ("*Render* WindowRenderer")
end

function OnApplicationActivate ()
  game.application_deactivated = game.application_deactivated or 1
  
  game.application_deactivated = math.max (game.application_deactivated - 1, 0)

  if game.application_deactivated ~= 0 then return end
  if game.application_active then return end

  System.Application.PostNotification ("LocalNotifications RemoveAll|Login")
  System.Application.PostNotification ("LocalNotifications Schedule|Login|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.NO_LOGIN_3_DAYS") .. "|" .. 60 * 60 * 24 * 3)
  System.Application.PostNotification ("LocalNotifications Schedule|Login|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.NO_LOGIN_7_DAYS") .. "|" .. 60 * 60 * 24 * 7)
  System.Application.PostNotification ("LocalNotifications Schedule|Login|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.NO_LOGIN_14_DAYS") .. "|" .. 60 * 60 * 24 * 14)
  System.Application.PostNotification ("LocalNotifications Schedule|Login|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.NO_LOGIN_21_DAYS") .. "|" .. 60 * 60 * 24 * 21)
  System.Application.PostNotification ("LocalNotifications SchedulePeriodic|Login|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.NO_LOGIN_MORE_DAYS") .. "|week|" .. 60 * 60 * 24 * 28)

  local os_type = ApplicationDataGet (OS_TYPE)
  
  if os_type == "iOS" then
    System.Application.SystemProperties:SetInteger ("ApplicationIconBadgeNumber", 0)
  end
  
  OnActivate ()
  
  GetSocialManager ().OnActivate ()
  
  game.main_timer.Update ()

  if not game.application_active then
    SetProfileProperty (APPLICATION_ACTIVATIONS_COUNT_ATTRIBUTE, (GetProfileProperty (APPLICATION_ACTIVATIONS_COUNT_ATTRIBUTE) or 0) + 1)
  end
  
  game.application_active = true
end

function OnApplicationDeactivate ()
  game.application_deactivated = game.application_deactivated or 0
  
  game.application_deactivated = game.application_deactivated + 1 

  if not game.application_active then return end

  local os_type = ApplicationDataGet (OS_TYPE)
  
  if os_type == "iOS" then
    System.Application.SystemProperties:SetInteger ("ApplicationIconBadgeNumber", 0)
  end

  OnDeactivate ()
  
  game.application_active = nil
end

function OnDeactivate ()
  print ("!!!deactivate")

  PauseSounds (nil, true)
  
  local window = Engine.Windows.Find (WINDOW_NAME)
  
  if window then
    local is_full_screen = GetGameProperty ("Fullscreen") == "1" 
    
    if is_full_screen then
      window:Minimize ()
    end
  end

  if not IsGamePaused () and game.current_screen then
    game.current_screen.resume_game = true
  end

  PauseGame ()
  
  HsConnectionOnDeactivate ()
end

local function ResumeMusic ()
  if game.music_playing then return end

  local current_time = game.main_timer.GetTime ()

  if not game.dont_pause_music and game.music_emitter then
    game.music_emitter:Play (game.music_offset)
    game.music_start_time = current_time - game.music_offset
    game.music_playing = true 
  end
end 

function OnActivate ()
  print ("!!!activate")

  local window = Engine.Windows.Find (WINDOW_NAME)
  
  if window then
    local is_full_screen = GetGameProperty ("Fullscreen") == "1" 
    
    if is_full_screen then
      window:Maximize ()
    end
  end
  
  local resume = true
  
  if game.scripts and game.scripts.ui_windows then
    resume = not game.scripts.ui_windows.IsWindowPausedGame ()
  end
  
  if resume then
    if game.current_screen and game.current_screen.resume_game then
      ResumeSounds () 
    else
      ResumeMusic ()
    end
  
    ResumeGame ()
  end
  
  HsConnectionOnActivate ()
end

function IsGamePaused ()
  return game_paused
end

-- Обновление игры
function ExecuteOnNextUpdate (func, dont_cancel_on_set_screen)
  if not func then
    local status, error = pcall (error ("Null func at ExecuteOnNextUpdate"))
   
    print (tostring (error))
   
    return 
  end

  table.insert (game.next_update_actions, { func = func, dont_cancel_on_set_screen = dont_cancel_on_set_screen } )
end

function Update ()  
  game.main_timer.Update ()  

  if not game.application_active then
    System.Application.Sleep (100)
    return
  end

  if not update_info then
    update_info = {}
    
    update_info.time   = game.main_timer.GetTime ()
    update_info.frames = 0
  end

  update_info.frames = update_info.frames + 1

  if not IsGamePaused () and game.main_timer.GetTime () - update_info.time > 1 then
    GAME_FPS=string.format (" FPS: %.1f", update_info.frames / (game.main_timer.GetTime () - update_info.time))

--    print (GAME_FPS)
  
    if FPS_LINE then
      FPS_LINE.Text=GAME_FPS
    end

    if PROFILER_ENABLED then
      Profiler.Clean ()
      Profiler.Update ()
    
      print (Profiler.FlatString (50))
    end

    update_info.time   = game.main_timer.GetTime ()
    update_info.frames = 0
  end

  if game.current_screen then
    EvalAnimationScreenState (game.current_screen)
  elseif game.default_startup_screen then  --first screen start
    local last_logged_in_network = GetProfileProperty (LAST_LOGGED_IN_NETWORK)

    if last_logged_in_network then
      local default_startup_screen = game.default_startup_screen
      
      print ("Logging in to network " .. last_logged_in_network)
    
      SetCurrentScreen ("loader")
      
      ExecuteOnNextUpdate (function ()
        GetSocialManager ().SetEngine (last_logged_in_network)
      
        GetSocialManager ().Login (function ()
          HsConnectToServer ()
          
          if NeedsDownloadLobby () then
            return --loader already active
          else
            SetCurrentScreen ("lobby")
          end
        end, function ()
          GetSocialManager ().CancelLogin ()
          SetCurrentScreen (default_startup_screen, true, nil, true)
        end)
      end, true)
    else
      SetCurrentScreen (game.default_startup_screen, true, nil, true)
    end
  end
  
  if screen_to_set then
    local screen = screen_to_set 
    
    screen_to_set = nil
    
    SetCurrentScreen (screen.name, screen.skip_loading_screen, screen.finish_load, true)
  end
  
  PerformUnloadUnusedResources ()
  
  local actions = game.next_update_actions
  
  game.next_update_actions = {}
  
  for i = 1, #actions do
    actions [i].func ()
  end

    --Обработка очереди ввода
  if game.touch_begin_updated then
    if game.touch_begin_active and game.touch_end_active then --завершаем обработку касания
      local queue_size = #game.input_queue
  
      for i = 1, queue_size do
        local event = game.input_queue [1]

        table.remove (game.input_queue, 1)

        game.input.manager:ProcessEvent (event.event)

        if event.type == 'ended' then
          game.touch_begin_updated = false
          game.touch_begin_active  = false
          game.touch_end_active    = false
          break
        end
      end
    end
    
    local last_begin_touch_event  --поиск в очереди последнего события начала касания, все предыдущие события игнорируются
  
    for i = #game.input_queue, 1, -1 do
      if game.input_queue [i].type == 'began' then
        last_begin_touch_event = i
        break
      end
    end

    local pending_events = {}

    if last_begin_touch_event then
      for i = last_begin_touch_event, #game.input_queue do
        table.insert (pending_events, game.input_queue [i])
      end
    end

    game.input_queue = {}

    for i = 1, #pending_events do
      local event = pending_events [i]

      ProcessInputEvent (event.event, event.type)
    end
  end

  if game.touch_begin_active then
    ExecuteOnNextUpdate (function ()
      game.touch_begin_updated = true
    end)
  end
end

-- Перезагрузка скриптов
function ReloadScripts ()
  print ("Reload scripts...")

  game = nil  

  collectgarbage ("collect")

  dofile ("data/scripts/common/main.lua")
  
  Main ()
end

-- Перезагрузка скриптов и экранов
function ReloadAll ()
  print ("Reload resources...")

  global.bindings = {}

  collectgarbage ("collect")

  Engine.ResourceManager.FlushUnusedResources ()

  ReloadScripts ()  
end

-- Инициализация звука
local function InitSoundPlayer()
  local listener = Scene.Listener.Create () 
  
  listener:BindToScene (game.scene, Scene.NodeBindMode.AddRef) 
  listener:SetPosition(0, 0, -1)

  Engine.Listeners.Register ("Listener1", listener)

  game.listener = listener
end

-- Проигрывние звука
function PlaySound (sound_name, sound_volume, sound_group, offset)
  local emitter = Scene.SoundEmitter.Create (sound_name)

  emitter.Name = "SoundEmitter"
  
  if not sound_volume then
    sound_volume = 1
  end

  local sfx_volume = GetGameProperty (GAME_SFX_VOLUME_ATTRIBUTE)

  if not sfx_volume then
    sfx_volume = 1
  end

  if GetGameProperty (GAME_SOUND_ENABLED_ATTRIBUTE) ~= "true" then
    sfx_volume = 0
  end

  if not sound_group then
    sound_group = DEFAULT_SOUND_GROUP_NAME
  end
  
  local emitter_desc = {}
  
  emitter_desc.emitter    = emitter
  emitter_desc.gain       = sound_volume
  emitter_desc.start_time = game.main_timer.GetTime ()

  if not game.current_screen.sound_groups [sound_group] then
    game.current_screen.sound_groups [sound_group] = {}
    game.current_screen.sound_groups [sound_group].gain = 1
    game.current_screen.sound_groups [sound_group].emitters = {}
  end
  
  game.current_screen.sound_groups [sound_group].emitters [sound_name] = emitter_desc
    
  emitter.Gain = game.current_screen.sound_groups [sound_group].gain * emitter_desc.gain * sfx_volume

  if game_muted then
    emitter.Gain = 0
  end

  emitter:BindToParent (game.current_screen.node)

  emitter:Play (offset)

  return emitter
end

function UpdateSoundsVolume ()
  if not game.current_screen then return end

  local sfx_volume = GetGameProperty (GAME_SFX_VOLUME_ATTRIBUTE)

  if not sfx_volume then
    sfx_volume = 1
  end

  if GetGameProperty (GAME_SOUND_ENABLED_ATTRIBUTE) ~= "true" then
    sfx_volume = 0
  end

  ForEachInTable (game.current_screen.sound_groups, function (sound_group)
    ForEachInTable (sound_group.emitters, function (emitter_desc)
      emitter_desc.emitter.Gain = sound_group.gain * emitter_desc.gain * sfx_volume * game.current_screen.sound_volume_correction
    end)
  end)
end

function SetSoundGroupGain (group_name, gain)
  if not game.current_screen or not game.current_screen.sound_groups [group_name] then return end
  
  local sfx_volume = GetGameProperty (GAME_SFX_VOLUME_ATTRIBUTE)

  if not sfx_volume then
    sfx_volume = 1
  end

  game.current_screen.sound_groups [group_name].gain = gain
  
  ForEachInTable (game.current_screen.sound_groups [group_name].emitters, function (emitter_desc)
    emitter_desc.emitter.Gain = emitter_desc.gain * gain * sfx_volume
  end)
end

function PauseSounds (dont_pause_music, force_pause)
  local current_time = game.main_timer.GetTime ()

  game.dont_pause_music = dont_pause_music

  if not dont_pause_music and game.music_emitter then
    game.music_offset = current_time - game.music_start_time
  
    game.music_emitter:Stop ()
    game.music_playing = false
  end

  if game.current_screen then
    ForEachInTable (game.current_screen.sound_groups, function (sound_group)
      for name, emitter_desc in pairs (sound_group.emitters) do
        local pause = true
      
        for i = 1, #DONT_PAUSE_SOUNDS do
          if name == DONT_PAUSE_SOUNDS [i] then
            pause = false
            break
          end
        end
      
        if pause or force_pause then
          emitter_desc.offset = current_time - emitter_desc.start_time
          emitter_desc.emitter:Stop ()
        end
      end
    end)
  end
end

function ResumeSounds ()
  local current_time = game.main_timer.GetTime ()

  ResumeMusic ()

  if game.current_screen then
    ForEachInTable (game.current_screen.sound_groups, function (sound_group)
      ForEachInTable (sound_group.emitters, function (emitter_desc)
        if emitter_desc.offset then
          emitter_desc.emitter:Play (emitter_desc.offset)
          emitter_desc.start_time = current_time - emitter_desc.offset
        end
      end)
    end)
  end
end

function PlayStoredSound (sound_name, gain)
  game.current_screen.sound_emitters [sound_name] = PlaySound (sound_name, gain)
end

function StopStoredSound (sound_name)
  local emitter = game.current_screen.sound_emitters [sound_name]

  if not emitter then return end
  
  emitter:Stop ()
  
  game.current_screen.sound_emitters [sound_name] = nil
  game.current_screen.sound_groups [DEFAULT_SOUND_GROUP_NAME].emitters [sound_name] = nil
end

function PlayMusic (sound_name)
  if game.music_emitter then
    if game.music_emitter.SoundDeclarationName == sound_name then return end
  
    game.music_emitter:Stop ()
    game.music_playing = false
  end
  
  game.music_emitter = Scene.SoundEmitter.Create (sound_name)
  game.music_emitter.Name = "MusicEmitter"
  game.music_emitter:BindToScene (game.scene) 

  if game.music_fade_action then
    game.music_fade_action.Stop ()
    game.music_fade_action = nil
  end

  game.music_volume_correction = 0
  UpdateMusicVolume ()

  game.music_fade_action = SlowAction ("default", game.current_screen.sound_fade_in_time, function (t)
    game.music_volume_correction = t
    UpdateMusicVolume ()
  end)

  game.music_emitter:Play ()
  game.music_playing = true 

  game.music_start_time = game.main_timer.GetTime ()
end

function FadeOutSound (duration)
  local initial_sound_volume = game.current_screen.sound_volume_correction

  SlowAction ("default", duration, function (t)
    game.current_screen.sound_volume_correction = initial_sound_volume * (1 - t)
    UpdateSoundsVolume ()
  end)
end

function UpdateMusicVolume ()
  if not game.music_emitter then return end
  
  local music_volume = GetGameProperty (GAME_MUSIC_VOLUME_ATTRIBUTE) * (game.music_volume_correction or 1)
    
  if not music_volume then 
    return
  end
  
  if GetGameProperty (GAME_SOUND_ENABLED_ATTRIBUTE) ~= "true" then
    music_volume = 0
  end

  game.music_emitter.Gain = music_volume + 0.0
end

function FadeOutMusic (duration)
  local initial_music_volume = game.music_volume_correction

  if game.music_fade_action then
    game.music_fade_action.Stop ()
    game.music_fade_action = nil
  end

  game.music_fade_action = SlowAction ("default", duration, function (t)
    game.music_volume_correction = initial_music_volume * (1 - t)
    UpdateMusicVolume ()
  end)
end

function StopMusic ()
  if not game.music_emitter then return end

  game.music_emitter:Stop ()
  game.music_emitter = nil
  game.music_playing = false
end

-- Обновление системных параметров
function UpdateGameOptions ()
  local is_full_screen = GetGameProperty ("Fullscreen") == "1" 

  if game.is_full_screen ~= is_full_screen then
    Engine.SubsystemManager:Restart ("config.xml", "*Render* Window* *Input* ScenePlayer")

    if game.music_emitter then
      game.music_emitter:Play ()    
      game.music_playing = true 
    end
    
    ForEachInTable (game.current_screen.sound_groups, function (sound_group)
      ForEachInTable (sound_group.emitters, function (emitter_desc)
        emitter_desc.emitter:Play ()
        emitter_desc.emitter.Gain = 0
      end)
    end)
    
    DelayedAction ("ui", 2, function ()
      local sfx_volume = GetGameProperty (GAME_SFX_VOLUME_ATTRIBUTE)

      if not sfx_volume then
        sfx_volume = 1
      end

      SlowAction ("ui", 0.5, function (t)
        ForEachInTable (game.current_screen.sound_groups, function (sound_group)
          ForEachInTable (sound_group.emitters, function (emitter_desc)
            emitter_desc.emitter.Gain = sound_group.gain * emitter_desc.gain * t * sfx_volume
          end)
        end)
      end)
    end)
  end  

  game.is_full_screen = is_full_screen
end

-- Обработка события закрытия окна
function OnWindowClose ()
  print ("OnWindowClose")
  
  System.Application.Exit (0) 
end

-- Обработка события выхода из приложения
function OnApplicationExit ()
  if game.current_screen.on_lost_current then
    game.current_screen.on_lost_current(game.current_screen)
  end

  ForEachInTable (GetReverseTable (game.scripts), function (script)
    if script.OnLostScreen then
      script.OnLostScreen (game.current_screen)
    end
  end)  
end

-- Загрузка игры
function LoadGameScreens (game)
  game.particles_declarations = {}
  game.screen_decls           = {}
  game.include_list           = {}

  local root = Common.StringNode.LoadXml (GAME_DESC)
  
  if root.Name ~= "Game" then
    print (string.format ("No 'Game' root node in file '%s'", file_name))
    return
  end
  
  local children_count = root.ChildrenCount

  game.start_screen_name   = root:Get ("StartScreen")
  game.loading_screen_name = root:Get ("LoadingScreen", "")

  for i=0,children_count-1 do
    local node = root:Child (i)
    
    if node.Name == "AnimationScreens" then
      local animation_screens_file_name = node:Get ("Source")
    
      LoadAnimationScreens (game.screen_decls, animation_screens_file_name, game.main_timer)
    end
  end
end

function LoadGame ()
  local game = {}
  
  game.events                 = {}  
  game.scripts                = {}
  game.scene                  = Scene.Scene.Create ()    
  game.main_timer             = CreateTimer (false, true)  
  game.loaded_resources       = {}
  game.next_update_actions    = {}
  game.input_queue            = {}

  game.slot_pathes                = {}
  game.slot_mount_points          = {}
  game.slot_previous_pathes       = {}
  game.slot_previous_mount_points = {}
  
  game.font_library         = Engine.FontLibraries.Get ("FontLibrary")
  game.font_creation_params = Media.FontCreationParams.Create ()

  game.font_creation_params.FontSize    = 60  --TODO maybe this should be different for different resolutions
  game.font_creation_params.FontSizeEps = 60

  local scene_light = Scene.SpotLight.Create ()
  
  scene_light:BindToScene (game.scene, Scene.NodeBindMode.AddRef)    

--[[  FPS_LINE=CreateTextLine(vec3(45,45,1), vec3(20,-50,0))
  FPS_LINE.Text="GAMEFPS"
  FPS_LINE.Color=vec4(1,1,0,1)
  
  if GetProfileProperty ("ShowFps") == "true" then
    FPS_LINE:BindToScene(game.scene)
  end]]

  LoadGameScreens (game)
  
  return game
end

-- Добавление путей поиска файлов
local function GetResolutionSearchPathes ()
  local ART_RESOURCES = {
    {
      width  = 480,
      height = 320,
      device = "phone",
      path   = "480x320"
    },
    {
      width  = 960,
      height = 640,
      device = "phone",
      path   = "960x640"
    },
    {
      width  = 1024,
      height = 768,
      device = "pad",
      path   = "1024x768"
    },
  }

  local window      = Engine.Windows.Find (WINDOW_NAME)
  local width       = window.ClientWidth
  local height      = window.ClientHeight
  local side        = math.max (width, height)
  local platform    = GetPlatform ()
  local choosen_art = #ART_RESOURCES

  for i = 1, #ART_RESOURCES do
    if ART_RESOURCES [i].device == platform then
      choosen_art = i
    end
  end

  for i = 1, #ART_RESOURCES do
    if ART_RESOURCES [i].width >= side and ART_RESOURCES [i].device == platform then
      choosen_art = i
      break
    end
  end

  local best_texture_format = ApplicationDataGet (TEXTURE_FORMATS) [1]
    
  local resources = ART_RESOURCES [choosen_art]

  return { string.format ("resolutions/%s-%s-%s", resources.device, best_texture_format, resources.path), 
           string.format ("resolutions/%s-generic-%s", resources.device, resources.path) }
end

function AddResolutionSearchPathes (prefix)
  local search_pathes = GetResolutionSearchPathes ()

  for i = 1, #search_pathes do
    Common.File.AddSearchPath (prefix .. search_pathes [i])
    print ("Add search path " .. prefix .. search_pathes [i])
  end
end

function StartProfiler ()
  Profiler.Start ()
  
  PROFILER_ENABLED = true
end

function PrintStatCounters ()
  local function PrintCounter (group, name, value)
    print (string.format ("  name=%s, value=%u", name, value))
  end
  
  print ("Stat counters:")
  
  Debug.EnumStatCounters (Debug.CreateStatCountersEnumerator (PrintCounter))
end

local function OnOpenUrl (url)
  local REFERRERS = { "ADX1144://", "ADX1145://" }

  for i = 1, #REFERRERS do
    local start, finish = string.find (url, REFERRERS [i])
    
    if start then
      local referrer = string.sub (url, start)
      
      SetProfileProperty (REFERRER_ATTRIBUTE, referrer)
    end
  end
  
  if string.find (url, "ApplicationOpenURL fb") == 1 then
    local last_logged_in_network = GetProfileProperty (LAST_LOGGED_IN_NETWORK)

    if not last_logged_in_network or last_logged_in_network == SOCIAL_MANAGER_FACEBOOK_ENGINE then
      local fb_token = GetRequestParam (url, "access_token")
    
      if fb_token then
        SetProfileProperty (LAST_LOGGED_IN_NETWORK, SOCIAL_MANAGER_FACEBOOK_ENGINE)
        SetProfileProperty (FACEBOOK_LOGIN_TOKEN_ATTRIBUTE, fb_token)
      end
    end
  end  
end

local function OnSystemAlertButtonClicked (notification)
  notification = string.sub (notification, string.len ("SystemAlertButtonClicked ") + 1)
  
  if DEBUG then
    print ("system alert button clicked '" .. notification .. "'")
  end
  
  local args         = SplitString (notification, " ", true)  
  local alert_tag    = args [1]
  local button_index = args [2] + 0

  if alert_tag == "RateApplication" then
    if button_index == 1 then           --rate
      local os = ApplicationDataGet (OS_TYPE)

      local rate_url

      local system_properties          = System.Application.SystemProperties
      local app_version_property_name  = "ApplicationVersion"
      local app_version                = "" 

      if system_properties:IsPresent (app_version_property_name) then
        app_version = system_properties:GetString (app_version_property_name)
      end
      
      if os == "iOS" then
        local os_version = SplitString (system_properties:GetString ("OSVersion"), ".", true) [1] + 0
      
        if os_version > 6 then
          if GetPlatform () == "pad" then
            rate_url = "itms-apps://itunes.apple.com/app/id586574454"
          else
            rate_url = "itms-apps://itunes.apple.com/app/id543831789"
          end
        else
          if GetPlatform () == "pad" then
            rate_url = "itms-apps://itunes.apple.com/WebObjects/MZStore.woa/wa/viewContentsUserReviews?type=Purple+Software&id=586574454"
          else
            rate_url = "itms-apps://itunes.apple.com/WebObjects/MZStore.woa/wa/viewContentsUserReviews?type=Purple+Software&id=543831789"
          end
        end
      elseif os == "Android" then
        if GetStoreManager ().GetStoreName () == AMAZON_STORE_NAME then
          rate_url = "amzn://apps/android?p=com.plarium.playslots"
        else
          rate_url = "market://details?id=com.plarium.playslots"
        end
      end

      SetProfileProperty (APPLICATION_RATED_ATTRIBUTE, app_version)

      if rate_url then
        System.Application.OpenUrl (rate_url)
      end
    elseif button_index == 0 then
      SetProfileProperty (APPLICATION_RATE_AFTER_ACTIVATIONS_ATTRIBUTE, (GetProfileProperty (APPLICATION_ACTIVATIONS_COUNT_ATTRIBUTE) or 0) + 4)
    else
      print ("Unknown button index received from rate alert")
    end
  end
end

local function OnGooglePlusClickedClicked ()
  local lobby_bonuses = GetLobbyBonuses ()

  for i = 1, #lobby_bonuses do
    local bonus = lobby_bonuses [i]
      
    if bonus.type == 3 and not bonus.expired then
      HsConnectionSendMessage (1, { event = "bonus.collect", type = 3, isCollect = true })
      return
    end
  end    
end

local function OnAdvertisingIdReceived (notification)
  ApplicationDataSet (ADVERTISING_IDENTIFIER, string.sub (notification, string.len ("AdvertisingIdRecieved ") + 1))
end

local function OnGooglePlayServicesStatusReceived (notification)
  ApplicationDataSet (GOOGLE_PLAY_SERVICES_STATUS, string.sub (notification, string.len ("GooglePlayServicesStatus ") + 1) == "1")
end

-- Точка входа
function Main()
  if not global then global = {} end   

  local cache_version_path = CACHE_PATH .. "cache_version.txt"
  local cache_version      = 0

  if Common.File.IsFileExist (cache_version_path) then
    cache_version = Common.File.LoadString (cache_version_path) + 0
  end
  
  if cache_version < DOWNLOADS_VERSION then
    if Common.File.IsFileExist (CACHE_PATH) then
      Common.File.Remove (CACHE_PATH)
    end

    if not Common.File.IsFileExist (CACHE_PATH) then
      Common.File.Mkdir (CACHE_PATH)
    end

    Common.File.PostString (cache_version_path, tostring (DOWNLOADS_VERSION))
  end  

  --Поиск поддерживаемых форматов текстур

  print ("System properties: ")
  
  local system_properties = System.Application.SystemProperties
  local gl_extensions     = ""

  system_properties:SetInteger ("ApplicationIconBadgeNumber", 0);

  local os_type = system_properties:GetString ("Operating System")

  if os_type == "Android" then
    Social.Session.Create ("Facebook")  --init facebook here so login will work correctly
  end
  
  ApplicationDataSet (OS_TYPE, os_type)

  for i = 0, system_properties.Size - 1 do
    local name = system_properties:GetPropertyName (i)

    if string.match (name, "GlDevice%d%d\.Extensions") then
      gl_extensions = system_properties:GetString (name)
    end

    print (string.format ("Property '%s' = '%s'", name, system_properties:GetString (name)))
  end

  local GL_TEXTURE_FORMATS = {}
  
  if os_type == "iOS" then
    GL_TEXTURE_FORMATS = {
      {
        extension = { "GL_IMG_texture_compression_pvrtc" },
        path      = "pvr",
      },
    }
  elseif os_type == "Android" then
    GL_TEXTURE_FORMATS = {
      {
        extension = { "GL_OES_compressed_ETC1_RGB8_texture" },
        path      = "etc",
      },
    }
  else
    GL_TEXTURE_FORMATS = {
      {
        extension = { "GL_OES_compressed_ETC1_RGB8_texture" },
        path      = "etc",
      },
      {
        extension = { "GL_EXT_texture_compression_s3tc" },
        path      = "dxt",
      },
      {
        extension = { "GL_IMG_texture_compression_pvrtc" },
        path      = "pvr",
      },
    }
  end

  gl_extensions = SplitString (gl_extensions, " ", true)
  
  local texture_formats = {}

  for i = 1, #gl_extensions do
    local extension = gl_extensions [i]
    
    for j = 1, #GL_TEXTURE_FORMATS do
      local format = GL_TEXTURE_FORMATS [j]
      
      for k = 1, #format.extension do
        if extension == format.extension [k] then
          table.insert (texture_formats, format.path)
        end
      end
    end
  end

  table.insert (texture_formats, "png")
  
  ApplicationDataSet (TEXTURE_FORMATS, texture_formats)
  
  AddResolutionSearchPathes ("")

  -- Подписка на выход из приложения

  System.Application.RegisterEventHandler (System.ApplicationEvent.OnExit, System.Application.CreateEventHandler ("OnApplicationExit"))  
  System.Application.RegisterEventHandler (System.ApplicationEvent.OnPause, System.Application.CreateEventHandler ("OnApplicationDeactivate"))    
  System.Application.RegisterEventHandler (System.ApplicationEvent.OnResume, System.Application.CreateEventHandler ("OnApplicationActivate"))
  System.Application.RegisterEventHandler (System.ApplicationEvent.OnStop, System.Application.CreateEventHandler ("OnApplicationDeactivate"))    
  System.Application.RegisterEventHandler (System.ApplicationEvent.OnStart, System.Application.CreateEventHandler ("OnApplicationActivate"))

  -- Загрузка профилей игроков

  LoadPlayersProfiles ()
  DeleteProfileProperties ("inventory")  --удаление предметов инвентаря.
  
  local localization
  
  if system_properties:IsPresent ("Language") then
    localization = system_properties:GetString ("Language")
  end

  localization = GetProfileProperty (LANGUAGE_ATTRIBUTE) or localization

  LoadLocalization (localization)

  -- Загрузка игры
  
  game = LoadGame ()

  -- Инициализация графики

  InitRender ()
  
  if not (is_initialized) then       
    -- Инициализация ввода
    
    InitInput ()  
    
    -- Рестарт указанных выше служб не нужен
    
    is_initialized = true
  end  
  
  -- Инициализация звука

  InitSoundPlayer()

  -- Запуск игровой обработки  
  game.idle_connection = System.Application.RegisterEventHandler (System.ApplicationEvent.OnIdle, System.Application.CreateEventHandler ("Update"))    
  
  System.Application.RegisterNotificationHandler ("SystemAlertButtonClicked *", System.Application.CreateNotificationHandler (OnSystemAlertButtonClicked))
  System.Application.RegisterNotificationHandler ("ApplicationOpenURL *", System.Application.CreateNotificationHandler (OnOpenUrl))
  System.Application.RegisterNotificationHandler ("GooglePlusClicked", System.Application.CreateNotificationHandler (OnGooglePlusClickedClicked))
  System.Application.RegisterNotificationHandler ("AdvertisingIdRecieved *", System.Application.CreateNotificationHandler (OnAdvertisingIdReceived))

  System.Application.PostNotification ("GetAdvertisingId")

  if ApplicationDataGet (OS_TYPE) == "Android" then
    System.Application.RegisterNotificationHandler ("GooglePlayServicesStatus *", System.Application.CreateNotificationHandler (OnGooglePlayServicesStatusReceived))
    System.Application.PostNotification ("GetGooglePlayServicesStatus")
  end

  game.is_full_screen = GetGameProperty ("Fullscreen") == "1"
  
  game.default_startup_screen = game.start_screen_name
  
  collectgarbage ("setpause", 100)  
  
  if DEBUG then
    System.Application.PostNotification ("DebugMode#1")
  end

  ApplicationDataSet (PUSH_NOTIFICATIONS_SERVER_URL, "http://109.234.156.148:5222/")
  
  GetStoreManager () --initialize store

  if GetHsPort () ~= HS_LIVE_PORT and GetProfileProperty (OVERRIDE_PORT_ATTRIBUTE) then
    SetHsPort (GetProfileProperty (OVERRIDE_PORT_ATTRIBUTE))
  end    

  InitFlurry ()

  System.Application.PostNotification ("LocalNotifications RemoveAll|Bonus")

  game.main_timer.Update ()
  
  OnApplicationActivate ()
end
