SOCIAL_MANAGER_FACEBOOK_ENGINE = "facebook"
SOCIAL_MANAGER_GUEST_ENGINE    = "guest"
SOCIAL_MANAGER_VK_ENGINE       = "vk"

SHARE_NEW_LEVEL_TYPE                    = 1
SHARE_FIVE_IN_ROW_TYPE                  = 2
SHARE_BONUS_GAME_TYPE                   = 3
SHARE_MEGA_BONUS_TYPE                   = 4
SHARE_ACHIEVEMENTS_GROUP_COINS_TYPE     = 5
SHARE_ACHIEVEMENT_COINS_TYPE            = 6
SHARE_BATTLE_WIN_TYPE                   = 7
SHARE_LEADER_WIN_COINS_TYPE             = 8
SHARE_TOP_COINS_100_TYPE                = 10
SHARE_TOP_COINS_10_TYPE                 = 11
SHARE_TOP_COINS_1_TYPE                  = 12
SHARE_ACHIEVEMENTS_GROUP_DIAMONDS_TYPE  = 105
SHARE_ACHIEVEMENT_DIAMONDS_TYPE         = 106
SHARE_LEADER_WIN_DIAMONDS_TYPE          = 108
SHARE_TOP_DIAMONDS_100_TYPE             = 110
SHARE_TOP_DIAMONDS_10_TYPE              = 111
SHARE_TOP_DIAMONDS_1_TYPE               = 112

local SHARE_VALUES = {}

SHARE_VALUES [SHARE_NEW_LEVEL_TYPE] = {
  players_count = 5,
  chips_count = 300 
}

SHARE_VALUES [SHARE_FIVE_IN_ROW_TYPE] = {
  players_count = 5,
  chips_count = 200 
}

SHARE_VALUES [SHARE_BONUS_GAME_TYPE] = {
  players_count = 5,
  chips_count = 200 
}

SHARE_VALUES [SHARE_MEGA_BONUS_TYPE] = {
  players_count = 5,
  chips_count = 200 
}

SHARE_VALUES [SHARE_ACHIEVEMENTS_GROUP_COINS_TYPE] = {
  players_count = 5,
  chips_count = 1000 
}

SHARE_VALUES [SHARE_ACHIEVEMENT_COINS_TYPE] = {
  players_count = 5,
  chips_count = 200 
}

SHARE_VALUES [SHARE_BATTLE_WIN_TYPE] = {
  players_count = 5,
  chips_count = 100 
}

SHARE_VALUES [SHARE_LEADER_WIN_COINS_TYPE] = {
  players_count = 5,
  chips_count = 1000 
}

SHARE_VALUES [SHARE_LEADER_WIN_DIAMONDS_TYPE] = {
  players_count = 5,
  chips_count = 5
}

SHARE_VALUES [SHARE_TOP_COINS_100_TYPE] = {
  players_count = 5,
  chips_count = 200 
}

SHARE_VALUES [SHARE_TOP_COINS_10_TYPE] = {
  players_count = 5,
  chips_count = 300 
}

SHARE_VALUES [SHARE_TOP_COINS_1_TYPE] = {
  players_count = 5,
  chips_count = 500 
}

SHARE_VALUES [SHARE_ACHIEVEMENTS_GROUP_DIAMONDS_TYPE] = {
  players_count = 5,
  chips_count = 10 
}

SHARE_VALUES [SHARE_ACHIEVEMENT_DIAMONDS_TYPE] = {
  players_count = 5,
  chips_count = 3 
}

SHARE_VALUES [SHARE_TOP_DIAMONDS_100_TYPE] = {
  players_count = 5,
  chips_count = 1 
}

SHARE_VALUES [SHARE_TOP_DIAMONDS_10_TYPE] = {
  players_count = 5,
  chips_count = 2 
}

SHARE_VALUES [SHARE_TOP_DIAMONDS_1_TYPE] = {
  players_count = 5,
  chips_count = 3 
}

local function OnActivate (impl) 
  if not impl.engine then return end
  
  if impl.engine.OnActivate then
    impl.engine.OnActivate ()
  end
end

local function Logout (impl, on_logout_handler)
  impl.engine.Logout (on_logout_handler)
end

local function Login (impl, on_login_handler, on_cancel_handler, logout_before_login)
  local function PerformLogin ()
    impl.engine.Login (function ()
      SetProfileProperty (LAST_LOGGED_IN_NETWORK, impl.engine_name)
    
      if on_login_handler then
        on_login_handler () 
        
        if impl.engine_name == SOCIAL_MANAGER_VK_ENGINE then
          System.Application.PostNotification ("OnWindowFocusChangedAndroid5")
        end
      end
    end, on_cancel_handler)
  end

  LogFlurryEvent ("LogIn", { NetworkName = impl.engine.GetNetworkName () })
  LogFlurryEvent ("LogInTo" .. impl.engine.GetNetworkName ())

  if logout_before_login then
    impl.engine.Logout (function ()
      PerformLogin ()
    end)
  else
    PerformLogin ()
  end
end

local function GetFirstName (impl)
  return impl.engine.GetFirstName ()
end

local function GetLastName (impl)
  return impl.engine.GetLastName ()
end

local function PreShareChips (impl)
  if impl.engine_name ~= SOCIAL_MANAGER_GUEST_ENGINE then return true end
  
  game.scripts.ui_windows.ShowLockedFunctionalityWindow (8)

  return false
end

local function GetShareValues (impl, share_type)
  return SHARE_VALUES [share_type]
end

local function ShareBattleWin (impl, on_shared_handler) 
  local share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_BATTLE_WIN_DESCRIPTION_FORMAT"), {
    { token = "{userName}",  replacement = GetFirstName (impl) }
  })

  game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)

  impl.engine.ShareChips (GetLocalizedString ("SHARE_BATTLE_WIN_MESSAGE_FORMAT"), SHARE_BATTLE_WIN_TYPE, nil, function ()
    DelayedAction ("ui", 1, function ()
      game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("SHARE_PUBLISHED") }, nil, nil, true, nil, nil, on_shared_handler)
    end)
  end, share_description, nil, on_shared_handler)
end

local function ShareNewLevel (impl, level, coins)
  if not PreShareChips (impl) then return end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_NEW_LEVEL_DESCRIPTION_FORMAT"), {
    { token = "{userName}",  replacement = GetFirstName (impl) },
    { token = "{coins}",     replacement = FormatMoneyString (coins - coins % 100) },
    { token = "{level}",     replacement = level }
  })
  impl.share_message         = string.format (GetLocalizedString ("SHARE_NEW_LEVEL_MESSAGE_FORMAT"), SHARE_VALUES [SHARE_NEW_LEVEL_TYPE].players_count, SHARE_VALUES [SHARE_NEW_LEVEL_TYPE].chips_count)
  impl.share_type            = SHARE_NEW_LEVEL_TYPE
  impl.share_additional_info = level + 0

  HsConnectionSendMessage (1, { event = "share.post", type = SHARE_NEW_LEVEL_TYPE })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareBonusEnded (impl, bonus_name, coins, hs_path)
  if not PreShareChips (impl) then return end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_BONUS_ENDED_DESCRIPTION_FORMAT"), {
    { token = "{userName}",  replacement = GetFirstName (impl) },
    { token = "{coins}",     replacement = FormatMoneyString (coins - coins % 100) },
    { token = "{bonusName}", replacement = bonus_name }
  })
  impl.share_message         = string.format (GetLocalizedString ("SHARE_BONUS_ENDED_MESSAGE_FORMAT"), SHARE_VALUES [SHARE_BONUS_GAME_TYPE].players_count, SHARE_VALUES [SHARE_BONUS_GAME_TYPE].chips_count)
  impl.share_type            = SHARE_BONUS_GAME_TYPE
  impl.share_additional_info = { hs_path = hs_path }

  HsConnectionSendMessage (1, { event = "share.post", type = SHARE_BONUS_GAME_TYPE })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareFiveInRow (impl, coins, hs_path, symbol_id)
  if not PreShareChips (impl) then return end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_FIVE_IN_ROW_DESCRIPTION_FORMAT"), {
    { token = "{userName}",  replacement = GetFirstName (impl) },
    { token = "{coins}",     replacement = FormatMoneyString (coins - coins % 100) }
  })
  impl.share_message         = string.format (GetLocalizedString ("SHARE_FIVE_IN_ROW_MESSAGE_FORMAT"), SHARE_VALUES [SHARE_FIVE_IN_ROW_TYPE].players_count, SHARE_VALUES [SHARE_FIVE_IN_ROW_TYPE].chips_count)
  impl.share_type            = SHARE_FIVE_IN_ROW_TYPE
  impl.share_additional_info = { hs_path = hs_path, symbol_id = symbol_id }

  HsConnectionSendMessage (1, { event = "share.post", type = SHARE_FIVE_IN_ROW_TYPE })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareLeaderWin (impl, win_diamonds, place, rating_name) 
  if not PreShareChips (impl) then return end

  local share_type           = SHARE_LEADER_WIN_COINS_TYPE
  local message_format_index = 1
  
  if win_diamonds then
    share_type           = SHARE_LEADER_WIN_DIAMONDS_TYPE
    message_format_index = 2
  end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_LEADER_WIN_DESCRIPTION_FORMAT"), {
    { token = "{userName}",   replacement = GetFirstName (impl) },
    { token = "{ratingName}", replacement = rating_name },
    { token = "{place}",      replacement = tostring (place) }
  })
  impl.share_message         = string.format (GetLocalizedString ("SHARE_LEADER_WIN_MESSAGE_FORMAT_" .. message_format_index), SHARE_VALUES [share_type].players_count, SHARE_VALUES [share_type].chips_count)
  impl.share_type            = share_type
  impl.share_additional_info = place

  HsConnectionSendMessage (1, { event = "share.post", type = share_type })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareTop (impl, share_type, place, rating_name) 
  if not PreShareChips (impl) then return end

  local type

  if place < 4 then
    type = 1
  elseif place < 11 then
    type = 2
  else
    type = 3
  end

  local message_format_index = 1
  
  if share_type == SHARE_TOP_DIAMONDS_100_TYPE or share_type == SHARE_TOP_DIAMONDS_10_TYPE or share_type == SHARE_TOP_DIAMONDS_1_TYPE then
    message_format_index = 2
  end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_TOP_DESCRIPTION_FORMAT_" .. type), {
    { token = "{userName}",    replacement = GetFirstName (impl) },
    { token = "{rating_name}", replacement = rating_name },
    { token = "{place}",       replacement = tostring (place) }
  })
  impl.share_message         = string.format (GetLocalizedString ("SHARE_TOP_MESSAGE_FORMAT_" .. message_format_index), SHARE_VALUES [share_type].players_count, SHARE_VALUES [share_type].chips_count)
  impl.share_type            = share_type
  impl.share_additional_info = place

  HsConnectionSendMessage (1, { event = "share.post", type = share_type })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareMegaBonus (impl, coins)
  if not PreShareChips (impl) then return end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_MEGA_BONUS_DESCRIPTION_FORMAT"), {
    { token = "{userName}",  replacement = GetFirstName (impl) },
    { token = "{coins}",     replacement = FormatMoneyString (coins - coins % 100) }
  })
  impl.share_message     = string.format (GetLocalizedString ("SHARE_MEGA_BONUS_MESSAGE_FORMAT"), SHARE_VALUES [SHARE_MEGA_BONUS_TYPE].players_count, SHARE_VALUES [SHARE_MEGA_BONUS_TYPE].chips_count)
  impl.share_type        = SHARE_MEGA_BONUS_TYPE

  HsConnectionSendMessage (1, { event = "share.post", type = SHARE_MEGA_BONUS_TYPE })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareAchievement (impl, id, current, share_type) 
  if not PreShareChips (impl) then return end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_ACHIEVEMENT_" .. id .. "_DESCRIPTION_FORMAT"), {
    { token = "{userName}", replacement = GetFirstName (impl) },
    { token = "{money}",    replacement = FormatMoneyString (current, 5) },
    { token = "{x}",        replacement = tostring (current) }
  })
  impl.share_message     = string.format (GetLocalizedString ("SHARE_ACHIEVEMENT_MESSAGE_FORMAT_" .. share_type), SHARE_VALUES [share_type].players_count, SHARE_VALUES [share_type].chips_count)
  impl.share_type        = share_type
  impl.share_additional_info = id

  HsConnectionSendMessage (1, { event = "share.post", type = share_type })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function ShareAchievementGroup (impl, id, prize, share_type) 
  if not PreShareChips (impl) then return end

  impl.share_description = ProcessStringWithReplacements (GetLocalizedString ("SHARE_ACHIEVEMENT_GROUP_DESCRIPTION_FORMAT"), {
    { token = "{userName}",  replacement = GetFirstName (impl) },
    { token = "{coins}",  replacement = FormatMoneyString (prize) },
    { token = "{groupName}", replacement = GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. id) }
  })
  impl.share_message     = string.format (GetLocalizedString ("SHARE_ACHIEVEMENT_GROUP_MESSAGE_FORMAT_" .. share_type), SHARE_VALUES [share_type].players_count, SHARE_VALUES [share_type].chips_count)
  impl.share_type        = share_type
  impl.share_additional_info = id

  HsConnectionSendMessage (1, { event = "share.post", type = share_type })
  
  if game.scripts.ui_windows then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("SHARE_IN_PROGRESS") }, true)
  end
end

local function OnSharePostIdReceived (impl, id, type)
  if not impl.share_type then return end

  impl.engine.ShareChips (impl.share_message, impl.share_type, id, function ()
    DelayedAction ("ui", 1, function ()
      game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("SHARE_PUBLISHED") }, nil, nil, true)
    end)
  end, impl.share_description, impl.share_additional_info)
  
  impl.share_additional_info = nil
  impl.share_type            = nil
end

local function LoadUserInfo (impl, handler)
  impl.engine.LoadUserInfo (handler)
end

local function GetNetworkName (impl)
  if not impl.engine then
    return ""
  end
  
  return impl.engine.GetNetworkName ()
end

local function GetNetworkHost (impl)
  return impl.servers [impl.engine_name].host
end

local function GetNetworkPort (impl, type)
  return impl.servers [impl.engine_name].port
end

local function SetServers (impl, vk, fb, guest) 
  impl.servers = {}
  
  local function ParseServer (url)
    local components = SplitString (url, ":", true)
  
    if #components ~= 2 then
      error ("Invalid server address (" .. url .. ") format, must be host:port")
    end
  
    return { host = components [1], port = components [2] }
  end
  
  impl.servers [SOCIAL_MANAGER_VK_ENGINE]       = ParseServer (vk)
  impl.servers [SOCIAL_MANAGER_FACEBOOK_ENGINE] = ParseServer (fb)
  impl.servers [SOCIAL_MANAGER_GUEST_ENGINE]    = ParseServer (guest)
  
  impl.servers_setted = true
  
  HsConnectToServer ()
end

local function IsServersSetted (impl) 
  return impl.servers_setted
end

local function GetPhoto (impl)
  return impl.engine.GetPhoto ()
end

local function GetUserGender (impl)
  if not impl.engine or not impl.engine.GetUserGender then
   return nil
  end

  return impl.engine.GetUserGender ()
end

local function GetUserId (impl)
  if not impl.engine then
    return nil
  end

  return impl.engine.GetUserId ()
end

local function GetUserInfo (impl)
  if not impl.engine then
    return nil
  end

  return impl.engine.GetUserInfo ()
end

local function GetFriends (impl)
  return impl.engine.GetFriends ()
end

local function HasInviteFriendsWindow (impl) 
  return impl.engine.HasInviteFriendsWindow ()
end

local function InviteFriends (impl, friends_list, finish_handler)
  impl.engine.InviteFriends (friends_list, finish_handler)
end

local function ShowInviteFriendsWindow (impl)
  impl.engine.ShowInviteFriendsWindow (impl)
end

local function IsLoggedIn (impl)
  return impl.engine.IsLoggedIn ()
end

local function FillAuthMessage (impl, message)
  impl.engine.FillAuthMessage (message)
end

local function CancelLogin (impl)
  if impl.engine.CancelLogin then
    impl.engine.CancelLogin ()
  end
end

local function SetEngine (impl, engine_name)
  if impl.engine_name == engine_name then return end
  
  if impl.engine then
    impl.engine.Logout ()
    impl.engine = nil
  end

  if engine_name == SOCIAL_MANAGER_GUEST_ENGINE then
    impl.engine = CreateSocialGuestEngine ()
  elseif engine_name == SOCIAL_MANAGER_VK_ENGINE then
    impl.engine = CreateSocialVKEngine ()
  elseif engine_name == SOCIAL_MANAGER_FACEBOOK_ENGINE then
    impl.engine = CreateSocialFacebookEngine ()
  elseif engine_name == nil then
    -- Do nothing
  else
    error ("Unknown social engine " .. engine_name)
  end
  
  impl.engine_name = engine_name
end

local function CreateSocialManager ()
  local impl = {}

  local public_methods = {}

  public_methods.CancelLogin             = function () CancelLogin (impl) end
  public_methods.FillAuthMessage         = function (message) FillAuthMessage (impl, message) end
  public_methods.GetFirstName            = function () return GetFirstName (impl) end
  public_methods.GetFriends              = function () return GetFriends (impl) end
  public_methods.GetLastName             = function () return GetLastName (impl) end
  public_methods.GetNetworkHost          = function () return GetNetworkHost (impl) end
  public_methods.GetNetworkName          = function () return GetNetworkName (impl) end
  public_methods.GetNetworkPort          = function (type) return GetNetworkPort (impl, type) end
  public_methods.GetPhoto                = function () return GetPhoto (impl) end
  public_methods.GetShareValues          = function (share_type) return GetShareValues (impl, share_type) end
  public_methods.GetUserGender           = function () return GetUserGender (impl) end
  public_methods.GetUserId               = function () return GetUserId (impl) end
  public_methods.GetUserInfo             = function () return GetUserInfo (impl) end
  public_methods.LoadUserInfo            = function (handler) LoadUserInfo (impl, handler) end
  public_methods.HasInviteFriendsWindow  = function () return HasInviteFriendsWindow (impl) end
  public_methods.InviteFriends           = function (friends_list, finish_handler) InviteFriends (impl, friends_list, finish_handler) end
  public_methods.IsInGuestMode           = function () return impl.engine_name == SOCIAL_MANAGER_GUEST_ENGINE end
  public_methods.IsLoggedIn              = function () return IsLoggedIn (impl) end
  public_methods.IsServersSetted         = function () return IsServersSetted (impl) end
  public_methods.Login                   = function (login_handler, cancel_handler, logout_before_login) Login (impl, login_handler, cancel_handler, logout_before_login) end
  public_methods.Logout                  = function (handler) Logout (impl, handler) end
  public_methods.OnActivate              = function () OnActivate (impl) end
  public_methods.OnSharePostIdReceived   = function (id, type) OnSharePostIdReceived (impl, id, type) end
  public_methods.SetEngine               = function (engine) SetEngine (impl, engine) end
  public_methods.SetServers              = function (vk, fb, guest) SetServers (impl, vk, fb, guest) end
  public_methods.ShareAchievement        = function (id, current, share_type) ShareAchievement (impl, id, current, share_type) end
  public_methods.ShareAchievementGroup   = function (id, prize, share_type) ShareAchievementGroup (impl, id, prize, share_type) end
  public_methods.ShareBattleWin          = function (on_shared_handler) ShareBattleWin (impl, on_shared_handler) end
  public_methods.ShareNewLevel           = function (level, coins) ShareNewLevel (impl, level, coins) end
  public_methods.ShareFiveInRow          = function (coins, hs_path, symbol_id) ShareFiveInRow (impl, coins, hs_path, symbol_id) end
  public_methods.ShareBonusEnded         = function (bonus_name, coins, hs_path) ShareBonusEnded (impl, bonus_name, coins, hs_path) end
  public_methods.ShareLeaderWin          = function (win_diamonds, place, rating_name) ShareLeaderWin (impl, win_diamonds, place, rating_name) end
  public_methods.ShareMegaBonus          = function (coins) ShareMegaBonus (impl, coins) end
  public_methods.ShareTop                = function (share_type, place, rating_name) ShareTop (impl, share_type, place, rating_name) end
  public_methods.ShowInviteFriendsWindow = function () ShowInviteFriendsWindow (impl) end
    
  return public_methods
end

local SocialManagerInstance

function GetSocialManager ()
  if not SocialManagerInstance then
    SocialManagerInstance = CreateSocialManager ()
  end
  
  return SocialManagerInstance
end
