local NETWORK_NAME = "GUEST"

local function Logout (impl, on_logout_handler)
  impl.is_logged_in = false

  if on_logout_handler then
    on_logout_handler ()
  end
end

local function Login (impl, on_login_handler)
  SetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE)

  impl.is_logged_in = true
  
  if on_login_handler then
    on_login_handler ()
  end
end

local function LoadUserInfo (impl, handler)
  local os = ApplicationDataGet (OS_TYPE)

  if string.find (os, "Windows") ~= nil or string.find (os, "Mac OS X") ~= nil or string.find (os, "TabletOS") ~= nil then
    impl.user_id = GUEST_LOGIN_ID or "GuestTestId"
  else
    local system_properties = System.Application.SystemProperties
    local property_name     = "UUID"

    if system_properties:IsPresent (property_name) then
      impl.user_id = system_properties:GetString (property_name)
    else
      impl.user_id = "GuestTestId"
    end
  end

  handler ()
end

local function GetFirstName (impl)
  return GetLocalizedString ("GUEST_FIRST_NAME")
end

local function GetLastName (impl)
  return GetLocalizedString ("GUEST_LAST_NAME")
end

local function GetNetworkName (impl)
  return NETWORK_NAME
end

local function GetPhoto (impl)
  return ""
end

local function GetUserId (impl)
  return impl.user_id
end

local function GetUserInfo (impl)
  return nil
end

local function IsLoggedIn (impl)
  return impl.is_logged_in
end

local function FillAuthMessage (impl, message)
  message.network = NETWORK_NAME
  message.loginId = impl.user_id
end

function CreateSocialGuestEngine ()
  local impl = {}
 
  local public_methods = {}

  public_methods.FillAuthMessage        = function (message) FillAuthMessage (impl, message) end
  public_methods.GetFirstName           = function () return GetFirstName (impl) end
  public_methods.GetFriends             = function () return {} end
  public_methods.GetLastName            = function () return GetLastName (impl) end
  public_methods.GetNetworkName         = function () return GetNetworkName (impl) end
  public_methods.GetPhoto               = function () return GetPhoto (impl) end
  public_methods.GetUserId              = function () return GetUserId (impl) end
  public_methods.GetUserInfo            = function () return GetUserInfo (impl) end
  public_methods.LoadUserInfo           = function (handler) LoadUserInfo (impl, handler) end
  public_methods.HasInviteFriendsWindow = function () return false end
  public_methods.IsLoggedIn             = function () return IsLoggedIn (impl) end
  public_methods.Login                  = function (handler) Login (impl, handler) end
  public_methods.Logout                 = function (handler) Logout (impl, handler) end
    
  return public_methods
end

