local NETWORK_NAME = "VK"

local SHARE_NEW_LEVEL_IMAGE          = "-40786235_305706762"
local SHARE_FIVE_IN_ROW_IMAGE        = "-40786235_305706760"
local SHARE_BONUS_GAME_IMAGE         = "-40786235_305706758"
local SHARE_MEGA_BONUS_IMAGE         = "-40786235_305706766"
local SHARE_ACHIEVEMENTS_GROUP_IMAGE = "-40786235_305706768"
local SHARE_BATTLE_WIN_IMAGE         = "-40786235_305706757"
local SHARE_TOP_1_IMAGE              = "-40786235_310475146"
local SHARE_TOP_2_IMAGE              = "-40786235_310475148"
local SHARE_TOP_3_IMAGE              = "-40786235_310475150"
local SHARE_TOP_10_IMAGE             = "-40786235_310475151"
local SHARE_TOP_100_IMAGE            = "-40786235_310475152"

local SHARE_ACHIEVEMENT_IMAGE = {
  "-40786235_305706712",
  "-40786235_305706713",
  "-40786235_305706715",
  "-40786235_305706716",
  "-40786235_305706717",
  "-40786235_305706718",
  "-40786235_305706719",
  "-40786235_305706720",
  "-40786235_305706721",
  "-40786235_305706722",
  "-40786235_305706723",
  "-40786235_305706724",
  "-40786235_305706725",
  "-40786235_305706727",
  "-40786235_305706729",
  "-40786235_305706730",
  "-40786235_305706731",
  "-40786235_305706732",
  "-40786235_305706733",
  "-40786235_305706734",
  "-40786235_305706735",
  "-40786235_305706737",
  "-40786235_308456265",
  "-40786235_308456266",
  "-40786235_308456268",
  "-40786235_308456269",
  "-40786235_308456270"
}

local SHARE_LEADER_WIN_IMAGE = {
  "214078742_307456179",
  "214078742_307456180",
  "214078742_307456181"
}

local VK_CLIENT_ID      = "3178038"
local VK_GROUP_ID       = "plariumslotsgroup"
local VK_LOGIN_REQUEST  = "https://oauth.vk.com/authorize?client_id=" .. VK_CLIENT_ID .. "&scope=friends,wall,messages,groups&redirect_uri=https://oauth.vk.com/blank.html&display=touch&response_type=token"
local VK_LOGOUT_REQUEST = "https://oauth.vk.com/logout?client_id=" .. VK_CLIENT_ID
local VK_REQUEST_FORMAT = "https://api.vk.com/method/%s?%s&access_token=%s"

local on_plarium_token_obtained_handler

local function GetVkShareAppId ()
  local hs_port = GetHsPort ()

  if hs_port == HS_LIVE_PORT then
    return "3037478"
  elseif hs_port == HS_QA_PORT then
    return "3034728"
  else
    return "2890865"
  end
end

function OnPlariumTokenObtained (token)
  SetProfileProperty (PLARIUM_TOKEN_ATTRIBUTE, token)
  
  if on_plarium_token_obtained_handler then
    on_plarium_token_obtained_handler ()
    
    on_plarium_token_obtained_handler = nil
  end
end

local function SetLoginInfo (impl, uid, token)
  impl.uid   = uid
  impl.token = token
  
  if not uid then
    impl.user_info    = nil
    impl.is_logged_in = false
  else
    impl.is_logged_in = true
  end
end

local function CanLogin ()
  local os = ApplicationDataGet (OS_TYPE)

  return string.find (os, "Windows") == nil
end

local function Logout (impl, on_logout_handler)
  impl.is_logged_in = false

  SetProfileProperty (PLARIUM_TOKEN_ATTRIBUTE, nil)
  SetProfileProperty (VK_LOGIN_ID_ATTRIBUTE, nil)
  SetProfileProperty (VK_LOGIN_TOKEN_ATTRIBUTE, nil)

  SetLoginInfo (impl, nil, nil)

  if CanLogin () then
    System.CookieManager.DeleteAllCookies ()
  end
  
  if on_logout_handler then
    on_logout_handler ()
  end
end

local function OnActivate (impl)
  DelayedAction (game.main_timer, 0.1, function ()
    if not impl.login_view then return end

    local login_view_window = impl.login_view.Window
  
    login_view_window.Visible = true
    login_view_window.Active  = true
    login_view_window.Focus   = true
    login_view_window:Maximize ()
  end)
end

local function CancelLogin (impl)
  if not impl.login_view then return end

  if impl.load_failed_connection then
    impl.load_failed_connection:Disconnect ()
    impl.load_failed_connection = nil
  end

  if impl.login_view_filter then
    impl.login_view_filter:Disconnect ()
    impl.login_view_filter  = nil
  end

  if impl.show_login_window_action then
    impl.show_login_window_action.Stop ()
    impl.show_login_window_action = nil
  end
        
  ExecuteOnNextUpdate (function ()
    ExecuteOnNextUpdate (function ()
      if not impl.login_view then return end

      local login_view_window = impl.login_view.Window

      login_view_window.Visible = false
      
      impl.login_view = nil
      collectgarbage ("collect")
    end, true)
  end, true)
end

local function ProcessLoginRequest (impl, request, on_login_handler, on_cancel_handler)
  local login_view_window 

  if impl.login_view then
    login_view_window = impl.login_view.Window
  end

  if DEBUG then
    print ("request = " .. request)
  end

  if string.find (request, "logout") or string.find (request, "error") or string.find (request, "act=blocked") then
    if on_cancel_handler then
      on_cancel_handler ()
    end
    
    return false
  elseif string.find (request, "https://vk.com/app") then
    if DEBUG then
      print ("request blocked")
    end

    return false
  elseif string.find (request, "https://m.vk.com/") and ApplicationDataGet (OS_TYPE) == "Android" then
    if DEBUG then
      print ("request blocked")
    end

    return false
  elseif string.find (request, "https://oauth.vk.com/blank.html") == 1 then
    -- print ("access_token = " .. GetRequestParam (request, "access_token"))
    local error
        
    pcall (function ()
      error = GetRequestParam (request, "error")
    end)
       
    if error then
      impl.login_view:LoadRequest (VK_LOGIN_REQUEST)
    else
      local user_id      = GetRequestParam (request, "user_id")
      local access_token = GetRequestParam (request, "access_token")
        
      SetProfileProperty (VK_LOGIN_ID_ATTRIBUTE, user_id)
      SetProfileProperty (VK_LOGIN_TOKEN_ATTRIBUTE, access_token)

      SetLoginInfo (impl, user_id, access_token)

      impl.is_logged_in = true

      if on_login_handler then
        on_login_handler ()
      end

      CancelLogin (impl)
    end

    return false
  end

  if not string.find (request, "https://oauth.vk.com/grant_access") and request ~= VK_LOGIN_REQUEST then
    OnActivate (impl)
  end

  if DEBUG then
    print ("request allowed")
  end
    
  return true
end

local function GetNetworkName (impl)
  return NETWORK_NAME
end

local function Login (impl, on_login_handler, on_cancel_handler, ask_for_token)
  System.Application.PostNotification ("LocalNotifications RemoveAll|SocialLogin")
  LogFlurryEvent ("SocialLogIn", { NetworkName = GetNetworkName () })  

  if not CanLogin () then
    if VK_LOGIN_TOKEN and VK_LOGIN_ID then
      SetProfileProperty (PLARIUM_TOKEN_ATTRIBUTE, VK_LOGIN_TOKEN)
      SetProfileProperty (VK_LOGIN_ID_ATTRIBUTE, VK_LOGIN_ID)
    end
  end

  local is_logged_in = GetProfileProperty (PLARIUM_TOKEN_ATTRIBUTE) and GetProfileProperty (VK_LOGIN_ID_ATTRIBUTE) and GetProfileProperty (VK_LOGIN_TOKEN_ATTRIBUTE)
  
  if not CanLogin () then
    is_logged_in = true
  end  
  
  if is_logged_in then
    SetLoginInfo (impl, GetProfileProperty (VK_LOGIN_ID_ATTRIBUTE), GetProfileProperty (VK_LOGIN_TOKEN_ATTRIBUTE))
  
    if ask_for_token then
      on_plarium_token_obtained_handler = on_login_handler
      System.Application.PostNotification ("GetPlariumToken#" .. GetProfileProperty (VK_LOGIN_ID_ATTRIBUTE))
    elseif on_login_handler then
      on_login_handler ()
    end
    
    return
  end

  impl.login_view = System.WebView.Create ()
  
  local login_view_window = impl.login_view.Window
  local handler           = on_login_handler
  
  if ask_for_token then
    handler = function ()
      on_plarium_token_obtained_handler = on_login_handler
      System.Application.PostNotification ("GetPlariumToken#" .. GetProfileProperty (VK_LOGIN_ID_ATTRIBUTE))
    end
  end
  
  impl.event_handler = impl.login_view.CreateEventHandler (function ()
    ProcessLoginRequest (impl, impl.login_view.Request, handler, on_cancel_handler)
  end)
    
  impl.login_view:RegisterEventHandler (System.WebViewEvent.OnLoadStart, impl.event_handler)
  impl.load_failed_connection = impl.login_view:RegisterEventHandler (System.WebViewEvent.OnLoadFail, impl.login_view.CreateEventHandler (function ()
    game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
      HsConnectionClearStoredMessages ()
      SetCurrentScreen ("main_menu")
    end, nil, GetLocalizedString ("CLOSE"), { GetLocalizedString ("CONNECTION_ERROR") })
    
    CancelLogin (impl)
  end))
    
  local login_view_filter = impl.login_view.CreateFilter (function (request)
    return ProcessLoginRequest (impl, request, handler, on_cancel_handler)
  end)
    
  impl.login_view_filter = impl.login_view:RegisterFilter (login_view_filter)
  
  impl.login_view:LoadRequest (VK_LOGIN_REQUEST)
  
  impl.show_login_window_action = DelayedAction (game.main_timer, 2, function ()
    OnActivate (impl)
  end)
end

local function FindAttribute (node, attribute_name)
  if node.Name == attribute_name then
    if node.AttributesCount > 0 then
      return node:Attribute (0)
    else
      return nil
    end
  end
  
  for current_child = 0, node.ChildrenCount - 1, 1 do
    local return_value = FindAttribute (node:Child (current_child), attribute_name)
    
    if return_value then 
      return return_value
    end
  end

  return nil
end

local function PerformVKRequest (impl, method_name, params, handler, retry_count)
  local url = string.format (VK_REQUEST_FORMAT, method_name, params, impl.token)
 
  if DEBUG then
    print ("Sending request to VK to url " .. url)
  end   

  url = GetPercentEscapedUrl (url)
 
  GetDownloadsManager ().LoadString (url, function (json_string)
    if DEBUG then
      print ("Answer from VK: " .. json_string)
    end

    local answer = Common.StringNode.LoadJsonFromString (json_string)
    
    local error = FindAttribute (answer, "error_code")
    
    if error then
      local function TryAgain (after_logout)
        SetProfileProperty (VK_LOGIN_TOKEN_ATTRIBUTE, nil)
      
        Login (impl, function ()
          if after_logout then
            on_plarium_token_obtained_handler = function ()
              HsConnectToServer ()
              SetCurrentScreen ("lobby")
            end
          
            System.Application.PostNotification ("GetPlariumToken#" .. GetProfileProperty (VK_LOGIN_ID_ATTRIBUTE))
          end
          
          if retry_count then
            retry_count = retry_count + 1
          else
            retry_count = 1
          end
          
          PerformVKRequest (impl, method_name, params, handler, retry_count)
        end, function ()
          HsConnectionClearStoredMessages ()
          SetCurrentScreen ("main_menu")
        end)
      end
    
      if retry_count and retry_count > 1 then
        if game.scripts.ui_windows then
          game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
            HsConnectionPopStoredMessages (AUTH_MESSAGE)
            HsConnectionReconnect ()              
            end, nil, GetLocalizedString ("RECONNECT"), { GetLocalizedString ("CONNECTION_ERROR") })
        else
          Logout (impl, function ()
            TryAgain (true)
          end)
        end
        
        return
      end
    
      if error == "7" or error == "14" then            --TODO captcha
        Logout (impl, function ()
          TryAgain (true)
        end)
      elseif error == "5" then
        if retry_count then
          Logout (impl, function ()
            TryAgain (true)
          end)
        else
          TryAgain ()
        end 
      elseif error == "125" then  -- invalid group
        handler (answer)
      elseif error == "9" then  --flood control
        handler (answer)
      else
        TryAgain ()
      end
    else
      handler (answer)
    end
  end, function ()
    game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
      HsConnectionPopStoredMessages (AUTH_MESSAGE)
      HsConnectionReconnect ()              
    end, nil, GetLocalizedString ("RECONNECT"), { GetLocalizedString ("CONNECTION_ERROR") })
  end)
end

local function ShareChips (impl, message, share_type, share_id, handler, description, additional_info)
  local url

  if share_type == SHARE_BATTLE_WIN_TYPE then
    url = "http://vk.com/app" .. GetVkShareAppId ()
  
    message = message .. "\n\n" .. description .. " " .. url
  else
    url = "http://vk.com/app" .. GetVkShareAppId () .. "?from_id=" .. impl.uid .. "%26loc=" .. share_id
  
    message = description .. "\n\n" .. message .. " " .. url
  end
  
  local share_image

  if share_type == SHARE_NEW_LEVEL_TYPE then
    share_image = SHARE_NEW_LEVEL_IMAGE
  elseif share_type == SHARE_FIVE_IN_ROW_TYPE then
    share_image = SHARE_FIVE_IN_ROW_IMAGE
  elseif share_type == SHARE_BONUS_GAME_TYPE then
    share_image = SHARE_BONUS_GAME_IMAGE
  elseif share_type == SHARE_MEGA_BONUS_TYPE then
    share_image = SHARE_MEGA_BONUS_IMAGE
  elseif share_type == SHARE_ACHIEVEMENT_COINS_TYPE or share_type == SHARE_ACHIEVEMENT_DIAMONDS_TYPE then
    share_image = SHARE_ACHIEVEMENT_IMAGE [additional_info + 1]
  elseif share_type == SHARE_ACHIEVEMENTS_GROUP_COINS_TYPE or share_type == SHARE_ACHIEVEMENTS_GROUP_DIAMONDS_TYPE then
    share_image = SHARE_ACHIEVEMENTS_GROUP_IMAGE
  elseif share_type == SHARE_BATTLE_WIN_TYPE then
    share_image = SHARE_BATTLE_WIN_IMAGE
  elseif share_type == SHARE_LEADER_WIN_COINS_TYPE or share_type == SHARE_LEADER_WIN_DIAMONDS_TYPE then
    share_image = SHARE_LEADER_WIN_IMAGE [additional_info]
  elseif share_type == SHARE_TOP_COINS_100_TYPE or share_type == SHARE_TOP_COINS_10_TYPE or share_type == SHARE_TOP_COINS_1_TYPE or
         share_type == SHARE_TOP_DIAMONDS_100_TYPE or share_type == SHARE_TOP_DIAMONDS_10_TYPE or share_type == SHARE_TOP_DIAMONDS_1_TYPE then
    if additional_info == 1 then
      share_image = SHARE_TOP_1_IMAGE
    elseif additional_info == 2 then
      share_image = SHARE_TOP_2_IMAGE
    elseif additional_info == 3 then
      share_image = SHARE_TOP_3_IMAGE
    elseif additional_info < 11 then
      share_image = SHARE_TOP_10_IMAGE
    else
      share_image = SHARE_TOP_100_IMAGE
    end
  else
    error ("Unknown share type " .. tostring (share_type))
  end

  PerformVKRequest (impl, "wall.post", string.format ("message=%s&attachments=photo%s", message, share_image), handler)
end

local function LoadUserInfo (impl, handler)
  if not CanLogin () then
    if handler then
      handler ()
    end

    return
  end

  if not impl.uid or not impl.token then
    error ("Can't get user info, VK token and uid not setted")
  end

  if impl.user_info then
    if handler then
      handler ()
    end
    
    return
  end
  
  PerformVKRequest (impl, "users.get", string.format ("uids=%s&fields=first_name,last_name,photo,sex,bdate", impl.uid), function (node)
    impl.user_info = {}
    
    impl.user_info.first_name = FindAttribute (node, "first_name")
    impl.user_info.last_name  = FindAttribute (node, "last_name")
    impl.user_info.photo      = FindAttribute (node, "photo")
    impl.user_info.birthday   = FindAttribute (node, "bdate")
    impl.user_info.sex        = FindAttribute (node, "sex")
    impl.user_info.friends    = {}

    if impl.user_info.sex then
      impl.user_info.sex = impl.user_info.sex - 1
    end

    PerformVKRequest (impl, "groups.isMember", "extended=2&gid=" .. VK_GROUP_ID, function (node)
      impl.user_info.in_group = FindAttribute (node, "member")
    
      PerformVKRequest (impl, "friends.get", "fields=first_name,last_name,photo", function (node)
        local response = node:Find ("response")
        
        if response then
          if response.ChildrenCount > 0 then
            local friends_list  = response:Child (0)
            local friends_count = friends_list.ChildrenCount
            
            for i = 0, friends_count - 1 do
              local friend_node = friends_list:Child (i)
              local friend      = {}
            
              local uid        = friend_node:Find ("uid")
              local first_name = friend_node:Find ("first_name")
              local last_name  = friend_node:Find ("last_name")
              local photo      = friend_node:Find ("photo")
            
              if uid and uid.AttributesCount > 0 then
                friend.uid = uid:Attribute (0)
              else
                friend.uid = 0
              end
            
              if first_name and first_name.AttributesCount > 0 then
                friend.first_name = first_name:Attribute (0)
              else
                friend.first_name = ""
              end
            
              if last_name and last_name.AttributesCount > 0 then
                friend.last_name = last_name:Attribute (0)
              else
                friend.last_name = ""
              end
            
              if photo and photo.AttributesCount > 0 then
                friend.photo = photo:Attribute (0)
              else
                friend.photo = ""
              end
            
              table.insert (impl.user_info.friends, friend)
            end
          end
        end
        
        if handler then
          handler ()
        end
      end)
    end)     
  end)
end

local function InviteFriends (impl, friends_list, finish_handler, index)
  if not CanLogin () then
    return
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
--[[  if not index then
    index = 1
  end]] 

  math.randomseed (game.main_timer.GetTime ())
    
  local invite_request_id = tostring (math.random (2000000000))
  local invite_url        = "http://vk.com/app" .. GetVkShareAppId () .. "?request_key=" .. invite_request_id    
  
  if #friends_list == 1 then
    PerformVKRequest (impl, "messages.send", string.format ("uid=%s&message=%s", friends_list [1], string.format (GetLocalizedString ("INVITE_FRIEND_MESSAGE_VK"), invite_url)), function ()
      HsReportFriendsInvited (invite_request_id, friends_list)

      if finish_handler then
        finish_handler ()
      end
    end)
  else
    local uids = ""
  
    for i = 1, #friends_list do
      if i > 1 then
        uids = uids .. ","
      end 
  
      uids = uids .. friends_list [i]
    end

    PerformVKRequest (impl, "messages.createChat", string.format ("uids=%s&title=%s", uids, GetLocalizedString ("INVITE_FRIEND_MESSAGE_VK")), function (node)
      local chat_id = node:Find ("response"):Attribute (0)
    
      PerformVKRequest (impl, "messages.send", string.format ("chat_id=%s&message=%s", chat_id, GetLocalizedString ("INVITE_FRIEND_MESSAGE_VK")), function ()
        HsReportFriendsInvited (invite_request_id, friends_list)

        if finish_handler then
          finish_handler ()
        end
      end)
    end)
  end
  
--[[  if index > #friends_list then
    if finish_handler then
      finish_handler ()
    end
     
    return 
  end
  
  PerformVKRequest (impl, "messages.send", string.format ("uid=%s&message=%s", friends_list [index], GetLocalizedString ("INVITE_FRIEND_MESSAGE_VK")), function ()
    InviteFriends (impl, friends_list, index + 1)
  end)]]
end

local function GetFirstName (impl)
  if not CanLogin () then
    return "Windows"
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
  return impl.user_info.first_name
end

local function GetLastName (impl)
  if not CanLogin () then
    return "Windows"
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
  return impl.user_info.last_name
end

local function GetPhoto (impl)
  if not CanLogin () then
    return "http://cs912.userapi.com/u93339557/a_57700b24.jpg"
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
  return impl.user_info.photo
end

local function GetUserGender (impl)
  if not CanLogin () then
    return nil
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
  return impl.user_info.sex
end

local function GetUserId (impl)
  if not CanLogin () then
    return VK_LOGIN_ID or ""
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
  return impl.uid
end

local function GetUserInfo (impl)
  return impl.user_info
end

local function GetFriends (impl)
  if not CanLogin () then
    return ""
  end

  if not impl.user_info then
    error ("User info not obtained")
  end
  
  return impl.user_info.friends
end

local function FillAuthMessage (impl, message)
  message.lastName  = GetLastName (impl)
  message.firstName = GetFirstName (impl)
  message.token     = GetProfileProperty (PLARIUM_TOKEN_ATTRIBUTE)
  message.photo     = GetPhoto (impl)
  message.network   = NETWORK_NAME
  message.loginId   = impl.uid
  
  if CanLogin () then
    message.sex       = impl.user_info.sex
    message.birthday  = impl.user_info.birthday
    message.inGroup   = impl.user_info.in_group == "1"
    
    local friends = {}
    
    for i = 1, #impl.user_info.friends do
      table.insert (friends, impl.user_info.friends [i].uid)
    end
    
    message.friends = friends
  end
end

local function IsLoggedIn (impl)
  return impl.is_logged_in
end

function CreateSocialVKEngine ()
  local impl = {}

  local public_methods = {}

  public_methods.CancelLogin            = function () CancelLogin (impl) end
  public_methods.FillAuthMessage        = function (message) FillAuthMessage (impl, message) end
  public_methods.GetFirstName           = function () return GetFirstName (impl) end
  public_methods.GetFriends             = function () return GetFriends (impl) end
  public_methods.GetLastName            = function () return GetLastName (impl) end
  public_methods.GetNetworkName         = function () return GetNetworkName (impl) end
  public_methods.GetPhoto               = function () return GetPhoto (impl) end
  public_methods.GetUserGender          = function () return GetUserGender (impl) end
  public_methods.GetUserId              = function () return GetUserId (impl) end
  public_methods.GetUserInfo            = function () return GetUserInfo (impl) end
  public_methods.HasInviteFriendsWindow = function () return false end
  public_methods.InviteFriends          = function (friends_list, finish_handler) InviteFriends (impl, friends_list, finish_handler) end 
  public_methods.IsLoggedIn             = function () return IsLoggedIn (impl) end
  public_methods.LoadUserInfo           = function (handler) LoadUserInfo (impl, handler) end
  public_methods.Login                  = function (login_handler, cancel_handler) Login (impl, login_handler, cancel_handler, true) end
  public_methods.Logout                 = function (handler) Logout (impl, handler) end
  public_methods.OnActivate             = function () OnActivate (impl) end
  public_methods.ShareChips             = function (message, share_type, share_id, handler, description, additional_info) ShareChips (impl, message, share_type, share_id, handler, description, additional_info) end    
  
  return public_methods
end
