-- используется для ввода текста в строку

TextLineInput={}

-- основные рабочие переменные
--local screen_keyboard
local text_line=nil   -- текстовая линия, в которую вводится текст (графический обьект)
local text=""         -- текст, введенный пользователем (строка)
local on_stop=nil     -- функция, вызываемая при завершении ввода
local on_edit=nil     -- функция, вызываемая при редактировании строки
local on_new_symbol=nil -- функция, вызываемая при добавлении символа
local max_text_length=nil -- максимальная допустимая длина строки текста пр вводе

-- настройки мигалки в конце строки
local blink=nil         -- периодический вызов мигалки в конце строки
local blink_delay=0.25  -- частота мигалки
local blink_char="|"    --символы, мигающие в конце строки
local blink_state=0     --текущее состояние мигалки(0\1)

local charcode_function={} -- список функций для обработки особых кодов клавиш(Enter, Backspace, ...)

local function UpdateTextline(show_blink)
  if show_blink then
    text_line:SetCharsColorFactors (text_line.TextLength - 1, 1, vec4 (0))
    blink_state=1 -- записываем новое состояние мигалки
  else
    text_line:SetCharsColorFactors (text_line.TextLength - 1, 1, vec4 (1))
    blink_state=0 -- записываем новое состояние мигалки
  end
end

local function ApplyBackspace()
  if text_line.TextLength == 0 then return end

  local last_symbol_start = #text
  
  while last_symbol_start > 0 and string.byte (text, last_symbol_start) > 0x7f and string.byte (text, last_symbol_start) < 0xc0 do
    last_symbol_start = last_symbol_start - 1
  end

  text = string.sub (text, 1, math.max (last_symbol_start - 1, 0))

  text_line.Text = text
end

local function ApplyEnter()
  TextLineInput.Stop ()
end

local function Utf32ToUtf8 (code)
  local return_value = {}

  if code < 0x80 then
    table.insert (return_value, code) 
  elseif code < 0x800 then
    table.insert (return_value, BitwiseOr (0xc0, BitwiseShiftRight (code, 6)))
    table.insert (return_value, BitwiseOr (0x80, BitwiseAnd (code, 0x3f)))
  elseif code < 0x10000 then
    table.insert (return_value, BitwiseOr (0xc0, BitwiseShiftRight (code, 12)))
    table.insert (return_value, BitwiseOr (0x80, BitwiseAnd (code, 0xfc0)))
    table.insert (return_value, BitwiseOr (0x80, BitwiseAnd (code, 0x3f)))
  elseif code < 0x200000 then
    table.insert (return_value, BitwiseOr (0xc0, BitwiseShiftRight (code, 18)))
    table.insert (return_value, BitwiseOr (0x80, BitwiseAnd (code, 0x3f000)))
    table.insert (return_value, BitwiseOr (0x80, BitwiseAnd (code, 0xfc0)))
    table.insert (return_value, BitwiseOr (0x80, BitwiseAnd (code, 0x3f)))
  else
    return "?"
  end

  return string.char (unpack (return_value))
end

charcode_function[8]  = ApplyBackspace
charcode_function[10] = ApplyEnter
charcode_function[13] = ApplyEnter

-- используется для проверки, работает ли ввод
TextLineInput.Online=function()
  if text_line==nil then
    return false
  end
  return true
end

TextLineInput.Stop=function()
  -- завершение ввода в строку
  -- если ввод уже завершен - выходим
  if text_line==nil then
    return
  end

--  print ("Stop text = '" .. text .. "'")
  -- забываем про строку
  text_line=nil
  -- если нужно запустить функцию при остановке ввода - пускаем ее
  if on_stop then
    on_stop(text)
  end
  return text
end

TextLineInput.Start=function(line, line_max_text_length, stop_function,edit_function,new_symbol_function)
  -- если идет ввод в другую строку - завершаем его
  if text_line then
    TextLineInput.Stop()
  end

--[[  if not screen_keyboard then
    screen_keyboard = System.ScreenKeyboard.Create (Engine.Windows.Find ("MainWindow"), System.ScreenKeyboardType.Ascii)
  end
  
  screen_keyboard:Show ()]]

  -- старт нового ввода 
  --assert(line==nil, "ERROR: TextLineInput.Start(line, stop_function) - line is nil")
  -- запоминаем новые устновки ввода
  text_line=line
  text = line.Text 
  on_stop=stop_function
  on_edit=edit_function
  on_new_symbol=new_symbol_function

  if line_max_text_length==nil then
    max_text_length=nil
  else
    if type(line_max_text_length)=="number" then
      if line_max_text_length > 0 then
        max_text_length = line_max_text_length
      end
    end
  end
end

TextLineInput.GetBlinkChar=function(c)
  return blink_char
end

-- получение нового символа
TextLineInput.NewCharcode=function(c)
  -- если не выбрана строка для ввода - игнорируем символ
  if text_line==nil then
    return
  end

  print ("NEW CHAR CODE: " .. c)
  
  if type (c) == "string" then
    c = string.byte (c)
  end
  
  -- обработка введенного символа
  local func=charcode_function[c]

  if func==nil then
    if (c < 32) then return end
  
    -- если это просто обычный символ - дописываем его в конец строки
    -- если строка ограничена в длине - проверяем ограничение
    if max_text_length then
      if text_line.TextLength > max_text_length then
        -- при выходе за пределы ограничения - не добавляем новый символ в строку, а просто выходим из функции
        return
      end
    end

    local utf8 = Utf32ToUtf8 (c)

    if on_new_symbol then
      if not on_new_symbol (utf8) then
         return
      end
    end

    text = text .. utf8

    text_line.Text = text
    
    if on_edit then
      on_edit (text)
    end
  else
    -- если есть специальная функция обработки для этого символа - применяем ее и выходим
    func()
    
    if on_edit and text_line then
      on_edit (text)
    end
    
    return
  end
end

TextLineInput.OnKeyboardHidden = function ()
  TextLineInput.Stop ()
end
