local DEFAULT_SET_NUMBER_DURATION   = 1
local LARGEST_PHONE_SCREEN_DIAGONAL = 5.4

---------------------------------------------------------------------------------------------------------------------------------
-- Разное
---------------------------------------------------------------------------------------------------------------------------------

function TryCallFunction(func) --попытка вызвать переданную функцию, если она есть
  if type(func)=="function" then
    func()
  end
end

---------------------------------------------------------------------------------------------------------------------------------
-- Работа с таблицами 
---------------------------------------------------------------------------------------------------------------------------------

-- выполнение действия для всех элементов таблицы
function ForEachInTable (t, func, ...)
  if not (t) then return end
  local index, value = next (t)
  local result=nil
  while index do
    result = func (value, ...)
    if result~=nil then
      return result
    end
    index, value = next (t, index)
  end
end

function ForEachIndexInTable (t, func, ...)
  if not (t) then return end
  local index, value = next (t)
  local result=nil
  while index do
    result = func (index, ...)
    if result~=nil then
      return result
    end
    index, value = next (t, index)
  end
end

function GetRandomFromTable (t)
  local value=0
  local MaxValue=-255 --любой положительный рандом будет больше этого числа
  local MaxElement=nil
  local index, element=next(t)
  while index do
    value=math.random()
    if value>MaxValue then
      MaxElement=element
      MaxValue=value
    end
    index, element=next(t, index)
  end
  return MaxElement
end

-- вынимает из таблицы 'a' все поля совпадающие по индексу и значению с таблицей 'b'
function SubtractTables(a, b)
  ForEachIndexInTable(b, function(index)
    if a[index]==b[index] then
      a[index]=nil
    end
  end)
end

-- дополнение талицы 'a' полями из таблицы 'b' - без изменения уже существующих полей 'а'
function SupplementTable(a, b)
  ForEachIndexInTable(b, function(index)
    if a[index]==nil then
      a[index]=b[index]
    end
  end)
end

function GetTableCopy (source)
  local return_value = {}

  SupplementTable (return_value, source)

  return return_value
end

-- добивание талицы 'a' полями из таблицы 'b' - без изменения уже существующих полей 'а'
function PushBackTable (a, b)
  for i = 1, #b do
    table.insert (a, b [i])
  end
end

--Копирование таблицы
function CopyArrayTable (source)
  local return_value = {}
  
  for i = 1, #source do
    table.insert (return_value, source [i])
  end
  
  return return_value
end

-- созданиетаблицы с элементами в обратном порядке
function GetReverseTable(t)
  local rev_t={}
  
  local count=0
  
  -- подсчет количества элементов в таблице
  ForEachInTable(t, function()
    count=count+1
  end)
  
  
  ForEachInTable(t,  function(value)
    rev_t[count]=value
    count=count-1
  end)
  
  return rev_t
end  

---------------------------------------------------------------------------------------------------------------------------------
-- работа со строками
---------------------------------------------------------------------------------------------------------------------------------

function AnyToString(value) 
  local t=type(value)
  if t=="string" then
    return value
  end
  if t=="userdata" then
    local s=tostring(value)
    if s==nil then
      return "[userdata]"
    else
      return s
    end
  end
  return tostring(value)
end

-- новая команда print 
-- для использования не надо преобразовывать типы
-- а переменные и строки просто перечисляются как параметры через запятую
local p=print
local function myprint(...)
  --[[
  создает строку из нескольких принимаемых параметров и печатает ее, например:
  myprint("a=",1," b=",2," count:",3,4,5) ==> "a=1 b=2 count:345"
  ]]
  local params={...}
  local value=nil
  local index=nil
  index, value = next(params)
  local s=""
  while index do
    value=AnyToString(value)
    s=s..value
    index, value = next(params, index)
  end
  p(s)
  return s
end
print=myprint

--генерация строки, для печати входных параметров функции
function FunctionCallString(function_name,...)
  local params={...}
  local str=function_name.."( "
  local divide_params=false
  ForEachInTable(params, function(param)
    if divide_params then
      str=str..", "
    end
    str=str..tostring(param)
    divide_params=true
  end)
  str=str..")"
  return str
end

-- преобразует принятые строки в числа
function tonumbers (...)
  local results={}
  local ResultNumber=0
  local params={...}
  local index, value = next(params)
  while index do
    value=tonumber(value)
    ResultNumber=ResultNumber+1
    results[ResultNumber]=value
    index, value = next(params, index)
  end
  return unpack(results)
end

-- разбивает строку на несколько в оостветствии с разделителями (terminal)
function SplitString (str, terminal, return_as_table)
  if not str then return nil end

  str = str .. terminal --присоединяем еще один терминал в конец строки, чтобы не пришлось 

   --послений найденный элемент записывать в результаты вне основного цикла

  local results       = {}
  local result_number = 0
  local init          = 1
  local found         = string.find (str, terminal, init, true)
  local terminal_len  = string.len (terminal)

  while found do
    local substr = string.sub (str, init, found - 1)

    if substr ~= "" then
      result_number           = result_number + 1    
      results [result_number] = substr
    end
    
    init  = found + terminal_len
    found = string.find (str, terminal, init)
  end

  if return_as_table then return results end

  return unpack (results)
end

function StringToVec(str)
  local a,b,c,d = tonumbers( SplitString( str, ";"))
  if b==nil then
    error("StringToVec(str) - unable to convert string: "..str)
  end
  if c==nil then
    return vec2(a,b)
  end
  if d==nil then
    return vec3(a,b,c)
  end
  return vec4(a,b,c,d)
end

-- Поиск ключей, подходящих по времени для анимации
function FindKeyRange (track, time)
  local index, key = next (track.keys)
  local first_key  = nil

  while index do
    if key.time < time then  
      if not (first_key) then
        first_key = key
      else
        if key.time < time and key.time > first_key.time then
          first_key = key
        end
      end
    end

    index, key = next (track.keys, index)
  end
  
  if not (first_key) then
    index, key = next (track.keys)    
  
    while index do          
      if not (first_key) then
        first_key = key
      else
        if key.time < first_key.time then
          first_key = key
        end
      end

      index, key = next (track.keys, index)
    end  

    if not (first_key) then return nil end
  
    return {first_key, first_key}
  end
  
  index, key = next (track.keys)  
  local second_key = nil

  while index do
    if key.time > first_key.time then
      if not (second_key) then
        second_key = key
      elseif key.time > time and key.time < second_key.time then
        second_key = key
      end
    end

    index, key = next (track.keys, index)
  end  
  
  if not (second_key) then return {first_key, first_key} end

  return {first_key, second_key}
end


---------------------------------------------------------------------------------------------------------------------------------
-- Эффекты
---------------------------------------------------------------------------------------------------------------------------------


-- плавное изменение цвета спрайта с вызовом заданной функции в конце изменений
local sprite_color_changes={} -- текущие изменения цвета спрайтов, используется чтобы избежать изменения цвета двумя процессами сразу

function ChangeSpriteColorSlowly (timer_name, sprite, target_color, change_time, on_change_func)
  -- отключение изменения цвета для заданного спрайта, если оно есть
  local current_change=sprite_color_changes[sprite]
  if current_change then
    current_change.Stop()
  end

  -- старт нового изменения цвета
  local start_color
  if type (sprite) == 'table' then
    start_color = sprite.node.Color
  else
    start_color = sprite.Color
  end
  
  local change={} -- периодический вызов перерасчета цвета спрайта
  
  change=SlowAction(timer_name, change_time, function(status)
    local new_color=start_color*(1-status) + target_color*status
    if type (sprite) == 'table' then
      sprite.node.Color = new_color
    else
      sprite.Color = new_color
    end
    
    -- если действие было окончено
    if not(status<1) then
      if on_change_func then
        on_change_func()
      end
      sprite_color_changes[sprite]=nil
    end
    
  end)
  
  -- регистрация нового изменения цвета по времени
  sprite_color_changes[sprite]=change
end

-- поочередно изменяет цвет нескольких спрайтов после чего запускает заданную функцию
function ChangeSpritesColorSlowly(timer_name, sprites, target_color, change_time, on_change_func)
  local index=0
  local function GetNextSprite()
    index=index+1
    return sprites[index]
  end
  
  local function ChangeNextSpriteColor()
    -- выбор следующего спрайта из списка
    local sprite=GetNextSprite()
    sprite=GetSprite(sprite)
    if sprite==nil then
      -- если спрайта нет, значит они закончились и можно вызвать функцию окончания
      if on_change_func then
        on_change_func()
      end
    else
      -- если спрайт найден - изменяем его цвет
      ChangeSpriteColorSlowly(timer_name, sprite.node, target_color, change_time, ChangeNextSpriteColor)
    end
  end
  
  ChangeNextSpriteColor()
end
  
function SetSpritesColor(sprites, color)
  ForEachInTable(sprites, function(sprite_name)    
    local sprite = GetSprite (sprite_name)
    sprite.tracks["alpha"] = nil
    sprite.tracks["intensity"] = nil    
    sprite.node.Color=color
  end)
end

function CreateTextLine(size, position)
  local line = Scene.TextLine.Create (game.font_library)
  
  line.Font                = "data/fonts/MyriadPro-BoldCond.xfont"
  line.Color               = vec4 (1,1,1,1)
  line.FontCreationParams  = game.font_creation_params
  line.Material            = "text_material"
  
  if size then
    line.Scale = vec3 (size.x, size.y, 1)
  end
  
  if position then
    if position.z then
      line.Position = vec3 (position.x, position.y, position.z)
    else
      line.Position = vec3 (position.x, position.y, 0)
    end
  end
  
  line:SetAlignment (Scene.TextLineAlignment.Left, Scene.TextLineAlignment.BaseLine)
  
  return line
end

function CopyTextLine (source)
  local line = Scene.TextLine.Create (game.font_library)

  line.Text                = source.Text
  line.Font                = source.Font
  line.HorizontalAlignment = source.HorizontalAlignment
  line.VerticalAlignment   = source.VerticalAlignment
  line.FontCreationParams  = game.font_creation_params
  line.Material            = "text_material"
  
  return line
end

function IsTableEmpty (t)
  if not t then return true end

  local index, value = next (t)

  return index == nil
end

function PrintRelativeCoordinates(sprite, x, y)
  local relative_x = (x - sprite.node.WorldPosition.x) / sprite.node.WorldScale.x
  local relative_y = (y - sprite.node.WorldPosition.y) / sprite.node.WorldScale.y
  
  print (string.format ("relative coordinates '%s' x=%.3f y=%.3f", sprite.name, relative_x, relative_y))
end

---------------------------------------------------------------------------------------------------------------------------------
-- Чтение данных
---------------------------------------------------------------------------------------------------------------------------------
function LoadVar (node, vars)
  local name       = node:Get ("Name")
  local expression = node:Get ("Expression", "")
  local value
  
  if expression ~= "" then
    value = loadstring ("return function (var) return " .. expression .. " end")()(vars)
  else
    local str = node:Get ("String", "")
    
    if str ~= "" then
      value = str
    else
      local value_string = node:Get ("Value")      
    
      if string.find (value_string, ";") == nil then
        -- Конвертация скалярного значения
        
        value = value_string + 0.0
      else
        -- Конвертация вектора
        
        value = StringToVec (value_string)
      end                
    end      
  end

  vars [name] = value
end

function LoadVars (xml_node, vars)
  local children_count = xml_node.ChildrenCount
  
  for i=0, children_count-1 do
    local child_node = xml_node:Child (i)
    
    if child_node.Name == "Var" then
      LoadVar (child_node, vars)
    end
  end
end

function ChangeVarBySine (timer_name, var_name, time, min_value, max_value)
  local value_difference = max_value - min_value

  SlowAction (timer_name, time, function (t)
    game.current_screen.vars [var_name] = min_value + math.sin (t * math.pi) * value_difference
  end)
end

function SetNumberInStringSlowly (text, start_val, end_val, duration, format_function, on_finish, prev_change_aware)
  local text_line = GetTextLine (text)
  local node      = text_line.node

  if prev_change_aware then
    if text_line.set_number_slowly then
      if text_line.set_number_slowly.action then
        text_line.set_number_slowly.action.Stop ()
      end

      start_val = text_line.set_number_slowly.current_value
    end

    text_line.set_number_slowly = {}
  end

  local function UpdateText  (t)
    local value = start_val + math.floor ((end_val - start_val) * t)
    local str
      
    if prev_change_aware then
      text_line.set_number_slowly.current_value = value
    end
      
    if format_function then
      str = format_function (value)
    else
      str = string.format ('%d', value)
    end
      
    node.Text = str
    
    if t == 1 then
      if on_finish then
        on_finish ()
      end
      
      text_line.set_number_slowly = nil
    end
  end
  
  UpdateText (0)

  if end_val ~= start_val then 
    local action = SlowAction ("ui", duration or DEFAULT_SET_NUMBER_DURATION, UpdateText)
    
    if prev_change_aware then
      text_line.set_number_slowly.action = action
    end
  end  
end

function FormatMoneyString (value, max_symbols_count, force_show_cents, dont_round_if_exact_length)
  max_symbols_count = max_symbols_count or 99999

  if not value then
    return ""
  end

  value = value + 0

  if value == 0 then
    return "0"
  end

  local cents   = value % 100
  local dollars = math.floor (value / 100)

  local return_value = ""
  
  local show_cents = cents ~= 0
  
  if force_show_cents == false then
    show_cents = false
  elseif force_show_cents == true then
    show_cents = true
  end
  
  if show_cents then
    return_value = string.format (".%02d", cents)
  end

  if dollars == 0 then
    return_value = "0" .. return_value
  end
  
  while dollars > 0 do
    local rest = dollars % 1000
    
    dollars = math.floor (dollars / 1000)
    
    if dollars > 0 then
      return_value = string.format (",%03d%s", rest, return_value)
    else
      return_value = rest .. return_value
    end
  end

  local has_cents_part = show_cents

  if #return_value > max_symbols_count and show_cents then
    return_value = string.sub (return_value, 1, -4)
    
    has_cents_part = false
  end 

  local needs_round

  if dont_round_if_exact_length then
    needs_round = #return_value > max_symbols_count and value > 100000
  else
    needs_round = #return_value >= max_symbols_count and value > 1000000
  end

  if needs_round then
    if has_cents_part then
      return_value = string.sub (return_value, 1, -4)   -- 1,000,000.00 -> 1,000,000
    end
  
    return_value   = string.sub (return_value, 1, -2) .. "K"                                    -- 1,000,000 -> 1,000,00k 
    return_value   = string.sub (return_value, 1, -5) .. "." .. string.sub (return_value, -3)   -- 1,000,00k -> 1,000.00k
    has_cents_part = true
  end 

  if #return_value > max_symbols_count then
    return_value   = string.sub (return_value, 1, -5) .. "K"  -- 1,000.00k -> 1,000k
    has_cents_part = false
  end 

  local function RoundToNextStep (return_value, next_step_suffix, has_cents_part, needs_round)
    if #return_value > 5 and string.sub (return_value, -4, -2) == "000" then --1,000K needs to be rounded
      needs_round = true 
    end
  
    if needs_round then
      if has_cents_part then
        return_value = string.sub (return_value, 1, -6)
      else
        return_value = string.sub (return_value, 1, -3)  -- 1,000k -> 1,00
      end

      return_value = string.sub (return_value, 1, -4) .. "." .. string.sub (return_value, -2) .. next_step_suffix -- 1,00 -> 1.00M
      
      has_cents_part = true
    end

    if #return_value > max_symbols_count then
      return_value = string.sub (return_value, 1, -5) .. next_step_suffix
      return false, return_value
    else
      return has_cents_part, return_value
    end 
  end

  if dont_round_if_exact_length then
    needs_round = #return_value > max_symbols_count and value >= 100000000
  else
    needs_round = #return_value >= max_symbols_count and value >= 100000000
  end

  has_cents_part, return_value = RoundToNextStep (return_value, "M", has_cents_part, needs_round)

  if dont_round_if_exact_length then
    needs_round = #return_value > max_symbols_count and value >= 100000000000
  else
    needs_round = #return_value >= max_symbols_count and value >= 100000000000
  end

  has_cents_part, return_value = RoundToNextStep (return_value, "B", has_cents_part, needs_round)

  if dont_round_if_exact_length then
    needs_round = #return_value > max_symbols_count and value >= 100000000000000
  else
    needs_round = #return_value >= max_symbols_count and value >= 100000000000000
  end

  has_cents_part, return_value = RoundToNextStep (return_value, "T", has_cents_part, needs_round)
  
  if has_cents_part and not force_show_cents then
    if #return_value > 4 and string.sub (return_value, -3, -2) == "00" then
      return_value = string.sub (return_value, 1, -5) .. string.sub (return_value, -1) 
    end
  end

  return return_value
end

function FormatWageringString (value)
  if value == 0 then
    return "0"
  end

  local return_value = ""

  value = value / 100

  while value > 0 do
    local rest = value % 1000
    
    value = math.floor (value / 1000)
    
    if value > 0 then
      return_value = string.format (",%03d%s", rest, return_value)
    else
      return_value = rest .. return_value
    end
  end

  return return_value
end

function UpdateGameCameraSize (window_width, window_height, min_camera_width, min_camera_height, min_camera_center_x, min_camera_center_y, min_ar, 
                               max_camera_width, max_camera_height, max_camera_center_x, max_camera_center_y, max_ar, camera, 
                               reference_width, reference_height, min_zone_touch_inside, constant_height)
  if reference_width == window_width and reference_height == window_height then return end
  
  local camera_center_x = min_camera_center_x

  if reference_height == window_height then  --Рисуем пиксел в пиксел
    camera.Left   = math.floor (camera_center_x - window_width / 2)
    camera.Right  = camera.Left + window_width
  elseif constant_height then  --Сохраняем высоту
    local window_ar    = window_width / window_height
    local camera_width = math.floor ((camera.Top - camera.Bottom) * window_ar)
    
    camera.Left   = math.floor (camera_center_x - camera_width / 2)
    camera.Right  = camera.Left + camera_width
  else
    local top_margin    = 0
    local bottom_margin = 0

    if game.scripts.bottom_ui then
      game.scripts.bottom_ui.OnSizeChanged (window_width, window_height)
      bottom_margin = game.scripts.bottom_ui.GetHeight ()
    end
    if game.scripts.top_ui then
      game.scripts.top_ui.OnSizeChanged (window_width, window_height)
      top_margin = game.scripts.top_ui.GetHeight ()
    end

    local frame_height    = math.max (1, window_height - top_margin - bottom_margin)
    local camera_width    = math.max (window_width, min_camera_width)
    local camera_height   = math.max (frame_height, min_camera_height)
    local frame_ar        = window_width / frame_height
    local window_ar       = window_width / window_height
    local camera_center_y = min_camera_center_y

    local center_camera_in_frame = true

    if min_zone_touch_inside then
      if frame_ar > min_ar then
        camera_height = min_camera_height
        camera_width  = camera_height * frame_ar
      else
        camera_width  = min_camera_width
        camera_height = camera_width / frame_ar
      end
    else
        --calculate camera size for frame
      if window_width < min_camera_width or frame_height < min_camera_height then
        if frame_ar > min_ar then
          camera_height = min_camera_height
          camera_width  = camera_height * frame_ar
        else
          camera_width  = min_camera_width
          camera_height = camera_width / frame_ar
        end
    
        print ("shrink!!!!!!!!!!!!!!!!!!!!!!!!!")
      elseif window_width > max_camera_width and window_height > max_camera_height then
        if window_ar > max_ar then
          camera_height = max_camera_height
          camera_width  = camera_height * window_ar
        else
          camera_width  = max_camera_width
          camera_height = camera_width / window_ar
        end
    
        camera_center_x = max_camera_center_x
        camera_center_y = max_camera_center_y
        center_camera_in_frame = false
      end
    end
  
      --calculate camera height for window
    if center_camera_in_frame then
      camera_height = camera_height / ((window_height - top_margin - bottom_margin) / window_height)
    end
  
    camera_width  = math.floor (camera_width)
    camera_height = math.floor (camera_height)
  
    camera.Left   = math.floor (camera_center_x - camera_width / 2)
    camera.Right  = camera.Left + camera_width

    if center_camera_in_frame then
      camera.Top = math.floor (camera_center_y + camera_height * (top_margin + frame_height / 2) / window_height)
    else
      camera.Top = math.floor (camera_center_y + camera_height / 2)
    end
  
    camera.Bottom = camera.Top - camera_height
  end                               
end

function PerformUnloadUnusedResources ()
  if game and not game.needs_unload_unused_resources then return end
  
  if game then 
    game.needs_unload_unused_resources = false 
  end
  
  collectgarbage ("collect")
    
  Engine.ResourceManager.FlushUnusedResources ()  
end

function LoadResource (resource_path)
  if string.find (resource_path, ".xmtl", 1, true) then
    --compatibility with previous code - load xreslist together with xmtl
  
    local resources = {}

    local xreslist
    
    if string.find (resource_path, "lobby/slot") then
      xreslist = string.gsub (string.gsub (resource_path, "/materials", ""), ".xmtl", ".png.tga")
    else
      xreslist = string.gsub (string.gsub (resource_path, "/materials", ""), ".xmtl", ".xreslist")
    end
    
    table.insert (resources, xreslist)
    table.insert (resources, resource_path)
    
    return LoadResources (resources)
  end

  PerformUnloadUnusedResources ()

  local binding = Engine.ResourceManager.CreateBinding (Engine.ResourceManager.Group.Create (resource_path))
    
  binding:Load () 

  return binding    
end

function LoadResources (resources_paths)
  PerformUnloadUnusedResources ()

  local group = Engine.ResourceManager.Group.Create ()
  
  for i = 1, #resources_paths do
    group:AddFileGroup (resources_paths [i])
  end

  local binding = Engine.ResourceManager.CreateBinding (group)
    
  binding:Load () 

  return binding    
end

function UnloadUnusedResources ()
  if not game then
    PerformUnloadUnusedResources ()
  else
    game.needs_unload_unused_resources = true
  end
end

function NeedsDownloadSlots ()
  local os = ApplicationDataGet (OS_TYPE)

  return string.find (os, "Windows") == nil and string.find (os, "Mac OS X") == nil
end

function NeedsDownloadLobby ()
  local os = ApplicationDataGet (OS_TYPE)

  return string.find (os, "Windows") == nil and string.find (os, "Mac OS X") == nil
end

function HasUnlimitedInternet ()
  local system_properties = System.Application.SystemProperties
  local property_name     = "CellularOnlyInternet"

  if system_properties:IsPresent (property_name) and system_properties:GetInteger (property_name) == 1 then
    return false
  end
  
  return true
end

local saved_device_platform

function GetPlatform ()
  if saved_device_platform then 
    return saved_device_platform
  end
  
  local os = ApplicationDataGet (OS_TYPE)

  if string.find (os, "Windows") ~= nil or string.find (os, "Mac OS X") ~= nil then
    local window = Engine.Windows.Find (WINDOW_NAME)
    local height = window.ClientHeight
    
    if height > 640 then
      return "pad"
    else
      return "phone"
    end
  end

  if os == "TabletOS" then
    return "pad"
  end

  local window = Engine.Windows.Find (WINDOW_NAME)
  local screen = window.ContainingScreen
  
  local screen_width    = screen.CurrentWidth / screen.CurrentXdpi
  local screen_height   = screen.CurrentHeight / screen.CurrentYdpi
  local screen_diagonal = math.sqrt (screen_width * screen_width + screen_height * screen_height)
  
  if screen_diagonal > LARGEST_PHONE_SCREEN_DIAGONAL then
    saved_device_platform = "pad"
  else
    saved_device_platform = "phone"
  end
  
  return saved_device_platform
end

function GetRequestParam (request, param)
  local param_name_start, param_start = string.find (request, param .. "=", 1, true)
        
  if not param_start or param_name_start == 1 then
    print ("No " .. param .. " !!!!!!")
    return nil
  end
  
  local previous_symbol = string.sub (request, param_name_start - 1, param_name_start - 1)

  if previous_symbol ~= "?" and previous_symbol ~= "&" and previous_symbol ~= "#" then
    return GetRequestParam (string.sub (request, param_name_start + 1), param)
  end
    
  param_start = param_start + 1
    
  local _, param_end      = string.find (request, "&", param_start)
  local _, fragment_start = string.find (request, "#", param_start)

  if fragment_start then
    if param_end then
      param_end = math.min (param_end, fragment_start)
    else
      param_end = fragment_start
    end 
  end

  if not param_end then 
    param_end = #request
  else
    param_end = param_end - 1
  end
    
  return string.sub (request, param_start, param_end)
end

function GetPercentEscapedUrl (url)
  url = string.gsub (url, "([%c\128-\255])", function (c)
        return string.format("%%%02X", string.byte(c))
      end)
      
  url = string.gsub (url, " ", "+")
  
  return url
end

function UrlToFileName (url)
  url = string.gsub (url, "([%c\128-\255])", function (c)
        return "_"
      end)
      
  url = string.gsub (url, " ", "_")
  url = string.gsub (url, ":", "_")
  url = string.gsub (url, "/", "_")
  url = string.gsub (url, "%%", "_")
  url = string.gsub (url, "&", "_")
  url = string.gsub (url, "?", "_")
  
  if string.len (url) > 200 then
    url = string.sub (url, -200)
  end
  
  return url
end

function GetLastPathComponent (path)
  local start = 0
    
  while true do
    local next_start = string.find (path, "/", start + 1, true)
     
    if next_start then
      start = next_start
    else
      break
    end
  end
    
  if start == 0 then
    return path
  else
    return string.sub (path, start + 1) 
  end  
end

function GetPathExtension (path)
  local args_start = string.find (path, "?")
  
  if args_start then
    path = string.sub (path, 1, args_start - 1)
  end
  
  local start = 0
    
  while true do
    local next_start = string.find (path, ".", start + 1, true)
     
    if next_start then
      start = next_start
    else
      break
    end
  end
    
  if start == 0 then
    return path
  else
    return string.sub (path, start + 1) 
  end  
end

function GetPathWithoutExtension (path)
  local start = 0
    
  while true do
    local next_start = string.find (path, ".", start + 1, true)
     
    if next_start then
      start = next_start
    else
      break
    end
  end
    
  if start == 0 then
    return path
  else
    return string.sub (path, 1, start - 1) 
  end  
end

function PerformLogOut (logout_handler)
  local function SendPushLogOutEvent ()
    local hs_port = GetHsPort ()
    local url     = ApplicationDataGet (PUSH_NOTIFICATIONS_SERVER_URL) .. "logout?app=slots&env="

    if hs_port == HS_LIVE_PORT then
      url = url .. "live"
    elseif hs_port == HS_QA_PORT then
      url = url .. "qa"
    else
      url = url .. "dev"
    end

    local system_properties = System.Application.SystemProperties
    local uid_property_name = "UUID"

    if system_properties:IsPresent (uid_property_name) then
      url = url .. "&uid=" .. system_properties:GetString (uid_property_name)
    end

    local platform_code = "iphone"

    if ApplicationDataGet (OS_TYPE) == "Android" then
      platform_code = "android"
    elseif GetPlatform () == "pad" then
      platform_code = "ipad"
    end

    url = url .. "&platform=" .. platform_code

    if DEBUG then
      print ("Unregistering from push notification with url '" .. url .. "'")
    end

    Common.File.AsyncLoadString (url, Common.File.CreateAsyncLoadHandler (function (result)
      if DEBUG then
        print ("Push unregistration result '" .. result .. "'")
      end
    end))
  end

  System.Application.PostNotification ("LocalNotifications RemoveAll|Bonus")

  HsConnectionDisconnect ()
  HsConnectionClearStoredMessages ()

  ApplicationDataSet (GIFTS_EVENT, nil)
  ApplicationDataSet (PROMO_ACTIONS_SHOWN, nil)
  ApplicationDataSet (LOBBY_ROOM, nil)
  ApplicationDataSet (SLOTS_DESCS, nil)
  ApplicationDataSet (REGISTERED_FOR_PUSH_NOTIFICATIONS, nil)

  SetProfileProperty (LAST_LOGGED_IN_NETWORK, nil)

  SetCurrentScreen ("loader")

  GetSocialManager ().Logout (function ()
    local status, return_value = pcall (SendPushLogOutEvent)
    
    if not status then 
      print ("LogOut event error: " .. status)
    end 
      
    SetProfileProperty (REFERRER_ATTRIBUTE, nil)

    ExecuteOnNextUpdate (function () ExecuteOnNextUpdate (logout_handler, true) end, true)
  end)
end

local function LoginToNetwork ()
  if NeedsDownloadLobby () then
    SetCurrentScreen ("loader")
  else
    SetCurrentScreen ("lobby")
  end

  ExecuteOnNextUpdate (function () ExecuteOnNextUpdate (function ()
    GetSocialManager ().Login (function ()
      HsConnectToServer ()
    end, function ()
      GetSocialManager ().CancelLogin ()
      SetProfileProperty (LAST_LOGGED_IN_NETWORK, nil)
      SetCurrentScreen ("main_menu")
    end, true)
  end, true) end, true)
end

function VKLogin ()
  if GetSocialManager ().IsInGuestMode () then
    SetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE, "true")
  end
  
  GetSocialManager ().SetEngine (SOCIAL_MANAGER_VK_ENGINE)
  
  LoginToNetwork ()  
end

function FacebookLogin ()
  if GetSocialManager ().IsInGuestMode () then
    SetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE, "true")
  end
  
  GetSocialManager ().SetEngine (SOCIAL_MANAGER_FACEBOOK_ENGINE)
  
  LoginToNetwork ()  
end

function GuestLogin ()
  SetProfileProperty (ACCOUNT_MIGRATION_IN_PROGRESS_ATTRIBUTE)

  GetSocialManager ().SetEngine (SOCIAL_MANAGER_GUEST_ENGINE)

  LoginToNetwork ()
end

function GetTextLength (text)
  local text_line = Scene.TextLine.Create (game.font_library)
  
  text_line.Text = text

  return text_line.TextLength
end

function GetSpriteOffset (from, to)
  local from_center = vec3 (from:Get ("Left") + from:Get ("Width") / 2, -from:Get ("Top") - from:Get ("Height") / 2, 0) 
  local to_center   = vec3 (to:Get ("Left") + to:Get ("Width") / 2, -to:Get ("Top") - to:Get ("Height") / 2, 0) 

  return to_center - from_center 
end

function LaunchSlot (slot_name, slot_id)
  ApplicationDataSet (FREE_SPINS_GAIN, nil)

  HsConnectionPopStoredMessages (AUTH_MESSAGE)

  HsConnectionSendStoredMessage (SLOT_INIT_MESSAGE, 1, { event = "slot.init", slotId = slot_id + 0 }, function ()
    SetSlotScreen (slot_name, true)
  end)
  
  SetProfileProperty ("CurrentSlotId", slot_id)
  SetSlotScreen (slot_name)
end

function IsSlotAvailable (slot_id)
  local slots      = ApplicationDataGet (SLOTS_DESCS)
  local user_level = GetUserLevel ()

  for i = 1, #slots do
    local slot_desc = slots [i]

    if slot_desc.id and slot_desc.id + 0 == slot_id + 0 then
      return slot_desc.level and slot_desc.level + 0 <= user_level.id + 0 and not slot_desc.coming_soon
    end
  end
end

local buy_chips_window_mounted_version
local buy_chips_window_search_pathes = {}

function UnmountBuyChipsWindow ()
  if not buy_chips_window_mounted_version then return end

  local mount_point = "/buy_chips"
  
  Common.File.Unmount (mount_point)
    
  for i = 1, #buy_chips_window_search_pathes do
    Common.File.RemoveSearchPath (buy_chips_window_search_pathes [i])
  end
    
  buy_chips_window_search_pathes = {}
  buy_chips_window_mounted_version = nil
end

function MountBuyChipsWindow (on_mounted_callback, on_progress_callback, on_error_callback)
  local buy_chips_resource_data_all = ApplicationDataGet (BUY_CHIPS_RESOURCE_DATA)
  local buy_chips_resource_data     = buy_chips_resource_data_all ["default"]
  local hs_promo_actions            = GetPromoActions ()
  local current_time                = game.main_timer.GetTime ()

  for i = 1, #hs_promo_actions do
    local action = hs_promo_actions [i]
  
    if buy_chips_resource_data_all [action.id] then
      if action.time_left < 0 or action.time_end > current_time then
        buy_chips_resource_data = buy_chips_resource_data_all [action.id]
        break
      end
    end
  end

  local function AfterMountHandler (mount_point)
    local after_mount_script = mount_point .. "data/scripts/screens/buy_chips_window_after_mount.lua"
  
    if Common.File.IsFileExist (after_mount_script) then
      local status, error = pcall (function ()
        local mount_function = dofile (after_mount_script)
          
        mount_function ()
      end)
        
      if not status then
        print ("Buy chips after mount error " .. error)
      end
    end
  end
  
  if not NeedsDownloadSlots () then
    AfterMountHandler ("")
    on_mounted_callback () 
    return 
  end

  local mount_point = "/buy_chips"

  if buy_chips_window_mounted_version and buy_chips_resource_data.version == buy_chips_window_mounted_version then
    AfterMountHandler (mount_point .. "/")
    on_mounted_callback () 
    return 
  end

  UnmountBuyChipsWindow ()

  local downloaded_buy_chips_versions = SplitString (GetGameProperty (DOWNLOADED_BUY_CHIPS_ATTRIBUTE) or "", " ", true)
  
  local resources_path = DOWNLOADS_PATH .. "buy_chips"
  local save_path      = resources_path .. "/" .. buy_chips_resource_data.version .. ".zip"

  if not Common.File.IsFileExist (resources_path) then
    Common.File.Mkdir (resources_path)
  end
    
  GetDownloadsManager ().DownloadFile (buy_chips_resource_data.url, save_path, function (state, arg2, arg3)
    if state == Common.BackgroundCopyStateStatus.Finished then
      for i = 1, #downloaded_buy_chips_versions do
        local version = downloaded_buy_chips_versions [i]
    
        if version + 0 == buy_chips_resource_data.version then
          table.remove (downloaded_buy_chips_versions, i)
          break
        end
      end
    
      table.insert (downloaded_buy_chips_versions, buy_chips_resource_data.version)
        
      local MAX_BUY_CHIPS_WINDOWS_HISTORY = 3  
    
      if #downloaded_buy_chips_versions > MAX_BUY_CHIPS_WINDOWS_HISTORY then
        local oldest_window_path = resources_path .. "/" .. downloaded_buy_chips_versions [1] .. ".zip"
      
        Common.File.Remove (oldest_window_path)
      
        table.remove (downloaded_buy_chips_versions, 1)
      end
        
      local downloaded_versions_string = ""
        
      for i = 1, #downloaded_buy_chips_versions do
        if i == 1 then
          downloaded_versions_string = downloaded_buy_chips_versions [i]
        else
          downloaded_versions_string = downloaded_versions_string .. " " .. downloaded_buy_chips_versions [i]
        end
      end
        
      SetGameProperty (DOWNLOADED_BUY_CHIPS_ATTRIBUTE, downloaded_versions_string)

      Common.File.Mount (mount_point, save_path)
      
      table.insert (buy_chips_window_search_pathes, mount_point)
      table.insert (buy_chips_window_search_pathes, string.format (mount_point .. "/resolutions/%s-generic-%s", buy_chips_resource_data.platform, buy_chips_resource_data.resolution))
      
      for i = 1, #buy_chips_window_search_pathes do
        Common.File.AddSearchPath (buy_chips_window_search_pathes [i])
      end
  
      local path = LoadPackLocalization (mount_point)

      local add_pathes = { string.format (path .. "/resolutions/%s-generic-%s", buy_chips_resource_data.platform, buy_chips_resource_data.resolution) }

      for i = 1, #add_pathes do
        Common.File.AddSearchPath (add_pathes [i])
        table.insert (buy_chips_window_search_pathes, add_pathes [i])
      end
  
      AfterMountHandler (mount_point .. "/")
  
      buy_chips_window_mounted_version = buy_chips_resource_data.version
  
      on_mounted_callback ()
    elseif state == Common.BackgroundCopyStateStatus.Failed then
      on_error_callback ()
    elseif state == Common.BackgroundCopyStateStatus.InProgress then
      on_progress_callback (arg3 / arg2)
    end
  end, false, "buy_chips_window")
end

function IsLobbyMounted ()
  return game.lobby_mounted
end

function GetDownloadedLobbyInfo ()
  local downloaded_version_path = LOBBY_DOWNLOADED_VERSION_PATH
  
  if Common.File.IsFileExist (downloaded_version_path) then
    local status, return_value = pcall (function ()
      local lobby_info = Common.StringNode.LoadXmlFromFile (downloaded_version_path)

      local return_value = {}
    
      return_value.version    = lobby_info:Get ("version") + 0
      return_value.platform   = lobby_info:Get ("platform")
      return_value.resolution = lobby_info:Get ("resolution")
      return_value.format     = lobby_info:Get ("format")
    
      return return_value    
    end)
      
    if status then
      return return_value
    else
      print (return_value)
      return nil
    end
  end
end

function UnmountLobby ()
  local lobby_mounted_version = ApplicationDataGet (LOBBY_MOUNTED_VERSION)

  if not lobby_mounted_version then return end

  local mount_point = "/lobby"
  
  Common.File.Unmount (mount_point)
    
  local lobby_search_pathes = ApplicationDataGet (LOBBY_SEARCH_PATHES) or {}
    
  for i = 1, #lobby_search_pathes do
    Common.File.RemoveSearchPath (lobby_search_pathes [i])
  end
    
  ApplicationDataSet (LOBBY_SEARCH_PATHES, {})
  ApplicationDataSet (LOBBY_MOUNTED_VERSION, nil)
  
  game.lobby_mounted = false
end

function MountLobby (mount_handler)
  local function AfterMountHandler ()
    if game.scripts.ui_windows.IsAlertWindowShown () then return end --dont change screen if error alert shown

    if mount_handler then 
      mount_handler ()
    end    
  end

  local mount_point = "/lobby"

  local lobby_mounted_version = ApplicationDataGet (LOBBY_MOUNTED_VERSION)
  local lobby_info            = GetDownloadedLobbyInfo ()

  if lobby_mounted_version and lobby_info.version == lobby_mounted_version then
    AfterMountHandler ()
    return 
  end

  UnmountLobby ()

  local status, error = pcall (function ()
    Common.File.Mount (mount_point, LOBBY_DOWNLOAD_PAK_PATH)

    local lobby_search_pathes = {}

    table.insert (lobby_search_pathes, mount_point)
    table.insert (lobby_search_pathes, string.format (mount_point .. "/resolutions/%s-generic-%s", lobby_info.platform, lobby_info.resolution))
    table.insert (lobby_search_pathes, string.format (mount_point .. "/resolutions/%s-%s-%s", lobby_info.platform, lobby_info.format, lobby_info.resolution))

    for i = 1, #lobby_search_pathes do
      Common.File.AddSearchPath (lobby_search_pathes [i])
    end

    local path = LoadPackLocalization (mount_point)

    local add_pathes = { string.format (path .. "/resolutions/%s-generic-%s", lobby_info.platform, lobby_info.resolution) }

    for i = 1, #add_pathes do
      Common.File.AddSearchPath (add_pathes [i])
      table.insert (lobby_search_pathes, add_pathes [i])
    end

    ApplicationDataSet (LOBBY_MOUNTED_VERSION, lobby_info.version)
    ApplicationDataSet (LOBBY_SEARCH_PATHES, lobby_search_pathes)

    LoadLocalizationResource (mount_point .. "/locales/" .. GetCurrentLocale () .. "/data/strings/strings.xml", true)
  
    local after_mount_script = mount_point .. "/data/scripts/screens/lobby_after_mount.lua"
  
    if Common.File.IsFileExist (after_mount_script) then
      local status, error = pcall (function ()
        local mount_function = dofile (after_mount_script)
          
        mount_function ()
      end)
        
      if not status then
        print ("Lobby after mount error " .. error)
      end
    end

    AfterMountHandler ()
  end)
        
  if not status then
    print ("Lobby mount error " .. error)
  
    game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
      SetInputState (false)              
      HsConnectionDisconnect ()
      SetCurrentScreen ("main_menu")
    end, nil, GetLocalizedString ("CANCEL"), { GetLocalizedString ("CONNECTION_ERROR") })

    Common.File.Remove (LOBBY_DOWNLOAD_PATH)
  end

  game.lobby_mounted = true
end

function ProcessStringWithReplacements (str, replacements)
  for i = 1, #replacements do
    str = string.gsub (str, replacements [i].token, replacements [i].replacement)
  end
  
  return str
end

function SetTextLineTextWithReplacements (node, str, replacements, color_factor)
  local processed_text = ProcessStringWithReplacements (str, replacements)

  node.Text = processed_text

  if color_factor then
    node:SetCharsColorFactors (0, node.TextLength, color_factor)
  end

  for i = 1, #replacements do
    local replacement = replacements [i]
  
    if replacement.color_factor then
      local replacement_text     = string.gsub (replacement.replacement, "%%%%", "%%")
      local replacement_length   = GetTextLength (replacement_text)
      local current_search_index = 0
    
      while current_search_index do
        current_search_index = string.find (processed_text, replacement_text, current_search_index + 1, true)
    
        if current_search_index then
          local prev_text_length
    
          if current_search_index == 1 then
            prev_text_length = 0
          else
            prev_text_length = GetTextLength (string.sub (processed_text, 1, current_search_index - 1))
          end
    
          node:SetCharsColorFactors (prev_text_length, replacement_length, replacement.color_factor)
        end
      end
    end
  end
end

function GetNodeChildAttribute (node, child_name, default_value)
  if not node then return default_value end

  local child = node:Find (child_name)
  
  if not child or child.AttributesCount == 0 then return default_value end

  return child:Attribute (0)
end

function GetNodeChildArray (node, child_array_name, default_value)
  local child_array = {}

  local child_array_node = node:Find (child_array_name)

  if not child_array_node or child_array_node.ChildrenCount == 0 then return default_value end

  child_array_node = child_array_node:Child (0)

  for i = 0, child_array_node.AttributesCount - 1 do
    local count_attribute = child_array_node:Attribute (i)
    table.insert (child_array, count_attribute)
  end

  return child_array
end

function GetAspectFitSize (object_width, object_height, zone_width, zone_height)
  local object_ar = object_width / object_height
  local zone_ar   = zone_width / zone_height
  
  if object_ar > zone_ar then
    local scale = zone_width / object_width
    
    return { x = zone_width, y = object_height * scale }
  else
    local scale = zone_height / object_height
    
    return { x = object_width * scale, y = zone_height }
  end
end

function ShowPendingAchievementsCompleteWindows ()
  local pending_achievements = ApplicationDataGet  (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS) or {}
  
  if #pending_achievements == 0 then return false end
  
  for i = 1, #pending_achievements do
    local event = pending_achievements [i]
    
    if event.event == "achievement.complete" then
      game.scripts.ui_windows.ShowAchievementsCompleteWindow (event.data.id, event.data.group_id, event.data.current, event.data.prize_type)
    elseif event.event == "info.achievements" then
      game.scripts.ui_windows.ShowAchievementsGroupComplete (event.data.id, event.data.prize, event.data.prize_type)
    end 
  end
  
  ApplicationDataSet  (PENDING_ACHIEVEMENTS_COMPLETE_EVENTS, nil)
  
  return true
end

function GetBetMoney (bet_id)
  bet_id = bet_id + 0

  local bets = GetBets ()
  
  for i = 1, #bets do
    local bet = bets [i]
    
    if bet.id + 0 == bet_id then
      return bet.money
    end
  end
end

function ShowPendingPrizes ()
  if game.current_screen.name == "loader" then return false end
  if not game.scripts.ui_windows then return false end

  if ApplicationDataGet (IS_BATTLE_ACTIVE) then 
    if game.scripts.level and game.scripts.level.ShowPendingPrizesInBattle then
      game.scripts.level.ShowPendingPrizesInBattle ()
    end

    return false 
  end

  local pending_prizes = GetPrizes ()
  
  if #pending_prizes == 0 then return false end

  for i = 1, #pending_prizes do
    local prize = pending_prizes [i]  

    if not prize.shown then
      prize.shown = true

      local prize_value = prize.coins
      local prize_type  = 1

      if prize.diamonds > 0 then
        prize_type  = 2
        prize_value = prize.diamonds
      end

      if game.scripts.lobby then
        game.scripts.lobby.OpenPromoActions (false, true)
      end

      game.scripts.ui_windows.ShowWinnerOfBestPlayerWindow (prize.place, prize.type, prize_type, prize_value, prize.id)
    end
  end
  
  return true
end

function OnBackPressed ()
  if game.scripts.ui_windows and game.scripts.ui_windows.OnBackPressed then
    if game.scripts.ui_windows.OnBackPressed () then return end
  end

  if game.scripts.level and game.scripts.level.OnBackPressed then
    if game.scripts.level.OnBackPressed () then return end
  end

  if game.scripts.lobby and game.scripts.lobby.OnBackPressed then
    if game.scripts.lobby.OnBackPressed () then return end
  end
  
  if game.scripts.top_ui and game.scripts.top_ui.OnBackPressed then
    if game.scripts.top_ui.OnBackPressed () then return end
  end
end
