local EPS                             = 0.0001
local GAMES_RESOURCES_DOWNLOAD_PERIOD = 60 * 60
local PHOTO_ACTIVITY_INDICATOR_FACTOR = 0.4
local LOBBY_DOWNLOAD_PATH             = DOWNLOADS_PATH .. "lobby_resources/"

LOBBY_DOWNLOAD_PAK_PATH         = LOBBY_DOWNLOAD_PATH .. "lobby.zip"
LOBBY_DOWNLOADED_VERSION_PATH   = LOBBY_DOWNLOAD_PATH .. "version.txt"

local function CreateFramesNames (format, from, to)
  local return_value = {}
  
  for i = from, to, 2 do
    table.insert (return_value, string.format (format, i))
  end
  
  return return_value
end

local function SetAnimationPosition (animation, reference, z)
  local sprite = GetSprite (reference)
  local layout = sprite.layout
  
  animation.SetPosition (layout:Get ("Left") + layout:Get ("Width") / 2, -layout:Get ("Top") - layout:Get ("Height") / 2, z)
end

local function UpdateProgressScissor (impl)
  local camera_width  = impl.camera.Right - impl.camera.Left
  local camera_height = impl.camera.Top - impl.camera.Bottom
  
  if impl.is_battle_loader then
    local clip_width  = impl.opponent_progress_sprite_width * impl.opponent_progress
  
    BindOrUpdateScissor (impl.opponent_progress_sprite, impl.opponent_progress_sprite_left, 0, clip_width, camera_height)
  end

  local clip_width = impl.progress_sprite_width * impl.progress

  if not impl.is_battle_loader then
    local offset_x = impl.progress_sprite_width * impl.progress
    local spark    = GetSprite ("iiskra_anim_1").layout

    impl.spark_animation.SetPosition (((spark:Get ("Left") + spark:Get ("Width") / 2) + impl.progress_sprite_width * impl.progress), -spark:Get ("Top") - spark:Get ("Height") / 2, GetSprite ("iiskra_anim_1").z - EPS)
    
    game.current_screen.vars.percent_bar = vec2 (offset_x, 0)
    game.current_screen.vars.loading_percent_text = vec2 (offset_x, 0)    
    
    local progress_sprite = GetSprite ("percent_bg")

    EvalSpriteState (progress_sprite, game.current_screen.vars, progress_sprite.timer.GetTime ())

    local progress_text_line = GetTextLine ("loading_percent_text_zone")

    EvalTextLineState (progress_text_line, progress_text_line.timer.GetTime (), game.current_screen.vars)
  end

  BindOrUpdateScissor (impl.progress_sprite, impl.progress_sprite_left, 0, clip_width, camera_height)
end

local function UpdateLoadingProgress (impl)
  if impl.is_battle_loader then
    GetTextLine ("you_progress").node.Text      = math.floor (impl.progress * 100) .. "%"
    GetTextLine ("opponent_progress").node.Text = math.floor (impl.opponent_progress * 100) .. "%"
  else
    GetTextLine ("loading_percent_text_zone").node.Text = math.floor (impl.progress * 100) .. "%"
  end
  
  UpdateProgressScissor (impl)
end

local function SetNextScreenImpl (impl)
  local status, err = pcall (SetCurrentScreen, impl.next_screen, true, true, true)
  
  if impl.is_battle_loader then
    if impl.you_avatar then
      impl.you_avatar.OnHide (true)
    end
    
    if impl.opponent_avatar then
      impl.opponent_avatar.OnHide (true)
    end
  end 

  if not status then  --TODO show error window
    print (err)
  else
    if string.find (impl.next_screen, BONUS_SCREEN_NAME_SUFFIX) then
      local slot_id = GetProfileProperty ("CurrentSlotId")
    
      if slot_id then
        slot_id = slot_id + 0 
      end
    end
  end
end

local function SetNextScreen (impl)
  if not impl.next_screen then return end
  if not impl.resources_loaded then return end
  if impl.next_screen ~= "main_menu" and not HsConnectionIsConnected () then return end
  if not HsConnectionIsAllStoredMessagesLoaded () then return end

  if impl.is_battle_loader then
    HsConnectionSendMessage (1, { event = "battle.loading", progress = 100 })

    if impl.battle_interrupted then return end
    if not impl.opponent_ready then return end
  end

  local last_message = HsConnectionGetLastStoredMessageId ()

  if last_message == SLOT_INIT_MESSAGE then
    local slot_init_message = HsConnectionGetStoredResponse (SLOT_INIT_MESSAGE)
    local bonus_occured     = slot_init_message:Find ("bonus"):Attribute (0) == "true"
    
    if bonus_occured then
      local is_bonus = false
  
      local path = impl.next_screen
  
      if string.find (impl.next_screen, BONUS_SCREEN_NAME_SUFFIX) then
        path     = string.sub (impl.next_screen, 1, #impl.next_screen - #BONUS_SCREEN_NAME_SUFFIX)
        is_bonus = true
      end
    
      local bonus_screen = impl.next_screen .. BONUS_SCREEN_NAME_SUFFIX

      if is_bonus then
        bonus_screen = impl.next_screen
      end

      if not bonus_screen then     
        print ("Can't find bonus screen for screen ", impl.next_screen)
        SetCurrentScreen ('lobby')
        return
      end

      if NeedsDownloadSlots () and not GetDownloadsManager ().IsBonusDownloaded (path) then
        ActivateSprites ("bar", "overlay", "percent_bg")
        ActivateTextLine ("loading_percent_text_zone")
        
        impl.spark_animation.Show ("animate")

        impl.progress = 0
        UpdateLoadingProgress (impl)

        DeactivateTextLine (impl.loading_text_line)

        GetDownloadsManager ().DownloadBonus (path, function (progress)
            impl.progress = progress
            UpdateLoadingProgress (impl)
          end, function ()
            ActivateTextLine (impl.loading_text_line)

            SetNextScreen (impl)
          end)
        return
      end

      HsConnectionSendStoredMessage (BONUS_INIT_MESSAGE, 1, { event = "bonus.init", slotId = slot_init_message:Find ("slotId"):Attribute (0) + 0 },
        function ()
          SetSlotScreen (bonus_screen, true)
        end)
        
      SetProfileProperty (DONT_SHOW_BONUS_TUTORIAL_ATTRIBUTE, "1")
       
      SetSlotScreen (bonus_screen)
      
      return
    end
  end

  SetNextScreenImpl (impl)
end

local function OnAllStoredMessagesReceived (impl)
  SetNextScreen (impl)  
end

local function OnDownloadError ()
  game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
    HsConnectionClearStoredMessages ()
    SetCurrentScreen ("main_menu")
  end, nil, GetLocalizedString ("CLOSE"), { GetLocalizedString ("CONNECTION_ERROR") })
end

local function IsResourceOptimal (window_width, window_height, width, height, optimal_resource_width, optimal_resource_height)
  local prev_width_diff  = window_width - optimal_resource_width
  local prev_height_diff = window_height - optimal_resource_height
  local curr_width_diff  = window_width - width
  local curr_height_diff = window_height - height

  if prev_width_diff > 0 then                            --ширина предыдущего найденного ресурса меньше ширины окна
    if curr_width_diff < prev_width_diff then            --ширина текущего ресурса больше ширины предыдущего
      return true
    elseif curr_width_diff == prev_width_diff then       --ширины совпадают
      if prev_height_diff > 0 then                       --высота предыдущего ресурса меньше высоты окна
        if curr_height_diff < prev_height_diff then      --высота текущего ресурса больше высоты предыдущего
          return true
        end
      elseif prev_height_diff < 0 and curr_height_diff < 0 and curr_height_diff > prev_height_diff then --высоты обоих ресурсов больше высоты окна, но разница текущего меньше
        return true
      end
    end
  elseif prev_width_diff == curr_width_diff then       --ширины совпадают, подбор по высоте
    if prev_height_diff > 0 then                       --высота предыдущего ресурса меньше высоты окна
      if curr_height_diff < prev_height_diff then      --высота текущего ресурса больше высоты предыдущего
        return true
      end
    elseif prev_height_diff < 0 and curr_height_diff < 0 and curr_height_diff > prev_height_diff then --высоты обоих ресурсов больше высоты окна, но разница текущего меньше
      return true
    end
  elseif prev_width_diff <= 0 and curr_width_diff <= 0 then  --ширина обоих ресурсов больше ширины окна, подбор по соотношению сторон
    local prev_ar   = optimal_resource_width / optimal_resource_height
    local curr_ar   = width / height
    local window_ar = window_width / window_height
                
    if math.abs (window_ar - curr_ar) < math.abs (window_ar - prev_ar) then
      return true
    end
  end
end

local function IsLobbyDownloaded (version, update_from_version)
  local pak_path = LOBBY_DOWNLOAD_PATH
  local pak_flag = LOBBY_DOWNLOAD_PAK_PATH .. "_downloaded"
    
  if not Common.File.IsFileExist (pak_path) or not Common.File.IsFileExist (pak_flag) then return false end --Check that data files actually exists
  
  local lobby_info = GetDownloadedLobbyInfo ()
  
  if not lobby_info then return false end
  
  return lobby_info.version > update_from_version and lobby_info.platform == GetPlatform () and lobby_info.format == ApplicationDataGet (TEXTURE_FORMATS) [1]
end

local function DownloadLobby (impl)
  local window        = Engine.Windows.Find (WINDOW_NAME)
  local window_width  = window.ClientWidth
  local window_height = window.ClientHeight

  local slots_resources_node     = ApplicationDataGet (GAMES_RESOURCES_XML)  
  local resources_versions_node  = slots_resources_node:Find ("AppResources"):Find ("Lobby"):Find ("ResourcesVersions")
  local resources_versions_count = resources_versions_node.ChildrenCount

  local platform       = GetPlatform ()
  local texture_format = ApplicationDataGet (TEXTURE_FORMATS) [1]
  
  local latest_app_engine_version = ApplicationDataGet (LATEST_APP_ENGINE_VERSION)
  
  for i = 0, resources_versions_count - 1 do
    local version_node = resources_versions_node:Child (i)
    
    if version_node.Name == "Version" then
      local min_app_engine_version = version_node:Get ("MinAppEngineVersion") + 0
      local max_app_engine_version = version_node:Get ("MaxAppEngineVersion") + 0
      local version                = version_node:Get ("Version") + 0
      local update_from_version    = version_node:Get ("UpdateFromVersion") + 0

      local is_supported              = min_app_engine_version <= APP_ENGINE_VERSION and max_app_engine_version >= APP_ENGINE_VERSION
      local is_supported_after_update = min_app_engine_version <= latest_app_engine_version and max_app_engine_version >= latest_app_engine_version

      if is_supported or is_supported_after_update then
        local resources_node  = version_node:Find ("Resources")
        local constants_node  = version_node:Find ("Constants")
        local resources_count = resources_node.ChildrenCount
        
        local optimal_resource        = -1
        local optimal_resource_width  = -1
        local optimal_resource_height = -1

        for j = 0, resources_count - 1 do
          local resource_node = resources_node:Child (j)
         
          if resource_node.Name == "Resource" then
            local resource_platform        = resource_node:Get ("Platform")
            local resource_texture_formats = SplitString (resource_node:Get ("TextureFormats"), " ", true)
            local has_texture_format
            
            for k = 1, #resource_texture_formats do
              if texture_format == resource_texture_formats [k] then
                has_texture_format = true
                break
              end
            end
            
            if platform == resource_platform and has_texture_format then
              local resolutions = SplitString (resource_node:Get ("Resolutions"), " ", true)
              
              for k = 1, #resolutions do
                local components = SplitString (resolutions [k], "x", true)
                local width      = components [1]
                local height     = components [2]
            
                local is_new_optimal = false
            
                if optimal_resource < 0 then
                  is_new_optimal = true
                else
                  is_new_optimal = IsResourceOptimal (window_width, window_height, width, height, optimal_resource_width, optimal_resource_height)
                end
            
                if is_new_optimal then
                  optimal_resource        = j
                  optimal_resource_width  = width
                  optimal_resource_height = height
                end
              end
            end
          end
        end
        
        if optimal_resource >= 0 then
          if constants_node then
            ACHIEVEMENTS_GROUPS_COUNT = constants_node:Get ("AchievementsGroupsCount", ACHIEVEMENTS_GROUPS_COUNT) + 0
          end
        
          if not is_supported and is_supported_after_update then
            game.scripts.ui_windows.ShowUpdateApplicationWindow ()
            return
          end

          local resource_node = resources_node:Child (optimal_resource)
          local resolution    = optimal_resource_width .. "x" .. optimal_resource_height
          
          local path_replacements = {
            { token = "{version}",    replacement = version },
            { token = "{resolution}", replacement = resolution },
            { token = "{format}",     replacement = texture_format }
          } 
          
          local version             = version
          local update_from_version = update_from_version
          local path                = GetDownloadsManager ().GetGamesResourcesServerPath () .. ProcessStringWithReplacements (resource_node:Get ("Path"), path_replacements)
     
          local function mount_lobby_handler ()
            if impl.next_screen == "loader" then
              SetCurrentScreen ("lobby")
            else
              impl.resources_loaded = true

              ActivateTextLine (impl.loading_text_line)

              SetNextScreen (impl)
            end
          end
     
          if not IsLobbyDownloaded (version, update_from_version) then
            ActivateSprites ("bar", "overlay", "percent_bg")
            ActivateTextLine ("loading_percent_text_zone")
         
            impl.spark_animation.Show ("animate")

            impl.progress = 0
            UpdateLoadingProgress (impl)

            UnmountLobby ()

            Common.File.Remove (LOBBY_DOWNLOAD_PATH)
           
            if not Common.File.IsFileExist (LOBBY_DOWNLOAD_PATH) then
              Common.File.Mkdir (LOBBY_DOWNLOAD_PATH)
            end
            
            local last_progress = 0

            DeactivateTextLine (impl.loading_text_line)

            GetDownloadsManager ().DownloadFile (path, LOBBY_DOWNLOAD_PAK_PATH, function (status, arg2, arg3)
              if status == Common.BackgroundCopyStateStatus.InProgress then
                local progress = math.floor (arg3 / arg2 * 100)

                if progress > last_progress then
                  last_progress = progress
                
                  impl.progress = progress / 100
                  UpdateLoadingProgress (impl)
                end
              elseif status == Common.BackgroundCopyStateStatus.Finished then
                Common.File.PostString (LOBBY_DOWNLOADED_VERSION_PATH, 
                  string.format ("<?xml version='1.0' encoding='utf-8'?> <info version='%s' platform='%s' resolution='%s' format='%s'/>", tostring (version), GetPlatform (), resolution, ApplicationDataGet (TEXTURE_FORMATS) [1]))
                ExecuteOnNextUpdate (function ()
                  MountLobby (mount_lobby_handler)
                end)
              elseif status == Common.BackgroundCopyStateStatus.Failed then
                game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
                  SetInputState (false)              
                  HsConnectionDisconnect ()
                  SetCurrentScreen ("main_menu")
                end, nil, GetLocalizedString ("CANCEL"), { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
              end
            end, nil, nil, true)
          else
            MountLobby (mount_lobby_handler)
          end

          return     
        end
      end
    end
  end
  
  game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_ERROR, function ()
    SetInputState (false)              
    HsConnectionDisconnect ()
    SetCurrentScreen ("main_menu")
  end, nil, GetLocalizedString ("CANCEL"), { GetLocalizedString ("CONNECTION_ERROR") })
end

local function LoadResourcesIndex (impl, suffix)
  local games_resources_index
  local promo_actions_index
  
  if suffix then
    games_resources_index = GetPathWithoutExtension (GAMES_RESOURCES_INDEX) .. suffix .. "." .. GetPathExtension (GAMES_RESOURCES_INDEX) 
    promo_actions_index   = GetPathWithoutExtension (PROMO_ACTIONS_INDEX) .. suffix .. "." .. GetPathExtension (PROMO_ACTIONS_INDEX)
  else
    games_resources_index = GAMES_RESOURCES_INDEX
    promo_actions_index   = PROMO_ACTIONS_INDEX
  end

  GetDownloadsManager ().LoadString (GetDownloadsManager ().GetGamesResourcesServerPath () .. games_resources_index, function (xml_string)
    if impl.screen_lost then return end
  
    local slots_resources_node = Common.StringNode.LoadXmlFromString (xml_string)
    local os_type              = ApplicationDataGet (OS_TYPE)
        
    if string.find (os_type, "Mac OS X") or string.find (os_type, "Windows") or string.find (os_type, "TabletOS") then
      os_type = "iOS" 
    end
        
    ApplicationDataSet (LATEST_APP_ENGINE_VERSION, slots_resources_node:Find ("LatestAppEngine"):Get (os_type) + 0)
    ApplicationDataSet (LATEST_PROTOCOL_ENGINE_VERSION, slots_resources_node:Find ("LatestProtocolEngine"):Get (os_type) + 0)
    ApplicationDataSet (GAMES_RESOURCES_XML, slots_resources_node)

    GetDownloadsManager ().LoadString (GetDownloadsManager ().GetGamesResourcesServerPath () .. promo_actions_index, function (xml_string)
      if impl.screen_lost then return end
  
      local promo_actions_node = Common.StringNode.LoadXmlFromString (xml_string)

      ApplicationDataSet (PROMO_ACTIONS_XML, promo_actions_node)
      ApplicationDataSet (LAST_GAMES_RESOURCES_DOWNLOAD_TIME, game.main_timer.GetTime ())

      if NeedsDownloadLobby () then
        DownloadLobby (impl)
      else
        impl.resources_loaded = true
        
        SetNextScreen (impl)
      end
    end, OnDownloadError)
  end, OnDownloadError)
end

local function LoadServerSettings (handler)
  local hs_port = GetHsPort ()

  local suffix_url

  if hs_port == HS_LIVE_PORT then
    suffix_url = "https://slots.plrm.in/mobile/mobilesettings.ashx?e=l"
  elseif hs_port == HS_QA_PORT then
    suffix_url = "https://slots.plrm.in/mobile/mobilesettings.ashx?e=q"
  else
    suffix_url = "https://slots.plrm.in/mobile/mobilesettings.ashx?e=d"
  end

  local system_properties          = System.Application.SystemProperties
  local app_version_property_name  = "ApplicationVersion"
  local device_model_property_name = "DeviceModel"
  local uid_property_name          = "UUID"
  local cellular_property_name     = "CellularOnlyInternet"

  if system_properties:IsPresent (app_version_property_name) then
    suffix_url = suffix_url .. "&v=" .. system_properties:GetString (app_version_property_name)
  end

  if system_properties:IsPresent (device_model_property_name) then
    suffix_url = suffix_url .. "&dv=" .. system_properties:GetString (device_model_property_name)
  end

  if system_properties:IsPresent (uid_property_name) then
    suffix_url = suffix_url .. "&uid=" .. system_properties:GetString (uid_property_name)
  end

  if system_properties:IsPresent (cellular_property_name) and system_properties:GetInteger (cellular_property_name) == 1 then
    suffix_url = suffix_url .. "&wf=0"
  else
    suffix_url = suffix_url .. "&wf=1"
  end

  local os_code = "i"
  
  if ApplicationDataGet (OS_TYPE) == "Android" then
    if GetStoreManager ().GetStoreName () == AMAZON_STORE_NAME then
      os_code = "amz"
    else
      os_code = "a"
    end
  end

  suffix_url = suffix_url .. "&d=" .. os_code

  GetDownloadsManager ().LoadString (GetPercentEscapedUrl (suffix_url), function (json_string)
    local status, err = pcall (function ()
      local json     = Common.StringNode.LoadJsonFromString (json_string)
      local path     = json:Find ("path"):Attribute (0)
      local push_url = GetNodeChildAttribute (json, "push")

      GetSocialManager ().SetServers (GetNodeChildAttribute (json, "vk"), GetNodeChildAttribute (json, "fb"), GetNodeChildAttribute (json, "guest"))

      if push_url then
        ApplicationDataSet (PUSH_NOTIFICATIONS_SERVER_URL, push_url)
      end
      
      handler (path)
    end)
      
    if not status then
      print (err)
      OnDownloadError ()
    end
  end, OnDownloadError)
end

local function OnResourcesLoaded (impl, screen_name)
  impl.next_screen = screen_name

  if screen_name == "main_menu" then
    UnmountLobby ()
  end

  DeactivateSprites ("bar", "overlay", "percent_bg")
  DeactivateTextLine ("loading_percent_text_zone")

  if impl.is_battle_loader then
    --TODO
  else
    impl.spark_animation.Hide ()

    ActivateTextLine (impl.loading_text_line)
    impl.loading_text_line.node.Text = GetLocalizedString ("LOADER.CONNECTING")
  end
  
  if impl.next_screen == "lobby" or impl.next_screen == "loader" then
    if not ApplicationDataGet (LAST_GAMES_RESOURCES_DOWNLOAD_TIME) then
      ApplicationDataSet (LAST_GAMES_RESOURCES_DOWNLOAD_TIME, game.main_timer.GetTime () - GAMES_RESOURCES_DOWNLOAD_PERIOD * 2)
    end
    
    if game.main_timer.GetTime () > ApplicationDataGet (LAST_GAMES_RESOURCES_DOWNLOAD_TIME) + GAMES_RESOURCES_DOWNLOAD_PERIOD then
      LoadServerSettings (function (suffix)
        LoadResourcesIndex (impl, suffix)
      end)
      
      return
    else
      if NeedsDownloadLobby () then
        DownloadLobby (impl)
        return
      end
    end
  end
  
  impl.resources_loaded = true
        
  SetNextScreen (impl)  
end

local function OnSizeChanged (impl, window_width, window_height)
  UpdateGameCameraSize (window_width, window_height, impl.min_camera_width, impl.min_camera_height, 
                        impl.min_camera_center_x, impl.min_camera_center_y, impl.min_ar, 
                        impl.max_camera_width, impl.max_camera_height, impl.max_camera_center_x, 
                        impl.max_camera_center_y, impl.max_ar, impl.camera)

  UpdateProgressScissor (impl)
  
  if impl.is_battle_loader and not impl.you_avatar then
    local you_avatar_sprite      = GetSprite ("you_avatar")
    local opponent_avatar_sprite = GetSprite ("opponent_avatar")
    local you_avatar_layout      = you_avatar_sprite.layout
    local opponent_avatar_layout = opponent_avatar_sprite.layout
    local you_avatar_size        = { x = you_avatar_layout:Get ("Width") + 0, y = you_avatar_layout:Get ("Height") + 0 }
    local opponent_avatar_size   = { x = opponent_avatar_layout:Get ("Width") + 0, y = opponent_avatar_layout:Get ("Height") + 0 }
    local screen_scale           = window_height / (impl.camera.Top - impl.camera.Bottom)
    
    impl.you_avatar = CreateUrlSprite (URL_SPRITE_MODE_CACHE_FILE, impl.players_info.you.photo, {}, 
                                       { material = you_avatar_sprite.node.Material, tex_size = StringToVec (you_avatar_sprite.layout:Get ("TexSize")), tex_offset = StringToVec (you_avatar_sprite.layout:Get ("TexOffset")) },
                                       you_avatar_size, you_avatar_size.x * PHOTO_ACTIVITY_INDICATOR_FACTOR, screen_scale)
      
    local avatar_sprite = impl.you_avatar.GetSprite ()
      
    avatar_sprite.Position = vec3 (you_avatar_layout:Get ("Left") + you_avatar_size.x / 2, -you_avatar_layout:Get ("Top") - you_avatar_size.y / 2, you_avatar_sprite.z - EPS)    
      
    avatar_sprite:BindToParent (game.current_screen.node)
      
    impl.you_avatar.OnShow ()
    
    impl.opponent_avatar = CreateUrlSprite (URL_SPRITE_MODE_CACHE_FILE, impl.players_info.opponent.photo, {}, 
                                            { material = opponent_avatar_sprite.node.Material, tex_size = StringToVec (opponent_avatar_sprite.layout:Get ("TexSize")), tex_offset = StringToVec (opponent_avatar_sprite.layout:Get ("TexOffset")) }, 
                                            opponent_avatar_size, opponent_avatar_size.x * PHOTO_ACTIVITY_INDICATOR_FACTOR, screen_scale)
      
    local avatar_sprite = impl.opponent_avatar.GetSprite ()
      
    avatar_sprite.Position = vec3 (opponent_avatar_layout:Get ("Left") + opponent_avatar_size.x / 2, -opponent_avatar_layout:Get ("Top") - opponent_avatar_size.y / 2, opponent_avatar_sprite.z - EPS)    
      
    avatar_sprite:BindToParent (game.current_screen.node)
      
    impl.opponent_avatar.OnShow ()
  end
  
  if not impl.is_battle_loader then
    local camera_width  = impl.camera.Right - impl.camera.Left
    local camera_height = impl.camera.Top - impl.camera.Bottom
    
    local scissor = CreateScissor (impl.progress_sprite_left, 0, camera_width, camera_height)

    impl.spark_animation.GetNode ().Scissor = scissor
    
    if impl.spark_animation.GetNode ().Scene ~= nil then
      scissor:BindToScene (impl.spark_animation.GetNode ().Scene)
    end 
  end
end

local function SetProgress (impl, progress)
  if progress == 0 then return end

  impl.progress = progress
  
  UpdateLoadingProgress (impl)
  
  if impl.is_battle_loader then
    impl.last_battle_loading_progress = impl.last_battle_loading_progress or 0 
  
    if impl.progress - impl.last_battle_loading_progress >= 0.1 then
      impl.last_battle_loading_progress = impl.progress
      
      HsConnectionSendMessage (1, { event = "battle.loading", progress = math.floor (impl.progress * 90) })
    end 
  end
end

local function SetOpponentLoadingProgress (impl, progress)
  impl.opponent_progress = progress
  
  UpdateLoadingProgress (impl)
  
  if impl.opponent_progress > 1 - EPS then
    impl.opponent_ready = true
    
    SetNextScreen (impl)
  end
end

local function Update (impl)
  if impl.is_battle_loader then return end
  
  local time = game.current_screen.timers.default:GetTime ()
  
  local animation_time = time - impl.loading_animation_start_time
  
  if animation_time > game.current_screen.vars.text_animation_duration then
    animation_time                    = time % impl.text_animation_duration
    impl.loading_animation_phase      = not impl.loading_animation_phase
    impl.loading_animation_start_time = time - animation_time
  end 
  
  local phase = animation_time / impl.text_animation_duration
  
  if not impl.loading_animation_phase then
    phase = 1 - phase
  end
  
  phase = impl.text_min_phase + (phase * (1 - impl.text_min_phase))

  local loader_dt = game.current_screen.timers.default.GetTime () - impl.last_loader_sprite_change_time
  
  if loader_dt >= 1 / 20 then
    impl.last_loader_sprite_change_time = game.current_screen.timers.default.GetTime ()
  
    impl.loading_text_line.node.Color = impl.loading_text_line_color * phase
  
    DeactivateSprite (impl.loader_sprites [impl.active_loader_sprite % #impl.loader_sprites + 1])

    impl.active_loader_sprite = impl.active_loader_sprite + 1

    ActivateSprite (impl.loader_sprites [impl.active_loader_sprite % #impl.loader_sprites + 1])
  end
end

local function OnBattleInterrupt (impl)
  if not impl.is_battle_loader then return end
  
  StopMusic ()
  
  impl.battle_interrupted = true
  
  game.scripts.ui_windows.ShowAlertWindow (nil, nil, function ()
      SetCurrentScreen ("lobby")
    end, nil, { GetLocalizedString ("BATTLE.OPPONENT_LEAVE_ALERT") })
end

local function OnLostScreen (impl) 
  impl.screen_lost = true

  GetDownloadsManager ().CancelDownloadFile (LOBBY_DOWNLOAD_PAK_PATH)
end

local function CreateScreen (xml_desc)
  local impl = {}

  impl.is_battle_loader = GetNodeChildAttribute (xml_desc, "BattleLoader", "false") == "true"

  impl.progress = 0

  impl.camera = game.current_screen.cameras ["ui_camera"].node
  
  local min_camera_zone = GetSprite ("min_camera_zone")
  
  min_camera_zone = min_camera_zone.layout
  
  impl.min_camera_center_x = min_camera_zone:Get ("Left") + min_camera_zone:Get ("Width") / 2
  impl.min_camera_center_y = -min_camera_zone:Get ("Top") - min_camera_zone:Get ("Height") / 2
  impl.min_camera_width    = min_camera_zone:Get ("Width")  + 0
  impl.min_camera_height   = min_camera_zone:Get ("Height") + 0

  impl.min_ar = impl.min_camera_width / impl.min_camera_height

  local max_camera_zone = GetSprite ("max_camera_zone")
  
  max_camera_zone = max_camera_zone.layout
  
  impl.max_camera_center_x = max_camera_zone:Get ("Left") + max_camera_zone:Get ("Width") / 2
  impl.max_camera_center_y = -max_camera_zone:Get ("Top") - max_camera_zone:Get ("Height") / 2
  impl.max_camera_width    = max_camera_zone:Get ("Width")  + 0
  impl.max_camera_height   = max_camera_zone:Get ("Height") + 0
  
  impl.max_ar = impl.max_camera_width / impl.max_camera_height
  
  game.current_screen.timers.screen.Start ()
  game.current_screen.timers.default.Start ()

  if impl.is_battle_loader then
    ExecuteOnNextUpdate (function ()
      PlaySound ("battle_opponent_found")
      PlayMusic ("battle_loading")
    end)
  
    impl.opponent_progress = ApplicationDataGet (BATTLE_OPPONENT_LOADING_PROGRESS) or 0

    impl.players_info = ApplicationDataGet (BATTLE_PLAYERS_INFO)
    
    --[[DEBUG
    impl.players_info = {
      you = {
        first_name = "Ivanovich",
        last_name  = "Ivanovich",
        photo      = "http://cs306610.vk.me/u93339557/e_89c9bf5a.jpg",
        level      = 888,
        wins       = 888,
        loses      = 888
      },
      opponent = {
        first_name = "Ivanovich",
        last_name  = "Ivanovich",
        photo      = "http://cs306610.vk.me/u93339557/e_89c9bf5a.jpg",
        level      = 888,
        wins       = 888,
        loses      = 888
      }
    }
    --END DEBUG]]
    
    GetTextLine ("you_level").node.Text            = tostring (impl.players_info.you.level)
    GetTextLine ("opponent_level").node.Text       = tostring (impl.players_info.opponent.level)
    GetTextLine ("you_name").node.Text             = impl.players_info.you.first_name
    GetTextLine ("opponent_name").node.Text        = impl.players_info.opponent.first_name
    GetTextLine ("you_wins_value").node.Text       = tostring (impl.players_info.you.wins)
    GetTextLine ("opponent_wins_value").node.Text  = tostring (impl.players_info.opponent.wins)
    GetTextLine ("you_loses_value").node.Text      = tostring (impl.players_info.you.loses)
    GetTextLine ("opponent_loses_value").node.Text = tostring (impl.players_info.opponent.loses)
    
    impl.progress_sprite = GetSprite ("you_progress_bar")
  
    impl.progress_sprite_left  = impl.progress_sprite.layout:Get ("Left")
    impl.progress_sprite_width = impl.progress_sprite.layout:Get ("Width")    

    impl.opponent_progress_sprite = GetSprite ("opponent_progress_bar")
  
    impl.opponent_progress_sprite_left  = impl.opponent_progress_sprite.layout:Get ("Left")
    impl.opponent_progress_sprite_width = impl.opponent_progress_sprite.layout:Get ("Width")

    SetOpponentLoadingProgress (impl, impl.opponent_progress)    
  else
    if ApplicationDataGet (IS_SLOT_SCREEN_LOADING) then
      local slot_id = GetProfileProperty ("CurrentSlotId") + 0
      local slots   = GetSlots ()
      local features_in_slot = {}

      for i = 1, #slots do
        local slot_desc = slots [i]

        if slot_desc.id == slot_id then
          local features = slot_desc.features

          if features [SLOT_FEATURE_DOUBLE_GAME] then
            table.insert (features_in_slot, "double_game")
          end
          if features [SLOT_FEATURE_DOUBLE_WILD] then
            table.insert (features_in_slot, "double_wild")
          end
          if features [SLOT_FEATURE_TRIPLE_WILD] then
            table.insert (features_in_slot, "triple_wild")
          end
          if features [SLOT_FEATURE_BONUS_ANYWHERE] then
            table.insert (features_in_slot, "bonus_aniwhere")
          end
          if features [SLOT_FEATURE_EXPANDED_WILD] then
            table.insert (features_in_slot, "expanding_wild")
          end
          if features [SLOT_FEATURE_FREE_SPINS_SELECT] then
            table.insert (features_in_slot, "freespins_select")
          end
          if features [SLOT_FEATURE_BONUS_SPINS] then
            table.insert (features_in_slot, "bonus_spins")
          end

          break
        end
      end

      if #features_in_slot > 1 then
        local length_features = game.current_screen.vars.length_features
        local length_margin = game.current_screen.vars.length_margin
        local left_coordinates = - (((length_features + length_margin) * #features_in_slot - length_margin) / 2  - length_features / 2)

        for i = 1, #features_in_slot do
          game.current_screen.vars [features_in_slot [i] .. "_position"] = left_coordinates + (length_features + length_margin) * (i - 1)

          ActivateSprite (features_in_slot [i])
        end
      elseif #features_in_slot == 1 then
        ActivateSprite (features_in_slot [1])
      end
    end

    impl.text_animation_duration = game.current_screen.vars.text_animation_duration
    impl.text_min_phase          = game.current_screen.vars.text_min_phase
  
    impl.loading_text_line       = GetTextLine ("loading")
    impl.loading_text_line_color = impl.loading_text_line.node.Color
    
    impl.loading_animation_phase      = true
    impl.loading_animation_start_time = game.current_screen.timers.default:GetTime ()
  
    impl.progress_sprite = GetSprite ("bar")
  
    impl.progress_sprite_left  = impl.progress_sprite.layout:Get ("Left")
    impl.progress_sprite_width = impl.progress_sprite.layout:Get ("Width")

    impl.loader_sprites = {}
    
    for i = 1, 120, 2 do
      table.insert (impl.loader_sprites, GetSprite (string.format ("Loader_anim_%u", i)))
    end
       
    local spark_frames = CreateFramesNames ("iiskra_anim_%u", 1, 15)
    
    impl.spark_animation = CreateAnimationSpriteClass ( { animate = spark_frames }, spark_frames [1], 15 ).CreateInstance ()
    
    SetAnimationPosition (impl.spark_animation, spark_frames [1], GetSprite ("iiskra_anim_1").z - EPS)  
  
    impl.spark_animation.Show ("animate")
    
    ActivateSprite (impl.loader_sprites [1])
  
    impl.active_loader_sprite = 0
  
    impl.last_loader_sprite_change_time = game.current_screen.timers.default.GetTime ()
  end

  ApplicationDataSet (IS_SLOT_SCREEN_LOADING, nil)
  
  local methods = {}

  methods.OnAllStoredMessagesReceived = function () OnAllStoredMessagesReceived (impl) end
  methods.OnBattleInterrupt           = function () OnBattleInterrupt (impl) end
  methods.OnLostScreen                = function () OnLostScreen (impl) end
  methods.OnResourcesLoaded           = function (screen_name) OnResourcesLoaded (impl, screen_name) end
  methods.OnSizeChanged               = function (window_width, window_height) OnSizeChanged (impl, window_width, window_height) end
  methods.SetProgress                 = function (progress) SetProgress (impl, progress) end
  methods.SetOpponentLoadingProgress  = function (progress) SetOpponentLoadingProgress (impl, progress) end
  methods.Update                      = function (dt) Update (impl) end

  Update (impl)

  System.Application.ScreenSaverState = true  

  if ApplicationDataGet (SLOT_TO_DOWNLOAD) then
    local slot_to_download = ApplicationDataGet (SLOT_TO_DOWNLOAD)
  
    ApplicationDataSet (SLOT_TO_DOWNLOAD, nil)
  
    ActivateSprites ("bar", "overlay", "percent_bg")
    ActivateTextLine ("loading_percent_text_zone")
    DeactivateTextLine (impl.loading_text_line)
    
    impl.spark_animation.Show ("animate")

    impl.progress = 0
    UpdateLoadingProgress (impl)

    GetDownloadsManager ().DownloadSlot (slot_to_download, function (progress)
        impl.progress = progress
        UpdateLoadingProgress (impl)
      end, function ()
        SetSlotScreen (slot_to_download)
    end)
  end
  
  return methods
end

return CreateScreen
