local MINIMUM_FREE_SPACE_FOR_LOGIN = 1024 * 1024 * 15

local function CheckLoginAllowed ()
  if Common.File.GetFreeSpace (APP_DATA_PATH) > MINIMUM_FREE_SPACE_FOR_LOGIN and 
     Common.File.GetFreeSpace ("/system/temp/") > MINIMUM_FREE_SPACE_FOR_LOGIN then return true end
  
  game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
  
  return false
end

local function OnGuestLogin (impl)
  LogFlurryEvent ("ButtonClick", { ButtonName = "MainMenuGuestLogin" })  

  if not CheckLoginAllowed () then return end

  PlaySound ('main_menu_click')

  System.Application.PostNotification ("LocalNotifications SchedulePeriodic|SocialLogin|" .. GetLocalizedString ("LOCAL_NOTIFICATION_ALERT.SOCIAL_LOGIN") .. "|week|" .. 60 * 60 * 24 * 7)

  GuestLogin ()
end

local function OnVKLogin (impl)
  LogFlurryEvent ("ButtonClick", { ButtonName = "MainMenuVKLogin" })  

  if not CheckLoginAllowed () then return end

  PlaySound ('main_menu_click')

  VKLogin ()
end

local function OnFacebookLogin (impl)
  LogFlurryEvent ("ButtonClick", { ButtonName = "MainMenuFBLogin" })  

  if not CheckLoginAllowed () then return end

  PlaySound ('main_menu_click')

  FacebookLogin ()
end

local function UpdateSelectedLanguage (impl)
  DeactivateSprites ("language_button_ru-RU_l", "language_button_en-US_l", 
                     "language_button_de-DE_l", "language_button_fr-FR_l",  "language_button_it-IT_l", 
                     "language_button_es-ES_l", "language_button_tr-TR_l", "language_button_ru-RU_p", 
                     "language_button_en-US_p", "language_button_de-DE_p", "language_button_fr-FR_p", 
                     "language_button_it-IT_p", "language_button_es-ES_p", "language_button_tr-TR_p")
                     
  local current_locale = GetCurrentLocale ()
  
  ActivateSprite ("language_button_" .. current_locale .. "_l")
end

local function RepositionLanguagePanel (impl, window_width)
  local language_panel_sprites = { "language_button_underlay", "language_button_ru-RU_l", "language_button_en-US_l", 
                                   "language_button_de-DE_l", "language_button_fr-FR_l",  "language_button_it-IT_l", 
                                   "language_button_es-ES_l", "language_button_tr-TR_l", "language_button_ru-RU_p", 
                                   "language_button_en-US_p", "language_button_de-DE_p", "language_button_fr-FR_p", 
                                   "language_button_it-IT_p", "language_button_es-ES_p", "language_button_tr-TR_p" }

  local offset = impl.camera.Left

  for i = 1, #language_panel_sprites do
    local sprite = GetSprite (language_panel_sprites [i])
    
    SetSpritePosition (sprite, sprite.layout:Get ("Left") + sprite.layout:Get ("Width") / 2 + offset, -sprite.layout:Get ("Top") - sprite.layout:Get ("Height") / 2)
  end

  UpdateSelectedLanguage (impl)
end

local function OnSizeChanged (impl, window_width, window_height)
  local camera_width = window_width / window_height * impl.camera_height 

  impl.camera.Left  = math.floor (impl.camera_center_x - camera_width / 2)
  impl.camera.Right = impl.camera.Left + camera_width
  
  RepositionLanguagePanel (impl, window_width)
end

local function CreateFramesNames (format, from, to, step)
  local return_value = {}
  
  for i = from, to, step do
    table.insert (return_value, string.format (format, i))
  end
  
  return return_value
end

local function SetAnimationPosition (animation, reference, z)
  local sprite = GetSprite (reference)
  local layout = sprite.layout
  
  animation.SetPosition (layout:Get ("Left") + layout:Get ("Width") / 2, -layout:Get ("Top") - layout:Get ("Height") / 2, z or sprite.z)
end

local function CreateScreen (xml_desc)
  local impl = {}  

  HsConnectionDisconnect ()

  System.Application.PostNotification ("LocalNotifications RemoveAll|Bonus")

  impl.camera = game.current_screen.cameras ["ui_camera"].node
  
  impl.camera_center_x = impl.camera.Left + (impl.camera.Right - impl.camera.Left) / 2
  impl.camera_height   = impl.camera.Top - impl.camera.Bottom

  UpdateSelectedLanguage (impl)

  local public_methods = {}

  public_methods.FacebookLogin  = function () OnFacebookLogin (impl) end
  public_methods.GuestLogin     = function () OnGuestLogin (impl) end
  public_methods.OnSizeChanged  = function (window_width, window_height) OnSizeChanged (impl, window_width, window_height) end
  public_methods.ChangeLanguage = function () game.scripts.ui_windows.ShowChangeLanguageWindow () end
  public_methods.VKLogin        = function () OnVKLogin (impl) end

  if GetProfileProperty (WAS_LOGGED_IN_ATTRIBUTE) == "true" then
    ActivateSprite ("device_id_d")
    ActivateTextLine ("device_id_d_text")
  else
    ActivateSprite ("device_id_l")
    ActivateTextLine ("device_id_l_text")
    
    local button_frames = CreateFramesNames ("Play_btn_anim_%d", 1, 119, 2)

    impl.button_animation = CreateAnimationSpriteClass ( { animate = button_frames }, button_frames [1], 15 ).CreateInstance ()

    SetAnimationPosition (impl.button_animation, button_frames [1], GetSprite ("device_id_p").z - 0.0001)
    
    impl.button_animation.Show ("animate")
  end

  if GetCurrentLocale () == "ru-RU" then
    ActivateSprite ("vk_l_ru-RU")
  else
    ActivateSprite ("vk_l")
  end

  if GetHsPort () ~= HS_LIVE_PORT then
    local function UpdateDebugButtons ()
      local port = GetHsPort ()
      
      SetSpriteActive ("debug_dev", port ~= HS_DEV_PORT)
      SetSpriteActive ("debug_qa", port ~= HS_QA_PORT)
      SetSpriteActive ("debug_live", port ~= HS_LIVE_PORT)
    end

    public_methods.SetServerType = function (type)
      Common.File.Remove (DOWNLOADS_PATH)
      SetProfileProperty (OVERRIDE_PORT_ATTRIBUTE, type) 
      SetHsPort (type)
      
      UpdateDebugButtons () 
    end

    UpdateDebugButtons () 
  end

  return public_methods
end

return CreateScreen
