ALERT_TYPE_ERROR    = {}
ALERT_TYPE_GIFT     = {}
ALERT_TYPE_LANGUAGE = {}
ALERT_TYPE_QUESTION = {}

ALERT_TYPE_ERROR.resource    = "data/res/materials/alert_window_type_error_layers.xmtl"
ALERT_TYPE_ERROR.sprite      = {"alert_window_error_1"}
ALERT_TYPE_GIFT.resource     = "data/res/materials/alert_window_type_gift_layers.xmtl"
ALERT_TYPE_GIFT.sprite       = {"alert_window_gift_1"}
ALERT_TYPE_LANGUAGE.resource = "data/res/materials/alert_window_type_language_layers.xmtl"
ALERT_TYPE_LANGUAGE.sprite   = {"alert_window_language"}
ALERT_TYPE_QUESTION.resource = "data/res/materials/alert_window_type_question_layers.xmtl"
ALERT_TYPE_QUESTION.sprite   = {"alert_window_question_1"}

local ALERT_AUTOCLOSE_DELAY           = 3.5
local EPS                             = 0.0001
local MAX_INVITE_FRIENDS              = 200
local PHOTO_ACTIVITY_INDICATOR_FACTOR = 0.4
local RESUME_GAME_DELAY               = 0.3     --resume game after window hiding with delay, to prevent sound popping on window change

local local_methods = {}

local function OnSizeChanged (impl, window_width, window_height)
  if impl.window_width == window_width and impl.window_height == window_height then return end

  for i = 1, #impl.cameras do
    local camera        = impl.cameras [i]
    local camera_height = camera.min_height
    local camera_width  = camera_height * window_width / window_height

    if camera_width < camera.min_width then
      camera_width  = camera.min_width
      camera_height = camera_width / window_width * window_height
    end

    camera_width  = math.floor (camera_width)
    camera_height = math.floor (camera_height)

    camera.node.Left   = math.floor (camera.center_x - camera_width / 2)
    camera.node.Right  = camera.node.Left + camera_width
    camera.node.Bottom = math.floor (camera.center_y - camera_height / 2)
    camera.node.Top    = camera.node.Bottom + camera_height
  end

  impl.window_width  = window_width
  impl.window_height = window_height
end

local_methods.ProcessWindowQueue = function (impl, stack_name)
  local queue = impl.windows_queue

  if stack_name then
    impl.windows_queues = impl.windows_queues or {}

    if not impl.windows_queues [stack_name] then
      impl.windows_queues [stack_name] = {}
    end

    queue = impl.windows_queues [stack_name]
  end

  if #queue > 0 then
    local first_function = queue [1]

    table.remove (queue, 1)

    first_function ()
  end
end

local function HideWindow (impl, stack_name, dont_process_queue)
  ClearUrlSpritesCache ()

  local stack = impl.windows_stack
  
  if stack_name then
    impl.windows_stacks = impl.windows_stacks or {}
    
    if not impl.windows_stacks [stack_name] then
      impl.windows_stacks [stack_name] = {}
    end
    
    stack = impl.windows_stacks [stack_name]
  end
                           
  if not stack_name and impl.hide_window_function then
    impl.hide_window_function ()
  end

  if #stack == 0 then return end

  local top_windows_stack_entry = stack [#stack]

  local active_window = top_windows_stack_entry.window

  for i = 1, #active_window.all_text_lines do
    DeactivateTextLine (active_window.all_text_lines [i])
    active_window.all_text_lines [i].node.Text = ""
  end

  DeactivateTextLines (unpack (top_windows_stack_entry.additional_text_lines))

  for i = 1, #active_window.all_sprites do
    active_window.all_sprites [i].node.Material = "transparent"    
  end

  local active_additional_sprites = top_windows_stack_entry.additional_sprites

  ExecuteOnNextUpdate (function ()
    for i = 1, #active_window.all_sprites do
      if active_window.all_sprites [i].node.Material == "transparent" then
        local sprite_name = active_window.all_sprites [i].name
      
        ActivateSprite (sprite_name)
        
        if game.scripts.buttons then
          game.scripts.buttons.SetButtonActive (sprite_name, false)
        end
      end
    end

    ExecuteOnNextUpdate (function ()
      local current_active_window
      local current_active_additional_sprites

      if #stack > 0 then
        local current_top_windows_stack_entry = stack [#stack]

        current_active_window             = current_top_windows_stack_entry.window
        current_active_additional_sprites = current_top_windows_stack_entry.additional_sprites
      end

      if current_active_window then
        for i = 1, #active_window.sprites do
          local sprite = active_window.sprites [i]
          local active = false

          for j = 1, #current_active_window.sprites do
            if sprite == current_active_window.sprites [j] then
              active = true
              break
            end
          end
        
          if not active then
            DeactivateSprite (sprite)
          end
        end

        if active_additional_sprites then
          for i = 1, #active_additional_sprites do
            local sprite = active_additional_sprites [i]
            local active = false

            for j = 1, #current_active_additional_sprites do
              if sprite == current_active_additional_sprites [j] then
                active = true
                break
              end
            end

            if not active then
              DeactivateSprite (sprite)
            end
          end
        end
      else
        DeactivateSprites (unpack (active_window.sprites))

        if active_additional_sprites then
          DeactivateSprites (unpack (active_additional_sprites))
        end
      end

      local dont_deactivate_camera = false

      if current_active_window and active_window.camera == current_active_window.camera then
        dont_deactivate_camera = true
      end

      if not dont_deactivate_camera and impl.windows_stacks then
        for _, check_stack in pairs (impl.windows_stacks) do
          if #check_stack > 0 and active_window.camera == check_stack [#check_stack].window.camera then
            dont_deactivate_camera = true
            break
          end
        end
      end
        
      if not dont_deactivate_camera and not top_windows_stack_entry.dont_disable_camera then
        SetCameraState (active_window.camera, false)
      end

      for i = 1, #active_window.all_sprites do
        if active_window.all_sprites [i].node.Material == "transparent" then
          DeactivateSprite (active_window.all_sprites [i].name)
        end
      end

      collectgarbage ("collect")

      Engine.ResourceManager.FlushUnusedResources ()

      --restore sprites materials
      for i = 1, #active_window.all_sprites do
        local item = active_window.all_sprites [i]

        if item.node.Material == "transparent" then
          item.node.Material = item.material
        end
      end
    end)
  end)

  table.remove (stack)

  if active_window.on_hide then
    active_window.on_hide ()
  end

  if not stack_name and (#stack == 0 or stack [#stack].dont_pause_game) then
    impl.resume_game_action = DelayedAction ("ui", RESUME_GAME_DELAY, ResumeGame)
  end

  top_windows_stack_entry.additional_sprites           = nil
  top_windows_stack_entry.additional_text_lines        = nil
  top_windows_stack_entry.on_ok_handler                = nil
  top_windows_stack_entry.on_cancel_handler            = nil
  top_windows_stack_entry.resource_binding             = nil
  top_windows_stack_entry.additional_resource_bindings = nil

  if not stack_name and not top_windows_stack_entry.dont_disable_camera then
    SetInputState (true)
  end

  if not dont_process_queue then
    local_methods.ProcessWindowQueue (impl, stack_name)
  end
end

local function SendGifts (impl)
  if impl.gift_type == -1 then
    local friends_ids = {}

    for i = 1, #impl.send_gifts_selected_friends_cells do
      table.insert (friends_ids, impl.send_gifts_selected_friends_cells [i].friend.login_id)
    end

    HideWindow (impl)

    game.scripts.ui_windows.ShowAlertWindow (nil, nil, nil, nil, { GetLocalizedString ("FRIENDS_INVITE_IN_PROGRESS") }, true)

    DelayedAction ("ui", 1, function ()
      GetSocialManager ().InviteFriends (friends_ids, function ()
        game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("FRIENDS_INVITED") }, nil, nil, true)
      end)
    end)
  else
    local friends_ids = {}

    for i = 1, #impl.send_gifts_selected_friends_cells do
      table.insert (friends_ids, impl.send_gifts_selected_friends_cells [i].friend.login_id)
    end

    HsConnectionSendMessage (1, { event = "gifts.sent", friendsId = friends_ids, giftType = impl.gift_type })

    HideWindow (impl)

    if impl.gift_type == 1 and game.current_screen.name ~= "main_menu" then
      local slot_id = GetProfileProperty ("CurrentSlotId")

      if slot_id then
        slot_id = slot_id + 0
      end
    end

    if game.scripts.level and game.scripts.level.OnGiftsSent then
      game.scripts.level.OnGiftsSent ()
    end

    game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_GIFT, nil, nil, nil, { GetLocalizedString ("LOBBY.GIFT_SENT") }, nil, nil, true)
  end
end

local function OnSendGiftsFriendsScrolled (impl)
  game.current_screen.vars ["send_gifts_friends_scroll"] = impl.send_gifts_friends_scroll_view.RelativeScrollPosition (true).y
end

local function OnSendGiftsSelectedFriendsScrolled (impl)
  game.current_screen.vars ["send_gifts_selected_friends_scroll"] = impl.send_gifts_selected_friends_scroll_view.RelativeScrollPosition (true).y
end

local function IsAlertWindowShown (impl, type)
  if #impl.windows_stack == 0 or impl.windows_stack [#impl.windows_stack].window ~= impl.alert_window then return end
  
  if not type then return true end
  
  return type == impl.alert_window_type
end

local function ShowWindow (impl, window, additional_sprites, additional_text_lines, text_lines, on_window_shown,
                           dont_disable_camera, dont_pause_game, dont_hide_previous_windows, on_ok_handler, on_cancel_handler,
                           additional_resource_bindings, enqueue, stack_name, before_window_show, params)
  params = params or {}
                           
  if IsAlertWindowShown (impl, ALERT_TYPE_ERROR) and window ~= impl.alert_window then return end                         

  local stack = impl.windows_stack
  
  if stack_name then
    impl.windows_stacks = impl.windows_stacks or {}
    
    if not impl.windows_stacks [stack_name] then
      impl.windows_stacks [stack_name] = {}
    end
    
    stack = impl.windows_stacks [stack_name]
  end
  
  enqueue = enqueue or params.enqueue or params.enqueue_first
                           
  if enqueue and #stack > 0 then
    local queue = impl.windows_queue
  
    if stack_name then
      impl.windows_queues = impl.windows_queues or {}

      if not impl.windows_queues [stack_name] then
        impl.windows_queues [stack_name] = {}
      end

      queue = impl.windows_queues [stack_name]
    end
  
    local enqueue_position = #queue + 1
    
    if params.enqueue_first then
      enqueue_position = 1
    end
  
    table.insert (queue, enqueue_position, function ()
      ShowWindow (impl, window, additional_sprites, additional_text_lines, text_lines, on_window_shown,
                  dont_disable_camera, dont_pause_game, dont_hide_previous_windows, on_ok_handler, on_cancel_handler,
                  additional_resource_bindings, enqueue, stack_name, before_window_show, params)
    end)
    return
  end

  if dont_pause_game and #stack > 0 and not stack [#stack].dont_pause_game then
    dont_pause_game = false
  end

  if dont_hide_previous_windows and #stack > 0 and stack [#stack].window == window then
    HideWindow (impl, stack_name, true)
  end

  if not dont_hide_previous_windows then
    local windows_count = #stack

    for i = 1, windows_count do
      HideWindow (impl, stack_name, true)
    end
  end

  if before_window_show then
    before_window_show ()
  end

  local previous_windows_stack_entry

  if #stack then
    previous_windows_stack_entry = stack [#stack]
  end

  local windows_stack_entry = {}

  windows_stack_entry.dont_disable_camera = dont_disable_camera

  if previous_windows_stack_entry then
    windows_stack_entry.dont_disable_camera = windows_stack_entry.dont_disable_camera and previous_windows_stack_entry
  end

  windows_stack_entry.resource_binding             = LoadResource (window.override_resource or window.resource)
  windows_stack_entry.window                       = window
  windows_stack_entry.additional_sprites           = additional_sprites
  windows_stack_entry.additional_text_lines        = additional_text_lines
  windows_stack_entry.additional_resource_bindinds = additional_resource_bindings
  windows_stack_entry.on_ok_handler                = on_ok_handler
  windows_stack_entry.dont_process_queue_on_ok     = params.dont_process_queue_on_ok
  windows_stack_entry.on_cancel_handler            = on_cancel_handler
  windows_stack_entry.dont_pause_game              = dont_pause_game
  windows_stack_entry.hide_on_back                 = params.hide_on_back
  windows_stack_entry.on_back_handler              = params.on_back_handler
  
  window.ok_pressed = nil  

  for i = 1, #window.all_sprites do
    local item = window.all_sprites [i]

    if item.node.Material == "transparent" then
      local active = not params.dont_restore_sprites_materials
    
      for j = 1, #window.sprites do
        if item.name == window.sprites [j] then
          active = true
          break
        end
      end
      
      for j = 1, #additional_sprites do
        if item.name == additional_sprites [j] then
          active = true
          break
        end
      end
      
      if active then
        item.node.Material = item.material
      end
    end
        
    if game.scripts.buttons then
      game.scripts.buttons.SetButtonActive (item.name, true)
    end
  end

  for i = 1, #window.sprites do
    if not IsSpriteActive (window.sprites [i]) then
      ActivateSprite (window.sprites [i])
    end
  end
  
  for i = 1, #additional_sprites do
    if not IsSpriteActive (additional_sprites [i]) then
      ActivateSprite (additional_sprites [i])
    end
  end

  ActivateTextLines (unpack (additional_text_lines))

  for i = 1, #window.text_lines do
    local line = window.text_lines [i]

    line.node.Text = ""
    ActivateTextLine (line)
  end

  if #window.text_lines < #text_lines then
    print ("Can't show all text, window doesn't have sufficient text lines count")
  end

  local first_line_index = math.max (0, math.floor ((#window.text_lines - #text_lines) / 2))

  for i = 1, #text_lines do
    if i + first_line_index > #window.text_lines then
      break
    end

    window.text_lines [i + first_line_index].node.Text = text_lines [i]
  end

  SetCameraState (window.camera, true)

  if not stack_name then
    SetInputState (false)
  end

  for i = 1, #window.camera.viewports do
    window.camera.viewports [i].InputState = true
  end

  if impl.resume_game_action then
    impl.resume_game_action.Stop ()
    impl.resume_game_action = nil
  end

  if not dont_pause_game then
    PauseGame ()
  end

  if on_window_shown then
    on_window_shown ()
  end

  if not params.dont_play_sound and not window.dont_play_sound then
    ExecuteOnNextUpdate (function ()
      PlaySound ('good_ui_window')
    end)
  end

  collectgarbage ("collect")

  table.insert (stack, windows_stack_entry)
end

local_methods.OnBackPressed = function (impl)
  local stack = impl.windows_stack
  local stack_name
  
  if #stack == 0 and impl.windows_stacks then
    for name, check_stack in pairs (impl.windows_stacks) do
      if #check_stack > 0 then
        stack      = check_stack
        stack_name = name
        break
      end
    end
  end
                           
  if #stack == 0 then return false end

  local top_stack_entry   = stack [#stack]
  local on_cancel_handler = top_stack_entry.on_cancel_handler

  if top_stack_entry.on_back_handler then
    top_stack_entry.on_back_handler ()
  elseif on_cancel_handler then
    game.scripts.ui_windows.OnCancelPressed ()
  elseif top_stack_entry.hide_on_back then
    HideWindow (impl, stack_name)
  end

  return true
end

local_methods.IsWindowPausedGame = function (impl)
  if #impl.windows_stack == 0 then return false end
  
  return not impl.windows_stack [#impl.windows_stack].dont_pause_game
end

local function OnCancelPressed (impl)
  local on_cancel_handler = impl.windows_stack [#impl.windows_stack].on_cancel_handler

  HideWindow (impl)

  if on_cancel_handler then
    on_cancel_handler ()
  end
end

local function OnOkPressed (impl)
  local top_stack_entry = impl.windows_stack [#impl.windows_stack]
  local on_ok_handler   = top_stack_entry.on_ok_handler

  top_stack_entry.window.ok_pressed = true

  HideWindow (impl, nil, top_stack_entry.dont_process_queue_on_ok)

  if on_ok_handler then
    on_ok_handler ()
  end
end

local function OnHideInfoWindow (impl)
  impl.info_window.override_resource = nil

  if impl.info_window.expire_time_update_action then
    impl.info_window.expire_time_update_action.Stop ()
    impl.info_window.expire_time_update_action = nil
  end

  if game.scripts.level and game.scripts.level.OnGiftsSent then
    game.scripts.level.OnGiftsSent ()
  end

  if impl.info_window.frame_animation then
    impl.info_window.frame_animation.Hide ()
    impl.info_window.frame_animation = nil
  end

  if impl.info_window.happy_time_window_close_handler then
    if not impl.info_window.ok_pressed then
      impl.info_window.happy_time_window_close_handler ()
    end
    
    impl.info_window.happy_time_window_close_handler = nil
  end
  
  if impl.info_window.hide_handler then
    impl.info_window.hide_handler ()
    impl.info_window.hide_handler = nil
  end
end

local function OnHideAlertWindow (impl)
  if impl.alert_activity_indicator then
    impl.alert_activity_indicator.GetSprite ():Unbind ()
    impl.alert_activity_indicator.Stop ()
    impl.alert_activity_indicator = nil
  end

  if impl.alert_autoclose_action then
    impl.alert_autoclose_action.Stop ()
    impl.alert_autoclose_action = nil
  end
  
  if impl.alert_window.client_on_hide_handler then
    impl.alert_window.client_on_hide_handler ()
    impl.alert_window.client_on_hide_handler = nil
  end
end

local function MakeAllSprites (sprites)
  local return_value = {}

  for i = 1, #sprites do
    local sprite = GetSprite (sprites [i])

    local item = {}

    item.material = sprite.original_material or sprite.node.Material
    item.node     = sprite.node
    item.name     = sprites [i]

    sprite.original_material = item.material -- we needs to save original material in sprite, because node material can be setted to transparent on window hide

    table.insert (return_value, item)
  end

  return return_value
end

local function GetTextLines (lines)
  local return_value = {}

  for i = 1, #lines do
    local text_line = GetTextLine (lines [i])

    if not text_line then
      print ("text line '" .. lines [i] .. "' not found")
    end

    table.insert (return_value, text_line)
  end

  return return_value
end

local function BattleResultTryAgain (impl) 
  if #impl.windows_stack == 0 or impl.windows_stack [#impl.windows_stack].window ~= impl.battle_result_window then return end

  local share_checked = impl.battle_result_win_check
  
  HideWindow (impl, nil, share_checked)
 
  local function BattleResultTryAgainImpl ()
    local battle_init_data    = ApplicationDataGet (BATTLE_INIT_DATA)
    local battle_players_info = ApplicationDataGet (BATTLE_PLAYERS_INFO)
    local battles             = GetBattles ()

    for i = 1, #battles do
      local modes = battles [i].modes

      for j = 1, #modes do
        local mode = modes [j]

        if battle_init_data.mode_id == modes [j].id then
          if mode.coins_fee > GetUserMoney () then
            ApplicationDataSet (NEEDS_SHOW_NO_MONEY_WINDOW_ON_LOBBY_RETURN, true)
          end

          if mode.coins_fee > GetUserMoney () then
            SetCurrentScreen ("lobby")
            return
          end

          break
        end
      end
    end

    if battle_init_data and battle_players_info then
      game.scripts.ui_windows.ShowBattleSearchWindow (battle_init_data.mode_id, battle_init_data.slot_id, battle_init_data.slot_path, battle_players_info.opponent.user_id)
    else
      if not battle_init_data then
        print ("ERROR: Battle init data not found!!!!")
      end

      if not battle_players_info then
        print ("ERROR: Battle players_info not found!!!!")
      end

      if game.current_screen.name ~= "lobby" then
        SetCurrentScreen ("lobby")
      else
        HideWindow (impl)
      end
    end
  end

  if share_checked then
    GetSocialManager ().ShareBattleWin (BattleResultTryAgainImpl)
  else
    BattleResultTryAgainImpl () 
  end
end

local_methods.BattleResultShowHideRatingPopup = function (impl)
  local state = IsSpriteActive ("battle_result_win_rating_popup")
  
  state = not state
  
  SetSpritesActive (state, "battle_result_win_rating_popup")
  SetTextLinesActive (state, "battle_result_window_rating_popup_line_1", "battle_result_window_rating_popup_line_2") 
end

local function CloseBattleResultWindow (impl) 
  HideWindow (impl)
  
  if game.current_screen.name == "lobby" or not HsConnectionIsConnected () or impl.dont_return_to_lobby then return end
   
  SetCurrentScreen ("lobby")
end

local function LoadBattleResultWindow (impl)
  if impl.battle_result_window then return end

  PrepareAnimationScreenPart ("data/animations/battle_result_window.xml")

  local battle_result_window_sprite_list = { "battle_result_window_bg", "battle_result_window_close_button_l", 
    "battle_result_window_close_button_p", "battle_result_window_one_button_l", "battle_result_window_one_button_p",
    "battle_result_window_one_button_pink_l", "battle_result_window_one_button_pink_p", 
    "battle_result_defeat_overlay_1", 
    "battle_result_win_overlay_1", "battle_result_win_overlay_2", "battle_result_draw_overlay_1",
    "battle_result_win_rating_popup", "battle_result_window_rating_popup_zone", "screen_lock",
    "battle_result_win_check_zone", "battle_result_win_check", "battle_result_win_check_bg" }

  local battle_result_window_text_line_list = {"battle_result_window_one_button_text", "battle_result_win_title", "battle_result_win_coins_count", "battle_result_win_title_interrupted",
  "battle_result_win_text_1", "battle_result_win_text_2", "battle_result_win_text_3", "battle_result_win_text_4",
  "battle_result_win_rating", "battle_result_defeat_one_button", "battle_result_defeat_title", "battle_result_defeat_text_1", "battle_result_defeat_text_2", "battle_result_defeat_text_3",
  "battle_result_draw_one_button", "battle_result_draw_title", "battle_result_draw_coins_count", "battle_result_draw_text_1",
  "battle_result_draw_text_2", "battle_result_draw_text_3", "battle_result_draw_text_4", "battle_result_draw_rating", "battle_result_window_rating_popup_line_1",
  "battle_result_window_rating_popup_line_2"}
    
  impl.battle_result_window = {}

  impl.battle_result_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.battle_result_window.resource       = "data/res/materials/battle_result_window_layers.xmtl"
  impl.battle_result_window.sprites        = { "battle_result_window_bg", "battle_result_window_close_button_l", "screen_lock" }
  impl.battle_result_window.text_lines     = {}
  impl.battle_result_window.all_text_lines = GetTextLines (battle_result_window_text_line_list)
  impl.battle_result_window.all_sprites    = MakeAllSprites (battle_result_window_sprite_list)
end

local_methods.LockedFunctionalityWindow = function (impl)
  if impl.locked_functionality_window then return end

  PrepareAnimationScreenPart ("data/animations/locked_functionality_window.xml")

  local locked_functionality_window_sprite_list = { "locked_functionality_window_background", "locked_functionality_window_screen_lock",
    "locked_functionality_window_close_l", "locked_functionality_window_close_p", "screen_lock",
    "locked_functionality_window_fb_button_l", "locked_functionality_window_fb_button_p",
    "locked_functionality_window_vk_button_l", "locked_functionality_window_vk_button_p",
    "locked_functionality_window_center_button_l", "locked_functionality_window_center_button_p",
    "locked_functionality_window_combined", "locked_functionality_window_hall_of_fame", "locked_functionality_window_gifts",
    "locked_functionality_window_coins", "locked_functionality_window_social",
    "locked_functionality_window_battle", "locked_functionality_window_friends", "locked_functionality_window_boosters" }

  impl.locked_functionality_window = {}

  impl.locked_functionality_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.locked_functionality_window.resource       = "data/res/materials/locked_functionality_window_layers.xmtl"
  impl.locked_functionality_window.sprites        = { "locked_functionality_window_background", "locked_functionality_window_screen_lock", "screen_lock" }
  impl.locked_functionality_window.text_lines     = {}
  impl.locked_functionality_window.all_text_lines = GetTextLines ({ "locked_functionality_window_text_vk_button", "locked_functionality_window_text_fb_button", "locked_functionality_window_text_center_button" })
  impl.locked_functionality_window.all_sprites    = MakeAllSprites (locked_functionality_window_sprite_list)
  impl.locked_functionality_window.on_hide        = function ()
      if impl.locked_functionality_window.on_hide_handler then
        impl.locked_functionality_window.on_hide_handler ()
        impl.locked_functionality_window.on_hide_handler = nil
      end
    end
end

local function LoadInfoWindow (impl)
  if impl.info_window then return end

  PrepareAnimationScreenPart ("data/animations/info_window.xml")

  local info_window_sprite_list = { "info_window_background", "info_window_one_button", "info_window_one_button_pressed",
    "info_window_two_button_left", "info_window_two_button_left_pressed",
    "info_window_two_button_right", "info_window_two_button_right_pressed",
    "info_window_close_button", "info_window_close_button_pressed", "info_window_share_free_spins_overlay1",
    "info_window_share_free_spins_overlay2", "info_window_share_free_spins_overlay3",
    "info_window_no_money_overlay1", "info_window_no_money_overlay2",
    "info_window_new_game_overlay1", "info_window_new_game_overlay2",
    "info_window_bonus_by_level_girl", "info_window_bonus_by_level_present",
			"info_window_share_bonus_overlay1", "info_window_share_bonus_overlay2", "info_window_share_bonus_icon",
    "info_window_share_bonus_check_bg", "info_window_share_bonus_check", "info_window_share_bonus_check_zone",
    "info_window_share_five_in_row_overlay1", "info_window_share_five_in_row_overlay2", "info_window_share_five_in_row_overlay3", "info_window_share_five_in_row_icon",
    "info_window_share_five_in_row_check_bg", "info_window_share_five_in_row_check", "info_window_share_five_in_row_check_zone",
    "info_window_new_game_icon_zone", "screen_lock",
    "info_window_no_diamond_overlay1", "info_window_no_diamond_overlay2",
    "info_window_battles_bonus_diamonds", "info_window_screen_lock",
    "info_window_happy_hour_blocker", "info_window_happy_hour_top_overlay_time_1", "info_window_happy_hour_top_overlay_time_2",
    "info_window_happy_hour_overlay_1", "info_window_happy_hour_overlay_2", "info_window_happy_hour_overlay_3", "info_window_happy_hour_overlay_4",
    "info_window_happy_hour_icon_zone", "info_window_happy_hour_top_overlay",
    "info_window_get_coins_overlay_1", "info_window_get_coins_overlay_2", "info_window_action_is_over_overlay"}

  impl.info_window = {}

  impl.info_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.info_window.resource       = "data/res/materials/info_window_layers.xmtl"
  impl.info_window.sprites        = { "info_window_background", "screen_lock" }
  impl.info_window.text_lines     = {}
  impl.info_window.all_text_lines = GetTextLines ({ "info_window_one_button", "info_window_two_button_left", "info_window_two_button_right" })
  impl.info_window.all_sprites    = MakeAllSprites (info_window_sprite_list)
  impl.info_window.on_hide        = function () OnHideInfoWindow (impl) end
end

local function LoadGiftWindow (impl)
  if impl.gift_window then return end

  PrepareAnimationScreenPart ("data/animations/gift_select_window.xml")

  local gift_window_sprite_list = { "gift_background", "window_gift_select_btn_1_zone", "window_gift_select_btn_2_zone", 
    "window_gift_select_btn_1_l", "gift_window_close_button", "gift_window_close_button_pressed",
    "window_gift_select_btn_2_l", "window_gift_select_btn_1_p", "window_gift_select_btn_2_p",
    "screen_lock" }

  impl.gift_window = {}

  impl.gift_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.gift_window.resource       = "data/res/materials/gift_select_window_layers.xmtl"
  impl.gift_window.sprites        = {"gift_background", "gift_window_close_button", "screen_lock"}
  impl.gift_window.text_lines     = {}
  impl.gift_window.all_text_lines = GetTextLines ({"gift_window_send_gifts_title_line_1"})
  impl.gift_window.all_sprites    = MakeAllSprites (gift_window_sprite_list)
end

local function LoadNoCuurentBetWindow (impl)
  if impl.no_current_bet_window then return end

  PrepareAnimationScreenPart ("data/animations/info_no_current_bet.xml")

  local info_no_current_bet_window_sprite_list = { "info_no_current_bet_background", "info_no_current_bet_close_button", "info_no_current_bet_close_button_pressed",
    "info_no_current_bet_two_button_right_pressed", "info_no_current_bet_two_button_right",
    "info_no_current_bet_two_button_left_pressed", "info_no_current_bet_two_button_left", "screen_lock"}

  impl.no_current_bet_window = {}

  impl.no_current_bet_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.no_current_bet_window.resource       = "data/res/materials/info_no_current_bet_layers.xmtl"
  impl.no_current_bet_window.sprites        = {"info_no_current_bet_background", "info_no_current_bet_close_button"}
  impl.no_current_bet_window.text_lines     = {}
  impl.no_current_bet_window.all_text_lines = GetTextLines ({"info_no_current_bet_no_money_title_line_1"})
  impl.no_current_bet_window.all_sprites    = MakeAllSprites (info_no_current_bet_window_sprite_list)
end

local function OnSendGiftsFriendsWindowHide (impl)
  for i = 1, #impl.send_gifts_selected_friends_cells do
    local cell = impl.send_gifts_selected_friends_cells [i]

    for j = 1, #cell.handlers do
      local handler = cell.handlers [j]

      for k = 1, #handler do
        handler [k].connection:Disconnect ()
        handler [k].callback = nil
      end
    end
  end

  impl.send_gifts_friends_scroll_view = nil
  impl.send_gifts_friends_cells_node:Unbind ()
  impl.send_gifts_friends_cells_node  = nil
  impl.send_gifts_friends_cells = nil
  impl.send_gifts_selected_friends_scroll_view = nil
  impl.send_gifts_selected_friends_cells_node:Unbind ()
  impl.send_gifts_selected_friends_cells_node  = nil
  impl.send_gifts_window_selected_friends_scissor:Unbind ()
  impl.send_gifts_window_selected_friends_scissor = nil
  impl.send_gifts_friends_cell_data.cell_scissor:Unbind ()
  impl.send_gifts_friends_cell_data.cell_scissor = nil
  impl.send_gifts_selected_friends_cells = nil
  impl.friends_list = nil
  impl.selected_friends_list = nil
  impl.send_gifts_friends_cell_data = nil

  collectgarbage ("collect")
  Engine.ResourceManager.FlushUnusedResources ()

  if game.scripts.level and game.scripts.level.OnGiftsSent then
    game.scripts.level.OnGiftsSent ()
  end
end

local function LoadSendGiftWindow (impl)
  if impl.send_gift_friends_window then return end

  PrepareAnimationScreenPart ("data/animations/send_gift_window.xml")

  impl.send_gift_friends_window = {}

  impl.send_gift_friends_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.send_gift_friends_window.resource       = "data/res/materials/send_gift_window_layers.xmtl"
  impl.send_gift_friends_window.sprites        = { "send_gift_window_background", "send_gift_window_close_l", "window_gift_screen_lock", "send_gift_window_send_d", "send_gift_window_check_all_friends_zone",
    "send_gift_window_friends_list_roll_slide_l", "send_gift_window_selected_friends_list_roll_slide_l",
    "send_gift_window_friends_list_zone", "send_gift_window_selected_friends_list_zone" }
  impl.send_gift_friends_window.text_lines     = GetTextLines ({ "send_gift_window_friends", "send_gift_window_selected_friends" })
  impl.send_gift_friends_window.all_text_lines = GetTextLines ({ "send_gift_window_friends", "send_gift_window_selected_friends", "send_gift_window_send",
    "send_gift_window_send_d" })
  impl.send_gift_friends_window.all_sprites    = MakeAllSprites ({ "send_gift_window_background", "send_gift_window_close_l", "send_gift_window_close_p", "send_gift_window_check_all_friends_zone",
    "send_gift_window_send_l", "send_gift_window_send_p", "send_gift_window_send_d", "send_gift_window_check_all_friends",
    "send_gift_window_friends_list_roll_slide_l", "send_gift_window_friends_list_roll_slide_p",
    "send_gift_window_selected_friends_list_roll_slide_l", "send_gift_window_selected_friends_list_roll_slide_p",
    "send_gift_window_friends_list_zone", "send_gift_window_selected_friends_list_zone", "window_gift_screen_lock" })
  impl.send_gift_friends_window.on_hide        = function () OnSendGiftsFriendsWindowHide (impl) end
end

local function LoadNewLevelWindow (impl)
  if impl.level_up_window then return end

  PrepareAnimationScreenPart ("data/animations/new_level_window.xml")

  impl.level_up_window = {}

  impl.level_up_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.level_up_window.resource       = "data/res/materials/new_level_window_layers.xmtl"
  impl.level_up_window.sprites        = { "new_level_window_background", "new_level_window_one_button",
    "screen_lock" , "new_level_window_check", "new_level_window_check_bg", "new_level_window_check_zone",
    "new_level_window_girl" }
  impl.level_up_window.text_lines     = GetTextLines ({ "new_level_window_title_line_1", "new_level_window_title_line_2", 
    "new_level_window_title_line_2_1", "new_level_window_title_line_2_2", "new_level_window_max_bet", 
    "new_level_window_super_bonus", "new_level_window_mega_bonus", "new_level_window_share_button",
    "new_level_window_level_value", "new_level_window_max_bet_value", "new_level_window_super_bonus_value", 
    "new_level_window_mega_bonus_value", "new_level_window_share_friends" })
  impl.level_up_window.all_text_lines = GetTextLines ({ "new_level_window_title_line_1", "new_level_window_title_line_2", 
    "new_level_window_title_line_2_1", "new_level_window_title_line_2_2", "new_level_window_max_bet", "new_level_window_super_bonus",
    "new_level_window_mega_bonus", "new_level_window_share_button", "new_level_window_bonus_value_diamonds_long", "new_level_window_bonus_value_long",
    "new_level_window_level_value", "new_level_window_max_bet_value", "new_level_window_super_bonus_value", "new_level_window_mega_bonus_value",
    "new_level_window_bonus_value_diamonds", "new_level_window_bonus_value", "new_level_window_share_friends" })
  impl.level_up_window.all_sprites    = MakeAllSprites ({ "new_level_window_background", "new_level_window_one_button",
    "new_level_window_one_button_pressed", "new_level_window_check", "new_level_window_check_bg", "screen_lock",
    "new_level_window_girl", "new_level_window_long_plates", "new_level_window_short_plates" })
end

local function LoadNinetyNineLinesWindow (impl)
  if impl.ninety_nine_lines then return end

  PrepareAnimationScreenPart ("data/animations/ninety_nine_lines_window.xml")

  impl.ninety_nine_lines = {}

  impl.ninety_nine_lines.camera         = game.current_screen.cameras.modal_window_camera
  impl.ninety_nine_lines.resource       = "data/res/materials/ninety_nine_lines_window_layers.xmtl"
  impl.ninety_nine_lines.sprites        = { "ninety_nine_lines_window_bg", "ninety_nine_lines_window_button_l", "ninety_nine_lines_window_screen_lock"}
  impl.ninety_nine_lines.text_lines     = GetTextLines ({ "ninety_nine_lines_window_button_text" })
  impl.ninety_nine_lines.all_text_lines = GetTextLines ({ "ninety_nine_lines_window_button_text" })
  impl.ninety_nine_lines.all_sprites    = MakeAllSprites ({ "ninety_nine_lines_window_bg", "ninety_nine_lines_window_button_l",
    "ninety_nine_lines_window_button_p" , "ninety_nine_lines_window_screen_lock" })
end

local function OnTakeFreeSpinsWindowHide (impl)
  local next_page_zone_sprite                = GetSprite ("next_page_button")
  local next_page_button_sprite              = GetSprite ("next_page_l")
  local next_page_button_disabled_sprite     = GetSprite ("next_page_d")
  local next_page_button_pressed_sprite      = GetSprite ("next_page_p")
  local previous_page_zone_sprite            = GetSprite ("previous_page_button")
  local previous_page_button_sprite          = GetSprite ("previous_page_l")
  local previous_page_button_disabled_sprite = GetSprite ("previous_page_d")
  local previous_page_button_pressed_sprite  = GetSprite ("previous_page_p")
  local slots_drag_zone_sprite               = GetSprite ("slots_drag_zone")

  next_page_zone_sprite.z                = impl.lobby_next_page_zone_z
  next_page_button_disabled_sprite.z     = impl.lobby_next_page_button_disabled_z
  next_page_button_sprite.z              = impl.lobby_next_page_button_z
  next_page_button_pressed_sprite.z      = impl.lobby_next_page_button_pressed_z
  previous_page_zone_sprite.z            = impl.lobby_previous_page_zone_z
  previous_page_button_disabled_sprite.z = impl.lobby_previous_page_button_disabled_z
  previous_page_button_sprite.z          = impl.lobby_previous_page_button_z
  previous_page_button_pressed_sprite.z  = impl.lobby_previous_page_button_pressed_z
  slots_drag_zone_sprite.z               = impl.lobby_slots_drag_zone_z

  if IsSpriteActive ("next_page_l") then
    DeactivateSprite ("next_page_l")
    ActivateSprite ("next_page_l")
  end
  if IsSpriteActive ("next_page_d") then
    DeactivateSprite ("next_page_d")
    ActivateSprite ("next_page_d")
  end
  if IsSpriteActive ("next_page_button") then
    DeactivateSprite ("next_page_button")
    ActivateSprite ("next_page_button")
  end
  if IsSpriteActive ("previous_page_l") then
    DeactivateSprite ("previous_page_l")
    ActivateSprite ("previous_page_l")
  end
  if IsSpriteActive ("previous_page_d") then
    DeactivateSprite ("previous_page_d")
    ActivateSprite ("previous_page_d")
  end
  if IsSpriteActive ("previous_page_button") then
    DeactivateSprite ("previous_page_button")
    ActivateSprite ("previous_page_button")
  end

  DeactivateSprite ("slots_drag_zone")
  ActivateSprite ("slots_drag_zone")

  game.scripts.lobby.SetSlotsZ (impl.lobby_slots_z)
  game.scripts.lobby.SetHappyTimeHelperVisible (true)

  game.current_screen.vars.previous_page_button_offset = vec2 (0, 0)
  game.current_screen.vars.next_page_button_offset     = vec2 (0, 0)
  
  if impl.take_free_spins_window.close_handler then
    impl.take_free_spins_window.close_handler ()
    impl.take_free_spins_window.close_handler = nil
  end
end

local function LoadTakeFreeSpinsWindow (impl)
  if impl.take_free_spins_window then return end

  PrepareAnimationScreenPart ("data/animations/take_free_spins.xml")

  impl.take_free_spins_window = {}

  impl.take_free_spins_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.take_free_spins_window.resource       = "data/res/materials/take_free_spins_layers.xmtl"
  impl.take_free_spins_window.sprites        = { "take_free_spins_background", "take_free_spins_close", "take_free_spins_background_top",
    "take_free_spins_background_bottom", "take_free_spins_background_left", "take_free_spins_background_right" }
  impl.take_free_spins_window.text_lines     = GetTextLines ({ "take_free_spins_text" })
  impl.take_free_spins_window.all_text_lines = GetTextLines ({ "take_free_spins_text" })
  impl.take_free_spins_window.all_sprites    = MakeAllSprites ({ "take_free_spins_background", "take_free_spins_close" })
  impl.take_free_spins_window.on_hide        = function () OnTakeFreeSpinsWindowHide (impl) end
end

local function LoadAlertWindow (impl)
  if impl.alert_window then return end

  PrepareAnimationScreenPart ("data/animations/alert_window.xml")

  local all_sprite_list = { "alert_window_background", "alert_window_button_l", "alert_window_button_p",
    "alert_window_close_button", "alert_window_close_button_pressed", "alert_window_close_underlay",
    "alert_window_error_1", "alert_window_gift_1", "alert_window_question_1", "alert_window_language",
    "alert_window_screen_lock" }

  impl.alert_window = {}

  impl.alert_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.alert_window.resource       = "data/res/materials/alert_window_layers.xmtl"
  impl.alert_window.sprites        = { "alert_window_background", "alert_window_screen_lock" }
  impl.alert_window.text_lines     = GetTextLines ({ "alert_window_line_1", "alert_window_line_2", "alert_window_line_3", "alert_window_line_4", "alert_window_line_5" })
  impl.alert_window.all_text_lines = GetTextLines ({ "alert_window_line_1", "alert_window_line_2", "alert_window_line_3", "alert_window_line_4", "alert_window_line_5",
    "alert_window_ok_button" })
  impl.alert_window.all_sprites    = MakeAllSprites (all_sprite_list)
  impl.alert_window.on_hide        = function () OnHideAlertWindow (impl) end

end

local function ShowAlertWindow (impl, type, on_ok_handler, on_cancel_handler, ok_text, text_lines, show_progress_indicator, dont_disable_camera, autoclose, dont_pause_game, enqueue, on_hide_handler)
  local additional_sprites    = {}
  local additional_text_lines = {}
  local additional_resource_bindings = {}

  LoadAlertWindow (impl)

  impl.alert_window_type = type

  if type then
    table.insert (additional_resource_bindings, LoadResource (type.resource))
    additional_sprites = CopyArrayTable (type.sprite)
  end

  if on_ok_handler then
    table.insert (additional_sprites, "alert_window_button_l")
    table.insert (additional_text_lines, "alert_window_ok_button")
  end

  if on_cancel_handler then
    table.insert (additional_sprites, "alert_window_close_button")
    table.insert (additional_sprites, "alert_window_close_underlay")
  end

  ShowWindow (impl, impl.alert_window, additional_sprites, additional_text_lines, text_lines, function ()
    impl.alert_window.client_on_hide_handler = on_hide_handler
  
    if ok_text then
      GetTextLine ("alert_window_ok_button").node.Text = ok_text
    end
  end, dont_disable_camera, dont_pause_game, true, on_ok_handler, on_cancel_handler, additional_resource_bindings, enqueue)

  if show_progress_indicator then
    impl.alert_activity_indicator = CreateActivityIndicator (game.current_screen.vars.activity_indicator_size, "ui")

    local activity_indicator_node = impl.alert_activity_indicator.GetSprite ()
    
    activity_indicator_node.Position = vec3 (0, -0.2, -0.01)    
    
    activity_indicator_node:BindToParent (GetSprite ("alert_window_background").node)
  end

  if autoclose then
    impl.alert_autoclose_action = DelayedAction ("ui", ALERT_AUTOCLOSE_DELAY, function ()
      game.scripts.ui_windows.OnCancelPressed ()
    end) 
  end
end

local function OnHideBattleInfoWindow (impl)
  if impl.battle_info_window.show_action then
    impl.battle_info_window.show_action.Stop ()
    impl.battle_info_window.show_action = nil
  end 
  
  if impl.battle_info_window.auto_hide_action then
    impl.battle_info_window.auto_hide_action.Stop ()
    impl.battle_info_window.auto_hide_action = nil
  end
end

local function HideBattleInfoWindow (impl)
  if impl.battle_info_window.hiding then return end

  impl.battle_info_window.hiding = true

  OnHideBattleInfoWindow (impl)
  
  impl.battle_info_window.show_action = SlowAction ("ui", game.current_screen.vars.battle_info_window_animation_duration, function (t)
    game.current_screen.vars.battle_info_window_offset = vec2 (0, t * game.current_screen.vars.battle_info_window_max_offset)
    
    if t == 1 then
      game.current_screen.vars.battle_info_window_offset = vec2 (0, 0)
      HideWindow (impl, "battle_info_window")
    end
  end)
end

local function ShowBattleInfoWindow (impl, icon, title, texts, show_duration)
  if not impl.battle_info_window then
    PrepareAnimationScreenPart ("data/animations/battle_info_window.xml")

    local battle_info_window_sprite_list = { "battle_info_window_close_l", "battle_info_window_close_p", 
      "battle_info_window_background", "battle_info_window_achievement", "battle_info_window_new_battle",
      "battle_info_window_achievement_group", "battle_info_window_level_up",
      "battle_info_window_icon", "battle_info_window_top_1", "battle_info_window_top_2", "battle_info_window_top_3",
      "battle_info_window_top_10", "battle_info_window_top_100", "battle_info_window_prize_1", 
      "battle_info_window_prize_2", "battle_info_window_prize_3" }

    impl.battle_info_window = {}

    impl.battle_info_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.battle_info_window.resource       = "data/res/materials/battle_info_window_layers.xmtl"
    impl.battle_info_window.sprites        = { "battle_info_window_background", "battle_info_window_close_l" }
    impl.battle_info_window.text_lines     = GetTextLines ({'battle_info_window_title', 'battle_info_window_line_1', 'battle_info_window_line_2', 'battle_info_window_line_3'})
    impl.battle_info_window.all_text_lines = GetTextLines ({'battle_info_window_title', 'battle_info_window_line_1', 'battle_info_window_line_2', 'battle_info_window_line_3'})
    impl.battle_info_window.all_sprites    = MakeAllSprites (battle_info_window_sprite_list)
    impl.battle_info_window.on_hide        = function () OnHideBattleInfoWindow (impl) end
  end
  
  local additional_sprites = {}
  local additional_resource_bindings
  
  if type (icon) == "string" then
    additional_sprites = { icon }
  else
    additional_sprites = { "battle_info_window_icon" }
    
    if icon.additional_resources then
      additional_resource_bindings = {}
    
      for i = 1, #icon.additional_resources do
        table.insert (additional_resource_bindings, LoadResource (icon.additional_resources [i]))
      end
    end
  end

  ShowWindow (impl, impl.battle_info_window, additional_sprites, {}, {}, function ()
    impl.battle_info_window.hiding = false

    OnHideBattleInfoWindow (impl)

    GetTextLine ('battle_info_window_title').node.Text = title

    for i = 1, 3 do
      GetTextLine ('battle_info_window_line_' .. i).node.Text = ""
    end

    for i = 1, #texts do
      GetTextLine ('battle_info_window_line_' .. i).node.Text = texts [i]
    end

    if type (icon) ~= "string" then
      local icon_sprite = GetSprite ("battle_info_window_icon")
      local icon_zone   = GetSprite ("battle_info_window_icon_zone")
      local width       = icon_zone.layout:Get ("Width") + 0
      local height      = icon_zone.layout:Get ("Height") + 0

      icon_sprite.node.Material = icon.material
      
      local icon_ref = GetSprite (icon.material, true)
      
      if icon_ref then
        icon_sprite.node.TexSize   = StringToVec (icon_ref.layout:Get ("TexSize"))
        icon_sprite.node.TexOffset = StringToVec (icon_ref.layout:Get ("TexOffset"))
      else
        icon_sprite.node.TexSize   = vec2 (1, 1)
        icon_sprite.node.TexOffset = vec2 (0, 0)
      end 

      local icon_scale = GetAspectFitSize (icon.scale.x, icon.scale.y, width, height)
      local force_scale_multiplier = icon.force_scale_multiplier or vec3 (1, 1, 1)

      icon_sprite.node.WorldScale = vec3 (icon_scale.x, icon_scale.y, 1) * force_scale_multiplier  

      icon_sprite.tracks ["size"] = nil
      icon_sprite.tracks ["texsize"] = nil
      icon_sprite.tracks ["texoffset"] = nil
    end

     impl.battle_info_window.show_action = SlowAction ("ui", game.current_screen.vars.battle_info_window_animation_duration, function (t)
      game.current_screen.vars.battle_info_window_offset = vec2 (0, (1 - t) * game.current_screen.vars.battle_info_window_max_offset)
    end)

    impl.battle_info_window.auto_hide_action = DelayedAction ("ui", show_duration or game.current_screen.vars.battle_info_window_show_duration, function ()
      HideBattleInfoWindow (impl)
    end)
  end, true, true, true, nil, nil, additional_resource_bindings, true, "battle_info_window", function ()
    game.current_screen.vars.battle_info_window_offset = vec2 (0, game.current_screen.vars.battle_info_window_max_offset)
  end)
end

local function ShowBuyDiamondsWindow (impl)
  if not GetStoreManager ().CanBuyProducts () then
    local texts = {}

    local os = ApplicationDataGet (OS_TYPE)

    if string.find (os, "iOS") ~= nil then
      texts = { GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_1"), GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_2"), GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_3") }
    else
      texts = { GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_1") }
    end

    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, texts)
    return
  end

  local buy_chips_resource_data = ApplicationDataGet (BUY_CHIPS_RESOURCE_DATA)

  if not buy_chips_resource_data then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
    return
  end

  local show_downloading_window_action

  local function OnMounted ()
    if show_downloading_window_action then
      show_downloading_window_action.Stop ()
      show_downloading_window_action = nil
    end

    LoadLocalizationResource ("buy_chips")

    PrepareAnimationScreenPart ("data/animations/buy_chips.xml")

    game.scripts.buy_chips_window.ShowBuyDiamonds ()
  end

  show_downloading_window_action = DelayedAction ("default", 0.01, function ()
    game.scripts.ui_windows.ShowLoadingWindow (0, GetLocalizedString ("DOWNLOADING"))
  end)

  MountBuyChipsWindow (OnMounted, function (progress)
    game.scripts.ui_windows.ShowLoadingWindow (progress * 100, GetLocalizedString ("DOWNLOADING"))
  end, function ()
    game.scripts.ui_windows.HideWindow ()
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
  end)
end

local function CreateFramesNames (format, from, to)
  local return_value = {}

  for i = from, to do
    table.insert (return_value, string.format (format, i))
  end

  return return_value
end

local function SetAnimationPosition (animation, reference, z)
  local sprite = GetSprite (reference)
  local layout = sprite.layout
  animation.SetPosition (layout:Get ("Left") + layout:Get ("Width") / 2, -layout:Get ("Top") - layout:Get ("Height") / 2, z or sprite.z)
end

local function ShowBuyChipsWindow (impl)
  if not GetStoreManager ().CanBuyProducts () then
    local texts = {}

    local os = ApplicationDataGet (OS_TYPE)

    if string.find (os, "iOS") ~= nil then
      texts = { GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_1"), GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_2"), GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_3") }
    else
      texts = { GetLocalizedString ("CANT_MAKE_PURCHASES_IOS_LINE_1") }
    end

    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, texts)
    return
  end

  local buy_chips_resource_data = ApplicationDataGet (BUY_CHIPS_RESOURCE_DATA)

  if not buy_chips_resource_data then
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
    return
  end

  local show_downloading_window_action

  local function OnMounted ()
    if show_downloading_window_action then
      show_downloading_window_action.Stop ()
      show_downloading_window_action = nil
    end
   
    LoadLocalizationResource ("buy_chips")

    PrepareAnimationScreenPart ("data/animations/buy_chips.xml")
  
    game.scripts.buy_chips_window.ShowBuyChips ()
  end

  show_downloading_window_action = DelayedAction ("default", 0.01, function ()
    game.scripts.ui_windows.ShowLoadingWindow (0, GetLocalizedString ("DOWNLOADING"))
  end)
  
  MountBuyChipsWindow (OnMounted, function (progress)
    game.scripts.ui_windows.ShowLoadingWindow (progress * 100, GetLocalizedString ("DOWNLOADING"))
  end, function ()
    game.scripts.ui_windows.HideWindow ()
    game.scripts.ui_windows.ShowAlertWindow (nil, nil, function () end, nil, { GetLocalizedString ("DOWNLOADING_ERROR_LINE1"), GetLocalizedString ("DOWNLOADING_ERROR_LINE2"), GetLocalizedString ("DOWNLOADING_ERROR_LINE3"), GetLocalizedString ("DOWNLOADING_ERROR_LINE4"), GetLocalizedString ("DOWNLOADING_ERROR_LINE5") })
  end)
end

local function ShowNotEnoughMoneyWindow (impl, play_handler)
  LoadNoCuurentBetWindow (impl)

  local additional_sprites    = { "info_no_current_bet_two_button_right", "info_no_current_bet_two_button_left", "info_no_current_bet_close_button", "screen_lock" }
  local additional_text_lines = { "info_no_current_bet_no_money_title_line_1", "info_no_current_bet_no_money_title_line_2",
                                  "info_no_current_bet_no_money_text_line_1", "info_no_current_bet_no_money_text_line_2",
                                  "info_no_current_bet_two_button_left", "info_no_current_bet_two_button_right"}

  GetTextLine ("info_no_current_bet_two_button_left").node.Text  = GetLocalizedString ("INFO_WINDOW.BUY_CHIPS")
  GetTextLine ("info_no_current_bet_two_button_right").node.Text = GetLocalizedString ("INFO_WINDOW.PLAY")

  ShowWindow (impl, impl.no_current_bet_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, function ()
    ShowBuyChipsWindow (impl) 
  end, play_handler, nil, nil, nil, nil, { on_back_handler = function ()
    game.scripts.ui_windows.HideWindow ()
  end })
end

local_methods.ShowHappyTimeActionEndedWindow  = function (impl, day_or_hour, close_handler)
  if #impl.windows_stack > 0 and impl.windows_stack [#impl.windows_stack].window == impl.action_ended_window then return end --ignore duplicate show of this window

  if not impl.action_ended_window then
    PrepareAnimationScreenPart ("data/animations/action_ended_window.xml")

    local action_ended_window_sprite_list = { "action_ended_window_background", "action_ended_window_overlay", "action_ended_window_close_btn_l", "action_ended_window_screen_lock", "action_ended_window_overlay_" .. day_or_hour }

    impl.action_ended_window = {}

    impl.action_ended_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.action_ended_window.resource       = "data/res/materials/action_ended_window_layers.xmtl"
    impl.action_ended_window.sprites        = action_ended_window_sprite_list
    impl.action_ended_window.text_lines     = {}
    impl.action_ended_window.all_text_lines = GetTextLines ({"action_ended_window_text_1", "action_ended_window_text_2", "action_ended_window_button_text"})
    impl.action_ended_window.all_sprites    = MakeAllSprites (action_ended_window_sprite_list)
    
    impl.action_ended_window.on_hide = function ()
      impl.action_ended_window.frame_animation.Hide ()
      if impl.action_ended_window.close_handler then
        impl.action_ended_window.close_handler ()
        impl.action_ended_window.close_handler = nil
      end
    end
  end

  impl.action_ended_window.close_handler = close_handler

  local z_position_frame_animation = GetSprite  ("action_ended_border_anim_0").z - EPS
  local frame_animations_sprite = CreateFramesNames ("action_ended_border_anim_%d", 0, 45);
  impl.action_ended_window.frame_animation = CreateAnimationSpriteClass ( { animation = frame_animations_sprite }, frame_animations_sprite [1], 30 ).CreateInstance ()
  SetAnimationPosition (impl.action_ended_window.frame_animation, frame_animations_sprite [1], z_position_frame_animation)

  GetTextLine ("action_ended_window_text_2").node.Text = ProcessStringWithReplacements (GetLocalizedString("ACTION_ENDED_WINDOW.TEXT_2"), {{ token = "{promoName}",  replacement = GetLocalizedString("INFO_WINDOW.HAPPY_HOUR_TITLE_DAY_OR_HOUR_" .. day_or_hour) }})

  ShowWindow (impl, impl.action_ended_window, {}, {"action_ended_window_text_1", "action_ended_window_text_2", "action_ended_window_button_text"}, {},
              function () impl.action_ended_window.frame_animation.Show ('animation', nil, nil, 2) end, nil, nil, nil, nil, nil, nil, nil, nil, nil,
              { on_back_handler = function () game.scripts.ui_windows.OnOkPressed () end } ) --game should be paused to prevent freespins autospinning
end

local_methods.ShowHappyTimerUpdate = function (impl, time_left)
  local seconds = time_left % 60
  local minutes = time_left / 60
  local hours   = time_left / 60 / 60
  local days    = math.floor (time_left / 60 / 60 / 24)

  minutes = minutes % 60
  hours   = hours % 24

  local text

  if days > 0 then
    text = string.format ("%d%s %02d:%02d:%02d", days, GetLocalizedString ("TIME_FORMAT_DAY_SUFFIX"), hours, minutes, seconds)
  else
    text = string.format ("%02d:%02d:%02d", hours, minutes, seconds)
  end

  GetTextLine ("info_window_happy_hour_time_before_closure_value").node.Text = text

  if time_left == 0 then
    HideWindow (impl)
  end
end

local_methods.GetSaleCoins = function (impl)
 local payments = GetPayments ()
 local happy_time_actions = {}
 local promo_actions = GetPromoActions ()
 local sale_hour_line

 for i = 1, #promo_actions do
   local action = promo_actions [i]
   if action.id == 28 then
     table.insert (happy_time_actions, action)
   end
 end

 for i = 1, #happy_time_actions do
   local action = happy_time_actions [i]

   local additional = action.additional:Child ( 0 )
   local type       = GetNodeChildAttribute (additional, "type") + 0

   if type == 4 then
     sale_hour_line = 7 - (GetNodeChildAttribute (additional, "value2") + 0)
   end
 end

 local payment = GetPayments () [sale_hour_line]
 
 return FormatMoneyString (payment.total_bonus_money)
end

local_methods.ShowInfoSocialConnectingWindow = function (impl)
  if not impl.info_social_connecting_window then
    PrepareAnimationScreenPart ("data/animations/info_social_connecting_window.xml")

    local sprite_list = { "info_social_connecting_window_background", "info_social_connecting_window_screen_lock",
      "info_social_connecting_window_close_l", "info_social_connecting_window_close_p", "screen_lock",
      "info_social_connecting_window_fb_button_l", "info_social_connecting_window_fb_button_p",
      "info_social_connecting_window_vk_button_l", "info_social_connecting_window_vk_button_p"}

    local locked_functionality_friends_text_line_list = {"info_social_connecting_window_text_vk_button", "info_social_connecting_window_text_fb_button",
      "info_social_connecting_window_hello_user_text_1", "info_social_connecting_window_hello_user_text_2",
      "info_social_connecting_window_hello_user_text_3", "info_social_connecting_window_hello_user_text_4", "info_social_connecting_window_hello_user_text_5",
      "info_social_connecting_window_hello_user_text_6"}

		if impl.phone_ui then
			table.insert (locked_functionality_friends_text_line_list, "info_social_connecting_window_hello_user_title_1")
			table.insert (locked_functionality_friends_text_line_list, "info_social_connecting_window_hello_user_title_2")
		else
			table.insert (locked_functionality_friends_text_line_list, "info_social_connecting_window_hello_user_title")
		end

    impl.info_social_connecting_window = {}

    impl.info_social_connecting_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.info_social_connecting_window.resource       = "data/res/materials/info_social_connecting_window_layers.xmtl"
    impl.info_social_connecting_window.sprites        = { "info_social_connecting_window_background", "info_social_connecting_window_screen_lock", "screen_lock",
      "info_social_connecting_window_close_l", "info_social_connecting_window_fb_button_l", "info_social_connecting_window_vk_button_l" }
    impl.info_social_connecting_window.text_lines     = GetTextLines (locked_functionality_friends_text_line_list)
    impl.info_social_connecting_window.all_text_lines = GetTextLines (locked_functionality_friends_text_line_list)
    impl.info_social_connecting_window.all_sprites    = MakeAllSprites (sprite_list)
    impl.info_social_connecting_window.on_hide        = function () end
  end
  
  ShowWindow (impl, impl.info_social_connecting_window, {}, {}, {}, nil, nil, nil, nil, nil, nil, nil, true, nil, nil, { hide_on_back = true })
end

local_methods.ShowLockedFunctionalityFriendsWindow = function (impl)
  if not impl.phone_ui then return end

  local sprite_list = { "locked_functionality_friends_window_bg", "locked_functionality_friends_window_screen_lock",
    "locked_functionality_friends_window_close_l", "locked_functionality_friends_window_close_p", "screen_lock",
    "locked_functionality_friends_window_vk_button_l", "locked_functionality_friends_window_vk_button_p",
    "locked_functionality_friends_window_fb_button_l", "locked_functionality_friends_window_fb_button_p"}
  local locked_functionality_friends_text_line_list = {"locked_functionality_friends_window_text_vk_button", "locked_functionality_friends_window_text_fb_button",
    "locked_functionality_friends_window_text_title", "locked_functionality_friends_window_text_1", "locked_functionality_friends_window_text_2"}
  if not impl.locked_functionality_friends_window then

    PrepareAnimationScreenPart ("data/animations/locked_functionality_friends_window.xml")



    impl.locked_functionality_friends_window = {}

    impl.locked_functionality_friends_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.locked_functionality_friends_window.resource       = "data/res/materials/locked_functionality_friends_window_layers.xmtl"
    impl.locked_functionality_friends_window.sprites        = { "locked_functionality_friends_window_bg", "locked_functionality_window_screen_lock", "screen_lock",
      "locked_functionality_friends_window_close_l", "locked_functionality_friends_window_vk_button_l", "locked_functionality_friends_window_fb_button_l" }
    impl.locked_functionality_friends_window.text_lines     = {}
    impl.locked_functionality_friends_window.all_text_lines = GetTextLines ({})
    impl.locked_functionality_friends_window.all_sprites    = MakeAllSprites (sprite_list)
    impl.locked_functionality_friends_window.on_hide        = function () end
  end

  ShowWindow (impl, impl.locked_functionality_friends_window, {}, locked_functionality_friends_text_line_list, {}, nil, nil, nil, nil, nil, nil, nil, true, nil, nil, { hide_on_back = true })
end

local_methods.ShowLockedFunctionalityWindow = function (impl, type)
  local_methods.LockedFunctionalityWindow (impl)
  local type_window = { 'REMOVED(DIAMONDS)', 'battle', 'boosters', 'social', 'coins', 'hall_of_fame', 'gifts', 'friends' }

  local additional_sprites    = { "locked_functionality_window_vk_button_l", "locked_functionality_window_fb_button_l", "locked_functionality_window_close_l", "locked_functionality_window_" .. type_window[type] }
  local additional_text_lines = { "locked_functionality_window_more_options", "locked_functionality_window_text_vk_button", "locked_functionality_window_text_fb_button",
                      "locked_functionality_window_social_connect", "locked_functionality_window_" .. type_window[type] .. "_1", "locked_functionality_window_" .. type_window[type] .. "_2" }

  ShowWindow (impl, impl.locked_functionality_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, nil, nil, nil, true, nil, nil, { hide_on_back = true })
end

local_methods.ShowHelloCombinedWindow = function (impl, level, money, diamonds)
  local_methods.LockedFunctionalityWindow (impl)

  local social_name = GetSocialManager ().GetNetworkName ()

  local socials = {}

  socials ["VK"] = "Vkontakte"
  socials ["FB"] = "Facebook"

  local additional_sprites    = {  "locked_functionality_window_center_button_l", "locked_functionality_window_close_l", "locked_functionality_window_combined" }
  local additional_text_lines = { "locked_functionality_window_hello_user_title", "locked_functionality_window_text_center_button", "locked_functionality_window_hello_user_text_1",
    "locked_functionality_window_hello_user_text_1", "locked_functionality_window_hello_user_text_2", "locked_functionality_window_hello_user_text_3",
    "locked_functionality_window_hello_user_text_4", "locked_functionality_window_hello_user_text_5" }

  level    = tostring (level)
  money    = FormatMoneyString (money, 10, nil, true)
  diamonds = FormatMoneyString (diamonds * 100, 10, nil, true)
  
  ShowWindow (impl, impl.locked_functionality_window, additional_sprites, additional_text_lines, {}, function ()
    impl.locked_functionality_window.on_hide_handler = function ()
      HsConnectionSendMessage (1, { event = "migration.complete" })
    end
  
    GetTextLine('locked_functionality_window_hello_user_title').node.Text = ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TITLE"), {{ token = "{username}",  replacement = GetSocialManager ().GetFirstName () or "" }})
    GetTextLine('locked_functionality_window_hello_user_text_1').node.Text = ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TEXT_1"), {{ token = "{social_network}",  replacement = socials[social_name] }})
    GetTextLine('locked_functionality_window_hello_user_text_2').node.Text = ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TEXT_2"), {{ token = "{social_network}",  replacement = socials[social_name] }})
    
    SetTextLineTextWithReplacements (GetTextLine('locked_functionality_window_hello_user_text_3').node, GetLocalizedString("HELLO_USER.TEXT_3"), {{ token = "{level}",  replacement = level, color_factor = vec4 (1.09, 1.66, 2.22, 1) }})
    SetTextLineTextWithReplacements (GetTextLine('locked_functionality_window_hello_user_text_4').node, GetLocalizedString("HELLO_USER.TEXT_4"), {
      { token = "{coins}",  replacement = money, color_factor = vec4 (1.09, 1.66, 2.22, 1) },
      { token = "{diamonds}",  replacement = diamonds, color_factor = vec4 (1.09, 1.66, 2.22, 1) }})
  end, nil, nil, nil, nil, nil, nil, true, nil, nil, { hide_on_back = true })
end

local_methods.LoadDataTransferAndChangeLanguageWindow = function (impl)
  if not impl.data_transfer_window then

    local data_transfer_window_sprite_list = { "data_transfer_window_background", "data_transfer_window_screen_lock",
      "data_transfer_window_close_l", "data_transfer_window_close_p", "screen_lock",
      "data_transfer_window_one_button", "data_transfer_window_one_button_pressed"}
    local data_transfer_window_text_line_list = {"data_transfer_window_one_button_text", "data_transfer_window_text_line_1",
      "data_transfer_window_text_line_2", "data_transfer_window_text_line_3", "data_transfer_window_title_line",
      "data_transfer_window_text_line_1_1", "data_transfer_window_text_line_2_1", "data_transfer_window_text_line_3_1", "data_transfer_window_text_line_4_1"}

    PrepareAnimationScreenPart ("data/animations/data_transfer_window.xml")

    impl.data_transfer_window = {}

    impl.data_transfer_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.data_transfer_window.resource       = "data/res/materials/data_transfer_window_layers.xmtl"
    impl.data_transfer_window.sprites        = { "data_transfer_window_background", "data_transfer_window_screen_lock", "screen_lock", "data_transfer_window_close_l", "data_transfer_window_one_button" }
    impl.data_transfer_window.text_lines     = GetTextLines (data_transfer_window_text_line_list)
    impl.data_transfer_window.all_text_lines = GetTextLines (data_transfer_window_text_line_list)
    impl.data_transfer_window.all_sprites    = MakeAllSprites (data_transfer_window_sprite_list)
    impl.data_transfer_window.on_hide        = function ()
      if impl.data_transfer_window.hide_handler then
        impl.data_transfer_window.hide_handler ()
        impl.data_transfer_window.hide_handler = nil
      end 
    end
  end
end

local_methods.ShowDataTransferAndChangeLanguageWindow = function (impl, titlt, texts, start_fun, on_ok_handler)
  ShowWindow (impl, impl.data_transfer_window, {}, {}, {}, function ()
    if start_fun then start_fun () end
    GetTextLine('data_transfer_window_title_line').node.Text = titlt
    for i = 1, #texts do
      if #texts == 4 then
        GetTextLine('data_transfer_window_text_line_' .. i .. "_1").node.Text = texts[i]
      else
        GetTextLine('data_transfer_window_text_line_' .. i).node.Text = texts[i]
      end
    end
    GetTextLine('data_transfer_window_one_button_text').node.Text = GetLocalizedString("OK")
  end, nil, nil, nil, on_ok_handler, nil, nil, nil, nil, nil, { hide_on_back = true })
end

local_methods.ShowInfoWindowDataTransferWindow = function (impl)
  local_methods.LoadDataTransferAndChangeLanguageWindow(impl)

  local social_name = GetSocialManager ().GetNetworkName ()
  
  local socials = {}
 
  socials ["VK"] = "Vkontakte"
  socials ["FB"] = "Facebook"
  socials ["GUEST"] = "Guest"

  local texts = {}
  local title = ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TITLE"), {{ token = "{username}",  replacement = GetSocialManager ().GetFirstName () or "" }})

  table.insert(texts, ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TEXT_1_1"), {{ token = "{social_network}",  replacement = socials[social_name] }}))
  table.insert(texts, ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TEXT_1_2"), {{ token = "{social_network}",  replacement = socials[social_name] }}))
  table.insert(texts, ProcessStringWithReplacements (GetLocalizedString("HELLO_USER.TEXT_1_3"), {{ token = "{social_network}",  replacement = socials[social_name] }}))

  local_methods.ShowDataTransferAndChangeLanguageWindow (impl, title, texts, function ()
    impl.data_transfer_window.hide_handler = function ()
      HsConnectionSendMessage (1, { event = "migration.complete" })
    end
  end )
end

local_methods.ShowInfoChangeLanguageWindow = function (impl, on_ok_handler)
  local_methods.LoadDataTransferAndChangeLanguageWindow(impl)

  local texts = {}
  local title = GetLocalizedString("CHANGE_LANGUAGE.LINE_1")

  table.insert(texts, GetLocalizedString('CHANGE_LANGUAGE.LINE_2'))
  table.insert(texts, GetLocalizedString('CHANGE_LANGUAGE.LINE_3'))
  table.insert(texts, GetLocalizedString('CHANGE_LANGUAGE.LINE_4'))
  table.insert(texts, GetLocalizedString('CHANGE_LANGUAGE.LINE_5'))

  local_methods.ShowDataTransferAndChangeLanguageWindow (impl, title, texts, nil, on_ok_handler )
end

local_methods.ShowOpenDoorsActionIsOver = function (impl, level)
  LoadInfoWindow (impl)

  local additional_sprites    = { "info_window_action_is_over_overlay", "info_window_close_button" }
  local additional_text_lines = { "info_window_action_is_over_text_1", "info_window_action_is_over_text_2", "info_window_action_is_over_text_3", "info_window_action_is_over_text_title" }

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_action_is_over_layers.xmtl"))

  GetTextLine("info_window_action_is_over_text_1").node.Text = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.ACTION_IS_OVER_TEXT_1"), {{ token = "{level}",  replacement = tostring (level) }})
  GetTextLine("info_window_action_is_over_text_2").node.Text = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.ACTION_IS_OVER_TEXT_2"), {{ token = "{level}",  replacement = tostring (level) }})
  GetTextLine("info_window_action_is_over_text_3").node.Text = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.ACTION_IS_OVER_TEXT_3"), {{ token = "{level}",  replacement = tostring (level) }})

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, function ()
    impl.info_window.hide_handler = function ()
      SetCurrentScreen ("lobby")
    end
  end, nil, nil, nil, nil, nil, additional_resource_bindings, true, nil, nil, { hide_on_back = true })
end

local_methods.ShowBonusCoinsWindow = function (impl, coins, close_handler)
  LoadInfoWindow (impl)

  local additional_sprites    = { "info_window_one_button",  "info_window_get_coins_overlay_1", "info_window_get_coins_overlay_2", 'info_window_screen_lock' }
  local additional_text_lines = { "info_window_get_coins_one_button_text",  "info_window_get_coins_text_1", "info_window_get_coins_text_2", "info_window_get_coins_text_3", "info_window_get_coins_text_title" }

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_get_coins_layers.xmtl"))

  GetTextLine ("info_window_get_coins_text_1").node.Text = FormatMoneyString (coins, 10, nil, true)

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, close_handler ~= nil, true, true, function ()
    if close_handler then
      close_handler ()
    end
  end, nil, additional_resource_bindings, true, nil, nil, { dont_restore_sprites_materials = true, on_back_handler = function ()
    OnOkPressed (impl)
  end })
end

local function LoadBonusGoogle (impl)
  if impl.bonus_google_window then return end

  PrepareAnimationScreenPart ("data/animations/bonus_google_window.xml")

  local bonus_google_window_sprite_list = { "bonus_google_background", "google_label", "recommend_btn", "close"}

  local bonus_google_window_text_line_list = {"bonus_google_title_line", "bonus_google_text_line_1", "bonus_google_text_line_2", "bonus_google_text_line_3", 
                           "bonus_google_text_line_4", "bonus_google_text_line_5", "bonus_google_text_button"}
    
  impl.bonus_google_window = {}

  impl.bonus_google_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.bonus_google_window.resource       = "data/res/materials/bonus_google_window_layers.xmtl"
  impl.bonus_google_window.sprites        = { "bonus_google_background","google_label", "recommend_btn", "close", "bonus_google_window_screen_lock" }
  impl.bonus_google_window.text_lines     = {}
  impl.bonus_google_window.all_text_lines = GetTextLines (bonus_google_window_text_line_list)
  impl.bonus_google_window.all_sprites    = MakeAllSprites (bonus_google_window_sprite_list)
end
                           
local_methods.ShowBonusGoogle = function (impl, coins)
  LoadBonusGoogle (impl)
 
  local google_text     = {"bonus_google_title_line", "bonus_google_text_line_1", "bonus_google_text_line_2", "bonus_google_text_line_3", 
                           "bonus_google_text_line_4", "bonus_google_text_line_5", "bonus_google_text_button"}

  ShowWindow (impl, impl.bonus_google_window, {}, google_text, impl.bonus_google_window, nil, nil, nil, nil, nil, nil, nil, true)
  
  GetTextLine("bonus_google_text_line_4").node.Text = ProcessStringWithReplacements (GetLocalizedString("BONUS_GOOGLE.TEXT_LINE_4"), {{ token = "{amount}",  replacement = FormatMoneyString (coins) }})
end

local_methods.ShowHappyHourWindow = function (impl, level, day_or_hour, type, coins, slot_id, close_handler, close_ok_handler, time_end)
  LoadInfoWindow (impl)
  
  local additional_sprites    = { "info_window_one_button", "info_window_happy_hour_blocker", 'info_window_screen_lock', 
                                  "info_window_happy_hour_top_overlay", "info_window_happy_hour_top_overlay_time_" .. day_or_hour }
  local additional_text_lines = { "info_window_happy_hour_title", "info_window_happy_hour_text_button", "info_window_happy_hour_top_text" }

  if type ~= 3 and type ~= 2 then
    table.insert (additional_text_lines, "info_window_happy_hour_time_before_closure_value")
    table.insert (additional_text_lines, "info_window_happy_hour_time_before_closure_text")
  end

  if type ~= 2 then
    table.insert (additional_sprites, "info_window_close_button")
  end

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_happy_hour_layers.xmtl"))

  local ok_function

  local button_text = ""
  local text_1_1    = ""
  local text_1_2    = ""
  local text_1_3    = ""
  local text_2_1    = ""
  local text_2_2    = ""
  local text_2_3    = ""

  local slot_xmtl_path
  local slot_icon_material = "transparent"

  if slot_id then
    local slot_name          = ""
    local slot_path          = ""
    local slots              = ApplicationDataGet (SLOTS_DESCS)

    for i = 1, #slots do
      local slot_desc = slots [i]

      if slot_desc.id and slot_desc.id + 0 == slot_id + 0 then
        slot_name          = slot_desc.game_name
        slot_icon_material = slot_desc.icon_material
        slot_path          = slot_desc.path
        slot_xmtl_path     = slot_desc.xmtl_path
        break
      end
    end
    
    if slot_xmtl_path then
      table.insert (additional_resource_bindings, LoadResource (slot_xmtl_path))
    end

    if GetUserLevel ().id + 0 < level then
      button_text = GetLocalizedString("CLOSE")
      ok_function = close_handler
      close_ok_handler = nil
    else
      button_text = GetLocalizedString("INFO_WINDOW.PLAY")
      ok_function = function () LaunchSlot (slot_path, slot_id)  end
    end

    text_2_1 = slot_name
    text_2_2 = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.HAPPY_HOUR_LEVEL_1"), {{ token = "{level}",  replacement = tostring (level) }})
    text_2_3 = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.HAPPY_HOUR_LEVEL_2"), {{ token = "{level}",  replacement = tostring (level) }})

    table.insert (additional_text_lines, "info_window_happy_hour_blocker_text_2_1")
    table.insert (additional_text_lines, "info_window_happy_hour_blocker_text_2_2")
    table.insert (additional_text_lines, "info_window_happy_hour_blocker_text_2_3")

    if slot_xmtl_path and Common.File.IsFileExist (slot_xmtl_path) then
      table.insert (additional_sprites, "info_window_happy_hour_icon_zone")
    end
  else
    if type == 2 then
      ok_function = function ()
        local event = ApplicationDataGet (GIFTS_EVENT)
        
        if event then
          local gifts = event:Find ("receivedGifts"):Child (0)

          local gifts_count = gifts.ChildrenCount

          for i = gifts_count - 1, 0, -1 do
            local gift_node = gifts:Child (i)

            if gift_node:Find ("userId"):Attribute (0) + 0 == -1 and gift_node:Find ("type"):Attribute (0) + 0 == 0 then
              game.scripts.top_ui.MoneyForceNextUpdate ()
              HsConnectionSendMessage (1, { event = "gifts.take", giftId = gift_node:Find ("giftId"):Attribute (0) })
              break    
            end
          end
        end
        
        local_methods.ShowBonusCoinsWindow (impl, coins, close_handler)
        
        close_ok_handler = nil
      end
    elseif type == 3 then
      ok_function = function ()
        local event = ApplicationDataGet (GIFTS_EVENT)
        
        local gift_found = false
        
        if event then
          local gifts = event:Find ("receivedGifts"):Child (0)

          local gifts_count = gifts.ChildrenCount

          for i = gifts_count - 1, 0, -1 do
            local gift_node = gifts:Child (i)

            if gift_node:Find ("userId"):Attribute (0) + 0 == -1 and gift_node:Find ("type"):Attribute (0) + 0 == 1 then
              local gift_id = gift_node:Find ("giftId"):Attribute (0)
              
              gift_found = true
            
              if game.current_screen.name ~= "lobby" then
                ApplicationDataSet (LOBBY_SHOW_FREE_SPINS_TAKE_GIFT_ID, gift_id)
                SetCurrentScreen ("lobby")
                return
              else
                game.scripts.ui_windows.ShowTakeFreeSpinsWindow (gift_id, close_handler)
              end
              
              break    
            end
          end
        end
        
        if not gift_found then  --in case of server bug
          close_handler ()
        end
        
        close_ok_handler = nil
      end
    else
      ok_function = function ()
        ShowBuyChipsWindow (impl)
      end
    end

    local sale_coins = "0"

    if type == 4 then
      GetTextLine ("info_window_happy_hour_text_in_tag").node.Text = GetLocalizedString("INFO_WINDOW.HAPPY_HOUR_TITLE_SALE") .. " " .. coins
      table.insert (additional_text_lines, "info_window_happy_hour_text_in_tag")
      sale_coins = local_methods.GetSaleCoins (impl)
    end

    if type == 2 then
      text_1_1 = FormatMoneyString (coins, 10, nil, true)
      button_text = GetLocalizedString("LOBBY.MEGA_BONUS_COLLECT")
    elseif type == 3 then
      text_1_1 = coins
      button_text = GetLocalizedString("LOBBY.TAKE_GIFT_SPIN")
    else
      text_1_1 = coins
      button_text = GetLocalizedString("INFO_WINDOW.BUY_CHIPS")
    end
    
    text_1_2 = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.HAPPY_HOUR_TITLE_1_" .. type), {{ token = "{coins}",  replacement = sale_coins }})
    text_1_3 = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.HAPPY_HOUR_TITLE_2_" .. type), {{ token = "{coins}",  replacement = sale_coins }})

    table.insert (additional_text_lines, "info_window_happy_hour_blocker_text_1_1")
    table.insert (additional_text_lines, "info_window_happy_hour_blocker_text_1_2")
    table.insert (additional_text_lines, "info_window_happy_hour_blocker_text_1_3")
    table.insert (additional_sprites,  "info_window_happy_hour_overlay_" ..  type)
  end

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, function ()
    if slot_xmtl_path and Common.File.IsFileExist (slot_xmtl_path) then
      local icon = GetSprite ("info_window_happy_hour_icon_zone")
      icon.node.Material = slot_icon_material
      icon.node.TexSize  = vec2 (1, 1) 
      icon.node.TexOffset = vec2 (0, 0)
      
      icon.tracks ["texsize"] = nil
      icon.tracks ["texoffset"] = nil
    end

    if type ~= 3 or type ~= 2 then
      local_methods.ShowHappyTimerUpdate (impl, math.max (0, time_end - game.main_timer.GetTime ()))

      impl.info_window.expire_time_update_action = StartPeriodicCall ("ui", 1, function ()
        local_methods.ShowHappyTimerUpdate (impl, math.max (0, time_end - game.main_timer.GetTime ()))
      end)
    end

    local z_position_frame_animation = GetSprite  ("Border_anim_1").z - EPS
    local frame_animations_sprite    = CreateFramesNames ("Border_anim_%d", 1, 46)

    impl.info_window.frame_animation = CreateAnimationSpriteClass ( { animation = frame_animations_sprite }, frame_animations_sprite [1], 30 ).CreateInstance ()

    SetAnimationPosition (impl.info_window.frame_animation, frame_animations_sprite [1], z_position_frame_animation)

    GetTextLine("info_window_happy_hour_title").node.Text    = GetLocalizedString ("INFO_WINDOW.HAPPY_HOUR_TITLE_DAY_OR_HOUR_" .. day_or_hour)
    GetTextLine("info_window_happy_hour_top_text").node.Text = GetLocalizedString ("INFO_WINDOW.HAPPY_HOUR_TEXT_TOP_" .. day_or_hour)

    GetTextLine ("info_window_happy_hour_text_button").node.Text = button_text

    GetTextLine ("info_window_happy_hour_blocker_text_1_1").node.Text = text_1_1
    GetTextLine ("info_window_happy_hour_blocker_text_1_2").node.Text = text_1_2
    GetTextLine ("info_window_happy_hour_blocker_text_1_3").node.Text = text_1_3
    GetTextLine ("info_window_happy_hour_blocker_text_2_1").node.Text = text_2_1
    GetTextLine ("info_window_happy_hour_blocker_text_2_2").node.Text = text_2_2
    GetTextLine ("info_window_happy_hour_blocker_text_2_3").node.Text = text_2_3

    ExecuteOnNextUpdate (function ()
      impl.info_window.happy_time_window_close_handler = close_handler
    end)

    if impl.info_window.frame_animation then
      impl.info_window.frame_animation.Show ('animation', nil, nil, 2)
    end
  end, close_handler ~= nil, true, nil, function ()  --dont disable camera if close handler not nil (show all promos in lobby, regular promo will be hided if this is false)
    if ok_function then
      ok_function ()
    end

    if close_ok_handler then
      close_ok_handler ()
    end
  end, nil, additional_resource_bindings, true, nil, nil, { dont_restore_sprites_materials = true, hide_on_back = true })
end

local_methods.ShowHappyHourWindowAction = function (impl, action, closed_handler, closed_ok_handler) 
  local additional = action.additional:Child (0)
  local type       = GetNodeChildAttribute (additional, "type") + 0

  HsConnectionSendMessage (1, { event = "promo.check", id = action.unique_id })      

  if type == 1 then
    game.scripts.ui_windows.ShowHappyHourWindowCoins (GetNodeChildAttribute (additional, "duration", 1) + 0, GetNodeChildAttribute (additional, "value1") + 0, closed_handler, closed_ok_handler, action.time_end)
  elseif type == 2 then
    game.scripts.ui_windows.ShowHappyHourWindowFreeSpins (GetNodeChildAttribute (additional, "duration", 1) + 0, GetNodeChildAttribute (additional, "value1") + 0, closed_handler, closed_ok_handler, action.time_end)
  elseif type == 3 then
    game.scripts.ui_windows.ShowHappyHourWindowDeposit (GetNodeChildAttribute (additional, "duration", 1) + 0, GetNodeChildAttribute (additional, "value1") + 0, closed_handler, closed_ok_handler, action.time_end)
  elseif type == 4 then
    game.scripts.ui_windows.ShowHappyHourWindowDiscount (GetNodeChildAttribute (additional, "duration", 1) + 0, GetNodeChildAttribute (additional, "value1") + 0, closed_handler, closed_ok_handler, action.time_end)
  elseif type == 5 then
    game.scripts.ui_windows.ShowHappyHourWindowSlot (GetNodeChildAttribute (additional, "duration", 1) + 0, GetNodeChildAttribute (additional, "value1") + 0, GetNodeChildAttribute (additional, "value2", 1) + 0, closed_handler, closed_ok_handler, action.time_end)
  else
    print ("WARNING: unsupported happy time action type " .. type)
    return false
  end
  
  return true
end

local_methods.ShowNotHaveCoinsEnoughForBattleWindow = function (impl, battle_search)
  LoadInfoWindow (impl)

  local additional_sprites    = { "info_window_two_button_right", "info_window_two_button_left", "info_window_close_button", 'info_window_screen_lock'  }
  local additional_text_lines = { 'info_window_not_have_coins_enough_for_battle_title', 'info_window_not_have_coins_enough_for_battle_text_1',
                                  'info_window_not_have_coins_enough_for_battle_text_2', 'info_window_not_have_coins_enough_for_battle_text_3',
                                  'info_window_not_have_coins_enough_for_battle_left_button', 'info_window_not_have_coins_enough_for_battle_right' }

  local additional_resource_bindings = {}

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, true, true, function ()
    ShowBuyChipsWindow (impl)
  end, function () battle_search () end, nil, nil, nil, nil, { hide_on_back = true })
end

local function ShowNoMoneyWindow (impl)
  LoadInfoWindow (impl)
  
  local additional_sprites    = { "info_window_one_button", "info_window_close_button", "info_window_no_money_overlay1", "info_window_no_money_overlay2", 'info_window_screen_lock' }
  local additional_text_lines = { "info_window_zero_money_one_button", "info_window_zero_money_title_line_1", "info_window_zero_money_text_line_1", "info_window_zero_money_text_line_2", "info_window_zero_money_text_line_3" }

  local additional_resource_bindings = {}
  
  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_no_money_layers.xmtl"))

  GetTextLine ("info_window_zero_money_text_line_2").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.ZERO_MONEY_TEXT_LINE2"), GetGrantPeriod () / 60 / 60)
  GetTextLine ("info_window_zero_money_text_line_3").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.ZERO_MONEY_TEXT_LINE3"), GetGrantPeriod () / 60 / 60)

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, true, function ()
      ShowBuyChipsWindow (impl) 
    end, nil, additional_resource_bindings, nil, nil, nil, { hide_on_back = true })
end

local function ShowNoDiamondWindow (impl)
  LoadInfoWindow (impl)

  local additional_sprites    = { "info_window_one_button", "info_window_close_button", "info_window_no_diamond_overlay1", "info_window_no_diamond_overlay2", 'info_window_screen_lock' }
  local additional_text_lines = { "info_window_zero_diamond_one_button", "info_window_zero_diamond_title_line_1", "info_window_zero_diamond_text_line_1", "info_window_zero_diamond_text_line_2", "info_window_zero_diamond_text_line_3" }

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_no_diamond_layers.xmtl"))

  GetTextLine ("info_window_zero_diamond_text_line_2").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.ZERO_DIAMOND_TEXT_LINE2"), GetGrantPeriod () / 60 / 60)
  GetTextLine ("info_window_zero_diamond_text_line_3").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.ZERO_DIAMOND_TEXT_LINE3"), GetGrantPeriod () / 60 / 60)

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, true, function ()
    ShowBuyDiamondsWindow (impl)
  end, nil, additional_resource_bindings, nil, nil, nil, { hide_on_back = true })
end

local function OnLoadingWindowHide (impl)
  impl.loading_window_show_counter = impl.loading_window_show_counter - 1
  
  if impl.loading_window_show_counter == 0 and impl.loading_activity_indicator_action then
    impl.loading_activity_indicator_action.Stop ()
    impl.loading_activity_indicator_action = nil
  end
end

local function LoadLoadingWindow (impl)
  if impl.loading_window then return end

  PrepareAnimationScreenPart ("data/animations/loading_window.xml")

  impl.loading_window = {}

  impl.loading_window.camera          = game.current_screen.cameras.modal_window_camera
  impl.loading_window.resource        = "data/res/materials/loading_window_layers.xmtl"
  impl.loading_window.sprites         = {}
  impl.loading_window.text_lines      = GetTextLines ({ "loading_window_line_0", "loading_window_line_1", "loading_window_line_2", "loading_window_line_3" })
  impl.loading_window.all_text_lines  = GetTextLines ({ "loading_window_line_0", "loading_window_line_1", "loading_window_line_2", "loading_window_line_3" })
  impl.loading_window.all_sprites     = MakeAllSprites ({ "loading_window_background", "loading_window_circle", "loading_window_background_closable",
    "loading_window_close_l", "loading_window_close_p" })
  impl.loading_window.dont_play_sound = true
  impl.loading_window.on_hide         = function () OnLoadingWindowHide (impl) end
end

local function ShowLoadingWindow (impl, progress, first_line_text, closable, close_handler, slot_id)
  LoadLoadingWindow (impl)

  impl.loading_window_show_counter = (impl.loading_window_show_counter or 0) + 1

  if first_line_text then
    GetTextLine ("loading_window_line_0").text = first_line_text
  else
    GetTextLine ("loading_window_line_0").text = GetLocalizedString ("PLEASE_WAIT_GAME_DOWNLOADING_LINE0")
    GetTextLine ("loading_window_line_1").text = GetLocalizedString ("PLEASE_WAIT_GAME_DOWNLOADING_LINE1")
    GetTextLine ("loading_window_line_2").text = GetLocalizedString ("PLEASE_WAIT_GAME_DOWNLOADING_LINE2")
  end
  
  GetTextLine ("loading_window_line_3").text = string.format (GetLocalizedString ("PLEASE_WAIT_GAME_DOWNLOADING_PROGRESS"), progress)

  if not impl.loading_activity_indicator_action then
    impl.loading_activity_indicator_action = StartPeriodicCall ("ui", 1 / 10, function ()
      GetSprite ("loading_window_circle").node:Rotate (0, 0, -45)
    end)
  end

  local additional_sprites = {}
  
  if closable then
    additional_sprites = { "loading_window_background_closable", "loading_window_close_l" }
  else
    additional_sprites = { "loading_window_background" }
  end

  impl.loading_window.close_handler = close_handler

  local has_icon
  local additional_resource_bindings = {}

  if slot_id then
    local slot_path          = ""
    local slot_icon_material = "transparent"
    local slot_xmtl_path     = "data/res/materials/info_window_new_game_layers.xmtl"
    local slots              = ApplicationDataGet (SLOTS_DESCS)

    for i = 1, #slots do
      local slot_desc = slots [i]

      if slot_desc.id and slot_desc.id + 0 == slot_id + 0 then
        slot_icon_material = slot_desc.icon_material
        slot_path          = slot_desc.path
        slot_xmtl_path     = slot_desc.xmtl_path
        break
      end
    end

    if slot_xmtl_path and Common.File.IsFileExist (slot_xmtl_path) then
      has_icon = true
    
      table.insert (additional_resource_bindings, LoadResource (slot_xmtl_path))

      local icon_zone_name
      
      if string.len (GetTextLine ("loading_window_line_0").text) > 0 then
        icon_zone_name = "loading_window_icon_zone"
      else
        icon_zone_name = "loading_window_icon_zone_small"
      end

      local icon = GetSprite (icon_zone_name)
      
      icon.node.Material  = slot_icon_material
      icon.node.TexSize   = vec2 (1, 1)
      icon.node.TexOffset = vec2 (0, 0) 
      
      icon.tracks ["texsize"] = nil
      icon.tracks ["texoffset"] = nil

      table.insert (additional_sprites, icon_zone_name)
    end
  end

  if not has_icon then
    table.insert (additional_sprites, "loading_window_circle")
  end
  
  local params = {}
  
  if closable then
    params.on_back_handler = function () game.scripts.ui_windows.CloseLoadingWindow () end
  end
  
  ShowWindow (impl, impl.loading_window, additional_sprites, {}, {}, nil, nil, true, nil, nil, nil, additional_resource_bindings, nil, nil, nil, params)
end

local_methods.CloseLoadingWindow = function (impl)
  if impl.loading_window.close_handler then
    impl.loading_window.close_handler ()
    impl.loading_window.close_handler = nil
  end
 
  HideWindow (impl)
end

local function ShowLogoutWindow (impl, logout_handler, texts_override)
  HideWindow (impl)

  local texts = texts_override or { GetLocalizedString ("LOBBY.LOGOUT_LINE1") }

  game.scripts.ui_windows.ShowAlertWindow (nil, 
    function ()
      PerformLogOut (logout_handler)
    end, 
    function () end, 
    GetLocalizedString ("LOBBY.LOGOUT_OK"), 
    texts
  )
end

local function UpdateSocialStateButton (impl)
  local social_name = GetSocialManager ().GetNetworkName ()
  if social_name == "VK" then
    GetTextLine ('options_window_vk_text').node.Text = GetLocalizedString ("OPTIONS.CHANGE_ACCOUNT")
  else
    GetTextLine ('options_window_vk_text').node.Text = GetLocalizedString ("OPTIONS.CONNECT")
  end
  if social_name == "FB" then
    GetTextLine ('options_window_fb_text').node.Text = GetLocalizedString ("OPTIONS.CHANGE_ACCOUNT")
  else
    GetTextLine ('options_window_fb_text').node.Text = GetLocalizedString ("OPTIONS.CONNECT")
  end
end

local function UpdateSoundStateButton (impl, state)
  local key
  
  if state then
    key = "OPTIONS.SOUNDS_OFF"
  else
    key = "OPTIONS.SOUNDS_ON"
  end

  GetTextLine ("options_window_sound_text").node.Text = GetLocalizedString (key) 
end

local function LoadOptionsWindow (impl)

  if impl.options_window then return end

  PrepareAnimationScreenPart ("data/animations/options_window.xml")

  impl.options_window = {}

  impl.options_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.options_window.resource       = "data/res/materials/options_window_layers.xmtl"
  impl.options_window.sprites        = { "options_window_background", "options_window_sound_l",
    "options_window_change_language_l", "options_window_vk_login_l",
    "screen_lock", "options_window_fb_login_l", "options_window_close_l" }
  impl.options_window.text_lines     = GetTextLines ({ "options_window_title_text", "options_window_language_text", "options_window_vk_text", "options_window_fb_text", "options_window_sound_text" })
  impl.options_window.all_text_lines = GetTextLines ({ "options_window_title_text", "options_window_language_text", "options_window_vk_text", "options_window_fb_text", "options_window_sound_text" })
  impl.options_window.all_sprites    = MakeAllSprites ({ "options_window_background", "options_window_sound_l", "options_window_sound_p", "options_window_sound_off_icon", "options_window_sound_on_icon",
    "options_window_change_language_l", "options_window_change_language_p", "options_window_vk_login_l", "options_window_vk_login_p",
    "screen_lock", "options_window_fb_login_l", "options_window_fb_login_l", "options_window_fb_login_p", "options_window_close_l", "options_window_close_p" })

end

local function ShowOptionsWindow (impl)
  LoadOptionsWindow (impl)

  local additional_sprites = { }

  local state

  if GetGameProperty (GAME_SOUND_ENABLED_ATTRIBUTE) == "true" then
    state = true
    table.insert (additional_sprites, "options_window_sound_on_icon")
  else
    state = false
    table.insert (additional_sprites, "options_window_sound_off_icon")
  end

  ShowWindow (impl, impl.options_window, additional_sprites, {}, {}, function ()
    UpdateSocialStateButton (impl)
    UpdateSoundStateButton (impl, state)
  end, nil, true, nil, nil, nil, nil, nil, nil, nil, { hide_on_back = true })
end

local function ToggleSoundState (impl)
  local current_state = (GetGameProperty (GAME_SOUND_ENABLED_ATTRIBUTE) or "true") == "true"
  local state         = not current_state

  SetGameProperty (GAME_SOUND_ENABLED_ATTRIBUTE, state)

  UpdateMusicVolume ()
  UpdateSoundsVolume ()

  SetSpriteActive ("options_window_sound_on_icon", state)
  SetSpriteActive ("options_window_sound_off_icon", not state)
  
  PlaySound ("click")

  UpdateSoundStateButton (impl, state)
end

local function CloseTakeFreeSpinsWindow (impl)
  game.scripts.lobby.SetStartSlotHandler (nil)
  game.scripts.ui_windows.HideWindow ()
end

local function ShowTakeFreeSpinsWindow (impl, gift_id, close_handler)
  LoadTakeFreeSpinsWindow (impl)

  impl.take_free_spins_window.close_handler = close_handler

  ShowWindow (impl, impl.take_free_spins_window, {}, {}, {}, nil, nil, true, nil, nil, nil, nil, nil, nil, nil, 
              { on_back_handler = function () game.scripts.ui_windows.CloseTakeFreeSpinsWindow () end })

  game.scripts.lobby.SetHappyTimeHelperVisible (false)
  
  local background_z                         = GetSprite ("take_free_spins_background").z
  local next_page_zone_sprite                = GetSprite ("next_page_button")
  local next_page_button_sprite              = GetSprite ("next_page_l")
  local next_page_button_disabled_sprite     = GetSprite ("next_page_d")
  local next_page_button_pressed_sprite      = GetSprite ("next_page_p")
  local previous_page_zone_sprite            = GetSprite ("previous_page_button")
  local previous_page_button_sprite          = GetSprite ("previous_page_l")
  local previous_page_button_disabled_sprite = GetSprite ("previous_page_d")
  local previous_page_button_pressed_sprite  = GetSprite ("previous_page_p")
  local slots_drag_zone_sprite               = GetSprite ("slots_drag_zone")

  impl.lobby_next_page_zone_z                = next_page_zone_sprite.z  
  impl.lobby_next_page_button_disabled_z     = next_page_button_disabled_sprite.z  
  impl.lobby_next_page_button_z              = next_page_button_sprite.z  
  impl.lobby_next_page_button_pressed_z      = next_page_button_pressed_sprite.z  
  impl.lobby_previous_page_zone_z            = previous_page_zone_sprite.z  
  impl.lobby_previous_page_button_disabled_z = previous_page_button_disabled_sprite.z  
  impl.lobby_previous_page_button_z          = previous_page_button_sprite.z  
  impl.lobby_previous_page_button_pressed_z  = previous_page_button_pressed_sprite.z  
  impl.lobby_slots_drag_zone_z               = slots_drag_zone_sprite.z
  impl.lobby_slots_z                         = game.scripts.lobby.GetSlotsZ ()
  
  next_page_zone_sprite.z                = background_z - EPS
  next_page_button_disabled_sprite.z     = background_z - EPS
  next_page_button_sprite.z              = background_z - EPS
  next_page_button_pressed_sprite.z      = background_z - EPS * 2
  previous_page_zone_sprite.z            = background_z - EPS
  previous_page_button_disabled_sprite.z = background_z - EPS
  previous_page_button_sprite.z          = background_z - EPS
  previous_page_button_pressed_sprite.z  = background_z - EPS * 2
  slots_drag_zone_sprite.z               = background_z + EPS * 2
  
  if IsSpriteActive ("next_page_button") then
    DeactivateSprite ("next_page_button")
    ActivateSprite ("next_page_button")
  end
  if IsSpriteActive ("next_page_d") then
    DeactivateSprite ("next_page_d")
    ActivateSprite ("next_page_d")
  end
  if IsSpriteActive ("next_page_l") then
    DeactivateSprite ("next_page_l")
    ActivateSprite ("next_page_l")
  end
  if IsSpriteActive ("previous_page_button") then
    DeactivateSprite ("previous_page_button")
    ActivateSprite ("previous_page_button")
  end
  if IsSpriteActive ("previous_page_d") then
    DeactivateSprite ("previous_page_d")
    ActivateSprite ("previous_page_d")
  end
  if IsSpriteActive ("previous_page_l") then
    DeactivateSprite ("previous_page_l")
    ActivateSprite ("previous_page_l")
  end
  
  DeactivateSprite ("slots_drag_zone")
  ActivateSprite ("slots_drag_zone")

  game.scripts.lobby.SetSlotsZ (background_z + EPS)

  game.scripts.lobby.SetStartSlotHandler (function (slot_id)
    HsConnectionSendMessage (1, { event = "gifts.take", giftId = gift_id, slotId = slot_id })    
  end)
  
  game.current_screen.vars.previous_page_button_offset = game.current_screen.vars.previous_page_button_take_fs_offset
  game.current_screen.vars.next_page_button_offset     = game.current_screen.vars.next_page_button_take_fs_offset
end

local function OnSendGiftsChoosed (impl, gift_type)
  local on_ok_handler = impl.windows_stack [#impl.windows_stack].on_ok_handler

  HideWindow (impl)

  if on_ok_handler then
    on_ok_handler (gift_type)
  end
end

local function ShowSendGiftsWindow (impl, on_ok_handler)
  LoadGiftWindow (impl)

  local additional_sprites    = { "window_gift_select_btn_1_zone", "window_gift_select_btn_2_zone"}
  local additional_text_lines = { "gift_window_send_gifts_free_spins", "gift_window_send_gifts_chips", "gift_window_send_gifts_title_line_1",
                                  "gift_window_send_gifts_chips_1", "gift_window_send_gifts_chips_2", "gift_window_send_gifts_chips_3" }

  ShowWindow (impl, impl.gift_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, on_ok_handler, nil, nil, nil, nil, nil, { hide_on_back = true })
end

local function UpdateSendGiftsSelectedFriendsLayout (impl)
  if impl.send_gifts_update_layout_locked then return end

  local table_zone        = GetSprite ("send_gift_window_selected_friends_list_zone").layout
  local table_zone_height = table_zone:Get ("Height") + 0
  local first_cell_place  = GetSprite ("send_gift_window_selected_friends_list_devider")
  local cell_height       = first_cell_place.layout:Get ("Height") + 0
  local cells_margin      = game.current_screen.vars.send_gift_window_selected_friends_cells_margin
  local cell_interval     = vec3 (0, -cell_height - cells_margin, 0)

  table.sort (impl.send_gifts_selected_friends_cells, function (cell1, cell2)
    local cell1_name = cell1.friend.first_name .. cell1.friend.last_name
    local cell2_name = cell2.friend.first_name .. cell2.friend.last_name
    
    return cell1_name < cell2_name
  end)
      
  local scroll_range_x = 0
  local scroll_range_y = math.max (0, -#impl.send_gifts_selected_friends_cells * cell_interval.y - table_zone_height)

  impl.send_gifts_selected_friends_scroll_view.SetScrollRange (scroll_range_x, scroll_range_y)
  impl.send_gifts_selected_friends_scroll_view.SetTableParams (cell_interval.y, table_zone_height)

  for i = 1, #impl.send_gifts_selected_friends_cells do
    local cell = impl.send_gifts_selected_friends_cells [i]
    
    cell.node.Position = (i - 0.5) * cell_interval
  end
    
  OnSendGiftsSelectedFriendsScrolled (impl)
  
  SetSpritesActive (#impl.send_gifts_selected_friends_cells > 0, "send_gift_window_send_l")
  SetSpritesActive (#impl.send_gifts_selected_friends_cells < 1, "send_gift_window_send_d")
  SetTextLineActive ("send_gift_window_send", #impl.send_gifts_selected_friends_cells > 0)
  SetTextLineActive ("send_gift_window_send_d", #impl.send_gifts_selected_friends_cells < 1)
end

local function MakeButton (node, material, tex_size, tex_offset, pressed_material, pressed_tex_size, pressed_tex_offset, on_click, scale)
  local button_handlers = {}

  table.insert (button_handlers, AttachInputZoneToNode (node, Scene.InputZoneNotification.OnPress, function ()
    node.Material  = pressed_material
    node.TexSize   = pressed_tex_size
    node.TexOffset = pressed_tex_offset
  end, scale))
  table.insert (button_handlers, AttachInputZoneToNode (node, Scene.InputZoneNotification.OnTouchLeave, function ()
    node.Material  = material
    node.TexSize   = tex_size
    node.TexOffset = tex_offset
  end))
  table.insert (button_handlers, AttachInputZoneToNode (node, Scene.InputZoneNotification.OnClick, function ()
    node.Material  = material
    node.TexSize   = tex_size
    node.TexOffset = tex_offset
    on_click ()
  end))
  
  return button_handlers
end

local_methods.AddSendGiftsSelectedFriendsCell = function (impl, friend, scissor)
  local cell = {}
  local cell_nodes = {}      
  
  cell.handlers = {}
  
  cell.friend = friend
  
  cell.node = Scene.Node.Create ()

  cell.node:BindToParent (impl.send_gifts_selected_friends_cells_node)

  local z = 0
  
  local first_cell_place = GetSprite ("send_gift_window_selected_friends_list_devider")
  local name_text        = GetSprite ("send_gift_window_selected_friends_name_text")
  local name_text_size   = game.current_screen.vars.send_gift_window_selected_friends_name_text_size
  local cell_width       = first_cell_place.layout:Get ("Width") + 0
  local cell_height      = first_cell_place.layout:Get ("Height") + 0
  local cells_margin     = game.current_screen.vars.send_gift_window_selected_friends_cells_margin
  local cell_interval    = vec3 (0, -cell_height - cells_margin, 0)
  local delete_button    = GetSprite ("send_gift_window_selected_friends_list_delete")

  cell.underlay = Scene.Sprite.Create ()
  
  cell.underlay.Scale     = vec3 (cell_width, cell_height, 1)
  cell.underlay.Material  = first_cell_place.node.Material
  cell.underlay.TexSize   = StringToVec (first_cell_place.layout:Get ("TexSize"))
  cell.underlay.TexOffset = StringToVec (first_cell_place.layout:Get ("TexOffset"))
  
  cell.underlay:BindToParent (cell.node)
  
  table.insert (cell_nodes, cell.underlay)

  z = z - EPS
  
  cell.delete_button = Scene.Sprite.Create ()
  
  cell.delete_button.Position  = GetSpriteOffset (first_cell_place.layout, delete_button.layout) + vec3 (0, 0, z)    
  cell.delete_button.Scale     = vec3 (delete_button.layout:Get ("Width") + 0, delete_button.layout:Get ("Height") + 0, 1)
  cell.delete_button.Material  = delete_button.node.Material
  cell.delete_button.TexSize   = StringToVec (delete_button.layout:Get ("TexSize"))
  cell.delete_button.TexOffset = StringToVec (delete_button.layout:Get ("TexOffset"))
  
  cell.delete_button:BindToParent (cell.node)

  cell.on_click = function ()
    if not impl.send_gifts_update_layout_locked then
      PlaySound ('click')
    end

    cell.node:Unbind ()

    for j = 1, #cell.handlers do
      local handler = cell.handlers [j]
    
      for k = 1, #handler do
        handler [k].connection:Disconnect ()
        handler [k].callback = nil
      end
    end
    
    for i = 1, #impl.send_gifts_selected_friends_cells do
      if cell == impl.send_gifts_selected_friends_cells [i] then
        table.remove (impl.send_gifts_selected_friends_cells, i)
        break
      end
    end
  
    UpdateSendGiftsSelectedFriendsLayout (impl)
    
    local_methods.AddSendGiftsFriendsCell (impl, cell.friend)

    DeactivateSprite('send_gift_window_check_all_friends')
  end
  
  table.insert (cell.handlers, MakeButton (cell.delete_button, delete_button.node.Material, StringToVec (delete_button.layout:Get ("TexSize")), StringToVec (delete_button.layout:Get ("TexOffset")), 
                                           delete_button.node.Material, StringToVec (delete_button.layout:Get ("TexSize")), StringToVec (delete_button.layout:Get ("TexOffset")), 
                                           function ()
    cell.on_click ()
  end, vec3 (2, 2, 1)))
  
  table.insert (cell_nodes, cell.delete_button)

  z = z - EPS

  cell.name_text = Scene.TextLine.Create (game.font_library)

  cell.name_text.Position            = GetSpriteOffset (first_cell_place.layout, name_text.layout) + vec3 (0, 0, z)    
  cell.name_text.Scale               = vec3 (name_text_size, name_text_size, 1)

  if GetCurrentLocaleIndex() ~= 1 or impl.phone_ui then
    cell.name_text.Font                = "data/fonts/MyriadPro-Bold.xfont"
  else
    cell.name_text.Font                = "data/fonts/MyriadPro-BoldCond.xfont"
  end
  
  cell.name_text.Color               = vec4 (0.88, 0.86, 0.62, 1)
  cell.name_text.HorizontalAlignment = Scene.TextLineAlignment.Left
  cell.name_text.VerticalAlignment   = Scene.TextLineAlignment.BaseLine
  cell.name_text.FontCreationParams  = game.font_creation_params
  cell.name_text.Material            = "text_material"

  cell.name_text.Text = friend.first_name .. " " .. friend.last_name

  if cell.name_text.TextLength > game.current_screen.vars.send_gift_window_friends_name_long_text then
    cell.name_text.Scale = vec3 (game.current_screen.vars.send_gift_window_friends_name_long_text_size, game.current_screen.vars.send_gift_window_friends_name_long_text_size, 1)
  end

  cell.name_text:BindToParent (cell.node)
  
  table.insert (cell_nodes, cell.name_text)      

  table.insert (impl.send_gifts_selected_friends_cells, cell)

  for j = 1, #cell_nodes do
    cell_nodes [j].Scissor = scissor 
  end

  UpdateSendGiftsSelectedFriendsLayout (impl)
end

local function UpdateSendGiftsFriendsLayout (impl)
  if impl.send_gifts_update_layout_locked then return end

  local table_zone        = GetSprite ("send_gift_window_friends_list_zone").layout
  local table_zone_height = table_zone:Get ("Height") + 0
  local first_cell_place = GetSprite ("send_gift_window_friends_list_devider")
  local cell_height      = first_cell_place.layout:Get ("Height") + 0
  local cells_margin     = game.current_screen.vars.send_gift_window_friends_cells_margin
  local cell_interval    = vec3 (0, -cell_height - cells_margin, 0)
      
  table.sort (impl.send_gifts_friends_cells, function (cell1, cell2)
    local cell1_name = cell1.friend.first_name .. cell1.friend.last_name
    local cell2_name = cell2.friend.first_name .. cell2.friend.last_name
    
    return cell1_name < cell2_name
  end)      

  local scroll_range_x = 0
  local scroll_range_y = math.max (0, -#impl.send_gifts_friends_cells * cell_interval.y - table_zone_height)

  impl.send_gifts_friends_scroll_view.SetScrollRange (scroll_range_x, scroll_range_y)
  impl.send_gifts_friends_scroll_view.SetTableParams (cell_interval.y, table_zone_height)

  for i = 1, #impl.send_gifts_friends_cells do
    local cell = impl.send_gifts_friends_cells [i]
    
    cell.node.Position = (i - 0.5) * cell_interval
  end
    
  OnSendGiftsFriendsScrolled (impl)
end

local_methods.AddSendGiftsFriendsCell = function (impl, friend, dont_update_layout)
  local cell = {}
  local cell_nodes = {}      
  
  cell.friend = friend
  
  cell.node = Scene.Node.Create ()

  cell.node:BindToParent (impl.send_gifts_friends_cells_node)

  cell.underlay = Scene.Sprite.Create ()
  
  cell.underlay.Scale     = impl.send_gifts_friends_cell_data.underlay_scale
  cell.underlay.Material  = impl.send_gifts_friends_cell_data.underlay_material
  cell.underlay.TexSize   = impl.send_gifts_friends_cell_data.underlay_tex_size
  cell.underlay.TexOffset = impl.send_gifts_friends_cell_data.underlay_tex_offset
  
  cell.underlay:BindToParent (cell.node)
  
  table.insert (cell_nodes, cell.underlay)

  cell.name_text = Scene.TextLine.Create (game.font_library)

  cell.name_text.Position            = impl.send_gifts_friends_cell_data.name_position
  cell.name_text.Scale               = impl.send_gifts_friends_cell_data.name_size

  if GetCurrentLocaleIndex() ~= 1 or impl.phone_ui then
    cell.name_text.Font                = "data/fonts/MyriadPro-Bold.xfont"
  else
    cell.name_text.Font                = "data/fonts/MyriadPro-BoldCond.xfont"
  end

  cell.name_text.Color               = impl.send_gifts_friends_cell_data.name_color
  cell.name_text.HorizontalAlignment = Scene.TextLineAlignment.Left
  cell.name_text.VerticalAlignment   = Scene.TextLineAlignment.BaseLine
  cell.name_text.FontCreationParams  = game.font_creation_params
  cell.name_text.Material            = "text_material"

  cell.name_text.Text = friend.first_name .. " " .. friend.last_name

  if cell.name_text.TextLength > game.current_screen.vars.send_gift_window_friends_name_long_text then
    cell.name_text.Scale = vec3 (game.current_screen.vars.send_gift_window_friends_name_long_text_size, game.current_screen.vars.send_gift_window_friends_name_long_text_size, 1)
  end

  cell.name_text:BindToParent (cell.node)
  
  table.insert (cell_nodes, cell.name_text)      

  cell.on_click = function ()
    if not impl.send_gifts_update_layout_locked then
      PlaySound ('click')
    end
  
    local_methods.AddSendGiftsSelectedFriendsCell (impl, friend, impl.send_gifts_window_selected_friends_scissor)
  
    cell.node:Unbind ()
    
    for i = 1, #impl.send_gifts_friends_cells do
      if cell == impl.send_gifts_friends_cells [i] then
        table.remove (impl.send_gifts_friends_cells, i)
        break
      end
    end
  
    UpdateSendGiftsFriendsLayout (impl)
    
    if #impl.send_gifts_friends_cells == 0 then

      ActivateSprite("send_gift_window_check_all_friends")

    end
  end

  table.insert (impl.send_gifts_friends_cells, cell)

  for j = 1, #cell_nodes do 
    cell_nodes [j].Scissor = impl.send_gifts_friends_cell_data.cell_scissor
  end
  
  if not dont_update_layout then
    UpdateSendGiftsFriendsLayout (impl)
  end 
end

local function SelectAllFriends (impl) 
  if #impl.send_gifts_friends_cells == 0 and #impl.send_gifts_selected_friends_cells == 0 then return end

  impl.send_gifts_update_layout_locked = true

  if #impl.send_gifts_friends_cells == 0 then 
    for i = #impl.send_gifts_selected_friends_cells, 1, -1 do
      impl.send_gifts_selected_friends_cells [i].on_click ()
    end
  else
    for i = #impl.send_gifts_friends_cells, 1, -1 do
      impl.send_gifts_friends_cells [i].on_click ()
    end
  end

  impl.send_gifts_update_layout_locked = false

  UpdateSendGiftsFriendsLayout (impl)
  UpdateSendGiftsSelectedFriendsLayout (impl)
end

local function ShowSendGiftsFriendsWindow (impl, gift_type, friends_list)
  local additional_text_lines

  LoadSendGiftWindow (impl)

  impl.gift_type             = gift_type
  impl.friends_list          = {}
  impl.selected_friends_list = {}

  if gift_type == -1 then
    additional_text_lines = { "send_gift_window_title_line_1", "send_gift_window_send_d" }  

    GetTextLine ("send_gift_window_title_line_1").node.Text = GetLocalizedString ("INVITE_FRIENDS_WINDOW.TITLE_LINE_1")
    GetTextLine ("send_gift_window_send").node.Text = GetLocalizedString ("INVITE_FRIENDS_WINDOW.INVITE")
    GetTextLine ("send_gift_window_send_d").node.Text = GetLocalizedString ("INVITE_FRIENDS_WINDOW.INVITE")
    
    local game_friends   = GetFriends ()
    local social_friends = GetSocialManager ().GetFriends ()

    for i = 1, math.min (#social_friends, MAX_INVITE_FRIENDS) do
      local social_friend = social_friends [i]
      
      local already_in_game = false
      
      for j = 1, #game_friends do
        if social_friend.uid == game_friends [j].login_id then
          already_in_game = true
        end
      end
      
      if not already_in_game then
        table.insert (impl.friends_list, { first_name = social_friend.first_name, last_name = social_friend.last_name, login_id = social_friend.uid })
      end
    end
  else
    additional_text_lines = { "send_gift_window_title_line_1", "send_gift_window_title_line_2",
                              "send_gift_window_rules_line_1", "send_gift_window_rules_line_2", 
                              "send_gift_window_send_d" }
                              
    GetTextLine ("send_gift_window_title_line_1").node.Text = GetLocalizedString ("SEND_GIFTS_WINDOW.TITLE_LINE_1")
    GetTextLine ("send_gift_window_send").node.Text = GetLocalizedString ("SEND_GIFTS_WINDOW.SEND")
    GetTextLine ("send_gift_window_send_d").node.Text = GetLocalizedString ("SEND_GIFTS_WINDOW.SEND")
    
    local friends    = GetFriends ()
    local sent_gifts = GetSentGifts ()
    
    for i = 1, #friends do
      local friend = friends [i]
      
      local already_sent = false
      
      for j = 1, #sent_gifts do
        if friend.login_id == sent_gifts [j] then
          already_sent = true
        end
      end
      
      if not already_sent and friend.user_id ~= GetUserId () then
        table.insert (impl.friends_list, { first_name = friend.first_name, last_name = friend.last_name, login_id = friend.login_id })
      end
    end
  end
 
  ShowWindow (impl, impl.send_gift_friends_window, {}, additional_text_lines, {}, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil,
              { hide_on_back = true })
  
  impl.send_gifts_friends_cells_node = Scene.Node.Create ()
  impl.send_gifts_friends_cells      = {}
  
  impl.send_gifts_friends_cells_node:BindToParent (game.current_screen.node)

  local friends_table_zone        = GetSprite ("send_gift_window_friends_list_zone").layout
  local friends_table_zone_left   = friends_table_zone:Get ("Left") + 0
  local friends_table_zone_top    = friends_table_zone:Get ("Top") + 0
  local friends_table_zone_width  = friends_table_zone:Get ("Width") + 0
  local friends_table_zone_height = friends_table_zone:Get ("Height") + 0
  local friends_first_cell_place  = GetSprite ("send_gift_window_friends_list_devider")
  local friends_cell_width        = friends_first_cell_place.layout:Get ("Width") + 0
  local friends_cell_height       = friends_first_cell_place.layout:Get ("Height") + 0
  local friends_cells_margin      = game.current_screen.vars.send_gift_window_friends_cells_margin
  local friends_cell_interval     = vec3 (0, -friends_cell_height - friends_cells_margin, 0)

  impl.send_gifts_friends_cells_node.Position = vec3 (friends_first_cell_place.layout:Get ("Left") + friends_cell_width / 2, -friends_table_zone_top + friends_cell_interval.y / 2, friends_first_cell_place.z) 

  impl.send_gifts_friends_scroll_view = CreateScrollView (impl.send_gifts_friends_cells_node, false, true, 0, 0, "send_gift_window_friends_list_zone", 
                                                          { vertical = { sprite = "send_gift_window_friends_list_roll_slide_l", pressed_sprite = "send_gift_window_friends_list_roll_slide_p", zone = "send_gift_window_friends_list_zone" } } )
  impl.send_gifts_friends_scroll_view.SetScrollCallback (function () OnSendGiftsFriendsScrolled (impl) end)
  impl.send_gifts_friends_scroll_view.SetTableDelegate ({
    OnCellClick = function (cell_index)
      impl.send_gifts_friends_cells [cell_index].on_click () 
    end
  })
  
  impl.send_gifts_selected_friends_cells_node = Scene.Node.Create ()
  impl.send_gifts_selected_friends_cells      = {}
  
  impl.send_gifts_selected_friends_cells_node:BindToParent (game.current_screen.node)

  local selected_friends_table_zone        = GetSprite ("send_gift_window_selected_friends_list_zone").layout
  local selected_friends_table_zone_left   = selected_friends_table_zone:Get ("Left") + 0
  local selected_friends_table_zone_top    = selected_friends_table_zone:Get ("Top") + 0
  local selected_friends_table_zone_width  = selected_friends_table_zone:Get ("Width") + 0
  local selected_friends_table_zone_height = selected_friends_table_zone:Get ("Height") + 0
  local selected_friends_first_cell_place  = GetSprite ("send_gift_window_selected_friends_list_devider")
  local selected_friends_cell_width        = selected_friends_first_cell_place.layout:Get ("Width") + 0
  local selected_friends_cell_height       = selected_friends_first_cell_place.layout:Get ("Height") + 0
  local selected_friends_cells_margin      = game.current_screen.vars.send_gift_window_selected_friends_cells_margin
  local selected_friends_cell_interval     = vec3 (0, -selected_friends_cell_height - selected_friends_cells_margin, 0)

  impl.send_gifts_selected_friends_cells_node.Position = vec3 (selected_friends_first_cell_place.layout:Get ("Left") + selected_friends_cell_width / 2, -selected_friends_table_zone_top + selected_friends_cell_interval.y / 2, selected_friends_first_cell_place.z) 

  impl.send_gifts_selected_friends_scroll_view = CreateScrollView (impl.send_gifts_selected_friends_cells_node, false, true, 0, 0, "send_gift_window_selected_friends_list_zone",
                                                                   { vertical = { sprite = "send_gift_window_selected_friends_list_roll_slide_l", pressed_sprite = "send_gift_window_selected_friends_list_roll_slide_p", zone = "send_gift_window_selected_friends_list_zone" } })
  impl.send_gifts_selected_friends_scroll_view.SetScrollCallback (function () OnSendGiftsSelectedFriendsScrolled (impl) end)

  impl.send_gifts_friends_cell_data = {}
  
  impl.send_gifts_window_selected_friends_scissor = CreateScissor (selected_friends_table_zone_left, -selected_friends_table_zone_top, selected_friends_table_zone_width, selected_friends_table_zone_height)
  
  impl.send_gifts_window_selected_friends_scissor:BindToParent (game.current_screen.node)
  
  impl.send_gifts_friends_cell_data.cell_scissor = CreateScissor (friends_table_zone_left, -friends_table_zone_top, friends_table_zone_width, friends_table_zone_height)
  
  impl.send_gifts_friends_cell_data.cell_scissor:BindToParent (game.current_screen.node)
  
  impl.send_gifts_friends_cell_data.underlay_scale      = vec3 (friends_cell_width, friends_cell_height, 1)
  impl.send_gifts_friends_cell_data.underlay_material   = friends_first_cell_place.node.Material
  impl.send_gifts_friends_cell_data.underlay_tex_size   = StringToVec (friends_first_cell_place.layout:Get ("TexSize"))
  impl.send_gifts_friends_cell_data.underlay_tex_offset = StringToVec (friends_first_cell_place.layout:Get ("TexOffset"))
  impl.send_gifts_friends_cell_data.name_position       = GetSpriteOffset (friends_first_cell_place.layout, GetSprite ("send_gift_window_friends_name_text").layout) + vec3 (0, 0, -EPS)
  impl.send_gifts_friends_cell_data.name_size           = vec3 (game.current_screen.vars.send_gift_window_friends_name_text_size, game.current_screen.vars.send_gift_window_friends_name_text_size, 1)
  impl.send_gifts_friends_cell_data.name_color          = vec4 (0.88, 0.86, 0.627, 1)

  for i = 1, #impl.friends_list do
    local_methods.AddSendGiftsFriendsCell (impl, impl.friends_list [i], true)
  end
  
  UpdateSendGiftsFriendsLayout (impl)
  UpdateSendGiftsSelectedFriendsLayout (impl)
end

local function ShowFreeSpinsEndedWindow (impl)
  LoadInfoWindow (impl)
  
  --Do not show this window if user can't send gifts right now
  local has_friends_without_gifts = false
  
  local friends    = GetFriends ()
  local sent_gifts = GetSentGifts ()
    
  for i = 1, #friends do
    local friend = friends [i]
      
    local already_sent = false
      
    for j = 1, #sent_gifts do
      if friend.login_id == sent_gifts [j] then
        already_sent = true
      end
    end
      
    if not already_sent and friend.user_id ~= GetUserId () then
      has_friends_without_gifts = true
      break
    end
  end

  if not has_friends_without_gifts then return end

  local additional_sprites    = { "info_window_one_button", "info_window_close_button", "info_window_share_free_spins_overlay1", "info_window_share_free_spins_overlay2", "info_window_share_free_spins_overlay3", 'info_window_screen_lock'}
  local additional_text_lines = { "info_window_share_free_spins_title_line_1", "info_window_share_free_spins_title_line_2", "info_window_share_free_spins_text_line_1" }

  local additional_resource_bindings = {}
  
  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_share_free_spins_layers.xmtl"))

  GetTextLine ("info_window_one_button").node.Text = GetLocalizedString ("INFO_WINDOW.SHARE_FREE_SPINS_BUTTON")

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, function () 
      if GetSocialManager ().IsInGuestMode () then
        game.scripts.ui_windows.ShowLockedFunctionalityWindow (7)
        return
      end
      
      game.scripts.ui_windows.ShowSendGiftsFriendsWindow (1)
    end, nil, additional_resource_bindings, nil, nil, nil, { hide_on_back = true })
end

local_methods.ShowBonusByLevelWindow = function  (impl, level, percent)
  LoadInfoWindow (impl)

  local additional_sprites    = { "info_window_one_button", "info_window_close_button", "info_window_bonus_by_level_girl", "info_window_bonus_by_level_present", }
  local additional_text_lines = { "info_window_bonus_by_level_title", "info_window_bonus_by_level_button", "info_window_bonus_by_level_text_line_1", "info_window_bonus_by_level_text_line_2", 
                                   "info_window_bonus_by_level_text_line_3", "info_window_bonus_by_level_text_line_4" }

  local additional_resource_bindings = {}
  
  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_bonus_by_level_layers.xmtl"))
  
  local replacements = {
      { token = "{level}",  replacement = tostring (level), color_factor = vec4 (1, 1, 1, 1) },
      { token = "{bonus}",  replacement = percent .. "%%", color_factor = vec4 (126 / 255, 162 / 255, 14 / 255, 1) }}

  for i = 1, 2 do
    GetTextLine ("info_window_bonus_by_level_text_line_" .. i).node.Text = ProcessStringWithReplacements (GetLocalizedString("INFO_WINDOW.BONUS_BY_LEVEL_TEXT_LINE_" .. i), replacements)
  end
    
  for i = 3, 4 do
    SetTextLineTextWithReplacements (GetTextLine ("info_window_bonus_by_level_text_line_" .. i).node, GetLocalizedString("INFO_WINDOW.BONUS_BY_LEVEL_TEXT_LINE_" .. i), replacements, vec4 (146 / 255, 135 / 255, 153 / 255, 1))
  end
    
  GetTextLine ("info_window_bonus_by_level_button").node.Text = GetLocalizedString ("INFO_WINDOW.BUY_CHIPS")
  
  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, true, true, function ()
    game.scripts.top_ui.BuyGold()
  end, nil, additional_resource_bindings, true, nil, nil, { hide_on_back = true })
end

local function ShowBonusEndedWindow (impl, bonus_game_name, icon_material, icon_tex_size, icon_tex_offset, icon_scale, hs_path)
  LoadInfoWindow (impl)
  
  impl.info_window_share_bonus_check = true
  
  local additional_sprites    = { "info_window_one_button", "info_window_share_bonus_overlay1", "info_window_share_bonus_overlay2", "info_window_share_bonus_icon",
                                   "info_window_share_bonus_check", "info_window_share_bonus_check_zone", "info_window_share_bonus_check_bg", 'info_window_screen_lock' }
  local additional_text_lines = { "info_window_share_bonus_title_line_1", "info_window_share_bonus_title_line_2", "info_window_share_bonus_text_line_1"}

	if impl.phone_ui then
		table.insert (additional_text_lines, "info_window_share_bonus_title_line_3")
		GetTextLine ("info_window_share_bonus_title_line_2").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.SHARE_BONUS_TITLE_LINE_PHONE_1"), bonus_game_name)
		GetTextLine ("info_window_share_bonus_title_line_3").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.SHARE_BONUS_TITLE_LINE_PHONE_2"), bonus_game_name)
	else
		GetTextLine ("info_window_share_bonus_title_line_2").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.SHARE_BONUS_TITLE_LINE2"), bonus_game_name)
	end

  GetTextLine ("info_window_one_button").node.Text = GetLocalizedString ("CLOSE")
  
  local share_values = GetSocialManager ().GetShareValues (SHARE_BONUS_GAME_TYPE)
  local text         = ProcessStringWithReplacements (GetLocalizedString ("INFO_WINDOW.SHARE_FRIENDS"), {
    { token = "{friends}", replacement = share_values.players_count },
    { token = "{coins}",   replacement = share_values.chips_count },
  })

  local info_window_text_line_1    = GetTextLine ("info_window_share_bonus_text_line_1")
  info_window_text_line_1.node.Text  = text

  local icon      = GetSprite ("info_window_share_bonus_icon")
  local icon_zone = GetSprite ("info_window_share_bonus_icon_zone")
  local left   = icon_zone.layout:Get ("Left") + 0
  local top    = icon_zone.layout:Get ("Top") + 0
  local width  = icon_zone.layout:Get ("Width") + 0
  local height = icon_zone.layout:Get ("Height") + 0

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_share_bonus_layers.xmtl"))

  icon.node.Material  = icon_material
  icon.node.TexSize   = icon_tex_size
  icon.node.TexOffset = icon_tex_offset
  
  icon_scale = GetAspectFitSize (icon_scale.x, icon_scale.y, width, height)
  
  icon.node.WorldScale = vec3 (icon_scale.x, icon_scale.y, 1)  

  icon.tracks ["size"] = nil
  icon.tracks ["texsize"] = nil
  icon.tracks ["texoffset"] = nil
  
  SetSpritePosition (icon, left + width / 2, -top - height / 2)  
  
  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, function () 
    if impl.info_window_share_bonus_check then
      GetSocialManager ().ShareBonusEnded (bonus_game_name, ApplicationDataGet (BONUS_GAME_MONEY), hs_path)
    else
      local_methods.ProcessWindowQueue (impl)
    end
  end, nil, additional_resource_bindings, nil, nil, nil, { dont_process_queue_on_ok = true })
end

local_methods.InfoBonusShare = function(impl)
  if not impl.info_window_share_bonus_check then
    impl.info_window_share_bonus_check = true
    ActivateSprite("info_window_share_bonus_check")
  else
    impl.info_window_share_bonus_check = nil
    DeactivateSprite("info_window_share_bonus_check")
  end
end

local function ShowFiveInRowWindow (impl, icon_material, icon_tex_size, icon_tex_offset, icon_scale, coins, hs_path, symbol_id)
  LoadInfoWindow (impl)
  
  impl.info_window_share_five_in_row_check = true
  
  local additional_sprites    = { "info_window_one_button", "info_window_share_five_in_row_overlay1", "info_window_share_five_in_row_icon",
                                  "info_window_share_five_in_row_overlay2", "info_window_share_five_in_row_overlay3",
                                  "info_window_share_five_in_row_check", "info_window_share_five_in_row_check_zone", "info_window_share_five_in_row_check_bg", 'info_window_screen_lock' }
  local additional_text_lines = { "info_window_share_five_in_row_title_line_1", "info_window_share_five_in_row_text_line_1", "info_window_share_five_in_row_text_line_2" }

  GetTextLine ("info_window_one_button").node.Text = GetLocalizedString ("CLOSE")

  local share_values = GetSocialManager ().GetShareValues (SHARE_BONUS_GAME_TYPE)
  local text         = ProcessStringWithReplacements (GetLocalizedString ("INFO_WINDOW.SHARE_FRIENDS"), {
    { token = "{friends}", replacement = share_values.players_count },
    { token = "{coins}",   replacement = share_values.chips_count },
  })

  local info_window_text_line_1    = GetTextLine ("info_window_share_five_in_row_text_line_2")
  info_window_text_line_1.node.Text  = text

  local icon      = GetSprite ("info_window_share_five_in_row_icon")
  local icon_zone = GetSprite ("info_window_share_five_in_row_icon_zone")
  local left   = icon_zone.layout:Get ("Left") + 0
  local top    = icon_zone.layout:Get ("Top") + 0
  local width  = icon_zone.layout:Get ("Width") + 0
  local height = icon_zone.layout:Get ("Height") + 0

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_share_five_in_row_layers.xmtl"))

  icon.node.Material  = icon_material
  icon.node.TexSize   = icon_tex_size
  icon.node.TexOffset = icon_tex_offset
  
  icon_scale = GetAspectFitSize (icon_scale.x, icon_scale.y, width, height)
  
  icon.node.WorldScale = vec3 (icon_scale.x, icon_scale.y, 1)  

  icon.tracks ["size"] = nil
  icon.tracks ["texsize"] = nil
  icon.tracks ["texoffset"] = nil
  
  SetSpritePosition (icon, left + width / 2, -top - height / 2)  
  
  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, function ()
    if impl.info_window_share_five_in_row_check then
      GetSocialManager ().ShareFiveInRow (coins, hs_path, symbol_id)
    else
      local_methods.ProcessWindowQueue (impl)
    end
  end, nil, additional_resource_bindings, nil, nil, nil, { dont_process_queue_on_ok = true })
end

local_methods.InfoFiveInRowShare = function(impl)
  if not impl.info_window_share_five_in_row_check then
    impl.info_window_share_five_in_row_check = true
    ActivateSprite("info_window_share_five_in_row_check")
  else
    impl.info_window_share_five_in_row_check = nil
    DeactivateSprite("info_window_share_five_in_row_check")
  end
end

local function ShowNewGameWindow (impl, slot_id)
  LoadInfoWindow (impl)
  
  impl.new_game_slot_id = slot_id 

  local additional_sprites    = { "info_window_two_button_right", "info_window_close_button", "info_window_two_button_left", "info_window_new_game_overlay1", "info_window_new_game_overlay2", 'info_window_screen_lock' }
  local additional_text_lines = { "info_window_new_game_title_line_1", "info_window_new_game_title_line_2", "info_window_new_game_right_button",
                                  "info_window_new_game_left_button" }

  local slot_name          = ""
  local slot_path          = ""
  local slot_icon_material = "transparent"
  local slot_xmtl_path     = "data/res/materials/info_window_new_game_layers.xmtl"
  local slots              = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots do
    local slot_desc = slots [i]
  
    if slot_desc.id and slot_desc.id + 0 == slot_id + 0 then
      slot_name          = slot_desc.game_name
      slot_icon_material = slot_desc.icon_material
      slot_path          = slot_desc.path
      slot_xmtl_path     = slot_desc.xmtl_path
      break
    end
  end

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource (slot_xmtl_path))
  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/info_window_new_game_layers.xmtl"))

	if impl.phone_ui then
		table.insert(additional_text_lines,"info_window_new_game_title_line_3")
		GetTextLine ("info_window_new_game_title_line_2").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.NEW_GAME_TITLE_LINE_PHONE_1"), slot_name)
		GetTextLine ("info_window_new_game_title_line_3").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.NEW_GAME_TITLE_LINE_PHONE_2"), slot_name)
	else
		GetTextLine ("info_window_new_game_title_line_2").node.Text = string.format (GetLocalizedString ("INFO_WINDOW.NEW_GAME_TITLE_LINE_2"), slot_name)
	end

	if slot_xmtl_path and Common.File.IsFileExist (slot_xmtl_path) then
    local icon = GetSprite ("info_window_new_game_icon_zone")
    
    icon.node.Material  = slot_icon_material
    icon.node.TexSize   = vec2 (1, 1)
    icon.node.TexOffset = vec2 (0, 0)

    icon.tracks ["texsize"] = nil
    icon.tracks ["texoffset"] = nil

    for i = 1, #impl.info_window.all_sprites do
      local item = impl.info_window.all_sprites [i]
      
      if item.name == "info_window_new_game_icon_zone" then
        item.material = slot_icon_material
      end
    end

    table.insert (additional_sprites, "info_window_new_game_icon_zone")
  end

  ShowWindow (impl, impl.info_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, function ()
      LaunchSlot (slot_path, slot_id)
    end, nil, additional_resource_bindings, true, nil, nil, { hide_on_back = true })
end

local function ShowLevelUpWindow (impl)
  LoadNewLevelWindow (impl)

  impl.new_level_window_check = true
    
  local additional_text_lines = {}

  local user_level = GetUserLevel ()

  local bet_money = GetBetMoney (user_level.max_bet_id)
  
  local share_values = GetSocialManager ().GetShareValues (SHARE_BONUS_GAME_TYPE)

  local additional_sprites = {}
  
  if user_level.bonus.coins > 990000 then
    table.insert (additional_sprites, "new_level_window_long_plates")

    GetTextLine ("new_level_window_bonus_value_diamonds_long").node.Text = tostring (user_level.bonus.diamonds)
    GetTextLine ("new_level_window_bonus_value_long").node.Text          = FormatMoneyString (user_level.bonus.coins, 17, nil, true)
    
    table.insert (additional_text_lines, "new_level_window_bonus_value_diamonds_long")
    table.insert (additional_text_lines, "new_level_window_bonus_value_long")
  else
    table.insert (additional_sprites, "new_level_window_short_plates")

    GetTextLine ("new_level_window_bonus_value_diamonds").node.Text = tostring (user_level.bonus.diamonds)
    GetTextLine ("new_level_window_bonus_value").node.Text          = FormatMoneyString (user_level.bonus.coins, 6, nil, true)
    
    table.insert (additional_text_lines, "new_level_window_bonus_value_diamonds")
    table.insert (additional_text_lines, "new_level_window_bonus_value")
  end

  ShowWindow (impl, impl.level_up_window, additional_sprites, additional_text_lines, {}, function ()
    GetTextLine ("new_level_window_share_friends").node.Text = ProcessStringWithReplacements (GetLocalizedString ("INFO_WINDOW.SHARE_FRIENDS"), {
      { token = "{friends}", replacement = share_values.players_count },
      { token = "{coins}",   replacement = share_values.chips_count },
    })

    GetTextLine ("new_level_window_level_value").node.Text          = tostring (user_level.id)
    GetTextLine ("new_level_window_max_bet_value").node.Text        = FormatMoneyString (bet_money, 10, nil, true)
    GetTextLine ("new_level_window_super_bonus_value").node.Text    = FormatMoneyString (user_level.super_bonus.coins, 10, nil, true)
    GetTextLine ("new_level_window_mega_bonus_value").node.Text     = FormatMoneyString (user_level.mega_bonus.coins, 10, nil, true)
    GetTextLine ("new_level_window_bonus_value_diamonds").node.Text = tostring (user_level.bonus.diamonds)
    GetTextLine ("new_level_window_bonus_value").node.Text          = FormatMoneyString (user_level.bonus.coins, 7, nil, true)
  end, nil, nil, nil, function ()
    if impl.new_level_window_check then 
      GetSocialManager ().ShareNewLevel (tostring (user_level.id), user_level.bonus.coins) 
    else
      local_methods.ProcessWindowQueue (impl)
    end
  end, nil, nil, nil, nil, nil, { dont_process_queue_on_ok = true })

  local user_level = GetUserLevel ()
  
  local level_bonuses = GetPaymentsLevelBonuses ()
  
  for i = 1, #level_bonuses do
    local bonus = level_bonuses [i]

    if bonus.bonus > 0 and bonus.level_from == user_level.id + 0 then
      game.scripts.ui_windows.ShowBonusByLevelWindow (bonus.level_from, bonus.bonus)
    end
  end
  
  if GetSocialManager ().IsInGuestMode () then
    if (user_level.id - 3) % 5 == 0 then
      local_methods.ShowInfoSocialConnectingWindow (impl)
    end
  end 
  
  local slots = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots do
    local slot_desc = slots [i]

    if slot_desc.id and slot_desc.level and slot_desc.level + 0 == user_level.id + 0 and not slot_desc.coming_soon then
      game.scripts.ui_windows.ShowNewGameWindow (slot_desc.id)
      return
    end
  end
end

local_methods.ShowNinetyNineLinesWindow = function (impl)
  LoadNinetyNineLinesWindow (impl)
  ShowWindow (impl, impl.ninety_nine_lines, {}, {}, {})
end

local_methods.NewLevelWindowShare = function(impl)
  if not impl.new_level_window_check then
    impl.new_level_window_check = true
    ActivateSprite("new_level_window_check")
  else
    impl.new_level_window_check = nil
    DeactivateSprite("new_level_window_check")
  end
end

local function ShowUpdateApplicationWindow (impl, cancellable)
  local ok_action     = nil
  local ok_text       = nil
  local cancel_action = nil
  
  if cancellable then
    cancel_action = function () end
  end
  
  local os = ApplicationDataGet (OS_TYPE)

  local update_url
  
  if os == "iOS" then
    if GetPlatform () == "pad" then
      update_url = "https://itunes.apple.com/app/playslots-hd/id586574454?l=en&mt=8"
    else
      update_url = "https://itunes.apple.com/app/playslots/id543831789?l=en&mt=8"
    end
  elseif os == "Android" then
    update_url = "market://details?id=com.plarium.playslots"
  end

  if update_url then
    ok_text   = GetLocalizedString ("UPDATE")
    ok_action = function ()
      System.Application.OpenUrl (update_url)
      SetCurrentScreen ("main_menu")
    end
  end
  
  game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_QUESTION, ok_action, cancel_action, ok_text, { GetLocalizedString ("NEW_APPLICATION_VERSION_AVAILABLE_LINE1"), GetLocalizedString ("NEW_APPLICATION_VERSION_AVAILABLE_LINE2"), GetLocalizedString ("NEW_APPLICATION_VERSION_AVAILABLE_LINE3"), GetLocalizedString ("NEW_APPLICATION_VERSION_AVAILABLE_LINE4"), GetLocalizedString ("NEW_APPLICATION_VERSION_AVAILABLE_LINE5") })
end

local function CloseBattleSearchWindow (impl)
  HsConnectionSendMessage (1, { event = "battle.interrupt" })

  HideWindow (impl)
end

local function BattleSearchWindowDeleteCounter (impl)
  if impl.battle_search_window.time_update_action then
    impl.battle_search_window.time_update_action.Stop ()
    impl.battle_search_window.time_update_action = nil
  end
  
  if impl.battle_search_window.digit_1_animation then
    impl.battle_search_window.digit_1_animation.Hide ()
    impl.battle_search_window.digit_1_animation = nil
  end
  
  if impl.battle_search_window.digit_2_animation then
    impl.battle_search_window.digit_2_animation.Hide ()
    impl.battle_search_window.digit_2_animation = nil
  end
end

local function UpdateBattleSearchWindowTimeLeft (impl, animated)
  if not impl.battle_search_window.digit_1_animation then
    local frames_0_9 = CreateFramesNames ("0_9_%d.png", 1, 30)
    local frames_1_0 = CreateFramesNames ("1_0_%d.png", 1, 30)
    local frames_2_1 = CreateFramesNames ("2_1_%d.png", 1, 30)
    local frames_3_2 = CreateFramesNames ("3_2_%d.png", 1, 30)
    local frames_4_3 = CreateFramesNames ("4_3_%d.png", 1, 30)
    local frames_5_4 = CreateFramesNames ("5_4_%d.png", 1, 30)
    local frames_6_5 = CreateFramesNames ("6_5_%d.png", 1, 30)
    local frames_7_6 = CreateFramesNames ("7_6_%d.png", 1, 30)
    local frames_8_7 = CreateFramesNames ("8_7_%d.png", 1, 30)
    local frames_9_8 = CreateFramesNames ("9_8_%d.png", 1, 30)
  
    local counter_animation = CreateAnimationSpriteClass ({ stay_0 = { frames_0_9 [1] },
      stay_1 = { frames_1_0 [1] },
      stay_2 = { frames_2_1 [1] },
      stay_3 = { frames_3_2 [1] },
      stay_4 = { frames_4_3 [1] },
      stay_5 = { frames_5_4 [1] },
      stay_6 = { frames_6_5 [1] },
      stay_7 = { frames_7_6 [1] },
      stay_8 = { frames_8_7 [1] },
      stay_9 = { frames_9_8 [1] }, 
      animate_9 = frames_0_9,
      animate_0 = frames_1_0,
      animate_1 = frames_2_1,
      animate_2 = frames_3_2,
      animate_3 = frames_4_3,
      animate_4 = frames_5_4,
      animate_5 = frames_6_5,
      animate_6 = frames_7_6,
      animate_7 = frames_8_7,
      animate_8 = frames_9_8 }, frames_0_9 [1], 30)
    
    impl.battle_search_window.digit_1_animation = counter_animation.CreateInstance (nil, "ui")
    impl.battle_search_window.digit_2_animation = counter_animation.CreateInstance (nil, "ui")

    SetAnimationPosition (impl.battle_search_window.digit_1_animation, "battle_search_window_counter_pivot_1")
    SetAnimationPosition (impl.battle_search_window.digit_2_animation, "battle_search_window_counter_pivot_2")
  end

  local first_digit_changed = math.min (9, math.floor ((impl.battle_search_window.time_left + 1) / 10)) ~= math.min (9, math.floor ((impl.battle_search_window.time_left) / 10))
  local digit_1             = math.min (9, math.floor (impl.battle_search_window.time_left / 10))
  local digit_2             = math.floor (impl.battle_search_window.time_left % 10)

  if animated then
    if first_digit_changed then
      impl.battle_search_window.digit_1_animation.Show ("animate_" .. digit_1)
    else
      impl.battle_search_window.digit_1_animation.Show ("stay_" .. digit_1)
    end
  
    impl.battle_search_window.digit_2_animation.Show ("animate_" .. digit_2)
  else
    impl.battle_search_window.digit_1_animation.Show ("stay_" .. digit_1)
    impl.battle_search_window.digit_2_animation.Show ("stay_" .. digit_2)
  end
end

local function OnHideBattleInviteWindow (impl)
  if impl.invite_avatar then
    impl.invite_avatar.OnHide ()
    impl.invite_avatar.GetSprite ():Unbind ()
    impl.invite_avatar = nil
  end

  if impl.battle_invite_window.time_update_action then
    impl.battle_invite_window.time_update_action.Stop ()
    impl.battle_invite_window.time_update_action = nil
  end
    
  if impl.battle_invite_window.timer_sound_action then
    impl.battle_invite_window.timer_sound_action.Stop ()
    impl.battle_invite_window.timer_sound_action = nil
  end
  
  if impl.battle_invite_window.search_wait_finished_action then
    impl.battle_invite_window.search_wait_finished_action.Stop ()
    impl.battle_invite_window.search_wait_finished_action = nil
  end
  
  if impl.battle_invite_window.digit_1_animation then
    impl.battle_invite_window.digit_1_animation.Hide ()
    impl.battle_invite_window.digit_1_animation = nil
  end
  
  if impl.battle_invite_window.digit_2_animation then
    impl.battle_invite_window.digit_2_animation.Hide ()
    impl.battle_invite_window.digit_2_animation = nil
  end  
  
  impl.battle_invite_window_shown = nil
end

local function BattleInviteAccept (impl) 
  if impl.battle_invite_window.accepted then return end

  impl.battle_invite_window.accepted = true

  HideWindow (impl)

  local battles = GetBattles ()
  local slot_id
  local slot_path

  for i = 1, #battles do
    local battle = battles [i]
    local modes  = battle.modes
    
    for j = 1, #modes do
      if modes [j].id == impl.battle_invite_window.mode_id then
        slot_id = battle.slot_id
        break
      end
    end
    
    if slot_id then
      break
    end
  end

  local slots = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots do
    local slot_desc = slots [i]

    if slot_desc.id and slot_desc.id == slot_id then
      slot_path          = slot_desc.path
      break
    end
  end

  game.scripts.ui_windows.ShowBattleSearchWindow (impl.battle_invite_window.mode_id, slot_id, slot_path, impl.battle_invite_window.user_id)
end

local_methods.BattleInviteWindowCancelHide = function (impl)
  if impl.battle_invite_window_shown_while_battle_results and impl.battle_result_window_data then
    if impl.battle_result_window_data.type == "win" then
      game.scripts.ui_windows.ShowBattleWinWindow (impl.battle_result_window_data.coins, impl.battle_result_window_data.rating, impl.battle_result_window_data.interrupted)
    elseif impl.battle_result_window_data.type == "draw" then
      game.scripts.ui_windows.ShowBattleDrawWindow (impl.battle_result_window_data.coins)
    else
      game.scripts.ui_windows.ShowBattleLoseWindow ()
    end
  else
    HideWindow (impl)

    if game.current_screen.name == "lobby" then
      game.scripts.lobby.ShowBattlesWindow ()
    else
      if impl.dont_return_to_lobby then return end
      
      SetCurrentScreen ("lobby")
    end
  end
end

local function BattleInviteCancel (impl) 
  if impl.battle_invite_window.accepted then return end

  HsConnectionSendMessage (1, { event = "battle.cancelinvite", userId = impl.battle_invite_window.user_id })
  
  local_methods.BattleInviteWindowCancelHide (impl)
end

local function UpdateBattleInviteWindowTimeLeft (impl, animated)
  if not impl.battle_invite_window.digit_1_animation then
    local frames_0_9 = CreateFramesNames ("invite_0_9_%d.png", 1, 30)
    local frames_1_0 = CreateFramesNames ("invite_1_0_%d.png", 1, 30)
    local frames_2_1 = CreateFramesNames ("invite_2_1_%d.png", 1, 30)
    local frames_3_2 = CreateFramesNames ("invite_3_2_%d.png", 1, 30)
    local frames_4_3 = CreateFramesNames ("invite_4_3_%d.png", 1, 30)
    local frames_5_4 = CreateFramesNames ("invite_5_4_%d.png", 1, 30)
    local frames_6_5 = CreateFramesNames ("invite_6_5_%d.png", 1, 30)
    local frames_7_6 = CreateFramesNames ("invite_7_6_%d.png", 1, 30)
    local frames_8_7 = CreateFramesNames ("invite_8_7_%d.png", 1, 30)
    local frames_9_8 = CreateFramesNames ("invite_9_8_%d.png", 1, 30)
  
    local counter_animation = CreateAnimationSpriteClass ({ stay_0 = { frames_0_9 [1] },
      stay_1 = { frames_1_0 [1] },
      stay_2 = { frames_2_1 [1] },
      stay_3 = { frames_3_2 [1] },
      stay_4 = { frames_4_3 [1] },
      stay_5 = { frames_5_4 [1] },
      stay_6 = { frames_6_5 [1] },
      stay_7 = { frames_7_6 [1] },
      stay_8 = { frames_8_7 [1] },
      stay_9 = { frames_9_8 [1] }, 
      animate_9 = frames_0_9,
      animate_0 = frames_1_0,
      animate_1 = frames_2_1,
      animate_2 = frames_3_2,
      animate_3 = frames_4_3,
      animate_4 = frames_5_4,
      animate_5 = frames_6_5,
      animate_6 = frames_7_6,
      animate_7 = frames_8_7,
      animate_8 = frames_9_8 }, frames_0_9 [1], 30)
    
    impl.battle_invite_window.digit_1_animation = counter_animation.CreateInstance (nil, "ui")
    impl.battle_invite_window.digit_2_animation = counter_animation.CreateInstance (nil, "ui")

    SetAnimationPosition (impl.battle_invite_window.digit_1_animation, "battle_invite_window_counter_pivot_1")
    SetAnimationPosition (impl.battle_invite_window.digit_2_animation, "battle_invite_window_counter_pivot_2")
  end

  local first_digit_changed = math.min (9, math.floor ((impl.battle_invite_window.time_left + 1) / 10)) ~= math.min (9, math.floor ((impl.battle_invite_window.time_left) / 10))
  local digit_1             = math.min (9, math.floor (impl.battle_invite_window.time_left / 10))
  local digit_2             = math.floor (impl.battle_invite_window.time_left % 10)

  if animated then
    if first_digit_changed then
      impl.battle_invite_window.digit_1_animation.Show ("animate_" .. digit_1)
    else
      impl.battle_invite_window.digit_1_animation.Show ("stay_" .. digit_1)
    end
  
    impl.battle_invite_window.digit_2_animation.Show ("animate_" .. digit_2)
  else
    impl.battle_invite_window.digit_1_animation.Show ("stay_" .. digit_1)
    impl.battle_invite_window.digit_2_animation.Show ("stay_" .. digit_2)
  end
end

local function ShowBattleInviteWindow (impl, mode_id, user_info)
  if impl.battle_search_window_shown and (impl.battle_search_window.time_left or 0) > 0 then return end
  
  impl.battle_invite_window_shown = true
  
  impl.battle_invite_window_shown_while_battle_results = false

  if #impl.windows_stack > 0 and (impl.windows_stack [#impl.windows_stack].window == impl.battle_result_window or 
    impl.windows_stack [#impl.windows_stack].window == impl.battle_search_window or (game.scripts.lobby and IsSpriteActive ("battles_window_background"))) then
    
    if impl.windows_stack [#impl.windows_stack].window == impl.battle_result_window then
      impl.battle_invite_window_shown_while_battle_results = true
    end
    
    HideWindow (impl, nil, true) 
  end

  if not impl.battle_invite_window then
    PrepareAnimationScreenPart ("data/animations/battle_invite_window.xml")

    local battle_invite_window_sprite_list = { "battle_invite_window_bg", "battle_invite_window_right_button_l", "battle_invite_window_right_button_p",
                                               "battle_invite_window_left_button_l", "battle_invite_window_left_button_p", "battle_invite_window_avatar",
                                               "battle_invite_window_lock", "battle_invite_window_overlay1", "battle_invite_window_overlay2", "battle_invite_window_game_icon_zone" }
    local battle_invite_window_text_line_list = { "battle_invite_window_level_text", "battle_invite_window_title", "battle_invite_window_name", "battle_invite_window_surname", "battle_invite_window_name_battle",
                                                  "battle_invite_window_win_text", "battle_invite_window_win_value", "battle_invite_window_lose_text", "battle_invite_window_lose_value",
                                                  "battle_invite_window_time_text", "battle_invite_window_bet", "battle_invite_window_entry", "battle_invite_window_prize",
                                                  "battle_invite_window_time_value", "battle_invite_window_bet_value", "battle_invite_window_entry_value", "battle_invite_window_prize_coin",
                                                  "battle_invite_window_total_battles", "battle_invite_window_left_button_text", "battle_invite_window_right_button_text" }
    impl.battle_invite_window = {}

    impl.battle_invite_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.battle_invite_window.resource       = "data/res/materials/battle_invite_window_layers.xmtl"
    impl.battle_invite_window.sprites        = { "battle_invite_window_bg", "battle_invite_window_right_button_l", "battle_invite_window_left_button_l",
                                                 "battle_invite_window_avatar", "battle_invite_window_lock", "battle_invite_window_overlay1", "battle_invite_window_overlay2" }
    impl.battle_invite_window.text_lines     = GetTextLines (battle_invite_window_text_line_list)
    impl.battle_invite_window.all_text_lines = GetTextLines (battle_invite_window_text_line_list)
    impl.battle_invite_window.all_sprites    = MakeAllSprites (battle_invite_window_sprite_list)
    impl.battle_invite_window.on_hide        = function () OnHideBattleInviteWindow (impl) end
  end

  OnHideBattleInviteWindow (impl)

  GetTextLine ("battle_invite_window_name_battle").node.Text = local_methods.GetBattleGameName (1)

  impl.battle_invite_window.accepted = nil
  impl.battle_invite_window.mode_id  = mode_id
  impl.battle_invite_window.user_id  = user_info.user_id

  local battles = GetBattles ()
  local mode_desc
  local slot_id
  
  for i = 1, #battles do
    local battle = battles [i]
    local modes  = battle.modes
    
    for j = 1, #modes do
      local mode = modes [j]

      if mode_id == modes [j].id then
        mode_desc = mode
        slot_id   = battle.slot_id
        break
      end
    end
    
    if mode_desc then
      break
    end
  end

  local slot_icon_material = "transparent"
  local slot_xmtl_path     = "data/res/materials/info_window_new_game_layers.xmtl"
  local slots              = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots do
    local slot_desc = slots [i]

    if slot_desc.id and slot_desc.id + 0 == slot_id + 0 then
      slot_icon_material = slot_desc.icon_material
      slot_xmtl_path     = slot_desc.xmtl_path
      break
    end
  end

  local additional_sprites           = {}
  local additional_resource_bindings = {}

    table.insert (additional_resource_bindings, LoadResource ("data/res/materials/battle_invite_window_animations_layers.xmtl"))

  if slot_xmtl_path and Common.File.IsFileExist (slot_xmtl_path) then
    table.insert (additional_resource_bindings, LoadResource (slot_xmtl_path))
  
    local icon = GetSprite ("battle_invite_window_game_icon_zone")
    
    icon.node.Material  = slot_icon_material
    icon.node.TexSize   = vec2 (1, 1)
    icon.node.TexOffset = vec2 (0, 0)
    
    icon.tracks ["texsize"] = nil
    icon.tracks ["texoffset"] = nil

    table.insert (additional_sprites, "battle_invite_window_game_icon_zone")
  end

  ShowWindow (impl, impl.battle_invite_window, additional_sprites, {}, {}, function ()
		if impl.phone_ui then
			GetTextLine ("battle_invite_window_name").node.Text           = user_info.first_name .. " " .. user_info.last_name
		else
			GetTextLine ("battle_invite_window_name").node.Text           = user_info.first_name
			GetTextLine ("battle_invite_window_surname").node.Text        = user_info.last_name
		end

    GetTextLine ("battle_invite_window_total_battles").node.Text  = tostring (user_info.rating)
    GetTextLine ("battle_invite_window_win_value").node.Text      = tostring (user_info.wins)
    GetTextLine ("battle_invite_window_lose_value").node.Text     = tostring (user_info.loses)
    GetTextLine ("battle_invite_window_level_text").node.Text     = tostring (user_info.level)
    GetTextLine ("battle_invite_window_name_battle").node.Text    = local_methods.GetBattleGameName (mode_id)
    GetTextLine ("battle_invite_window_time_value").node.Text     = mode_desc.time .. " " .. GetLocalizedString ("BATTLE_INVITE.TIME_SUFFIX")
    GetTextLine ("battle_invite_window_bet_value").node.Text      = FormatMoneyString (GetBetMoney (mode_desc.bet_id))
    GetTextLine ("battle_invite_window_entry_value").node.Text    = FormatMoneyString (mode_desc.coins_fee, game.current_screen.vars.battle_invite_window_coins_symbols_count)
    GetTextLine ("battle_invite_window_prize_coin").node.Text     = FormatMoneyString (mode_desc.coins_fee * 2, game.current_screen.vars.battle_invite_window_coins_symbols_count)

    local window        = Engine.Windows.Find (WINDOW_NAME)
    local camera        = impl.battle_invite_window.camera.node
    local camera_width  = camera.Right - camera.Left
    local camera_height = camera.Top - camera.Bottom
    local screen_scale  = window.ClientHeight / camera_height

    local avatar_ref  = GetSprite ("battle_invite_window_avatar")
    local avatar_size = { x = avatar_ref.layout:Get ("Width") + 0, y = avatar_ref.layout:Get ("Height") + 0 }
    local avatar      = CreateUrlSprite (URL_SPRITE_MODE_CACHE_FILE, user_info.photo, {}, 
                                         { material = avatar_ref.node.Material, tex_size = StringToVec (avatar_ref.layout:Get ("TexSize")), tex_offset = StringToVec (avatar_ref.layout:Get ("TexOffset")) }, 
                                         avatar_size, avatar_size.x * PHOTO_ACTIVITY_INDICATOR_FACTOR, screen_scale)

    local avatar_sprite = avatar.GetSprite ()

    avatar_sprite.Position = vec3 (avatar_ref.layout:Get ("Left") + avatar_size.x / 2, -avatar_ref.layout:Get ("Top") - avatar_size.y / 2, avatar_ref.z - EPS)

    avatar_sprite:BindToParent (game.current_screen.node)

    avatar.OnShow ()

    impl.invite_avatar = avatar
  
    impl.battle_invite_window.time_left = GetBattlesTimeout ()

    UpdateBattleInviteWindowTimeLeft (impl, false)

    impl.battle_invite_window.time_update_action = StartPeriodicCall ("ui", 1, function ()
      impl.battle_invite_window.time_left = impl.battle_invite_window.time_left - 1

      impl.battle_invite_window.timer_sound_action = DelayedAction ("ui", 0.3, function ()
        PlaySound ("battle_opponent_waiting")
      end)

      UpdateBattleInviteWindowTimeLeft (impl, true)

      if impl.battle_invite_window.time_left <= 0 then
        if impl.battle_invite_window.time_update_action then
          impl.battle_invite_window.time_update_action.Stop ()
          impl.battle_invite_window.time_update_action = nil
        end

        impl.battle_invite_window.search_wait_finished_action = DelayedAction ("ui", 1, function ()
          HsConnectionSendMessage (1, { event = "battle.interrupt" })
  
          local_methods.BattleInviteWindowCancelHide (impl)
        end)
      end
    end)
  end, nil, nil, nil, nil, nil, additional_resource_bindings, true)
end

local function BattleSearchWindowSearchAgain (impl, opponent_id)
  HsConnectionSendMessage (1, { event = "battle.search", modeId = impl.battle_search_window.mode_id, userId = opponent_id })

  impl.battle_search_window.time_left = GetBattlesTimeout ()

  DeactivateSprite ("battle_search_window_again_l")
  DeactivateTextLine ("battle_search_window_again")
  
  DeactivateTextLine ("battle_search_window_not_found")  
  ActivateTextLine ("battle_search_window_title")
  
  UpdateBattleSearchWindowTimeLeft (impl, false)

  impl.battle_search_window.time_update_action = StartPeriodicCall ("ui", 1, function ()
    impl.battle_search_window.time_left = impl.battle_search_window.time_left - 1
    
    impl.battle_search_window.timer_sound_action = DelayedAction ("default", 0.3, function ()
      PlaySound ("battle_opponent_waiting")
    end)

    UpdateBattleSearchWindowTimeLeft (impl, true)
    
    if impl.battle_search_window.time_left <= 0 then
      impl.battle_search_window.time_update_action.Stop ()
      impl.battle_search_window.time_update_action = nil
      
      impl.battle_search_window.search_wait_finished_action = DelayedAction ("ui", 1, function ()
        HsConnectionSendMessage (1, { event = "battle.interrupt" })

        BattleSearchWindowDeleteCounter (impl)

        if opponent_id then
          ActivateTextLine ("battle_search_window_dont_respond")
        else
          ActivateSprite ("battle_search_window_again_l")  
          ActivateTextLine ("battle_search_window_again")
          
          DeactivateTextLine ("battle_search_window_title") 
          ActivateTextLine ("battle_search_window_not_found") 
          
          PlaySound ("battle_wait_opponent_repeat_button_show")  
        end
      end)
    end
  end)
end

local function OnHideBattleSearchWindow (impl)
  impl.battle_search_window_shown       = nil
  impl.battle_search_window.opponent_id = nil

  if impl.battle_search_window.time_update_action then
    impl.battle_search_window.time_update_action.Stop ()
    impl.battle_search_window.time_update_action = nil
  end
  
  if impl.battle_search_window.search_wait_finished_action then
    impl.battle_search_window.search_wait_finished_action.Stop ()
    impl.battle_search_window.search_wait_finished_action = nil
  end
  
  if impl.battle_search_window.timer_sound_action then
    impl.battle_search_window.timer_sound_action.Stop ()
    impl.battle_search_window.timer_sound_action = nil
  end

  BattleSearchWindowDeleteCounter (impl)
  
  if game.current_screen.name == "lobby" then
    game.scripts.lobby.ShowBattlesWindow ()
  else
    if not impl.battle_invite_window_shown and HsConnectionIsConnected () then
      SetCurrentScreen ("lobby")
    end
  end
end

local function ShowBattleSearchWindow (impl, mode_id, slot_id, slot_path, opponent_id)
  local battles = GetBattles ()
  
  for i = 1, #battles do
    local modes = battles [i].modes
    
    for j = 1, #modes do
      local mode = modes [j]

      if mode_id == modes [j].id then
        if mode.coins_fee > GetUserMoney () then
          game.scripts.ui_windows.ShowNoMoneyWindow ()
          return
        end

        break
      end
    end
  end

  if not impl.battle_search_window then
    PrepareAnimationScreenPart ("data/animations/battle_search_window.xml")

    local battle_search_window_sprite_list = { "battle_search_window_close_l", "battle_search_window_close_p", 
      "battle_search_window_background", "battle_search_window_again_l", "battle_search_window_again_p", "screen_lock" }

    impl.battle_search_window = {}

    impl.battle_search_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.battle_search_window.resource       = "data/res/materials/battle_search_window_layers.xmtl"
    impl.battle_search_window.sprites        = { "battle_search_window_background", "battle_search_window_close_l", "screen_lock" }
    impl.battle_search_window.text_lines     = GetTextLines ({'battle_search_window_title',
                                                              'battle_search_window_power_ups_1_text_1', 'battle_search_window_power_ups_1_text_2', 'battle_search_window_power_ups_1_text_3',
                                                              'battle_search_window_power_ups_2_text_1', 'battle_search_window_power_ups_2_text_2', 'battle_search_window_power_ups_2_text_3',
                                                              'battle_search_window_power_ups_3_text_1', 'battle_search_window_power_ups_3_text_2', 'battle_search_window_power_ups_3_text_3'})
    impl.battle_search_window.all_text_lines = GetTextLines ({ "battle_search_window_again", 'battle_search_window_not_found', 'battle_search_window_title',
                                                               'battle_search_window_power_ups_1_text_1', 'battle_search_window_power_ups_1_text_2', 'battle_search_window_power_ups_1_text_3',
                                                               'battle_search_window_power_ups_2_text_1', 'battle_search_window_power_ups_2_text_2', 'battle_search_window_power_ups_2_text_3',
                                                               'battle_search_window_power_ups_3_text_1', 'battle_search_window_power_ups_3_text_2', 'battle_search_window_power_ups_3_text_3',
                                                               "battle_search_window_dont_respond", "battle_search_window_rejected" })
    impl.battle_search_window.all_sprites    = MakeAllSprites (battle_search_window_sprite_list)
    impl.battle_search_window.on_hide        = function () OnHideBattleSearchWindow (impl) end
  end

  impl.battle_search_window_shown = true

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/battle_search_window_animations_layers.xmtl"))

	local add_text_lines = {}
	if impl.phone_ui then
		add_text_lines = {'battle_search_window_power_ups_1_text_4','battle_search_window_power_ups_2_text_4','battle_search_window_power_ups_3_text_4',}
	end

  ShowWindow (impl, impl.battle_search_window, {}, add_text_lines, {}, nil, nil, true, nil, nil, nil, additional_resource_bindings, nil, nil, nil, { on_back_handler = function () game.scripts.ui_windows.CloseBattleSearchWindow () end })

  local boosters          = GetPowerupsStore ()
  local turbo_spins_count = 1
  
  for i = 1, #boosters do
    if boosters [i].type == 1 then
      turbo_spins_count = boosters [i].value
      break
    end
  end

	if impl.phone_ui then
		GetTextLine ('battle_search_window_power_ups_3_text_2').node.Text = string.format (GetLocalizedString ("BATTLE_SEARCH.POWER_UPS_3_1_PHONE"), turbo_spins_count)
		GetTextLine ('battle_search_window_power_ups_3_text_3').node.Text = string.format (GetLocalizedString ("BATTLE_SEARCH.POWER_UPS_3_2_PHONE"), turbo_spins_count)
		GetTextLine ('battle_search_window_power_ups_3_text_4').node.Text = string.format (GetLocalizedString ("BATTLE_SEARCH.POWER_UPS_3_3_PHONE"), turbo_spins_count)
	else
		GetTextLine ('battle_search_window_power_ups_3_text_2').node.Text = string.format (GetLocalizedString ("BATTLE_SEARCH.POWER_UPS_3_1"), turbo_spins_count)
		GetTextLine ('battle_search_window_power_ups_3_text_3').node.Text = string.format (GetLocalizedString ("BATTLE_SEARCH.POWER_UPS_3_2"), turbo_spins_count)
	end

  impl.battle_search_window.mode_id     = mode_id
  impl.battle_search_window.slot_id     = slot_id
  impl.battle_search_window.slot_path   = slot_path
  impl.battle_search_window.opponent_id = opponent_id

  ApplicationDataSet (BATTLE_OPPONENT_LOADING_PROGRESS, nil)

  BattleSearchWindowSearchAgain (impl, opponent_id)
end

local function OnBattleInviteCanceled (impl, user_id)
  if user_id ~= impl.battle_search_window.opponent_id then return end
  if not impl.battle_search_window.time_update_action then return end --timeout already occured
  
  if impl.battle_search_window.search_wait_finished_action then
    impl.battle_search_window.search_wait_finished_action.Stop ()
    impl.battle_search_window.search_wait_finished_action = nil
  end
  
  BattleSearchWindowDeleteCounter (impl)

  ActivateTextLine ("battle_search_window_rejected")
end

local function OnBattleSearchSucceeded (impl, event) 
  if impl.battle_search_window.search_wait_finished_action then
    impl.battle_search_window.search_wait_finished_action.Stop ()
    impl.battle_search_window.search_wait_finished_action = nil
  end

  ApplicationDataSet (IS_BATTLE_ACTIVE, true)
  ApplicationDataSet (IS_BATTLE_LOADING_STARTED, nil)
  ApplicationDataSet (BATTLE_INIT_DATA, { mode_id = impl.battle_search_window.mode_id, slot_id = impl.battle_search_window.slot_id, slot_path = impl.battle_search_window.slot_path } )
  
  LaunchSlot (impl.battle_search_window.slot_path, impl.battle_search_window.slot_id)
end

local_methods.GetBattleGameName = function (mode_id)
  if not mode_id then
    local battle_init_data = ApplicationDataGet (BATTLE_INIT_DATA)
  
    if not battle_init_data then return "" end
    
    mode_id = battle_init_data.mode_id
  end

  local battles = GetBattles ()
  
  for i = 1, #battles do
    local battle = battles [i]
    local modes  = battle.modes
    
    for j = 1, #modes do
      local mode = modes [j]

      if mode_id == modes [j].id then
        local slot_name = ""
        local slots     = ApplicationDataGet (SLOTS_DESCS)

        for i = 1, #slots do
          local slot_desc = slots [i]

          if slot_desc.id and slot_desc.id + 0 == battle.slot_id then
            slot_name = slot_desc.game_name
            break
          end
        end
        
        return slot_name .. " " .. GetLocalizedString ("BATTLES_WINDOW.BATTLE_" .. j .. "_TITLE")
      end
    end
  end
  
  return ""
end

local function ShowBattleWinWindow (impl, coins, rating, interrupted)
  impl.battle_result_window_data = { type = "win", coins = coins, rating = rating, interrupted = interrupted }

  LoadBattleResultWindow (impl)

  local additional_sprites    = { "battle_result_window_one_button_pink_l",
    "battle_result_win_overlay_1", "battle_result_win_overlay_2", "battle_result_window_rating_popup_zone","battle_result_win_check", "battle_result_win_check_zone", "battle_result_win_check_bg" }
  local additional_text_lines = { "battle_result_window_one_button_text", "battle_result_win_text_1", 
    "battle_result_win_text_2", "battle_result_win_text_3", "battle_result_win_text_4", "battle_result_win_title", 
    "battle_result_win_coins_count", "battle_result_win_rating", "battle_result_win_title_interrupted"}

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/battle_result_win_layers.xmtl"))

  local title = GetLocalizedString ("BATTLE_RESULT.WIN_TITLE")

  if interrupted then
		if impl.phone_ui then
			title = GetLocalizedString ("BATTLE_RESULT.WIN_TITLE_INTERRUPTED_PHONE")
		else
			title = GetLocalizedString ("BATTLE_RESULT.WIN_TITLE_INTERRUPTED")
		end

  end

  PlaySound ("battle_win_window")

	if game.current_screen.vars.battle_result_win_check_offset_param then
		game.current_screen.vars.battle_result_win_check_offset = { x = game.current_screen.vars.battle_result_win_check_offset_param, y = 0 }
	end

  ShowWindow (impl, impl.battle_result_window, additional_sprites, additional_text_lines, {}, function ()
    impl.battle_result_win_check = true
    
    ExecuteOnNextUpdate (function ()
      ExecuteOnNextUpdate (function ()
        ActivateSprite ("battle_result_win_check_zone") -- this sprite not active if same window was closed before
      end)
    end)

		if interrupted then
			GetTextLine ("battle_result_win_title_interrupted").node.Text          = title
		else
			GetTextLine ("battle_result_win_title").node.Text          = title
		end

    GetTextLine ("battle_result_win_coins_count").node.Text    = FormatMoneyString (coins, 5)
    GetTextLine ("battle_result_win_rating").node.Text         = "+" .. rating
    GetTextLine ("battle_result_win_text_2").node.Text         = local_methods.GetBattleGameName ()
  end, nil, nil, nil, nil, nil, additional_resource_bindings, nil, nil, nil, { on_back_handler = function () game.scripts.ui_windows.CloseBattleResultWindow () end })
end

local function BattleWinWindowShare (impl)
  if not impl.battle_result_win_check then
    impl.battle_result_win_check = true
    ActivateSprite("battle_result_win_check")
  else
    impl.battle_result_win_check = nil
    DeactivateSprite("battle_result_win_check")
  end
end

local function ShowBattleLoseWindow (impl)
  impl.battle_result_window_data = { type = "lose" }

  LoadBattleResultWindow (impl)

  local additional_sprites    = { "battle_result_window_one_button_l", "battle_result_defeat_overlay_1" }
  local additional_text_lines = { "battle_result_defeat_one_button", "battle_result_defeat_title", "battle_result_defeat_text_1", "battle_result_defeat_text_2", "battle_result_defeat_text_3" }

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/battle_result_defeat_layers.xmtl"))

  PlaySound ("battle_lose_window")
  
  impl.battle_result_win_check = false
  
  ShowWindow (impl, impl.battle_result_window, additional_sprites, additional_text_lines, {}, nil, nil, nil, nil, nil, nil, 
              additional_resource_bindings, nil, nil, nil, { on_back_handler = function () game.scripts.ui_windows.CloseBattleResultWindow () end })
end

local function ShowBattleDrawWindow (impl, coins)

  LoadBattleResultWindow (impl)

  local additional_sprites    = {  "battle_result_window_one_button_l", "battle_result_draw_overlay_1"}
  local additional_text_lines = {  "battle_result_draw_one_button", "battle_result_draw_title", "battle_result_draw_coins_count", "battle_result_draw_rating", "battle_result_draw_text_1",
                                   "battle_result_draw_text_2", "battle_result_draw_text_3", "battle_result_draw_text_4", "battle_result_draw_text_1_1", "battle_result_draw_text_1_2" }

  local additional_resource_bindings = {}

  table.insert (additional_resource_bindings, LoadResource ("data/res/materials/battle_result_draw_layers.xmtl"))

  PlaySound ("battle_draw_window")

  ShowWindow (impl, impl.battle_result_window, additional_sprites, additional_text_lines, {}, function () 
    impl.battle_result_win_check = false
    
    GetTextLine ("battle_result_draw_coins_count").node.Text    = FormatMoneyString (coins,5)
    GetTextLine ("battle_result_draw_text_2").node.Text         = local_methods.GetBattleGameName ()
  end, nil, nil, nil, nil, nil, additional_resource_bindings, nil, nil, nil, { on_back_handler = function () game.scripts.ui_windows.CloseBattleResultWindow () end })
end

local function OnHideBattleStartCounterWindow (impl)
  if impl.battle_start_counter.hide_action then
    impl.battle_start_counter.hide_action.Stop ()
    impl.battle_start_counter.hide_action = nil
  end
end

local function ShowBattleStartCounter (impl, on_hide_handler)
  if not impl.battle_start_counter then
    PrepareAnimationScreenPart ("data/animations/battle_start_counter.xml")

    local battle_start_counter_sprite_list = { "battle_start_counter_count_1", "battle_start_counter_count_2", "battle_start_counter_count_3", "battle_start_counter_count_go" }

    impl.battle_start_counter = {}

    impl.battle_start_counter.camera         = game.current_screen.cameras.modal_window_camera
    impl.battle_start_counter.resource       = "data/res/materials/battle_start_counter_layers.xmtl"
    impl.battle_start_counter.sprites        = { "battle_start_counter_count_1", "battle_start_counter_count_2", "battle_start_counter_count_3", "battle_start_counter_count_go"}
    impl.battle_start_counter.text_lines     = {}
    impl.battle_start_counter.all_text_lines = {}
    impl.battle_start_counter.all_sprites    = MakeAllSprites (battle_start_counter_sprite_list)
    impl.battle_start_counter.on_hide        = function () 
        OnHideBattleStartCounterWindow (impl)
      
        if on_hide_handler then
          on_hide_handler () 
        end
      end
  end

  ShowWindow (impl, impl.battle_start_counter, {}, {}, {}, function()
    impl.battle_start_counter.hide_action = DelayedAction(game.main_timer, 3, function ()
      if game.scripts.ui_windows.IsAlertWindowShown (ALERT_TYPE_ERROR) then return end
    
      HideWindow (impl)
    end)
  end, nil, true)
end

local function WinnerOfBestPlayerCheck (impl)
  if not impl.winner_of_best_player_window.checked then
    impl.winner_of_best_player_window.checked = true
    ActivateSprite("winner_of_best_player_window_check")
  else
    impl.winner_of_best_player_window.checked = nil
    DeactivateSprite("winner_of_best_player_window_check")
  end
end

local_methods.CloseWinnerOfBestPlayerWindow = function (impl) 
  if impl.current_winner_of_best_player_window_prize_id then
    impl.winner_of_best_player_window_shown_prizes [impl.current_winner_of_best_player_window_prize_id] = nil

    if game.scripts.top_ui then
      game.scripts.top_ui.MoneyForceNextUpdate ()
    end

    HsConnectionSendMessage (1, { event = "prize.collect", id = impl.current_winner_of_best_player_window_prize_id })
    impl.current_winner_of_best_player_window_prize_id = nil
  end
  
  local share_checked = impl.winner_of_best_player_window.checked
  
  HideWindow (impl, nil, share_checked)
  
  if share_checked then
    GetSocialManager ().ShareLeaderWin (impl.current_winner_of_best_player_window_share_data.win_diamonds, impl.current_winner_of_best_player_window_share_data.place, impl.current_winner_of_best_player_window_share_data.rating_name)
  end
end

local function ShowWinnerOfBestPlayerWindow (impl, place, type, coins_or_diamonds, count_coins_or_diamonds, prize_id)
  impl.winner_of_best_player_window_shown_prizes = impl.winner_of_best_player_window_shown_prizes or {}
  
  if impl.winner_of_best_player_window_shown_prizes [prize_id] then return end
  
  if prize_id then
    impl.winner_of_best_player_window_shown_prizes [prize_id] = true
  end

  if not impl.winner_of_best_player_window then
    PrepareAnimationScreenPart ("data/animations/winner_of_best_player_window.xml")

    local winner_of_best_player_window_sprite_list = { "winner_of_best_player_window_bg", "winner_of_best_player_window_get_button_l", 
      "winner_of_best_player_window_get_button_p", "winner_of_best_player_window_screen_lock", "winner_of_best_player_window_check", 
      "winner_of_best_player_window_button_check_zone", "winner_of_best_player_window_type_overlay_1", "winner_of_best_player_window_type_overlay_2",
      "winner_of_best_player_window_overlay_1", "winner_of_best_player_window_overlay_2", "winner_of_best_player_window_overlay_3" }
    local winner_of_best_player_window_text_line_list = { "winner_of_best_player_window_text_title", "winner_of_best_player_window_text_button",
      "winner_of_best_player_window_text_win_1", "winner_of_best_player_window_text_win_2", 
      "winner_of_best_player_window_text_win_3_1","winner_of_best_player_window_text_win_3_2","winner_of_best_player_window_text_win_3_3",
      "winner_of_best_player_window_text_share_1", "winner_of_best_player_window_text_share_2", "winner_of_best_player_window_text_share_3"}
    impl.winner_of_best_player_window = {}

    impl.winner_of_best_player_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.winner_of_best_player_window.resource       = "data/res/materials/winner_of_best_player_window_layers.xmtl"
    impl.winner_of_best_player_window.sprites        = { "winner_of_best_player_window_bg", "winner_of_best_player_window_get_button_l", "winner_of_best_player_window_check",
                                                         "winner_of_best_player_window_button_check_zone", "winner_of_best_player_window_screen_lock"}
    impl.winner_of_best_player_window.text_lines     = GetTextLines (winner_of_best_player_window_text_line_list)
    impl.winner_of_best_player_window.all_text_lines = GetTextLines (winner_of_best_player_window_text_line_list)
    impl.winner_of_best_player_window.all_sprites    = MakeAllSprites (winner_of_best_player_window_sprite_list)
  end

  if game.scripts.lobby and IsSpriteActive ("hall_of_fame_background") and GetSprite ("hall_of_fame_background").node.Material ~= "transparent" then
    HideWindow (impl)
  end

  ShowWindow (impl, impl.winner_of_best_player_window, { "winner_of_best_player_window_overlay_" .. place, "winner_of_best_player_window_type_overlay_" .. coins_or_diamonds }, {}, {}, function()
    ExecuteOnNextUpdate (function ()
      ExecuteOnNextUpdate (function ()
        ActivateSprite ("winner_of_best_player_window_button_check_zone") -- this sprite not active if same window was closed before
      end)
    end)
  
    impl.current_winner_of_best_player_window_share_data = {
        win_diamonds = coins_or_diamonds ~= 1,
        place        = place,
        rating_name  = GetLocalizedString ("WINNER_OF_BEST_PLAYER.RATING_NAME_" .. type),
      }
    
    impl.current_winner_of_best_player_window_prize_id = prize_id
  
    for i = 1, 3 do
      if i ~= place then
        DeactivateSprite ("winner_of_best_player_window_overlay_" .. i)
      end
    end
  
    impl.winner_of_best_player_window.checked = true
    
    for i=1,3 do
      GetTextLine ("winner_of_best_player_window_text_win_3_" .. i).node.Text = ProcessStringWithReplacements (GetLocalizedString ("WINNER_OF_BEST_PLAYER.WIN_3_" .. i), {
        { token = "{rating_name}",   replacement = GetLocalizedString ("WINNER_OF_BEST_PLAYER.RATING_NAME_" .. type) },
      })
    end

    if coins_or_diamonds == 1 then
      local share_values = GetSocialManager ().GetShareValues (SHARE_LEADER_WIN_COINS_TYPE)

      GetTextLine ("winner_of_best_player_window_text_share_1").node.Text = GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_0") .. " " .. FormatMoneyString (count_coins_or_diamonds, 9, nil, true)
      GetTextLine ("winner_of_best_player_window_text_share_2").node.Text = GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_1_1")
      
      if not impl.phone_ui then
        GetTextLine ("winner_of_best_player_window_text_share_3").node.Text = ProcessStringWithReplacements (GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_2_1"), {
          { token = "{friends}", replacement = share_values.players_count },
          { token = "{coins}",   replacement = share_values.chips_count },
        })
      else 
        GetTextLine ("winner_of_best_player_window_text_share_3").node.Text = ProcessStringWithReplacements (GetLocalizedString ("WINNER_OF_BEST_PLAYER.PHONE_SHARE_2_1"), {
          { token = "{coins}",   replacement = share_values.chips_count }})
      end
    else
      local share_values = GetSocialManager ().GetShareValues (SHARE_LEADER_WIN_DIAMONDS_TYPE)

      GetTextLine ("winner_of_best_player_window_text_share_1").node.Text = GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_0") .. " " .. FormatMoneyString (count_coins_or_diamonds * 100, 6)
			GetTextLine ("winner_of_best_player_window_text_share_2").node.Text = GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_1_2")
      
      if not impl.phone_ui then
        GetTextLine ("winner_of_best_player_window_text_share_3").node.Text = ProcessStringWithReplacements (GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_2_2"), {
          { token = "{friends}", replacement = share_values.players_count },
          { token = "{diamonds}",   replacement = share_values.chips_count },
        })
      else
        GetTextLine ("winner_of_best_player_window_text_share_3").node.Text = ProcessStringWithReplacements (GetLocalizedString ("WINNER_OF_BEST_PLAYER.PHONE_SHARE_2_2"), {
          { token = "{diamonds}",   replacement = share_values.chips_count }})
      end
    end
    
    GetTextLine ("winner_of_best_player_window_text_share_1").node:SetCharsColorFactors (0, GetTextLength (GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_0")), vec4 (0.57,0.52,0.60, 1))

--		GetTextLine ("winner_of_best_player_window_text_share_1").node:SetCharsColorFactors (GetTextLength (GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_0") .. " " .. FormatMoneyString (count_coins_or_diamonds, 9, nil, true) .. " "), GetTextLength (GetLocalizedString ("WINNER_OF_BEST_PLAYER.SHARE_1_1")), vec4 (0.57,0.52,0.60, 1))
  end, nil, nil, nil, nil, nil, nil, true)
end

local_methods.ShowWelcomeWindow = function (impl, coins, diamonds)
  if not impl.welcome_window then
    PrepareAnimationScreenPart ("data/animations/welcome_window.xml")

    local welcome_window_sprite_list = { "welcome_window_bg", "welcome_window_screen_lock",
      "welcome_window_button_l", "welcome_window_button_p",
      "welcome_window_close_l", "welcome_window_close_p" }
    local welcome_window_text_line_list = { "welcome_window_text_title", "welcome_window_text_top_1", "welcome_window_text_top_2",
                                            "welcome_window_text_coins", "welcome_window_text_diamonds", "welcome_window_text_button",
                                            "welcome_window_text_bot_1", "welcome_window_text_bot_2" }
    impl.welcome_window = {}

    impl.welcome_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.welcome_window.resource       = "data/res/materials/welcome_window_layers.xmtl"
    impl.welcome_window.sprites        = welcome_window_sprite_list
    impl.welcome_window.text_lines     = GetTextLines (welcome_window_text_line_list)
    impl.welcome_window.all_text_lines = GetTextLines (welcome_window_text_line_list)
    impl.welcome_window.all_sprites    = MakeAllSprites (welcome_window_sprite_list)
  end

  ShowWindow (impl, impl.welcome_window, {}, {}, {}, function ()
    GetTextLine ('welcome_window_text_coins').node.Text = FormatMoneyString (coins, 4)
    GetTextLine ('welcome_window_text_diamonds').node.Text = tostring (diamonds)

    GetTextLine ("welcome_window_text_bot_1").node.Text = ProcessStringWithReplacements (GetLocalizedString ("WELCOME.BOT_1"), {
      { token = "{hours}", replacement = GetGrantPeriod () / 60 / 60  }})
  end, nil, nil, nil, nil, nil, nil, nil, nil, nil, { hide_on_back = true })
end

local_methods.MoveEditText = function (text_length, len, text_line)
  if text_length < len then
    game.current_screen.vars.rating_window_edit_text_offset = { x = 0, y = 0 }
    text_line.node.HorizontalAlignment = Scene.TextLineAlignment.Left
  else
    game.current_screen.vars.rating_window_edit_text_offset = { x = game.current_screen.vars.offset_count, y = 0 }
    text_line.node.HorizontalAlignment = Scene.TextLineAlignment.Right
  end
end

local_methods.EditTextRatingWindow = function (impl)
  if impl.rating_window.text_editor.IsInputStarted () then return end

  DeactivateSprite ("rating_window_arrow")
  
  local camera        = game.current_screen.cameras.modal_window_camera.node
  local camera_height = camera.Top - camera.Bottom
  
  impl.rating_window.camera.node.Position = vec3 (0, game.current_screen.vars.offset_window, 0)
  
  impl.rating_window.text_editor.SetOnEdit (function() local_methods.MoveEditText (impl.rating_window.text_editor.TextLength (), game.current_screen.vars.symbol_count, GetTextLine ("rating_window_edit_text")) end)
  impl.rating_window.text_editor.SetOnEndEdit (function() impl.rating_window.camera.node.Position = vec3 (0, 0, 0) end)
  impl.rating_window.text_editor.StartInput ()
end

local_methods.OnHideRatingWindow = function (impl)
  impl.rating_window.text_editor.StopInput () 
  impl.rating_window.text_editor.SetOnEdit (nil)
  impl.rating_window.text_editor.SetOnEndEdit (nil)
end

local_methods.ShowRatingWindow = function (impl, slot_id)
  if not impl.rating_window then
    PrepareAnimationScreenPart ("data/animations/rating_window.xml")

    local rating_window_sprite_list = { "rating_window_bg", "rating_window_screen_lock",
      "rating_window_button_right", "rating_window_button_right_pressed", "rating_window_edit_text_button",
      "rating_window_button_left", "rating_window_button_left_pressed", "rating_window_edit_zone",
      "rating_window_game_icon_zone", "rating_window_button_right_disable", "rating_window_arrow",
      "rating_window_star_l_1", "rating_window_star_p_1",
      "rating_window_star_l_2", "rating_window_star_p_2",
      "rating_window_star_l_3", "rating_window_star_p_3",
      "rating_window_star_l_4", "rating_window_star_p_4",
      "rating_window_star_l_5", "rating_window_star_p_5" }
    local rating_window_sprite_list_show = { "rating_window_bg", "rating_window_screen_lock",
      "rating_window_button_left", "rating_window_button_right_disable", "rating_window_arrow",
      "rating_window_game_icon_zone", "rating_window_star_l_1", "rating_window_star_l_2",
      "rating_window_star_l_3", "rating_window_star_l_4", "rating_window_star_l_5", "rating_window_edit_text_button"}
    local rating_window_text_line_list = { "rating_window_title", "rating_window_text_1", "rating_window_text_2",
      "rating_window_send", "rating_window_cancel", "rating_window_edit_text" }

    impl.rating_window = {}

    impl.rating_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.rating_window.resource       = "data/res/materials/rating_window_layers.xmtl"
    impl.rating_window.sprites        = rating_window_sprite_list_show
    impl.rating_window.text_lines     = GetTextLines (rating_window_text_line_list)
    impl.rating_window.all_text_lines = GetTextLines (rating_window_text_line_list)
    impl.rating_window.all_sprites    = MakeAllSprites (rating_window_sprite_list)
    impl.rating_window.on_hide        = function () local_methods.OnHideRatingWindow (impl) end
  end

  local additional_resource_bindings = {}

  local slot_name          = ""
  local slot_path          = ""
  local slot_icon_material = "transparent"
  local slot_xmtl_path     = "data/res/materials/rating_window_layers.xmtl"
  local slots              = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots do
    local slot_desc = slots [i]

    if slot_desc.id and slot_desc.id + 0 == slot_id + 0 then
      slot_name          = slot_desc.game_name
      slot_icon_material = slot_desc.icon_material
      slot_path          = slot_desc.path
      slot_xmtl_path     = slot_desc.xmtl_path
      break
    end
  end

  impl.rating_window.text_editor = CreateAdvancedText (GetTextLine ("rating_window_edit_text"), 500)
  impl.rating_window.hall_of_fame_top_three_offset = 0;
  impl.rating_window.text_len = 0

  table.insert (additional_resource_bindings, LoadResource (slot_xmtl_path))

  if slot_xmtl_path and Common.File.IsFileExist (slot_xmtl_path) then
    local icon = GetSprite ("rating_window_game_icon_zone")
    
    icon.node.Material  = slot_icon_material
    icon.node.TexSize   = vec2 (1, 1)
    icon.node.TexOffset = vec2 (0, 0)
    
    icon.tracks ["texsize"] = nil
    icon.tracks ["texoffset"] = nil
  end

  impl.rating_window.count_star = 0

  ShowWindow (impl, impl.rating_window, {}, {}, {}, function ()
    impl.rating_window.text_editor.SetPlaceholder (GetLocalizedString ("RATING_WINDOW.FEEDBACK"))

    local edit_text_zone        = GetSprite ("rating_window_edit_zone").layout
    local edit_text_zone_left   = edit_text_zone:Get ("Left") + 0
    local edit_text_zone_top    = edit_text_zone:Get ("Top") + 0
    local edit_text_zone_width  = edit_text_zone:Get ("Width") + 0
    local edit_text_zone_height = edit_text_zone:Get ("Height") + 0

    BindOrUpdateScissor (GetTextLine ("rating_window_edit_text"), edit_text_zone_left, -edit_text_zone_top, edit_text_zone_width, edit_text_zone_height)
  end, nil, nil, nil, function () 
    local slots = GetSlots ()
    
    for i = 1, #slots do
      local slot_desc = slots [i]
      
      if slot_desc.id == slot_id then
        slot_desc.voted = true
        break
      end
    end
  
    HsConnectionSendMessage (1, { event = "slots.vote", slotId = slot_id, stars = impl.rating_window.count_star, message = impl.rating_window.text_editor.Text () })
    SetCurrentScreen ("lobby") 
  end, function ()
    local slots = GetSlots ()
    
    for i = 1, #slots do
      local slot_desc = slots [i]
      
      if slot_desc.id == slot_id then
        slot_desc.voted = true
        break
      end
    end
  
    HsConnectionSendMessage (1, { event = "slots.vote", slotId = slot_id, stars = 0, message = "" })
    SetCurrentScreen ("lobby") 
  end, additional_resource_bindings)
end

local_methods.SetStarCount = function (impl, count_star)
  ActivateSprite ("rating_window_button_right")
  DeactivateSprites ("rating_window_button_right_disable", "rating_window_arrow")
  impl.rating_window.count_star = count_star
  for i = 1, 5 do
    SetSpritesActive (i <= count_star, "rating_window_star_p_" .. i)
  end
end

local_methods.ShowBattleLeaveWindow = function (impl, on_ok_handler)
  if not impl.battle_leave_window then
    PrepareAnimationScreenPart ("data/animations/battle_leave_window.xml")

    local battle_leave_window_sprite_list = { "battle_leave_window_bg", "battle_leave_window_screen_lock",
                                              "battle_leave_window_right_button_l", "battle_leave_window_right_button_p",
                                              "battle_leave_window_left_button_l", "battle_leave_window_left_button_p" }
    local battle_leave_window_text_line_list = { "battle_leave_window_text_1", "battle_leave_window_text_2", "battle_leave_window_text_3",
                                                 "battle_leave_window_text_right_btn", "battle_leave_window_text_left_btn" }
    impl.battle_leave_window = {}

    impl.battle_leave_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.battle_leave_window.resource       = "data/res/materials/battle_leave_window_layers.xmtl"
    impl.battle_leave_window.sprites        = battle_leave_window_sprite_list
    impl.battle_leave_window.text_lines     = GetTextLines (battle_leave_window_text_line_list)
    impl.battle_leave_window.all_text_lines = GetTextLines (battle_leave_window_text_line_list)
    impl.battle_leave_window.all_sprites    = MakeAllSprites (battle_leave_window_sprite_list)
  end

  ShowWindow (impl, impl.battle_leave_window, {}, {}, {}, nil, nil, nil, nil, on_ok_handler, nil, nil, nil, nil, nil, { hide_on_back = true })
end

local_methods.ShowChangeLanguageWindow = function (impl)
  if not impl.change_language_window then
    PrepareAnimationScreenPart ("data/animations/change_language_window.xml")

    local all_sprite_list = { "change_language_window_background", "change_language_window_screen_lock",
      "change_language_window_close_button", "change_language_window_close_button_pressed",
      "change_language_window_language_button_l_1", "change_language_window_language_button_p_1", "change_language_window_language_button_zone_1", "change_language_window_language_choose_1",
      "change_language_window_language_button_l_2", "change_language_window_language_button_p_2", "change_language_window_language_button_zone_2", "change_language_window_language_choose_2",
      "change_language_window_language_button_l_3", "change_language_window_language_button_p_3", "change_language_window_language_button_zone_3", "change_language_window_language_choose_3",
      "change_language_window_language_button_l_4", "change_language_window_language_button_p_4", "change_language_window_language_button_zone_4", "change_language_window_language_choose_4",
      "change_language_window_language_button_l_5", "change_language_window_language_button_p_5", "change_language_window_language_button_zone_5", "change_language_window_language_choose_5",
      "change_language_window_language_button_l_6", "change_language_window_language_button_p_6", "change_language_window_language_button_zone_6", "change_language_window_language_choose_6",
      "change_language_window_language_button_l_7", "change_language_window_language_button_p_7", "change_language_window_language_button_zone_7", "change_language_window_language_choose_7" }
    local show_sprite_list = {"change_language_window_background", "change_language_window_screen_lock",
      "change_language_window_close_button", "change_language_window_close_button_pressed",
      "change_language_window_language_button_l_1", "change_language_window_language_button_l_2",
      "change_language_window_language_button_l_3", "change_language_window_language_button_l_4",
      "change_language_window_language_button_l_5", "change_language_window_language_button_l_6",
      "change_language_window_language_button_l_7", }
    local all_text_line_list = { "info_window_title_line" }
    impl.change_language_window = {}

    local all_text_line_list = { "change_language_window_title_line", "change_language_window_text_line",
      "change_language_window_language_text_1_select", "change_language_window_language_text_1_l", "change_language_window_language_text_1_p",
      "change_language_window_language_text_2_select", "change_language_window_language_text_2_l", "change_language_window_language_text_2_p",
      "change_language_window_language_text_3_select", "change_language_window_language_text_3_l", "change_language_window_language_text_3_p",
      "change_language_window_language_text_4_select", "change_language_window_language_text_4_l", "change_language_window_language_text_4_p",
      "change_language_window_language_text_5_select", "change_language_window_language_text_5_l", "change_language_window_language_text_5_p",
      "change_language_window_language_text_6_select", "change_language_window_language_text_6_l", "change_language_window_language_text_6_p",
      "change_language_window_language_text_7_select", "change_language_window_language_text_7_l", "change_language_window_language_text_7_p" }
    local show_text_line = { "change_language_window_title_line", "change_language_window_text_line",
      "change_language_window_language_text_1_l", "change_language_window_language_text_2_l",
      "change_language_window_language_text_3_l", "change_language_window_language_text_4_l",
      "change_language_window_language_text_5_l", "change_language_window_language_text_6_l",
      "change_language_window_language_text_7_l" }

    local button_index = {}
    local languages = { "en-US", "de-DE", "fr-FR", "ru-RU", "es-ES", "it-IT", "tr-TR"}

    button_index ["en-US"] = 1
    button_index ["fr-FR"] = 2
    button_index ["de-DE"] = 3
    button_index ["ru-RU"] = 4
    button_index ["it-IT"] = 5
    button_index ["es-ES"] = 6
    button_index ["tr-TR"] = 7

    local current_locale = GetCurrentLocale ()

    table.insert(show_sprite_list, "change_language_window_language_choose_" .. button_index[current_locale])
    table.insert(show_text_line, "change_language_window_language_text_" .. button_index[current_locale] .. "_select")
    for i = 1, #languages do
      if languages[i] ~=  current_locale then
        table.insert(show_sprite_list, "change_language_window_language_button_zone_" .. button_index[languages[i]])
      end
    end

    impl.change_language_window = {}
    impl.change_language_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.change_language_window.resource       = "data/res/materials/change_language_window_layers.xmtl"
    impl.change_language_window.sprites        = show_sprite_list
    impl.change_language_window.text_lines     = GetTextLines (show_text_line)
    impl.change_language_window.all_text_lines = GetTextLines (all_text_line_list)
    impl.change_language_window.all_sprites    = MakeAllSprites (all_sprite_list)
    impl.change_language_window.on_hide        = function ()
    DeactivateSprites ()
    end
  end

  ShowWindow (impl, impl.change_language_window, {}, {}, {}, nil, nil, nil, nil, nil, nil, nil, nil, nil, nil, { hide_on_back = true })
end

local function ShowAchievementsCompleteWindow (impl, id, group_id, current, prize_type)
  if id > 26 or group_id > ACHIEVEMENTS_GROUPS_COUNT then return end --IGNORE NEW ACHIEVEMENTS

  if not impl.achievements_complete_window then
    PrepareAnimationScreenPart ("data/animations/achievements_complete_window.xml")

    local achievements_complete_window_sprite_list = { "achievements_complete_window_bg", "achievements_complete_window_close_l", 
      "achievements_complete_window_close_p",
      "achievements_complete_window_right_button_l", "achievements_complete_window_right_button_p", "achievements_complete_window_screen_lock", 
      "achievements_complete_window_group_1", "achievements_complete_window_group_2", 
      "achievements_complete_window_group_3", "achievements_complete_window_group_4", "achievements_complete_window_group_5",
      "achievements_complete_window_check", "achievements_complete_window_button_check_zone" }
    local achievements_complete_window_text_line_list = { "achievements_complete_window_text_title", "achievements_complete_window_text_1", 
      "achievements_complete_window_text_2", "achievements_complete_window_text_3", "achievements_complete_window_text_share_friends", 
      "achievements_complete_window_left_button_text", "achievements_complete_window_right_button_text", "achievements_complete_window_text_complete"}
    impl.achievements_complete_window = {}

    impl.achievements_complete_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.achievements_complete_window.resource       = "data/res/materials/achievements_complete_window_layers.xmtl"
    impl.achievements_complete_window.sprites        = { "achievements_complete_window_bg", "achievements_complete_window_close_l", 
      "achievements_complete_window_right_button_l", "achievements_complete_window_screen_lock",
      "achievements_complete_window_check", "achievements_complete_window_button_check_zone"}
    impl.achievements_complete_window.text_lines     = GetTextLines (achievements_complete_window_text_line_list)
    impl.achievements_complete_window.all_text_lines = GetTextLines (achievements_complete_window_text_line_list)
    impl.achievements_complete_window.all_sprites    = MakeAllSprites (achievements_complete_window_sprite_list)
  end
  
  local share_type
  local share_key
  
  if prize_type == 1 then
    share_type = SHARE_ACHIEVEMENT_DIAMONDS_TYPE
    share_key  = "ACHIEVEMENTS_COMPLETE_WINDOW.SHARE_FRIENDS_DIAMONDS"
  else
    share_type = SHARE_ACHIEVEMENT_COINS_TYPE
    share_key  = "ACHIEVEMENTS_COMPLETE_WINDOW.SHARE_FRIENDS"
  end
  
  local share_values = GetSocialManager ().GetShareValues (share_type)
    
  local share_text = ProcessStringWithReplacements (GetLocalizedString (share_key), {
      { token = "{friends}", replacement = share_values.players_count },
      { token = "{coins}",   replacement = share_values.chips_count },
    })

  local text_replacement = { 
    { token = "{x}",     replacement = tostring (current) },
    { token = "{money}", replacement = FormatMoneyString (current, 5) }
  }
  
  local text_line_1 = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_COMPLETE_WINDOW.ACHIEVEMENTS_1_" .. id), text_replacement)
  local text_line_2 = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_COMPLETE_WINDOW.ACHIEVEMENTS_2_" .. id), text_replacement)
  local text_line_3 = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_COMPLETE_WINDOW.ACHIEVEMENTS_3_" .. id), text_replacement)

  ShowWindow (impl, impl.achievements_complete_window, { "achievements_complete_window_group_" .. group_id}, {}, {}, function ()
    impl.achievements_complete_window.checked = true
    
    ExecuteOnNextUpdate (function ()
      ExecuteOnNextUpdate (function ()
        ActivateSprite ("achievements_complete_window_button_check_zone") -- this sprite not active if same window was closed before
      end)
    end)

    for i = 1, ACHIEVEMENTS_GROUPS_COUNT do
      if i ~= group_id then
        DeactivateSprite ("achievements_complete_window_group_" .. i)
      end
    end
    
    impl.achievements_complete_window.id         = id
    impl.achievements_complete_window.group_id   = group_id
    impl.achievements_complete_window.current    = current
    impl.achievements_complete_window.share_type = share_type
  
    GetTextLine("achievements_complete_window_text_share_friends").node.Text = share_text
    GetTextLine("achievements_complete_window_text_1").node.Text             = text_line_1
    GetTextLine("achievements_complete_window_text_2").node.Text             = text_line_2
    GetTextLine("achievements_complete_window_text_3").node.Text             = text_line_3
  end, nil, nil, nil, nil, nil, nil, true, nil, nil, { hide_on_back = true })
end

local function AchievementCompleteShare (impl)
  if not impl.achievements_complete_window.checked then
    impl.achievements_complete_window.checked = true
    ActivateSprite("achievements_complete_window_check")
  else
    impl.achievements_complete_window.checked = nil
    DeactivateSprite("achievements_complete_window_check")
  end
end

local_methods.AchievementCompleteView = function (impl)
  HideWindow (impl, nil, true)
  
  if impl.achievements_complete_window.checked then
    GetSocialManager ().ShareAchievement (impl.achievements_complete_window.id, impl.achievements_complete_window.current, impl.achievements_complete_window.share_type)
  end

  game.scripts.ui_windows.ShowAchievementsWindow (GetUserId (), impl.achievements_complete_window.group_id, true)
end

local function CleanTextLines (text_lines)
  for i = 1, #text_lines do
    if text_lines [i] then
      GetTextLine (text_lines[i]).node.Text = ""
    end
  end
end

local_methods.OnUserDetailsLoaded = function (impl, event)
  local user_id = GetNodeChildAttribute (event, "userId") + 0


  if not impl.waiting_user_details_id or impl.waiting_user_details_id ~= user_id then return end

  impl.waiting_user_details_id = nil

  local details = {}

  details.firstName = GetNodeChildAttribute (event, "firstName", "")
  details.lastName  = GetNodeChildAttribute (event, "lastName", "")
  details.birthday  = GetNodeChildAttribute (event, "birthday", "")
  details.city      = GetNodeChildAttribute (event, "city", "-")
  details.level     = GetNodeChildAttribute (event, "level", "")
  details.photo     = GetNodeChildAttribute (event, "photo", "")

  local type_window
  local camera

  if impl.achievements_window and impl.achievements_window.is_shown then
    type_window   = "achievements"
    camera        = impl.achievements_window.camera.node
  elseif impl.profile_window and impl.profile_window.is_shown then
    type_window   = "profile"
    camera        = impl.profile_window.camera.node
  else
    return
  end

  GetTextLine(type_window .. "_window_text_name").node.Text           = details.firstName .. " " .. details.lastName
  GetTextLine(type_window .. "_window_text_value_birthday").node.Text = details.birthday
  GetTextLine(type_window .. "_window_text_value_city").node.Text     = details.city
  GetTextLine(type_window .. "_window_level").node.Text               = details.level

  if impl.profile_avatar then
    impl.profile_avatar.avatar.OnHide ()
    impl.profile_avatar.avatar.GetSprite ():Unbind ()
    impl.profile_avatar = nil
  end

  local avatar_ref  = GetSprite (type_window .. "_window_avatar_place")
  local avatar_size = { x = avatar_ref.layout:Get ("Width") + 0, y = avatar_ref.layout:Get ("Height") + 0 }

  if not impl.profile_avatar then
    impl.profile_avatar = {}
  
    impl.profile_avatar.user_id = user_id
  
    local window        = Engine.Windows.Find (WINDOW_NAME)
    local camera_width  = camera.Right - camera.Left
    local camera_height = camera.Top - camera.Bottom
    local screen_scale  = window.ClientHeight / camera_height
    
    impl.profile_avatar.avatar = CreateUrlSprite (URL_SPRITE_MODE_CACHE_FILE, details.photo, {}, 
                                                 { material = "transparent", tex_size = vec2 (1, 1), tex_offset = vec2 (0, 0) }, 
                                                 avatar_size, avatar_size.x * PHOTO_ACTIVITY_INDICATOR_FACTOR, screen_scale)
  end
  
  local avatar_sprite = impl.profile_avatar.avatar.GetSprite ()

  avatar_sprite.Position = vec3 (avatar_ref.layout:Get ("Left") + avatar_size.x / 2, -avatar_ref.layout:Get ("Top") - avatar_size.y / 2, avatar_ref.z - EPS)

  avatar_sprite:BindToParent (game.current_screen.node)

  impl.profile_avatar.avatar.OnShow ()
end

local_methods.OnUserProfileLoaded = function (impl, event)
  if not impl.waiting_user_profile_id or impl.waiting_user_profile_id ~= (GetNodeChildAttribute (event, "userId") + 0) then return end

  impl.waiting_user_profile_id = nil

  local profile = {}

  profile.register_day          = GetNodeChildAttribute (event, "registerDay", "")
  profile.last_login_day        = GetNodeChildAttribute (event, "lastLoginDay", "")
  profile.money                 = GetNodeChildAttribute (event, "money") + 0
  profile.diamonds              = GetNodeChildAttribute (event, "diamonds") + 0
  profile.wagering              = GetNodeChildAttribute (event, "wagering") + 0
  profile.friends_count         = GetNodeChildAttribute (event, "friendsCount", 0) + 0
  profile.friends_rating        = GetNodeChildAttribute (event, "friendsRating", 1) + 0
  profile.max_bonus_win         = GetNodeChildAttribute (event, "maxBonusWin", 0) + 0
  profile.max_bonus_win_slot_id = GetNodeChildAttribute (event, "maxBonusWinSlotId", 1) + 0
  profile.max_win               = GetNodeChildAttribute (event, "maxWin", 0) + 0
  profile.max_win_slot_id       = GetNodeChildAttribute (event, "maxWinSlotId", 1) + 0
  profile.favorite_slot_id      = GetNodeChildAttribute (event, "favoriteSlotId", 1) + 0
  profile.achievements          = GetNodeChildArray (event, 'achievements', {})
  profile.battle_win            = GetNodeChildAttribute (event, "battleWins", 1) + 0
  profile.battle_loses          = GetNodeChildAttribute (event, "battleLoses", 1) + 0
  profile.battle_rating         = GetNodeChildAttribute (event, "battleRating", 1) + 0
  profile.luck                  = GetNodeChildAttribute (event, "luck", 0) + 0

  for i = #profile.achievements, 1, -1 do
    if profile.achievements [i] + 0 > ACHIEVEMENTS_GROUPS_COUNT then
      table.remove (profile.achievements, i)
    end
  end

  GetTextLine("profile_window_text_value_fortune").node.Text           = tostring(profile.luck)
  GetTextLine("profile_window_text_value_reg_date").node.Text          = profile.register_day
  GetTextLine("profile_window_text_value_last_visit_day").node.Text    = profile.last_login_day
  GetTextLine("profile_window_text_value_experience").node.Text        = FormatMoneyString (profile.wagering, nil, false)
  GetTextLine("profile_window_text_value_battle_wins_loses").node.Text = tostring(profile.battle_win) .. "/" .. tostring(profile.battle_loses)
  GetTextLine("profile_window_text_value_battle_rating").node.Text     = tostring(profile.battle_rating)
  GetTextLine("profile_window_text_value_friends_rating").node.Text    = string.format (GetLocalizedString ("PROFILE.RATING"), profile.friends_rating , profile.friends_count)

  local profile_window_cons = {}
  local profile_window_get_cons_l = GetSprite ("profile_window_get_cons_l")
  local profile_window_get_cons_p = GetSprite ("profile_window_get_cons_p")
  local profile_window_diamond = GetSprite ("profile_window_diamond")
  local profile_window_get_diamonds_l = GetSprite ("profile_window_get_diamonds_l")
  local profile_window_get_diamonds_p = GetSprite ("profile_window_get_diamonds_p")

  if impl.profile_user_id == GetUserId () and not impl.phone_ui then
    ActivateSprite ("profile_window_get_cons_l")
    ActivateSprite ("profile_window_get_diamonds_l")
  end

  if not impl.phone_ui then
    GetTextLine("profile_window_text_value_balance").node.Text           = FormatMoneyString (profile.money, 14, nil, true)
    GetTextLine("profile_window_text_value_balance_diamonds").node.Text  = FormatMoneyString (profile.diamonds * 100, 7, nil, true)

    ActivateSprite("profile_window_diamond")
    ActivateSprite("profile_window_cons")
    ActivateSprite("profile_window_cons_underlay")
    ActivateSprite("profile_window_diamond_underlay")
  else
    GetTextLine("profile_window_text_value_balance").node.Text           = FormatMoneyString (profile.money, 8, nil, true)
    GetTextLine("profile_window_text_value_balance_diamonds").node.Text  = FormatMoneyString (profile.diamonds * 100, 5, nil, true)

    local scrollable_sprites_list = { "profile_window_diamond", "profile_window_cons" }

    if impl.profile_user_id == GetUserId () then
      table.insert (scrollable_sprites_list, "profile_window_get_cons_l")
      table.insert (scrollable_sprites_list, "profile_window_get_diamonds_l")
    end

    for i = 1, #scrollable_sprites_list do
      local sprite_node = GetSprite (scrollable_sprites_list [i]).node

      sprite_node:BindToParent (impl.profile_window.scroll_node)

      sprite_node.Scissor = impl.scissor_profile
    end
  end

  local slot_name_max_win = ""
  local slots             = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots do
    local slot_desc = slots [i]

    if slot_desc.id and slot_desc.id + 0 == profile.max_win_slot_id + 0 then
      slot_name_max_win = slot_desc.game_name
      break
    end
  end

  if slot_name_max_win ~= "" then
    GetTextLine("profile_window_text_value_maximum_gain").node.Text = string.format (GetLocalizedString ("PROFILE.CONS_IN"), FormatMoneyString (profile.max_win, 6) , slot_name_max_win)
  else
    GetTextLine("profile_window_text_value_maximum_gain").node.Text = "0"
  end

  local bonus_name = ""
  local bonus     = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #bonus do
    local bonus_desc = bonus [i]

    if bonus_desc.id and bonus_desc.id + 0 == profile.max_bonus_win_slot_id + 0 then
      bonus_name = bonus_desc.game_name
      break
    end
  end

  if bonus_name ~= "" then
    GetTextLine("profile_window_text_value_maximum_gain_in_bonus").node.Text = string.format (GetLocalizedString ("PROFILE.CONS_IN"), FormatMoneyString (profile.max_bonus_win, 6), bonus_name)
  else
    GetTextLine("profile_window_text_value_maximum_gain_in_bonus").node.Text = "0"
  end

  local slot_name_favorit = ""
  local slots_favorit     = ApplicationDataGet (SLOTS_DESCS)

  for i = 1, #slots_favorit do
    local slot_desc_favorit = slots_favorit [i]

    if slot_desc_favorit.id and slot_desc_favorit.id + 0 == profile.favorite_slot_id + 0 then
      slot_name_favorit = slot_desc_favorit.game_name
      break
    end
  end

  if slot_name_favorit ~= "" then
    GetTextLine("profile_window_text_value_favorite_slot").node.Text = slot_name_favorit
  else
    GetTextLine("profile_window_text_value_favorite_slot").node.Text = "-"
  end

  local achievements = ""

  if impl.phone_ui and profile.achievements and #profile.achievements > 2 then
    GetTextLine("profile_window_text_value_achievements_1").node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. tostring(profile.achievements[1])) .. "; " .. GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. tostring(profile.achievements[2])) .. ";"

    for i = 3, #profile.achievements do
      if i == #profile.achievements then
        achievements = achievements .. GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. tostring(profile.achievements[i]))
      else
        achievements = achievements .. GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. tostring(profile.achievements[i])) .. "; "
      end
    end

    GetTextLine("profile_window_text_value_achievements_2").node.Text = achievements
  else
    if profile.achievements and #profile.achievements ~= 0 then
      for i = 1, #profile.achievements do
        if i == #profile.achievements then
          achievements = achievements .. GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. tostring(profile.achievements[i]))
        else
          achievements = achievements .. GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. tostring(profile.achievements[i])) .. "; "
        end
      end
    else
      achievements = "-"
    end

    local text_line_node = GetTextLine("profile_window_text_value_achievements").node

    text_line_node.Text = achievements
    
    if not impl.phone_ui then
      local text_size = game.current_screen.vars.profile_window_achievements_text_size
    
      if text_line_node.TextLength > game.current_screen.vars.profile_window_achievements_long_text_length then
        text_size = game.current_screen.vars.profile_window_achievements_long_text_size
      end
    
      text_line_node.Scale = vec3 (text_size, text_size, 1)
    end
  end
end

local_methods.AchievementsWindowShowOverlay = function (impl)
	for i = 1, #impl.achievements_window.overlay_sprites_show do
		ActivateSprite (impl.achievements_window.overlay_sprites_show[i])
	end

	for i = 1, ACHIEVEMENTS_GROUPS_COUNT do

		local group = impl.achievements_window.achievements_data[i]

		if i == 5 then
			GetTextLine ("achievements_window_all_achiev_text_coin_" .. i).node.Text = FormatMoneyString (group.prize * 100)
			GetTextLine ("achievements_window_all_achiev_text_coin_" .. i .. "_p").node.Text = FormatMoneyString (group.prize * 100)
		else
			GetTextLine ("achievements_window_all_achiev_text_coin_" .. i).node.Text = FormatMoneyString (group.prize)
			GetTextLine ("achievements_window_all_achiev_text_coin_" .. i .. "_p").node.Text = FormatMoneyString (group.prize)
		end

		GetTextLine ("achievements_window_all_achiev_text_name_" .. i).node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. i)
		GetTextLine ("achievements_window_all_achiev_text_name_" .. i .. "_p").node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. i)
		GetTextLine ("achievements_window_all_achiev_text_progress_" .. i).node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.PAGE_PROGRESS_TITLE") .. group.completed_achievements .. "/" .. group.total_achievements
		GetTextLine ("achievements_window_all_achiev_text_progress_" .. i .. "_p").node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.PAGE_PROGRESS_TITLE") .. group.completed_achievements .. "/" .. group.total_achievements

		local page_desc  = impl.achievements_window.pages [i]

		BindOrUpdateScissor (page_desc.overlay_image_sprite, impl.achievements_window.clip_left, 0, impl.achievements_window.clip_width, EPS)

		local progress = 0

		if impl.achievements_window.achievements_data then
			progress = group.completed_achievements / group.total_achievements
		end

		SlowAction ("ui", 0.75, function (t)
		  local height = page_desc.overlay_image_sprite_scissor_height * t * progress 

      BindOrUpdateScissor (page_desc.overlay_image_sprite, impl.achievements_window.clip_left, page_desc.overlay_image_sprite_scissor_bottom + height, impl.achievements_window.clip_width, height)
		end)
	end
end

local function AchievementsScrollToPage (impl, page, animated)
  local new_position = (page - 1) * -impl.achievements_window.page_width

  if impl.achievements_window.page_scroll_action then
    impl.achievements_window.page_scroll_action.Stop ()
    impl.achievements_window.page_scroll_action = nil
  end

  local camera            = impl.achievements_window.camera.node
  local camera_width      = camera.Right - camera.Left
  local camera_height     = camera.Top - camera.Bottom
    
  local function UpdateProgressBarsScissors (impl)
    for name, scissor_width in pairs (impl.achievements_window.progress_bars_scissors) do
      local node       = GetSprite (name).node
      local clip_right = node.WorldPosition.x - node.WorldScale.x / 2 + scissor_width
      local clip_width = math.max (EPS, math.min (impl.achievements_window.clip_width, clip_right - impl.achievements_window.clip_left))
      
      BindOrUpdateScissor (GetSprite (name), impl.achievements_window.clip_left, 0, clip_width, camera_height)
    end
  end

  if animated then
    local current_position = impl.achievements_window.pages_node.Position.x
    local position_diff    = new_position - current_position
  
    impl.achievements_window.drag_locked = true
    impl.achievements_window.drag_x = nil  
  
    impl.achievements_window.page_scroll_action = SlowAction ("ui", 0.3, function (t)
      impl.achievements_window.pages_node.Position = vec3 (current_position + t * position_diff, 0, 0)
      UpdateProgressBarsScissors (impl) 
      
      if t == 1 then
        impl.achievements_window.drag_locked = false
      end
    end)
  else
    impl.achievements_window.pages_node.Position = vec3 (new_position, 0, 0)
  end

  UpdateProgressBarsScissors (impl) 

  impl.achievements_window.current_page = page

  local page_desc  = impl.achievements_window.pages [page]
    
  BindOrUpdateScissor (page_desc.image_sprite, impl.achievements_window.clip_left, 0, impl.achievements_window.clip_width, EPS)

  local progress = 0
  
  if impl.achievements_window.achievements_data then
    progress = impl.achievements_window.achievements_data [page].completed_achievements / impl.achievements_window.achievements_data [page].total_achievements
  end

  if impl.achievements_window.image_scissor_action then
    impl.achievements_window.image_scissor_action.Stop ()
    impl.achievements_window.image_scissor_action = nil
  end

  impl.achievements_window.image_scissor_action = SlowAction ("ui", 0.75, function (t)
    local height = page_desc.image_sprite_scissor_height * t * progress
  
    BindOrUpdateScissor (page_desc.image_sprite, impl.achievements_window.clip_left, page_desc.image_sprite_scissor_bottom + height, impl.achievements_window.clip_width, height)
  end)

  SetSpritesActive (page == 1, "achievements_window_back_button_d")
  SetSpritesActive (page > 1, "achievements_window_back_button_l")
  SetSpritesActive (page == ACHIEVEMENTS_GROUPS_COUNT, "achievements_window_next_button_d")
  SetSpritesActive (page < ACHIEVEMENTS_GROUPS_COUNT, "achievements_window_next_button_l")

  GetTextLine ("achievements_window_current_achievement_title").node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. page)
end

local_methods.ActivateAchievementsWindowPageSprite = function (impl, page_index, sprite_name)
   local page   = impl.achievements_window.pages [page_index]
   local sprite = GetSprite (sprite_name)

   ActivateSprite (sprite)

   sprite.node:BindToParent (page.node)
 end

local_methods.ActivateAchievementsWindowPageTextLine = function (impl, page_index, line_name)
  local page = impl.achievements_window.pages [page_index]
  local line = GetTextLine (line_name)

  ActivateTextLine (line)

  line.node:BindToParent (page.node)
end

local_methods.AchievementsWindowCheck = function (impl, index)
  local new_state = not IsSpriteActive ("achievements_window_pages_show_checked_" .. index)

  HsConnectionSendMessage (1, { event = "achievement.select", id = index, checked = new_state })

  for i = 1, ACHIEVEMENTS_GROUPS_COUNT do
    if i == index and new_state then
      local_methods.ActivateAchievementsWindowPageSprite (impl, i, "achievements_window_pages_show_checked_" .. i)
    else
      DeactivateSprite ("achievements_window_pages_show_checked_" .. i)
    end
  end
end

local_methods.OnUserAchievementsLoaded = function (impl, event)
  if not impl.waiting_user_achievements_id or impl.waiting_user_achievements_id ~= (GetNodeChildAttribute (event, "userId") + 0) then return end

  local is_current_user = impl.waiting_user_achievements_id == GetUserId ()

  local INCOMPLETE_ACHIEVEMENT_COLOR = vec4 (237/255, 232/255, 172/255, 1)
  
  for i = 1, #impl.achievements_window.pages do
    local page       = impl.achievements_window.pages [i]
    local text_lines = page.text_line_list
    
    for j = 1, #text_lines do
      local text_line = GetTextLine (text_lines [j])

      ActivateTextLine (text_line)

      text_line.node:BindToParent (page.node)
    end

    for j = 1, 5 do
      GetTextLine ("achievements_window_pages_achievement_title_" .. j .. "_line_1_" .. i).node.Color = INCOMPLETE_ACHIEVEMENT_COLOR
      GetTextLine ("achievements_window_pages_achievement_title_" .. j .. "_line_2_" .. i).node.Color = INCOMPLETE_ACHIEVEMENT_COLOR
      GetTextLine ("achievements_window_pages_achievement_title_" .. j .. "_line_3_" .. i).node.Color = INCOMPLETE_ACHIEVEMENT_COLOR
    end
  end
  
  impl.waiting_user_achievements_id = nil

  local achievements = {}
  local groups_node  = event:Find ("groups"):Child (0)
  local completed_achievements = 0
  local total_achievements     = 0

  local COMPLETED_ACHIEVEMENT_COLOR = vec4 (173/255, 169/255, 118/255, 1)
  
  for i = 0, groups_node.ChildrenCount - 1 do 
    local group      = {}
    local group_node = groups_node:Child (i)

    group.id = GetNodeChildAttribute (group_node, "id") + 0
    
    if group.id > 0 and group.id <= ACHIEVEMENTS_GROUPS_COUNT then
      group.prize    = GetNodeChildAttribute (group_node, "prize", 0) + 0
      group.selected = GetNodeChildAttribute (group_node, "selected", "false") == "true"

      group.achievements = {}

      local achievements_node = group_node:Find ("achievements"):Child (0)

      group.total_achievements     = achievements_node.ChildrenCount
      group.completed_achievements = 0

      total_achievements = total_achievements + group.total_achievements

      for j = 0, achievements_node.ChildrenCount - 1 do
        local achievement      = {}
        local achievement_node = achievements_node:Child (j)

        achievement.id      = GetNodeChildAttribute (achievement_node, "id") + 0
        achievement.type    = GetNodeChildAttribute (achievement_node, "type", "0") + 0
        achievement.current = GetNodeChildAttribute (achievement_node, "current", "0") + 0
        achievement.total   = GetNodeChildAttribute (achievement_node, "total", "0") + 0

        if achievement.current == achievement.total then
          completed_achievements = completed_achievements + 1
          group.completed_achievements = group.completed_achievements + 1

          if j < 5 then
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_1_" .. group.id).node.Color = COMPLETED_ACHIEVEMENT_COLOR
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_2_" .. group.id).node.Color = COMPLETED_ACHIEVEMENT_COLOR
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_3_" .. group.id).node.Color = COMPLETED_ACHIEVEMENT_COLOR

            local_methods.ActivateAchievementsWindowPageSprite (impl, group.id, "achievements_window_pages_achievement_" .. j + 1 .. "_checked_" .. group.id)
          end
        else
          if j < 5 then
            local filler_name = "achievements_window_pages_achievement_" .. achievement.id .. "_progress_bar_fill"

            for k = 1, #impl.achievements_window.progress_bars_sprites do
              if filler_name == impl.achievements_window.progress_bars_sprites [k] then
                impl.achievements_window.progress_bars_scissors [filler_name] = achievement.current / achievement.total * GetSprite (filler_name).layout:Get ("Width")

                local_methods.ActivateAchievementsWindowPageSprite (impl, group.id, "achievements_window_pages_achievement_" .. achievement.id .. "_progress_bar")
                local_methods.ActivateAchievementsWindowPageSprite (impl, group.id, filler_name)

                break
              end
            end

            local progress_text_line = GetTextLine ("achievements_window_pages_achievement_" .. achievement.id .. "_progress")

            if progress_text_line then
              local_methods.ActivateAchievementsWindowPageTextLine (impl, group.id, progress_text_line)

              if game.current_screen.vars ["show_only_current_for_achivement_" .. achievement.id] then
                progress_text_line.node.Text = tostring (achievement.current)
              else
                progress_text_line.node.Text = achievement.current .. "/" .. achievement.total
              end
            end

            if is_current_user then
              local button_sprite = GetSprite ("achievements_window_pages_achievement_" .. achievement.id .. "_button", true)

              if button_sprite then
                local_methods.ActivateAchievementsWindowPageTextLine (impl, group.id, "achievements_window_pages_achievement_" .. achievement.id .. "_button")
                local_methods.ActivateAchievementsWindowPageSprite (impl, group.id, button_sprite)
              end
            end
          end
        end

        if j < 5 then
          local replacements      = { { token = "{x}", replacement = tostring (achievement.total) },
            { token = "{money}", replacement = FormatMoneyString (achievement.total, 5) } }
          local achievement_index = (group.id - 1) * 5 + j + 1

          if impl.phone_ui then
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_1_" .. group.id).node.Text = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. achievement_index .. "_LINE_1_PHONE"), replacements)
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_2_" .. group.id).node.Text = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. achievement_index .. "_LINE_2_PHONE"), replacements)
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_3_" .. group.id).node.Text = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. achievement_index .. "_LINE_3_PHONE"), replacements)
          else
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_1_" .. group.id).node.Text = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. achievement_index .. "_LINE_1"), replacements)
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_2_" .. group.id).node.Text = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. achievement_index .. "_LINE_2"), replacements)
            GetTextLine ("achievements_window_pages_achievement_title_" .. j + 1 .. "_line_3_" .. group.id).node.Text = ProcessStringWithReplacements (GetLocalizedString ("ACHIEVEMENTS_WINDOW.ACHIEVEMENT_TITLE_" .. achievement_index .. "_LINE_3"), replacements)
          end

        end

        table.insert (group.achievements, achievement)
      end

      if group.total_achievements == group.completed_achievements and is_current_user then
        local_methods.ActivateAchievementsWindowPageTextLine (impl, group.id, "achievements_window_pages_show_" .. group.id)

        local_methods.ActivateAchievementsWindowPageSprite (impl, group.id, "achievements_window_pages_show_empty_" .. group.id)

        if group.selected then
          local_methods.ActivateAchievementsWindowPageSprite (impl, group.id, "achievements_window_pages_show_checked_" .. group.id)
        end
      end

      GetTextLine ("achievements_window_pages_progress_" .. group.id).node.Text = GetLocalizedString ("ACHIEVEMENTS_WINDOW.PAGE_PROGRESS_TITLE") .. group.completed_achievements .. "/" .. group.total_achievements

      local reward
      local reward_end
      local reward_start = GetLocalizedString ("ACHIEVEMENTS_WINDOW.PAGE_REWARD_START")
      local reward_node  = GetTextLine ("achievements_window_pages_reward_" .. group.id).node

      if group.id == 5 then
        reward     = FormatMoneyString (group.prize * 100)
        reward_end = GetLocalizedString ("ACHIEVEMENTS_WINDOW.PAGE_REWARD_END_DIAMONDS")

        if game.current_screen.vars.size_text_diamonds ~= nil then
          reward_node.Scale = vec3 (game.current_screen.vars.size_text_diamonds, game.current_screen.vars.size_text_diamonds, 1)
        end

        reward_node.Text = reward_start .. reward .. reward_end

        reward_node:SetCharsColorFactors (0, GetTextLength (reward_start), vec4 (0.13,0.63,0.76, 1))
        reward_node:SetCharsColorFactors (GetTextLength (reward_start) + #reward, GetTextLength (reward_end), vec4 (0.13,0.63,0.76, 1))
      else
        reward     = FormatMoneyString (group.prize)
        reward_end = GetLocalizedString ("ACHIEVEMENTS_WINDOW.PAGE_REWARD_END_COINS")

        reward_node.Text = reward_start .. reward .. reward_end

        reward_node:SetCharsColorFactors (0, GetTextLength (reward_start), vec4 (254/255,165/255,126/255, 1))
        reward_node:SetCharsColorFactors (GetTextLength (reward_start) + #reward, GetTextLength (reward_end), vec4 (254/255,165/255,126/255, 1))
      end

      table.insert (achievements, group)
    end
  end
  
  impl.achievements_window.achievements_data = achievements
  
  local achievements_progress = completed_achievements / total_achievements
  local progress_bar_sprite   = GetSprite ("achievements_window_total_progress_bar")
  local progress_bar_layout   = progress_bar_sprite.layout
  local camera                = impl.achievements_window.camera.node
  local camera_height         = camera.Top - camera.Bottom
  local scissor_left          = progress_bar_layout:Get ("Left") + 0
  local scissor_width         = progress_bar_layout:Get ("Width") * achievements_progress

  GetTextLine ("achievements_window_all_achievements_count").node.Text = tostring (completed_achievements) .. "/" .. tostring(total_achievements)

  impl.achievements_window.update_progress_action = SlowAction ("ui", 0.5, function (t)
    BindOrUpdateScissor (progress_bar_sprite, scissor_left, 0, t * scissor_width, camera_height)
  end)
	local_methods.AchievementsWindowShowOverlay (impl)
  AchievementsScrollToPage (impl, impl.achievements_window.current_page)
end

local_methods.LoadAchievements = function (impl, user_id)
  impl.waiting_user_achievements_id = user_id
  impl.waiting_user_details_id      = user_id

  HsConnectionSendMessage (1, { event  = "user.achievements", userId = user_id })
end

local_methods.LoadProfile = function (impl, user_id) 
  impl.waiting_user_profile_id = user_id
  impl.waiting_user_details_id = user_id

  HsConnectionSendMessage (1, { event  = "user.profile", userId = user_id })
end

local function CleanProfileWindowTab (impl)
  if impl.profile_avatar then
    impl.profile_avatar.avatar.OnHide ()
    impl.profile_avatar.avatar.GetSprite ():Unbind ()
  end
end

local function OnHideProfileWindow (impl)
  impl.profile_window.is_shown = false
  impl.waiting_user_profile_id = nil
  DeactivateSprites ("profile_window_diamond", "profile_window_cons", "profile_window_get_cons_l", "profile_window_get_diamonds_l", "profile_window_cons_underlay", "profile_window_diamond_underlay")
  CleanProfileWindowTab (impl)
  
  if impl.profile_window.scroll_node then
    impl.profile_window.scroll_node.Position = vec3 (0, 0, 0)
  
    local child = impl.profile_window.scroll_node.FirstChild
    
    while child do
      local next_child = child.NextChild
    
      child:BindToParent (game.current_screen.node, Scene.NodeBindMode.AddRef, Scene.NodeTransformSpace.World)
    
      local position = child.WorldPosition
    
      position.x = 100
      position.y = -100
    
      child.WorldPosition = position
    
      child = next_child
    end
  
    impl.profile_window.scroll_node:Unbind ()
    
    impl.profile_window.scroll_view = nil
  end

  if impl.profile_avatar then
    impl.profile_avatar.avatar.OnHide ()
    impl.profile_avatar.avatar.GetSprite ():Unbind ()
    impl.profile_avatar = nil
  end
  
  if impl.scissor_profile then
    impl.scissor_profile:Unbind ()
    impl.scissor_profile = nil
  end
end

local function ActivateDeactivateSprites (sprites_list)
  for i = 1, #sprites_list do
    local sprite = GetSprite(sprites_list [i])
    ActivateSprite (sprite)
    DeactivateSprite (sprite)
  end
end

local_methods.IsUserFriendAndWithoutGift = function (user_id)
  if GetUserId () == user_id then return false end

  local friends = GetFriends ()

  for i = 1, #friends do
    local friend = friends [i]
  
    if friend.user_id == user_id then
      local sent_gifts = GetSentGifts ()

      for j = 1, #sent_gifts do
        if friend.login_id == sent_gifts [j] then
          return false
        end
      end
    
      return true
    end
  end
  
  return false
end

local_methods.ProfileAchievementsSendGift = function (impl)
  local user_id = impl.profile_achievements_user_id
  local friends = GetFriends ()
  local login_id

  for i = 1, #friends do
    local friend = friends [i]
    
    if friend.user_id == user_id then
      login_id = friend.login_id
      break
    end
  end

  game.scripts.ui_windows.ShowSendGiftsWindow (function (gift_type)
    HsConnectionSendMessage (1, { event = "gifts.sent", friendsId = { login_id }, giftType = gift_type })

    game.scripts.ui_windows.ShowAlertWindow (ALERT_TYPE_GIFT, nil, nil, nil, { GetLocalizedString ("LOBBY.GIFT_SENT") }, nil, nil, true)
  end)
end

local_methods.ShowProfileWindow = function (impl, user_id)
  impl.profile_achievements_user_id = user_id
  
  if not impl.profile_window then
    PrepareAnimationScreenPart ("data/animations/profile_window.xml")

    local profile_window_sprite_list = { "profile_window_bg", "profile_window_bg_tab_profile_active", 
                                         "profile_window_show_profile_window", "profile_window_screen_lock",
                                         "profile_window_avatar_place", "profile_window_close_button", "profile_window_hand", "profile_window_bone_fortune",
                                         "profile_window_cons_underlay", "profile_window_diamond_underlay"}

    local profile_window_all_sprite_list = { "profile_window_bg", "profile_window_bg_tab_profile_active", "profile_window_screen_lock",
                                         "profile_window_add_friend_l", "profile_window_add_friend_p", "profile_window_send_gift_l", "profile_window_send_gift_p",
                                         "profile_window_get_cons_l", "profile_window_get_cons_p", "profile_window_get_diamonds_l", "profile_window_get_diamonds_p",
                                         "profile_window_avatar_place", "profile_window_diamond", "profile_window_cons", "profile_window_hand", "profile_window_bone_fortune",
                                         "profile_window_close_button", "profile_window_close_button_pressed", "profile_window_show_profile_window",
                                         "profile_window_cons_underlay", "profile_window_diamond_underlay" }

    if impl.phone_ui then
      PushBackTable (profile_window_sprite_list, { "profile_window_drag_zone", "profile_window_profile_line_1", 
        "profile_window_profile_line_2", "profile_window_profile_line_3", "profile_window_profile_line_4", 
        "profile_window_profile_line_5", "profile_window_profile_line_6", "profile_window_profile_line_7",
        "profile_window_profile_line_8", "profile_window_profile_line_9", "profile_window_profile_line_10", 
        "profile_window_profile_line_11", "profile_window_zone_no_pick_bank" } )
      PushBackTable (profile_window_all_sprite_list, { "profile_window_profile_line_1", "profile_window_profile_line_2", 
        "profile_window_profile_line_3", "profile_window_drag_zone", "profile_window_profile_line_4", "profile_window_profile_line_5", 
        "profile_window_profile_line_6", "profile_window_profile_line_7", "profile_window_profile_line_8", 
        "profile_window_profile_line_9", "profile_window_profile_line_10", "profile_window_profile_line_11", "profile_window_zone_no_pick_bank" })
    else
      table.insert (profile_window_sprite_list, "profile_window_profile_overlay")
      table.insert (profile_window_all_sprite_list, "profile_window_profile_overlay")
    end

    impl.profile_window = {}

    impl.profile_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.profile_window.resource       = "data/res/materials/profile_window_layers.xmtl"
    impl.profile_window.sprites        = profile_window_sprite_list
    impl.profile_window.text_lines     = game.scripts.ui_windows.GetTextLines ({ "profile_window_title_profile", "profile_window_text_birthday", "profile_window_text_city", "profile_window_title_achievements",
    "profile_window_text_name", "profile_window_text_value_birthday", "profile_window_text_value_city" })
    impl.profile_window.all_text_lines = game.scripts.ui_windows.GetTextLines ({ "profile_window_title_profile", "profile_window_text_birthday", "profile_window_text_city", "profile_window_title_achievements",
                                                                                 "profile_window_text_name", "profile_window_text_value_birthday", "profile_window_text_value_city", "profile_window_text_send_gift" })
    impl.profile_window.all_sprites    = game.scripts.ui_windows.MakeAllSprites (profile_window_all_sprite_list)
    impl.profile_window.on_hide        = function () OnHideProfileWindow (impl) end

    if impl.phone_ui then
      impl.profile_window.scroll_node = Scene.Node.Create ()
    end
  end

  if impl.phone_ui then
    impl.profile_window.scroll_node.Position = vec3 (0, 0, 0)
    impl.profile_window.scroll_view          = CreateScrollView (impl.profile_window.scroll_node, false, true, 0, game.current_screen.vars.profile_window_scroll_range, "profile_window_drag_zone")
  end

  local add_text_line_list = { "profile_window_text_reg_date", "profile_window_text_last_visit_day", "profile_window_text_balance", "profile_window_text_fortune", "profile_window_text_experience",
                               "profile_window_text_friends_rating", "profile_window_text_maximum_gain", "profile_window_text_maximum_gain_in_bonus", "profile_window_text_favorite_slot",
                               "profile_window_text_achievements", "profile_window_text_battle_wins_loses", "profile_window_text_battle_rating", "profile_window_text_value_balance_diamonds",
                               "profile_window_text_value_reg_date", "profile_window_text_value_last_visit_day", "profile_window_text_value_balance", "profile_window_text_value_fortune", "profile_window_text_value_experience", "profile_window_level",
                               "profile_window_text_value_friends_rating", "profile_window_text_value_maximum_gain", "profile_window_text_value_maximum_gain_in_bonus", "profile_window_text_value_favorite_slot",
                               "profile_window_text_value_achievements", "profile_window_text_value_battle_wins_loses", "profile_window_text_value_battle_rating"  }

  local add_text_line_list_value = { "profile_window_text_name", "profile_window_text_value_birthday", "profile_window_text_value_city", "profile_window_level", "profile_window_text_value_reg_date", "profile_window_text_value_last_visit_day",
                                     "profile_window_text_value_balance", "profile_window_text_value_fortune", "profile_window_text_value_experience", "profile_window_text_value_balance_diamonds",
                                     "profile_window_text_value_friends_rating", "profile_window_text_value_maximum_gain", "profile_window_text_value_maximum_gain_in_bonus", "profile_window_text_value_favorite_slot",
                                     "profile_window_text_value_achievements", "profile_window_text_value_battle_wins_loses", "profile_window_text_value_battle_rating" }

  local scrollable_sprites_list    = { "profile_window_profile_line_1", "profile_window_profile_line_2", "profile_window_profile_line_3", "profile_window_hand", "profile_window_bone_fortune",
                                       "profile_window_profile_line_4", "profile_window_profile_line_5", "profile_window_profile_line_6", "profile_window_profile_line_7",
                                       "profile_window_profile_line_8", "profile_window_profile_line_9", "profile_window_profile_line_10", "profile_window_profile_line_11",
                                       "profile_window_cons_underlay", "profile_window_diamond_underlay"}

  local scrollable_text_lines_list = { "profile_window_text_reg_date", "profile_window_text_last_visit_day", "profile_window_text_balance", "profile_window_text_fortune", "profile_window_text_experience",
                                       "profile_window_text_friends_rating", "profile_window_text_maximum_gain", "profile_window_text_maximum_gain_in_bonus", "profile_window_text_favorite_slot",
                                       "profile_window_text_achievements", "profile_window_text_battle_wins_loses", "profile_window_text_battle_rating", "profile_window_text_value_balance_diamonds",
                                       "profile_window_text_value_reg_date", "profile_window_text_value_last_visit_day", "profile_window_text_value_balance", "profile_window_text_value_fortune", "profile_window_text_value_experience",
                                       "profile_window_text_value_friends_rating", "profile_window_text_value_maximum_gain", "profile_window_text_value_maximum_gain_in_bonus", "profile_window_text_value_favorite_slot",
                                       "profile_window_text_value_achievements", "profile_window_text_value_battle_wins_loses", "profile_window_text_value_battle_rating" }

  if impl.phone_ui then
    table.insert (add_text_line_list, "profile_window_text_value_achievements_1")
    table.insert (add_text_line_list, "profile_window_text_value_achievements_2")
    table.insert (add_text_line_list_value, "profile_window_text_value_achievements_1")
    table.insert (add_text_line_list_value, "profile_window_text_value_achievements_2")
    table.insert (scrollable_text_lines_list, "profile_window_text_value_achievements_1")
    table.insert (scrollable_text_lines_list, "profile_window_text_value_achievements_2")
  end

  local add_sprite_list = {}

  impl.profile_user_id = user_id

  ExecuteOnNextUpdate (function ()
    ExecuteOnNextUpdate (function ()
      impl.profile_or_achievement_opened_from_tab = false
    end)
  end)

  ActivateSprite("profile_window_cons")
    impl.profile_window_cons = GetSprite ("profile_window_cons").position
  DeactivateSprite("profile_window_cons")

  ActivateDeactivateSprites ({"profile_window_diamond", "profile_window_get_cons_l", "profile_window_get_diamonds_l"})

  if local_methods.IsUserFriendAndWithoutGift (user_id) then
    table.insert (add_sprite_list, "profile_window_send_gift_l")
    table.insert (add_text_line_list, "profile_window_text_send_gift")
  end

  game.scripts.ui_windows.ShowWindow (impl.profile_window, add_sprite_list, add_text_line_list, {},function ()
    if impl.profile_window.scroll_node then
      impl.profile_window.scroll_node:BindToParent (game.current_screen.node)
      
      local scroll_zone        = GetSprite ("profile_window_drag_zone").layout
      local scroll_zone_top    = scroll_zone:Get ("Top") + 0
      local scroll_zone_height = scroll_zone:Get ("Height") + 0
      local camera             = impl.profile_window.camera.node
      local camera_width       = camera.Right - camera.Left

      impl.scissor_profile = CreateScissor (0, -scroll_zone_top, camera_width, scroll_zone_height)
      
      impl.scissor_profile:BindToParent (game.current_screen.node)

      for i = 1, #scrollable_text_lines_list do
        local text_line      = GetTextLine (scrollable_text_lines_list [i])
        local text_line_node = text_line.node
      
        text_line_node:BindToParent (impl.profile_window.scroll_node)
        
        text_line_node.Scissor = impl.scissor_profile
        
        if text_line.shadow_text_line then
          text_line.shadow_text_line.Scissor = impl.scissor_profile
        end
        
        if text_line.stroked_text_line then
          text_line.stroked_text_line.Scissor = impl.scissor_profile
        end
      end

      for i = 1, #scrollable_sprites_list do
        local sprite_node = GetSprite (scrollable_sprites_list [i]).node
      
        sprite_node:BindToParent (impl.profile_window.scroll_node)
        
        sprite_node.Scissor = impl.scissor_profile
      end
    end

    impl.profile_window.user_id  = user_id
    impl.profile_window.is_shown = true

    CleanTextLines (add_text_line_list_value)
    
    local_methods.LoadProfile (impl, user_id)
  end, nil, true, { dont_play_sound = impl.profile_or_achievement_opened_from_tab, hide_on_back = true })
end

local_methods.OnDragAchievements = function (impl, x) 
  if impl.achievements_window.drag_locked then return end

  local last_x = impl.achievements_window.drag_x

  if not last_x then return end 
  
  impl.achievements_window.drag_x = x  

  local dx = x - last_x

  local DRAG_THRESHOLD = 6

  if dx < -DRAG_THRESHOLD then 
    game.scripts.ui_windows.AchievementsWindowNextPage ()
  elseif dx > DRAG_THRESHOLD then
    game.scripts.ui_windows.AchievementsWindowPreviousPage ()
  end  
end

local_methods.OnDragAchievementsStarted = function (impl, x) 
  impl.achievements_window.drag_x = x  
end

local_methods.OnDragAchievementsEnded = function (impl) 
  impl.achievements_window.drag_x = nil  
end

local function CleanAchievementsWindowTab (impl)
  if impl.profile_avatar then
    impl.profile_avatar.avatar.OnHide ()
    impl.profile_avatar.avatar.GetSprite ():Unbind ()
  end
end

local function OnHideAchievementsWindow (impl)
  impl.waiting_user_achievements_id = nil
  impl.achievements_window.is_shown = false
  
  CleanAchievementsWindowTab (impl)
  
  if impl.achievements_window.update_progress_action then
    impl.achievements_window.update_progress_action.Stop ()
    impl.achievements_window.update_progress_action = nil
  end
  
  if impl.achievements_window.page_scroll_action then
    impl.achievements_window.page_scroll_action.Stop ()
    impl.achievements_window.page_scroll_action = nil
  end

  if impl.achievements_window.image_scissor_action then
    impl.achievements_window.image_scissor_action.Stop ()
    impl.achievements_window.image_scissor_action = nil
  end

  impl.achievements_window.pages_node:Unbind ()
  impl.achievements_window.page_scissor:Unbind ()

  if impl.profile_avatar then
    impl.profile_avatar.avatar.OnHide ()
    impl.profile_avatar.avatar.GetSprite ():Unbind ()
    impl.profile_avatar = nil
  end
end

local_methods.ShowAchievementsWindow = function (impl, user_id, group_id, enqueue_first)
  impl.profile_achievements_user_id = user_id

  if not impl.achievements_window then
    PrepareAnimationScreenPart ("data/animations/achievements_window.xml")

    local achievements_window_sprite_list = { "achievements_window_bg", "achievements_window_bg_tab_achievements_active", "achievements_window_screen_lock",
      "achievements_window_avatar_place", "achievements_window_close_button", "achievements_window_show_profile_window", "achievements_window_overlay_bg",
      "achievements_window_next_button_l", "achievements_window_back_button_l", "achievements_window_total_progress_bar", "achievements_window_pages_zone" }
    local achievements_window_all_sprite_list = { "achievements_window_bg", "achievements_window_bg_tab_achievements_active", "achievements_window_screen_lock",
      "achievements_window_add_friend_l", "achievements_window_add_friend_p", "achievements_window_send_gift_l", "achievements_window_send_gift_p",
      "achievements_window_avatar_place", "achievements_window_close_button", "achievements_window_close_button_pressed", "achievements_window_show_profile_window",
      "achievements_window_back_button_l", "achievements_window_back_button_p", "achievements_window_pages_zone",
      "achievements_window_next_button_l", "achievements_window_next_button_p", "achievements_window_total_progress_bar", "achievements_window_next_button_d" , "achievements_window_back_button_d",
			"achievements_window_overlay_bg", "achievements_window_bnt_l_overlay", "achievements_window_bnt_p_overlay" }

    local text_lines = { "achievements_window_title_profile", "achievements_window_text_birthday", "achievements_window_text_city",
      "achievements_window_text_name", "achievements_window_text_value_birthday", "achievements_window_text_value_city",
      "achievements_window_level", "achievements_window_title_achievements", "achievements_window_current_achievement_title",
      "achievements_window_all_achievements_text", "achievements_window_all_achievements_count" }
    local all_text_lines = { "achievements_window_title_profile", "achievements_window_text_birthday", "achievements_window_text_city", "achievements_window_text_send_gift",
      "achievements_window_text_name", "achievements_window_text_value_birthday", "achievements_window_text_value_city",
      "achievements_window_level", "achievements_window_title_achievements",
      "achievements_window_all_achievements_text", "achievements_window_all_achievements_count", "achievements_window_current_achievement_title" }

    local achievements_window_text_line_list = {}

    impl.achievements_window = {}

		impl.achievements_window.overlay_sprites_hide = {"achievements_window_overlay_bg" }
		impl.achievements_window.overlay_sprites_show = {"achievements_window_overlay_bg" }
		impl.achievements_window.overlay_text_line_hide = {}

		for i = 1, 5 do
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_image_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_image_underlay_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_image_small_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_coins_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_button_p_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_button_l_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_button_underlay_" .. i)
			table.insert(achievements_window_all_sprite_list, "achievements_window_all_achiev_coins_" .. i .. "_p")

			table.insert(all_text_lines, "achievements_window_all_achiev_text_coin_" .. i)
			table.insert(all_text_lines, "achievements_window_all_achiev_text_coin_3" .. i .. "_p")
			table.insert(all_text_lines, "achievements_window_all_achiev_text_progress_" .. i)
			table.insert(all_text_lines, "achievements_window_all_achiev_text_progress_" .. i .. "_p")
			table.insert(all_text_lines, "achievements_window_all_achiev_text_name_" .. i)
			table.insert(all_text_lines, "achievements_window_all_achiev_text_name_" .. i .. "_p")

			table.insert(achievements_window_sprite_list, "achievements_window_all_achiev_image_underlay_" .. i)
			table.insert(achievements_window_sprite_list, "achievements_window_all_achiev_image_small_" .. i)
			table.insert(achievements_window_sprite_list, "achievements_window_all_achiev_button_underlay_" .. i)

			table.insert(impl.achievements_window.overlay_text_line_hide, "achievements_window_all_achiev_text_coin_" .. i)
			table.insert(impl.achievements_window.overlay_text_line_hide, "achievements_window_all_achiev_text_coin_3" .. i .. "_p")
			table.insert(impl.achievements_window.overlay_text_line_hide, "achievements_window_all_achiev_text_progress_" .. i)
			table.insert(impl.achievements_window.overlay_text_line_hide, "achievements_window_all_achiev_text_progress_" .. i .. "_p")
			table.insert(impl.achievements_window.overlay_text_line_hide, "achievements_window_all_achiev_text_name_" .. i)
			table.insert(impl.achievements_window.overlay_text_line_hide, "achievements_window_all_achiev_text_name_" .. i .. "_p")

			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_image_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_image_underlay_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_image_small_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_coins_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_coins_" .. i .. "_p")
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_button_p_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_button_l_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_hide, "achievements_window_all_achiev_button_underlay_" .. i)

			table.insert(impl.achievements_window.overlay_sprites_show, "achievements_window_all_achiev_image_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_show, "achievements_window_all_achiev_image_underlay_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_show, "achievements_window_all_achiev_image_small_" .. i)
			table.insert(impl.achievements_window.overlay_sprites_show, "achievements_window_all_achiev_button_underlay_" .. i)
		end

    impl.achievements_window.progress_bars_sprites = { "achievements_window_pages_achievement_2_progress_bar_fill",
      "achievements_window_pages_achievement_3_progress_bar_fill", "achievements_window_pages_achievement_5_progress_bar_fill",
      "achievements_window_pages_achievement_6_progress_bar_fill", "achievements_window_pages_achievement_7_progress_bar_fill",
      "achievements_window_pages_achievement_8_progress_bar_fill", "achievements_window_pages_achievement_12_progress_bar_fill",
      "achievements_window_pages_achievement_15_progress_bar_fill", "achievements_window_pages_achievement_22_progress_bar_fill",
      "achievements_window_pages_achievement_23_progress_bar_fill", "achievements_window_pages_achievement_24_progress_bar_fill",
      "achievements_window_pages_achievement_25_progress_bar_fill"}

    impl.achievements_window.progress_bars_scissors = {}

    impl.achievements_window.pages_node = Scene.Node.Create ()

    impl.achievements_window.pages = {}

    local pages_zone_layout = GetSprite ("achievements_window_pages_zone").layout
    local pages_zone_width  = pages_zone_layout:Get ("Width") + 0
    local camera            = game.current_screen.cameras.modal_window_camera.node
    local camera_height     = camera.Top - camera.Bottom


    impl.achievements_window.clip_left  = pages_zone_layout:Get ("Left") + 0
    impl.achievements_window.clip_width = pages_zone_width

    local page_scissor = CreateScissor (impl.achievements_window.clip_left, 0, impl.achievements_window.clip_width, camera_height)

    for i = 1, ACHIEVEMENTS_GROUPS_COUNT do
      local page = {}

      page.node = Scene.Node.Create ()

      page.node.Position = vec3 (pages_zone_width * (i - 1), 0, 0)

      page.node:BindToParent (impl.achievements_window.pages_node)

      page.image_sprite = GetSprite ("achievements_window_page_image_" .. i)

			page.overlay_image_sprite =  GetSprite ("achievements_window_all_achiev_image_" .. i)

      page.overlay_image_sprite_scissor_height = page.overlay_image_sprite.layout:Get ("Height") + 0
			page.overlay_image_sprite_scissor_bottom = -page.overlay_image_sprite.layout:Get ("Top") - page.overlay_image_sprite_scissor_height

      page.image_sprite_scissor_height = page.image_sprite.layout:Get ("Height") + 0
      page.image_sprite_scissor_bottom = -page.image_sprite.layout:Get ("Top") - page.image_sprite_scissor_height

      local sprite_list = { "achievements_window_page_icon_" .. i, "achievements_window_page_image_" .. i,
        "achievements_window_page_background_" .. i, "achievements_window_pages_icon_underlay_" .. i,
        "achievements_window_pages_coin_" .. i, "achievements_window_pages_line_1_" .. i,
        "achievements_window_pages_line_2_" .. i, "achievements_window_pages_line_3_" .. i,
        "achievements_window_pages_line_4_" .. i, "achievements_window_pages_achievement_1_unchecked_" .. i,
        "achievements_window_pages_achievement_2_unchecked_" .. i, "achievements_window_pages_achievement_3_unchecked_" .. i,
        "achievements_window_pages_achievement_4_unchecked_" .. i, "achievements_window_pages_achievement_5_unchecked_" .. i,
					"achievements_window_bnt_l_overlay_" .. i}

      local optional_sprite_list = { "achievements_window_pages_show_checked_" .. i, "achievements_window_pages_show_empty_" .. i,
        "achievements_window_pages_achievement_1_checked_" .. i, "achievements_window_pages_achievement_2_checked_" .. i,
        "achievements_window_pages_achievement_3_checked_" .. i, "achievements_window_pages_achievement_4_checked_" .. i,
        "achievements_window_pages_achievement_5_checked_" .. i, "achievements_window_pages_achievement_2_progress_bar_fill",
        "achievements_window_pages_achievement_2_progress_bar", "achievements_window_pages_achievement_3_progress_bar_fill",
        "achievements_window_pages_achievement_3_progress_bar", "achievements_window_pages_achievement_5_progress_bar_fill",
        "achievements_window_pages_achievement_5_progress_bar", "achievements_window_pages_achievement_6_progress_bar_fill",
        "achievements_window_pages_achievement_6_progress_bar", "achievements_window_pages_achievement_7_progress_bar_fill",
        "achievements_window_pages_achievement_7_progress_bar", "achievements_window_pages_achievement_8_progress_bar_fill",
        "achievements_window_pages_achievement_8_progress_bar", "achievements_window_pages_achievement_12_progress_bar_fill",
        "achievements_window_pages_achievement_12_progress_bar", "achievements_window_pages_achievement_15_progress_bar_fill",
        "achievements_window_pages_achievement_15_progress_bar","achievements_window_pages_achievement_22_progress_bar_fill",
        "achievements_window_pages_achievement_22_progress_bar","achievements_window_pages_achievement_23_progress_bar_fill",
        "achievements_window_pages_achievement_23_progress_bar","achievements_window_pages_achievement_24_progress_bar_fill",
        "achievements_window_pages_achievement_24_progress_bar","achievements_window_pages_achievement_25_progress_bar_fill",
        "achievements_window_pages_achievement_25_progress_bar", "achievements_window_pages_achievement_7_button",
        "achievements_window_pages_achievement_7_button_p", "achievements_window_pages_achievement_19_button",
        "achievements_window_pages_achievement_19_button_p" }

      local text_line_list = { "achievements_window_pages_progress_" .. i, "achievements_window_pages_reward_" .. i,
        "achievements_window_pages_achievement_title_1_line_1_" .. i, "achievements_window_pages_achievement_title_1_line_2_" .. i,
        "achievements_window_pages_achievement_title_1_line_3_" .. i, "achievements_window_pages_achievement_title_2_line_1_" .. i,
        "achievements_window_pages_achievement_title_2_line_2_" .. i, "achievements_window_pages_achievement_title_2_line_3_" .. i,
        "achievements_window_pages_achievement_title_3_line_1_" .. i, "achievements_window_pages_achievement_title_3_line_2_" .. i,
        "achievements_window_pages_achievement_title_3_line_3_" .. i, "achievements_window_pages_achievement_title_4_line_1_" .. i,
        "achievements_window_pages_achievement_title_4_line_2_" .. i, "achievements_window_pages_achievement_title_4_line_3_" .. i,
        "achievements_window_pages_achievement_title_5_line_1_" .. i, "achievements_window_pages_achievement_title_5_line_2_" .. i,
        "achievements_window_pages_achievement_title_5_line_3_" .. i }

      local optional_text_line_list = { "achievements_window_pages_show_" .. i, "achievements_window_pages_achievement_2_progress",
        "achievements_window_pages_achievement_3_progress", "achievements_window_pages_achievement_5_progress",
        "achievements_window_pages_achievement_6_progress", "achievements_window_pages_achievement_8_progress",
        "achievements_window_pages_achievement_12_progress", "achievements_window_pages_achievement_15_progress",
        "achievements_window_pages_achievement_22_progress", "achievements_window_pages_achievement_23_progress",
        "achievements_window_pages_achievement_24_progress", "achievements_window_pages_achievement_25_progress",
        "achievements_window_pages_achievement_7_button", "achievements_window_pages_achievement_19_button" }

      page.sprite_list = sprite_list
      page.text_line_list = text_line_list

      local function ProcessSprite (sprite_name, active)
        GetSprite (sprite_name).node.Scissor = page_scissor

        if active then
          table.insert (achievements_window_sprite_list, sprite_name)
          table.insert (page.sprite_list, sprite_name)
        end

        table.insert (achievements_window_all_sprite_list, sprite_name)
      end

      local function ProcessTextLine (name, active)
        local line = GetTextLine (name)
        
        line.node.Scissor = page_scissor

        if line.shadow_text_line then
          line.shadow_text_line.Scissor = page_scissor
        end

        if line.stroked_text_line then
          line.stroked_text_line.Scissor = page_scissor
        end

        if active then
          table.insert (page.text_line_list, name)
        end

        table.insert (all_text_lines, name)
      end

      for j = 1, #sprite_list do
        ProcessSprite (sprite_list [j], true)
      end

      for j = 1, #optional_sprite_list do
        ProcessSprite (optional_sprite_list [j])
      end

      for j = 1, #text_line_list do
        ProcessTextLine (text_line_list [j], true)
      end

      for j = 1, #optional_text_line_list do
        ProcessTextLine (optional_text_line_list [j])
      end

      table.insert (impl.achievements_window.pages, page)
    end

    impl.achievements_window.page_scissor   = page_scissor
    impl.achievements_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.achievements_window.resource       = "data/res/materials/achievements_window_layers.xmtl"
    impl.achievements_window.sprites        = achievements_window_sprite_list
    impl.achievements_window.text_lines     = game.scripts.ui_windows.GetTextLines (text_lines)
    impl.achievements_window.all_text_lines = game.scripts.ui_windows.GetTextLines (all_text_lines)
    impl.achievements_window.all_sprites    = game.scripts.ui_windows.MakeAllSprites (achievements_window_all_sprite_list)
    impl.achievements_window.on_hide        = function () OnHideAchievementsWindow (impl) end
    impl.achievements_window.page_width     = pages_zone_width
  end

  local add_sprite_list = {}

  impl.achievements_window.user_id      = user_id
  impl.achievements_window.current_page = math.max (1, math.min (ACHIEVEMENTS_GROUPS_COUNT, group_id or 1))

  local add_text_line_list = {}

  local add_text_line_list_value = { "achievements_window_text_name",
    "achievements_window_text_value_birthday", "achievements_window_text_value_city", "achievements_window_level" }

  impl.achievements_window.pages_node:BindToParent (game.current_screen.node)
  impl.achievements_window.pages_node.Position = vec3 (0, 0, 0)

  impl.achievements_window.achievements_data = nil
  impl.achievements_window.drag_locked = nil

  ExecuteOnNextUpdate (function ()
    ExecuteOnNextUpdate (function ()
      impl.profile_or_achievement_opened_from_tab = false
    end)
  end)

  if local_methods.IsUserFriendAndWithoutGift (user_id) then
    table.insert (add_sprite_list, "achievements_window_send_gift_l")
    table.insert (add_text_line_list, "achievements_window_text_send_gift")
  end

  game.scripts.ui_windows.ShowWindow (impl.achievements_window, add_sprite_list, add_text_line_list, {}, function ()
    impl.achievements_window.page_scissor:BindToScene (impl.achievements_window.camera.node.Scene)
  
    for i = 1, #impl.achievements_window.pages do
      local page = impl.achievements_window.pages [i]
      
      for j = 1, #page.sprite_list do
        local sprite_name = page.sprite_list [j]
        local sprite_node = GetSprite (sprite_name).node
        
        sprite_node:BindToParent (page.node)
      end
    end
  
    impl.achievements_window.is_shown = true
    AchievementsScrollToPage (impl, impl.achievements_window.current_page)
    CleanTextLines (add_text_line_list_value)
    local_methods.LoadAchievements (impl, user_id)
  end, nil, true, { enqueue_first = enqueue_first, dont_play_sound = impl.profile_or_achievement_opened_from_tab, hide_on_back = true })
end

local_methods.AchievementsWindowNextPage = function (impl)
  if impl.achievements_window.current_page > ACHIEVEMENTS_GROUPS_COUNT - 1 then return end

  PlaySound ('click')

  AchievementsScrollToPage (impl, impl.achievements_window.current_page + 1, true)
end

local_methods.AchievementsWindowPreviousPage = function (impl)
  if impl.achievements_window.current_page < 2 then return end

  PlaySound ('click')

  AchievementsScrollToPage (impl, impl.achievements_window.current_page - 1, true)
end

local_methods.AchievementsWindowSetPage = function (impl, page)
	if page > ACHIEVEMENTS_GROUPS_COUNT then return end

	PlaySound ('click')
	for i = 1, #impl.achievements_window.overlay_text_line_hide do
		DeactivateTextLine (impl.achievements_window.overlay_text_line_hide[i])
	end

	for i = 1, #impl.achievements_window.overlay_sprites_hide do
		DeactivateSprite (impl.achievements_window.overlay_sprites_hide[i])
	end
	AchievementsScrollToPage (impl, page, true)

end

local_methods.AchievementsWindowSwitchToProfile = function (impl)
  if impl.profile_or_achievement_opened_from_tab then return end

  PlaySound ('click')

  impl.profile_or_achievement_opened_from_tab = true
  
  HideWindow (impl, nil, true)
  game.scripts.ui_windows.ShowProfileWindow (impl.achievements_window.user_id)
end

local_methods.ProfileWindowShowAchievements = function (impl)
  if impl.profile_or_achievement_opened_from_tab then return end

  PlaySound ('click')

  impl.profile_or_achievement_opened_from_tab = true
  
  HideWindow (impl, nil, true)
  game.scripts.ui_windows.ShowAchievementsWindow (impl.profile_window.user_id)
end

local function LoadTopNotificationWindow (impl)
  if impl.top_notification_window then return end

  PrepareAnimationScreenPart ("data/animations/top_notification.xml")

   local top_notification_all_text_list = { "top_notification_title", "top_notification_text_1", 
                                            "top_notification_text_2", "top_notification_text_3",
                                            "top_notification_text_4", "top_notification_text_5", 
                                            "top_notification_text_6", "top_notification_text_7", 
                                            "top_notification_text_6_small", "top_notification_text_7_small", 
                                            "top_notification_text_share_friends", "top_notification_text_button"
                                           }

  impl.top_notification_window = {}
                        
  impl.top_notification_window.camera         = game.current_screen.cameras.modal_window_camera
  impl.top_notification_window.resource       = "data/res/materials/top_notification_layers.xmtl"
  impl.top_notification_window.sprites        = { "top_notification_background", "top_notification_Girl", 
    "top_notification_close_btn", "top_notification_play_more_button", "top_notification_window_screen_lock", 
    "top_notification_check", "top_notification_check_box_zone"}
  impl.top_notification_window.text_lines     = GetTextLines (top_notification_all_text_list)
  impl.top_notification_window.all_text_lines = GetTextLines (top_notification_all_text_list)
  impl.top_notification_window.all_sprites    = MakeAllSprites ({ "top_notification_background", "top_notification_Girl", 
    "top_notification_close_btn", "top_notification_close_btn_press", "top_notification_play_more_button", 
    "top_notification_play_more_button_press", "top_notification_window_screen_lock", "top_notification_check", 
    "top_notification_check_box_zone" })
  impl.top_notification_window.on_hide        = function () HideWindow (impl) end
end

local_methods.ShowTopNotificationWindow = function (impl, info) 
  LoadTopNotificationWindow (impl)

  local type
  local share_key
  local share_type
  
  local text_share_coins    = GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_SHARE_FRIENDS_COINS")
  local text_share_diamonds = GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_SHARE_FRIENDS_DIAMONDS")
    
  if info.place < 4 then
    type = 1
    
    if info.top_type == 1 then
      share_type = SHARE_TOP_DIAMONDS_1_TYPE
      share_key  = text_share_diamonds
    else
      share_type = SHARE_TOP_COINS_1_TYPE
      share_key  = text_share_coins
    end
  elseif info.place < 11 then
    type = 2
    
    if info.top_type == 1 then
      share_type = SHARE_TOP_DIAMONDS_10_TYPE
      share_key  = text_share_diamonds
    else
      share_type = SHARE_TOP_COINS_10_TYPE
      share_key  = text_share_coins
    end
  else
    type = 3
    
    if info.top_type == 1 then
      share_type = SHARE_TOP_DIAMONDS_100_TYPE
      share_key  = text_share_diamonds
    else
      share_type = SHARE_TOP_COINS_100_TYPE
      share_key  = text_share_coins
    end
  end

  local share_values = GetSocialManager ().GetShareValues (share_type)
  
  ShowWindow (impl, impl.top_notification_window, {}, {}, {}, function ()
    impl.top_notification_window.checked = true
    
    ExecuteOnNextUpdate (function ()
      ExecuteOnNextUpdate (function ()
        ActivateSprites ("top_notification_check_box_zone") -- this sprite not active if same window was closed before
      end)
    end)

    impl.top_notification_window.share_type  = share_type
    impl.top_notification_window.place       = info.place
    impl.top_notification_window.rating_name = GetLocalizedString ("WINNER_OF_BEST_PLAYER.RATING_NAME_" .. info.top_type)

    GetTextLine ("top_notification_title").node.Text = GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_TITLE_" .. type)
    GetTextLine ("top_notification_text_button").node.Text = GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_BUTTON_" .. type)
  
    local replacements = {
      { token = "{place}",  replacement = tostring (info.place) },
      { token = "{rating_name}",  replacement = GetLocalizedString ("WINNER_OF_BEST_PLAYER.RATING_NAME_" .. info.top_type) },
      { token = "{friends}", replacement = share_values.players_count },
      { token = "{coins}",   replacement = share_values.chips_count }
    }

    GetTextLine ("top_notification_text_share_friends").node.Text = ProcessStringWithReplacements (share_key, replacements)

    for i = 1, 7 do
      GetTextLine("top_notification_text_" .. i).node.Text = ProcessStringWithReplacements (GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_LINE_".. i .. "_" .. type), replacements)
    end  
    
    if type == 1 then
      GetTextLine("top_notification_text_6_small").node.Text = ProcessStringWithReplacements (GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_LINE_6_1_SMALL"), replacements)
      GetTextLine("top_notification_text_7_small").node.Text = ProcessStringWithReplacements (GetLocalizedString ("BATTLE_INFO_WINDOW.TOP_NOTIFICATION_LINE_7_1_SMALL"), replacements)
    else
      GetTextLine("top_notification_text_6_small").node.Text = ""
      GetTextLine("top_notification_text_7_small").node.Text = ""
    end
  end, nil, nil, nil, nil, nil, nil, true, nil, nil, { hide_on_back = true })
end

local_methods.TopNotificationWindowHide = function (impl)
  local share_checked = impl.top_notification_window.checked
  local share_type    = impl.top_notification_window.share_type
  local place         = impl.top_notification_window.place
  local rating_name   = impl.top_notification_window.rating_name
 
  HideWindow (impl, nil, share_checked)
  
  if share_checked then
    GetSocialManager ().ShareTop (share_type, place, rating_name)
  end
end

local_methods.TopNotificationCheck = function (impl)
  if not impl.top_notification_window.checked then
    impl.top_notification_window.checked = true
    ActivateSprite("top_notification_check")
  else
    impl.top_notification_window.checked = nil
    DeactivateSprite("top_notification_check")
  end
end


local_methods.OnHidePopupTopWindow = function (impl)
  if not impl.phone_ui and game.current_screen.name == "lobby" then
    impl.popup_top_window.camera.node.Position = vec3 (0, 0, 0)
  end
end

local function ShowPopupTopWindow (impl)
  if not impl.popup_top_window then
    PrepareAnimationScreenPart ("data/animations/popup_top_window.xml")

    local popup_top_window_sprite_list = { "popup_top_window_bg", "popup_top_window_screen_lock" }
    local popup_top_window_text_line_list = { "popup_top_window_text_next_level", "popup_top_window_text_max_bet", "popup_top_window_text_bonus_next_level", "popup_top_window_text_super_bonus",
                                              "popup_top_window_text_mega_bonus", "popup_top_window_text_title", "popup_top_window_count_next_level", "popup_top_window_count_max_bet",
                                              "popup_top_window_count_bonus_next_level_diamond",
                                              "popup_top_window_count_bonus_next_level_chips", "popup_top_window_count_super_bonus_chips", "popup_top_window_count_mega_bonus_chips" }
    impl.popup_top_window = {}

    impl.popup_top_window.camera         = game.current_screen.cameras.modal_window_camera
    impl.popup_top_window.resource       = "data/res/materials/popup_top_window_layers.xmtl"
    impl.popup_top_window.sprites        = popup_top_window_sprite_list
    impl.popup_top_window.text_lines     = GetTextLines (popup_top_window_text_line_list)
    impl.popup_top_window.all_text_lines = GetTextLines (popup_top_window_text_line_list)
    impl.popup_top_window.all_sprites    = MakeAllSprites (popup_top_window_sprite_list)
    impl.popup_top_window.on_hide        = function () local_methods.OnHidePopupTopWindow (impl) end
  end
  
  local cline_text_line_list = {"popup_top_window_text_title", "popup_top_window_count_next_level", "popup_top_window_count_max_bet",
                               "popup_top_window_count_bonus_next_level_diamond",
                               "popup_top_window_count_bonus_next_level_chips", "popup_top_window_count_super_bonus_chips", "popup_top_window_count_mega_bonus_chips" }

  local user_level      = GetUserLevel ()
  local next_user_level = GetNextUserLevel ()
  local user            = GetUser

  local bet_money = GetBetMoney (user_level.max_bet_id)

  CleanTextLines (cline_text_line_list)

  ShowWindow (impl, impl.popup_top_window, {}, {}, {}, function ()
    GetTextLine ("popup_top_window_text_title").node.Text = FormatMoneyString (GetUserWagering (), nil, false) .. "/" .. FormatMoneyString (next_user_level.wagering)
    GetTextLine ("popup_top_window_count_next_level").node.Text = tostring (user_level.id + 1)
    GetTextLine ("popup_top_window_count_max_bet").node.Text = FormatMoneyString (bet_money)
    GetTextLine ("popup_top_window_count_bonus_next_level_diamond").node.Text = tostring (next_user_level.bonus.diamonds)
    GetTextLine ("popup_top_window_count_bonus_next_level_chips").node.Text = FormatMoneyString (next_user_level.bonus.coins, 6, nil, true)
    GetTextLine ("popup_top_window_count_super_bonus_chips").node.Text = FormatMoneyString (next_user_level.super_bonus.coins, 6, nil, true)
    GetTextLine ("popup_top_window_count_mega_bonus_chips").node.Text = FormatMoneyString (next_user_level.mega_bonus.coins, 6, nil, true)
  end, false, true)
end

local function OnHideAchievementsGroupCompleteWindow (impl)
  if impl.achievements_group_complete_check then
    HsConnectionSendMessage (1, { event = "achievement.select", id = impl.achievements_group_complete.group_id, checked = true })
  end

  if impl.achiev_group_avatar then
    impl.achiev_group_avatar.OnHide ()
    impl.achiev_group_avatar.GetSprite ():Unbind ()
    impl.achiev_group_avatar = nil
  end
  
  game.scripts.top_ui.MoneyForceUpdate (GetUserMoney ())
  game.scripts.top_ui.DiamondsForceUpdate (GetUserDiamonds ())
end

local function AchievementsGroupCompleteCheck (impl)
  if not impl.achievements_group_complete_check then
    impl.achievements_group_complete_check = true
    ActivateSprite("achievements_group_complete_check")
  else
    impl.achievements_group_complete_check = nil
    DeactivateSprite("achievements_group_complete_check")
  end
end

local function ShowAchievementsGroupComplete (impl, group_id, prize, prize_type)
  if group_id > ACHIEVEMENTS_GROUPS_COUNT then return end --IGNORE NEW ACHIEVEMENTS

  if not impl.achievements_group_complete then
    PrepareAnimationScreenPart ("data/animations/achievements_group_complete.xml")

    local achievements_group_complete_sprite_list = { "achievements_group_complete_bg",
      "achievements_group_complete_screen_lock", 'achievements_group_complete_coins', 'achievements_group_complete_diamond',
      "achievements_group_complete_right_button_l", "achievements_group_complete_right_button_p",
      "achievements_group_complete_check", "achievements_group_complete_avatar", "achievements_group_complete_check_button_zone",
      "achievements_group_complete_achievement_0", "achievements_group_complete_achievement_1", "achievements_group_complete_achievement_2",
      "achievements_group_complete_achievement_3", "achievements_group_complete_achievement_4", "achievements_group_complete_achievement_5",
      "achievements_group_complete_window_check", "achievements_group_complete_window_button_check_zone"}
    local achievements_group_complete_text_line_list = { "achievements_group_complete_text_title_1", "achievements_group_complete_text_title_2", "achievements_group_complete_text_title_3",
                                                         "achievements_group_complete_text_show_check", "achievements_group_complete_text_share_friends", "achievements_group_complete_text_chips",
                                                         "achievements_group_complete_text_left_button", "achievements_group_complete_text_right_button"}
    impl.achievements_group_complete = {}

    impl.achievements_group_complete.camera         = game.current_screen.cameras.modal_window_camera
    impl.achievements_group_complete.resource       = "data/res/materials/achievements_group_complete_layers.xmtl"
    impl.achievements_group_complete.sprites        = { "achievements_group_complete_bg",
      "achievements_group_complete_avatar", "achievements_group_complete_right_button_l", 
      "achievements_group_complete_screen_lock", "achievements_group_complete_check_button_zone", "achievements_group_complete_check",
      "achievements_group_complete_window_check", "achievements_group_complete_window_button_check_zone"}
    impl.achievements_group_complete.text_lines     = GetTextLines (achievements_group_complete_text_line_list)
    impl.achievements_group_complete.all_text_lines = GetTextLines (achievements_group_complete_text_line_list)
    impl.achievements_group_complete.all_sprites    = MakeAllSprites (achievements_group_complete_sprite_list)
    impl.achievements_group_complete.on_hide        = function () OnHideAchievementsGroupCompleteWindow (impl) end
  end

  local coin_or_diamond
  local share_key
  local share_type
  
  if prize_type == 1 then
    coin_or_diamond = 'achievements_group_complete_diamond'
    prize           = prize * 100
    share_type      = SHARE_ACHIEVEMENTS_GROUP_DIAMONDS_TYPE
    share_key       = "ACHIEVEMENTS_GROUP_COMPLETE.SHARE_FRIENDS_DIAMONDS"
  else
    coin_or_diamond = 'achievements_group_complete_coins'
    share_type      = SHARE_ACHIEVEMENTS_GROUP_COINS_TYPE
    share_key       = "ACHIEVEMENTS_GROUP_COMPLETE.SHARE_FRIENDS"
  end

  local share_values = GetSocialManager ().GetShareValues (share_type)

  local text         = ProcessStringWithReplacements (GetLocalizedString (share_key), {
    { token = "{friends}", replacement = share_values.players_count },
    { token = "{coins}",   replacement = share_values.chips_count },
  })

  ShowWindow (impl, impl.achievements_group_complete, { "achievements_group_complete_achievement_" .. group_id, coin_or_diamond }, {}, {},function ()
    impl.achievements_group_complete.group_id    = group_id
    impl.achievements_group_complete.prize       = prize
    impl.achievements_group_complete.share_type  = share_type
    impl.achievements_group_complete_check       = true
    impl.achievements_group_complete_share_check = true
    
    ExecuteOnNextUpdate (function ()
      ExecuteOnNextUpdate (function ()
        ActivateSprite ("achievements_group_complete_window_button_check_zone") -- this sprite not active if same window was closed before
      end)
    end)

    local window        = Engine.Windows.Find (WINDOW_NAME)
    local camera        = impl.achievements_group_complete.camera.node
    local camera_width  = camera.Right - camera.Left
    local camera_height = camera.Top - camera.Bottom
    local screen_scale  = window.ClientHeight / camera_height

    local avatar_ref  = GetSprite ("achievements_group_complete_avatar")
    local avatar_size = { x = avatar_ref.layout:Get ("Width") + 0, y = avatar_ref.layout:Get ("Height") + 0 }
    local avatar      = CreateUrlSprite (URL_SPRITE_MODE_CACHE_FILE, GetSocialManager ().GetPhoto (), {}, 
                                         { material = avatar_ref.node.Material, tex_size = StringToVec (avatar_ref.layout:Get ("TexSize")), tex_offset = StringToVec (avatar_ref.layout:Get ("TexOffset")) }, 
                                         avatar_size, avatar_size.x * PHOTO_ACTIVITY_INDICATOR_FACTOR, screen_scale)

    local avatar_sprite = avatar.GetSprite ()

    avatar_sprite.Position = vec3 (avatar_ref.layout:Get ("Left") + avatar_size.x / 2, -avatar_ref.layout:Get ("Top") - avatar_size.y / 2, avatar_ref.z - EPS)

    avatar_sprite:BindToParent (game.current_screen.node)

    avatar.OnShow ()

    impl.achiev_group_avatar = avatar

    GetTextLine("achievements_group_complete_text_chips").node.Text         = FormatMoneyString (impl.achievements_group_complete.prize, 6, nil, true)
    GetTextLine("achievements_group_complete_text_title_1").node.Text       = GetLocalizedString ("ACHIEVEMENTS_GROUP_COMPLETE.TITLE_1_" .. group_id)
    GetTextLine("achievements_group_complete_text_title_2").node.Text       = GetLocalizedString ("ACHIEVEMENTS_GROUP_COMPLETE.TITLE_2_" .. group_id)
    GetTextLine("achievements_group_complete_text_share_friends").node.Text = text
  end, nil, nil, nil, nil, nil, nil, true)
end

local function AchievementGroupCompleteWindowShare (impl)
  if not impl.achievements_group_complete_share_check then
    impl.achievements_group_complete_share_check = true
    ActivateSprite("achievements_group_complete_window_check")
  else
    impl.achievements_group_complete_share_check = nil
    DeactivateSprite("achievements_group_complete_window_check")
  end
end

local_methods.AchievementsGroupCompleteShare = function (impl) 
  local share_checked = impl.achievements_group_complete_share_check
  
  HideWindow (impl, nil, share_checked)
  
  if share_checked then
    GetSocialManager ().ShareAchievementGroup (impl.achievements_group_complete.group_id, impl.achievements_group_complete.prize, impl.achievements_group_complete.share_type)
  end
end

local function Update (impl)
  local current_time = game.main_timer.GetTime ()
  local dt           = current_time - impl.last_update_time

  impl.last_update_time = current_time

  if impl.send_gifts_friends_scroll_view then
    impl.send_gifts_friends_scroll_view.Update (dt)
  end

  if impl.send_gifts_selected_friends_scroll_view then
    impl.send_gifts_selected_friends_scroll_view.Update (dt)
  end
  
  if impl.profile_window and impl.profile_window.scroll_view then
    impl.profile_window.scroll_view.Update (dt)
  end
end

local_methods.LoginToFb = function (impl)
  LogFlurryEvent ("ButtonClick", { ButtonName = "OptionsFBLogin" })  

  local social_manager = GetSocialManager ()

  if social_manager.IsInGuestMode () then
    PerformLogOut (FacebookLogin)
    
    return
  end
  
  local texts_override
  
  if social_manager.GetNetworkName () == "FB" then
    local user_info = social_manager.GetUserInfo ()
    
    if user_info.is_facebook_app_installed then
      texts_override = { GetLocalizedString ("CHANGE_FB_ACCOUNT_LINE_1"), GetLocalizedString ("CHANGE_FB_ACCOUNT_LINE_2"), GetLocalizedString ("CHANGE_FB_ACCOUNT_LINE_3") }
    end
  end
  
  game.scripts.ui_windows.ShowLogoutWindow (FacebookLogin, texts_override)
end

local_methods.LoginToVk = function (impl)
  LogFlurryEvent ("ButtonClick", { ButtonName = "OptionsVKLogin" })  

  if GetSocialManager ().IsInGuestMode () then
    PerformLogOut (VKLogin)
    
    return
  end
  
  game.scripts.ui_windows.ShowLogoutWindow (VKLogin)
end

local_methods.GooglePlusRecommend = function (impl)
  game.scripts.ui_windows.HideWindow ()

  local button_index = {}

  button_index ["de-DE"] = "DE"
  button_index ["en-US"] = "US"
  button_index ["es-ES"] = "ES"
  button_index ["fr-FR"] = "FR"
  button_index ["it-IT"] = "IT"
  button_index ["ru-RU"] = "RU"
  button_index ["tr-TR"] = "TR"

  System.Application.PostNotification ("ShowOneButtonActivity " .. button_index [GetCurrentLocale ()])
end

local function LoadScript (impl, xml_desc)
  impl.phone_ui = game.current_screen.vars.phone_ui or 0 ~= 0

  local cameras = { "modal_window_camera" }

  impl.cameras = {}

  for i = 1, #cameras do
    local camera      = {}
    local camera_name = cameras [i]

    SetCameraState (camera_name, true)

    camera.node       = game.current_screen.cameras [camera_name].node
    camera.min_width  = camera.node.Right - camera.node.Left
    camera.min_height = camera.node.Top - camera.node.Bottom
    camera.center_x   = camera.node.Left + camera.min_width / 2
    camera.center_y   = camera.node.Bottom + camera.min_height / 2

    SetCameraState (camera_name, false)

    table.insert (impl.cameras, camera)
  end

  impl.windows_queue = {}
  impl.windows_stack = {}
end

local function CreateScript (xml_desc)
  local impl = {}

  LoadScript (impl, xml_desc)
  
  impl.last_update_time = game.main_timer.GetTime ()

  local methods = {}

  methods.AchievementCompleteShare          = function () AchievementCompleteShare (impl) end
  methods.AchievementCompleteView           = function () local_methods.AchievementCompleteView (impl) end
  methods.AchievementsGroupCompleteCheck    = function () AchievementsGroupCompleteCheck (impl) end
  methods.AchievementGroupCompleteWindowShare = function () AchievementGroupCompleteWindowShare (impl) end
  methods.AchievementsGroupCompleteShare    = function () local_methods.AchievementsGroupCompleteShare (impl) end
  methods.AchievementsWindowCheck           = function (index) local_methods.AchievementsWindowCheck (impl, index) end
  methods.AchievementsWindowNextPage        = function () local_methods.AchievementsWindowNextPage (impl) end
  methods.AchievementsWindowPreviousPage    = function () local_methods.AchievementsWindowPreviousPage (impl) end
	methods.AchievementsWindowSetPage					= function (page) local_methods.AchievementsWindowSetPage (impl, page) end
	methods.AchievementsWindowShowOverlay			= function () local_methods.AchievementsWindowShowOverlay (impl) end
  methods.AchievementsWindowSwitchToProfile = function () local_methods.AchievementsWindowSwitchToProfile (impl) end
  methods.BattleInviteAccept                = function () BattleInviteAccept (impl) end
  methods.BattleInviteCancel                = function () BattleInviteCancel (impl) end
  methods.BattleResultShowHideRatingPopup   = function () local_methods.BattleResultShowHideRatingPopup (impl) end
  methods.BattleResultTryAgain           = function () BattleResultTryAgain (impl) end
  methods.BattleSearchWindowSearchAgain  = function () BattleSearchWindowSearchAgain (impl) end
  methods.BattleWinWindowShare           = function () BattleWinWindowShare (impl) end
  methods.CloseBattleResultWindow        = function () CloseBattleResultWindow (impl) end
  methods.CloseBattleSearchWindow        = function () CloseBattleSearchWindow (impl) end
  methods.CloseLoadingWindow             = function () local_methods.CloseLoadingWindow (impl) end
  methods.CloseTakeFreeSpinsWindow       = function () CloseTakeFreeSpinsWindow (impl) end
  methods.CloseWinnerOfBestPlayerWindow  = function () local_methods.CloseWinnerOfBestPlayerWindow (impl) end
  methods.GooglePlusRecommend            = function () local_methods.GooglePlusRecommend (impl) end
  methods.HideBattleInfoWindow           = function () HideBattleInfoWindow (impl) end
  methods.HideWindow                     = function (stack_name, dont_process_queue) HideWindow (impl, stack_name, dont_process_queue) end
  methods.IsAlertWindowShown             = function (type) return IsAlertWindowShown (impl, type) end
  methods.IsWindowPausedGame             = function () return local_methods.IsWindowPausedGame (impl) end
  methods.InfoBonusShare                 = function () local_methods.InfoBonusShare(impl) end
  methods.InfoFiveInRowShare             = function () local_methods.InfoFiveInRowShare(impl) end
  methods.LoginToFb                      = function () local_methods.LoginToFb (impl) end
  methods.LoginToVk                      = function () local_methods.LoginToVk (impl) end
  methods.NewLevelWindowShare            = function () local_methods.NewLevelWindowShare (impl) end
  methods.OnBackPressed                  = function () return local_methods.OnBackPressed (impl) end
  methods.OnBattleInviteCanceled         = function (user_id) OnBattleInviteCanceled (impl, user_id) end
  methods.OnBattleSearchSucceeded        = function (event) OnBattleSearchSucceeded (impl, event) end
  methods.OnCancelPressed                = function () OnCancelPressed (impl) end
  methods.OnDragAchievements             = function (x) local_methods.OnDragAchievements (impl, x) end
  methods.OnDragAchievementsStarted      = function (x) local_methods.OnDragAchievementsStarted (impl, x) end
  methods.OnDragAchievementsEnded        = function () local_methods.OnDragAchievementsEnded (impl) end
  methods.OnOkPressed                    = function () OnOkPressed (impl) end
  methods.OnUserAchievementsLoaded       = function (event) local_methods.OnUserAchievementsLoaded (impl, event) end
  methods.OnUserDetailsLoaded            = function (event) local_methods.OnUserDetailsLoaded (impl, event) end
  methods.OnUserProfileLoaded            = function (event) local_methods.OnUserProfileLoaded (impl, event) end
  methods.OnSendGiftsChoosed             = function (gift_type) OnSendGiftsChoosed (impl, gift_type) end
  methods.OnSizeChanged                  = function (window_width, window_height) OnSizeChanged (impl, window_width, window_height) end
  methods.OptionsToggleSound             = function () ToggleSoundState (impl) end
  methods.ProfileWindowShowAchievements  = function () local_methods.ProfileWindowShowAchievements (impl) end
  methods.ProfileAchievementsSendGift    = function () local_methods.ProfileAchievementsSendGift (impl) end
  methods.Update                         = function () Update (impl) end
  methods.SelectAllFriends               = function () SelectAllFriends (impl) end
  methods.SendGifts                      = function () SendGifts (impl) end
  methods.SetHideWindowFunction          = function (new_function) impl.hide_window_function = new_function end
  methods.ShowAchievementsCompleteWindow = function (id, group_id, current, prize_type) ShowAchievementsCompleteWindow (impl, id, group_id, current, prize_type) end
  methods.ShowAchievementsGroupComplete  = function (group_id, prize, prize_type) ShowAchievementsGroupComplete (impl, group_id, prize, prize_type) end
  methods.ShowAchievementsWindow         = function (user_id, group_id, enqueue) local_methods.ShowAchievementsWindow (impl, user_id, group_id, enqueue) end
  methods.ShowOpenDoorsActionIsOver      = function (level) local_methods.ShowOpenDoorsActionIsOver (impl, level) end
  methods.ShowAlertWindow                = function (type, on_ok_handler, on_cancel_handler, ok_text, text_lines, show_progress_indicator, dont_disable_camera, autoclose, dont_pause_game, enqueue, on_hide_handler) ShowAlertWindow (impl, type, on_ok_handler, on_cancel_handler, ok_text, text_lines, show_progress_indicator, dont_disable_camera, autoclose, dont_pause_game, enqueue, on_hide_handler) end
  methods.ShowBattleInfoWindow           = function (icon, title, texts, show_duration) ShowBattleInfoWindow (impl, icon, title, texts, show_duration) end
  methods.ShowBattleInviteWindow         = function (mode_id, user_info, dont_return_to_lobby) ShowBattleInviteWindow (impl, mode_id, user_info); impl.dont_return_to_lobby = dont_return_to_lobby end
  methods.ShowBattleLeaveWindow          = function (on_ok_handler) local_methods.ShowBattleLeaveWindow (impl, on_ok_handler) end
  methods.ShowBattleSearchWindow         = function (mode_id, slot_id, slot_path, opponent_id) ShowBattleSearchWindow (impl, mode_id, slot_id, slot_path, opponent_id) end
  methods.ShowBattleStartCounter         = function (on_hide_handler) ShowBattleStartCounter (impl, on_hide_handler) end
  methods.ShowBattleWinWindow            = function (coins, rating, interrupted, dont_return_to_lobby) ShowBattleWinWindow (impl, coins, rating, interrupted); impl.dont_return_to_lobby = dont_return_to_lobby end
  methods.ShowBattleLoseWindow           = function (dont_return_to_lobby) ShowBattleLoseWindow (impl); impl.dont_return_to_lobby = dont_return_to_lobby end
  methods.ShowBattleDrawWindow           = function (coins, dont_return_to_lobby) ShowBattleDrawWindow (impl, coins); impl.dont_return_to_lobby = dont_return_to_lobby end
  methods.ShowBonusByLevelWindow         = function (level, percent) local_methods.ShowBonusByLevelWindow (impl, level, percent) end
  methods.ShowBonusCoinsWindow           = function (coins, close_handler) local_methods.ShowBonusCoinsWindow(impl, coins, close_handler) end
  methods.ShowBonusGoogle                = function (coins) local_methods.ShowBonusGoogle(impl, coins) end
  methods.ShowBonusEndedWindow           = function (bonus_game_name, icon_sprite, icon_tex_size, icon_tex_offset, icon_scale, hs_path) ShowBonusEndedWindow (impl, bonus_game_name, icon_sprite, icon_tex_size, icon_tex_offset, icon_scale, hs_path) end
  methods.ShowBuyChipsWindow             = function () ShowBuyChipsWindow (impl) end
  methods.ShowBuyDiamondsWindow          = function () ShowBuyDiamondsWindow (impl) end
  methods.ShowChangeLanguageWindow       = function () local_methods.ShowChangeLanguageWindow (impl) end
  methods.ShowFiveInRowWindow            = function (icon_sprite, icon_tex_size, icon_tex_offset, icon_scale, coins, hs_path, symbol_id) ShowFiveInRowWindow (impl, icon_sprite, icon_tex_size, icon_tex_offset, icon_scale, coins, hs_path, symbol_id) end
  methods.ShowFreeSpinsEndedWindow       = function () ShowFreeSpinsEndedWindow (impl) end
  methods.ShowHappyHourWindowAction      = function (action, closed_handler, closed_ok_handler) return local_methods.ShowHappyHourWindowAction (impl, action, closed_handler, closed_ok_handler) end
  methods.ShowHappyHourWindowCoins       = function (day_or_hour, coins, close_handler, close_ok_handler, time) local_methods.ShowHappyHourWindow (impl, nil, day_or_hour, 2, coins, nil, close_handler, close_ok_handler, time) end
  methods.ShowHappyHourWindowDeposit     = function (day_or_hour, coins, close_handler, close_ok_handler, time) local_methods.ShowHappyHourWindow (impl, nil, day_or_hour, 1, tostring (coins) .. "%", nil, close_handler, close_ok_handler, time) end
  methods.ShowHappyHourWindowDiscount    = function (day_or_hour, coins, close_handler, close_ok_handler, time) local_methods.ShowHappyHourWindow (impl, nil, day_or_hour, 4, tostring(coins) .. "%", nil, close_handler, close_ok_handler, time) end
  methods.ShowHappyHourWindowFreeSpins   = function (day_or_hour, coins, close_handler, close_ok_handler, time) local_methods.ShowHappyHourWindow (impl, nil, day_or_hour, 3, tostring(coins), nil, close_handler, close_ok_handler, time) end
  methods.ShowHappyHourWindowSlot        = function (day_or_hour, slot_id, level, close_handler, close_ok_handler, time) local_methods.ShowHappyHourWindow (impl, level, day_or_hour, nil, nil, slot_id, close_handler, close_ok_handler, time) end
  methods.ShowHappyTimeActionEndedWindow = function (day_or_hour, close_handler) local_methods.ShowHappyTimeActionEndedWindow (impl, day_or_hour, close_handler) end
  methods.ShowHelloCombinedWindow        = function (level, money, diamonds) local_methods.ShowHelloCombinedWindow (impl, level, money, diamonds) end
  methods.ShowLevelUpWindow              = function () ShowLevelUpWindow (impl) end
  methods.ShowLoadingWindow              = function (progress, first_line_text, closable, close_handler, slot_id) ShowLoadingWindow (impl, progress, first_line_text, closable, close_handler, slot_id) end
  methods.ShowLogoutWindow               = function (logout_handler, texts_override) ShowLogoutWindow (impl, logout_handler, texts_override) end
  methods.ShowLockedFunctionalityFriendsWindow = function () local_methods.ShowLockedFunctionalityFriendsWindow(impl) end
  methods.ShowLockedFunctionalityWindow  = function (type) local_methods.ShowLockedFunctionalityWindow (impl, type) end
  methods.ShowNewGameWindow              = function (slot_id) ShowNewGameWindow (impl, slot_id) end
  methods.ShowNoDiamondWindow            = function () ShowNoDiamondWindow (impl) end
  methods.ShowNoMoneyWindow              = function () ShowNoMoneyWindow (impl) end
  methods.ShowNotEnoughMoneyWindow       = function (play_handler) ShowNotEnoughMoneyWindow (impl, play_handler) end
  methods.ShowNotHaveCoinsEnoughForBattleWindow = function (battle_search) local_methods.ShowNotHaveCoinsEnoughForBattleWindow (impl, battle_search) end
  methods.ShowNinetyNineLinesWindow      = function () local_methods.ShowNinetyNineLinesWindow (impl) end
  methods.ShowInfoChangeLanguageWindow          = function (on_ok_handler) local_methods.ShowInfoChangeLanguageWindow (impl, on_ok_handler) end
  methods.ShowInfoSocialConnectingWindow        = function () local_methods.ShowInfoSocialConnectingWindow (impl) end
  methods.ShowInfoWindowDataTransferWindow      = function () local_methods.ShowInfoWindowDataTransferWindow (impl) end
  methods.ShowOptionsWindow              = function () ShowOptionsWindow (impl) end
  methods.ShowPopupTopWindow             = function () ShowPopupTopWindow (impl) end
  methods.ShowProfileWindow              = function (user_id) local_methods.ShowProfileWindow (impl, user_id) end
  methods.ShowRatingWindow               = function (slot_id) local_methods.ShowRatingWindow (impl, slot_id) end
  methods.EditTextRatingWindow           = function () local_methods.EditTextRatingWindow (impl) end
  methods.ShowSendGiftsWindow            = function (on_ok_handler) ShowSendGiftsWindow (impl, on_ok_handler) end
  methods.ShowSendGiftsFriendsWindow     = function (gift_type) ShowSendGiftsFriendsWindow (impl, gift_type) end
  methods.ShowTakeFreeSpinsWindow        = function (gift_id, close_handler) ShowTakeFreeSpinsWindow (impl, gift_id, close_handler) end
  methods.ShowTopNotificationWindow      = function (info) local_methods.ShowTopNotificationWindow (impl, info) end
  methods.ShowUpdateApplicationWindow    = function (cancellable) ShowUpdateApplicationWindow (impl, cancellable) end
  methods.ShowWelcomeWindow              = function (coins, diamonds) local_methods.ShowWelcomeWindow (impl, coins, diamonds) end
  methods.ShowWinnerOfBestPlayerWindow   = function (place, type, coins_or_diamonds, count_coins_or_diamonds, prize_id) ShowWinnerOfBestPlayerWindow (impl, place, type, coins_or_diamonds, count_coins_or_diamonds, prize_id) end
  methods.TopNotificationCheck           = function () local_methods.TopNotificationCheck (impl) end
  methods.TopNotificationWindowHide      = function () local_methods.TopNotificationWindowHide (impl) end
  methods.WinnerOfBestPlayerCheck        = function () WinnerOfBestPlayerCheck (impl) end
  methods.SetStarCount                   = function (count_star) local_methods.SetStarCount (impl, count_star) end

  --utility methods
  methods.GetTextLines   = function (list) return GetTextLines (list) end
  methods.MakeAllSprites = function (list) return MakeAllSprites (list) end
  methods.ShowWindow     = function (window, additional_sprites, additional_text_lines, text_lines, on_window_shown, dont_disable_camera, dont_pause_game, params) params = params or {}; ShowWindow (impl, window, additional_sprites, additional_text_lines, text_lines, on_window_shown, dont_disable_camera, dont_pause_game, params.dont_hide_previous_windows, params.on_ok_handler, params.on_cancel_handler, params.additional_resource_bindings, params.enqueue, params.stack_name, params.before_window_show, params) end

  return methods
end

return CreateScript
